package ghmessen;

import ghmessen.components.MessgeraeteToolBar;
import ghmessen.groesse.Darstellungsliste;
import ghmessen.groesse.Groessenliste;
import ghmessen.groesse.WerteTabelle;
import ghmessen.instrumente.MessgeraetStatus;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.HashMap;
import java.util.Map;
import java.util.Vector;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.sax.SAXTransformerFactory;
import javax.xml.transform.sax.TransformerHandler;
import javax.xml.transform.stream.StreamResult;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.AttributesImpl;
import org.xml.sax.helpers.DefaultHandler;

public class GHMXFile {

	private static final boolean DEBUG = false;
	/**
	 * GMXFile.version
	 * ===============
	 * Dateiveränderungen werden in "workaround(...) bearbeitet.
	 * 
	 * GHMX_1.0:
	 *   - Groesse.art wird beim Speichern von Messgroesse zu Parameter.
	 *           Daher muss beim Einlesen jeder Parameter testweise auf Messgroesse
	 *           geändert werden (siehe Ticket #28).
	 */
	public static String version = "GHMX_1.1";
	public static String[] compatibleVersions = new String[]{"GHMX_1.0"};
	
	private static enum BEREICH {
		ANFANG, GHMESSEN, KOMMENTAR, GROESSENLISTE, MESSPARAMETER, WERTETABELLE, DARSTELLUNGSLISTE, MESSGERAETE
	};

	public class Reader extends DefaultHandler {

		private SAXParserFactory spfac;
		private SAXParser sp;
		private File f;
		private String temp;
		private String kommentar;
		private Groessenliste gl;
		private Messparameter mp;
		private WerteTabelle gt;
		private Darstellungsliste ansicht;
		private boolean messgeraeteSichtbar;
		private Map<String, MessgeraetStatus> messgeraeteStatus;
		private int trenner;
		private String dateiVersion;

		private BEREICH bereich;

		public Reader(File f) throws IOException {
			this.f = f;
			spfac = SAXParserFactory.newInstance();
			spfac.setValidating(true);
			bereich = BEREICH.ANFANG;
			try {
				sp = spfac.newSAXParser();
			} catch (SAXException e) {
				throw new IOException(e.getMessage());
			} catch (ParserConfigurationException e) {
				throw new IOException(e.getMessage());
			}
		}

		public void parse() throws SAXException, IOException {
			if (sp == null || f == null)
				throw new IOException(
						"[GHMXFile] Reader ist nicht initialisiert.");
			sp.parse(f, this);
		}

		public void characters(char[] buffer, int start, int length) {
			temp = new String(buffer, start, length);
		}

		private Attributes workaround(String uri, String localName, String qName,
				Attributes attributes) {
			if (dateiVersion != null && dateiVersion.equals("GHMX_1.0") && qName.equals("Groesse")){
				AttributesImpl attributesImpl = new AttributesImpl(attributes);
				/* WORKAROUND: Ticket #28 */
				if (attributes.getValue("art").equals("Parameter")){
					attributesImpl.setValue(attributesImpl.getIndex("art"), "Messgroesse");
				}
				return attributesImpl;
			}
			return attributes;
		}
		
		public void startElement(String uri, String localName, String qName,
				Attributes attributes) throws SAXException {
			attributes = workaround(uri,localName,qName,attributes);
			switch (bereich) {
			case ANFANG:
				if (qName.equals("GHMessen")) {
					bereich = null;
					dateiVersion = attributes.getValue("Version");
					if(version.equals(dateiVersion))
						bereich = BEREICH.GHMESSEN;
					for(String cVersion : compatibleVersions)
						if( cVersion.equals(dateiVersion)) {
							bereich = BEREICH.GHMESSEN;
							break;
						}
					if(bereich == null)
						throw new SAXException("[GHMXFile] - Reader: Status("
								+ bereich + ") Start(" + qName + "): Falsche Datei-Version("+dateiVersion+")");						
				}
				else
					throw new SAXException("[GHMXFile] - Reader: Status("
							+ bereich + ") Start(" + qName + "): Keine GHMessen-Datei");
				break;
			case GHMESSEN:
				if (qName.equals("Messparameter")) {
					bereich = BEREICH.MESSPARAMETER;
					mp = new Messparameter(null, null);
					mp.startElement(uri, localName, qName, attributes, gl);
				} else if (qName.equals("Groessenliste")) {
					bereich = BEREICH.GROESSENLISTE;
					gl = new Groessenliste();
					gl.startElement(uri, localName, qName, attributes);
				} else if (qName.equals("WerteTabelle")) {
					bereich = BEREICH.WERTETABELLE;
					gt = new WerteTabelle(gl, null);
					gt.startElement(uri, localName, qName, attributes);
				} else if (qName.equals("Darstellungsliste")) {
					bereich = BEREICH.DARSTELLUNGSLISTE;
					ansicht = new Darstellungsliste();
					ansicht.startElement(uri, localName, qName, attributes);
				} else if (qName.equals("Kommentar")
						|| qName.equals("MessgeraeteSichtbar")
						|| qName.equals("Trenner")) {
					return;
				} else if (qName.equals("Messgeraete")) {
					bereich = BEREICH.MESSGERAETE;
					messgeraeteStatus = new HashMap<String, MessgeraetStatus>();
					MessgeraeteToolBar.startElement(uri, localName, qName,
							attributes, messgeraeteStatus);
				}
				break;
			case GROESSENLISTE:
				gl.startElement(uri, localName, qName, attributes);
				break;
			case MESSPARAMETER:
				mp.startElement(uri, localName, qName, attributes, gl);
				break;
			case WERTETABELLE:
				gt.startElement(uri, localName, qName, attributes);
				break;
			case DARSTELLUNGSLISTE:
				ansicht.startElement(uri, localName, qName, attributes);
				break;
			case MESSGERAETE:
				MessgeraeteToolBar.startElement(uri, localName, qName,
						attributes, messgeraeteStatus);
				break;
			default:
				break;
			}
			temp = "";
		}

		public void endElement(String uri, String localName, String qName)
				throws SAXException {
			switch (bereich) {
			case ANFANG:
				throw new SAXException("[GHMXFile] - Reader: Status(" + bereich
						+ ") Ende(" + qName + ")");
			case GHMESSEN:
				if (qName.equals("Kommentar"))
					kommentar = temp;
				else if (qName.equals("MessgeraeteSichtbar"))
					messgeraeteSichtbar = Boolean.parseBoolean(temp);
				else if (qName.equals("Trenner"))
					trenner = Integer.parseInt(temp);
				else if (qName.equals("GHMessen"))
					bereich = BEREICH.ANFANG;
				else
					throw new SAXException("[GHMXFile] - Reader: Status("
							+ bereich + ") Ende(" + qName + ")");
				break;
			case GROESSENLISTE:
				if (DEBUG)
					System.out.println("[GMXFile] - endElement: Status("
							+ bereich + ") Ende(" + qName + ")");
				gl.endElement(uri, localName, qName, temp);
				if (qName.equals("Groessenliste"))
					bereich = BEREICH.GHMESSEN;
				break;
			case MESSPARAMETER:
				mp.endElement(uri, localName, qName, temp);
				if (qName.equals("Messparameter"))
					bereich = BEREICH.GHMESSEN;
				break;
			case WERTETABELLE:
				gt.endElement(uri, localName, qName, temp);
				if (qName.equals("WerteTabelle"))
					bereich = BEREICH.GHMESSEN;
				break;
			case DARSTELLUNGSLISTE:
				ansicht.endElement(uri, localName, qName, temp);
				if (qName.equals("Darstellungsliste"))
					bereich = BEREICH.GHMESSEN;
				break;
			case MESSGERAETE:
				MessgeraeteToolBar.endElement(uri, localName, qName, temp);
				if (qName.equals("Messgeraete"))
					bereich = BEREICH.GHMESSEN;
				break;
			default:
				break;
			}
		}

		public String getKommentar() {
			return kommentar == null ? "" : kommentar;
		}

		public Messparameter getMessparameter() {
			return mp;
		}

		public Groessenliste getGroessenliste() {
			return gl;
		}

		@SuppressWarnings("rawtypes")
		public Vector<Vector> getWerte() {
			return gt == null ? null : gt.getDataVector();
		}

		public Darstellungsliste getDarstellungsliste() {
			return ansicht;
		}

		public boolean sindMessgeraeteSichtbar() {
			return messgeraeteSichtbar;
		}

		public Map<String, MessgeraetStatus> getMessgeraeteStatus() {
			return messgeraeteStatus;
		}

		public int getTrenner() {
			return trenner;
		}
	}

	public class Writer {
		private PrintWriter out;
		private StreamResult streamResult;
		private SAXTransformerFactory tf;
		private TransformerHandler hd;
		private Transformer serializer;

		public Writer(File file) throws IOException {
			try {
				out = new PrintWriter(file);
				streamResult = new StreamResult(out);
				tf = (SAXTransformerFactory) SAXTransformerFactory
						.newInstance();
				// SAX2.0 ContentHandler.
				hd = tf.newTransformerHandler();
				serializer = hd.getTransformer();
				serializer.setOutputProperty(OutputKeys.ENCODING, "UTF-8");
				serializer.setOutputProperty(OutputKeys.DOCTYPE_SYSTEM,
						"/usr/share/ghmessen/dtd/ghmessen.dtd");
				serializer.setOutputProperty(OutputKeys.INDENT, "yes");
				hd.setResult(streamResult);
			} catch (TransformerConfigurationException e) {
				throw new IOException(e.getMessage());
			}
		}

		public void close() {
			if (out != null)
				out.close();
			out = null;
		}

		public TransformerHandler getTransformerHandler() throws SAXException {
			if (hd == null)
				throw new SAXException(
						"[GHMXFile] Das Objekt ist nicht initialisiert.");
			return hd;
		}

		public void writeXmlProlog() throws SAXException {
			if (hd == null)
				throw new SAXException(
						"[GHMXFile] Objekt ist nicht initialisiert.");
			hd.startDocument();
			AttributesImpl atts = new AttributesImpl();
			atts.addAttribute("", "", "Version", "CDATA", version);
			hd.startElement("", "", "GHMessen", atts);
		}

		public void writeKommentar(String kommentar) throws SAXException {
			if (hd == null)
				throw new SAXException(
						"[GHMXFile] Objekt ist nicht initialisiert.");
			hd.startElement("", "", "Kommentar", null);
			if (kommentar != null)
				hd.characters(kommentar.toCharArray(), 0, kommentar.length());
			hd.endElement("", "", "Kommentar");
		}

		public void writeMessgeraeteSichtbar(boolean an) throws SAXException {
			if (hd == null)
				throw new SAXException(
						"[GHMXFile] Objekt ist nicht initialisiert.");
			hd.startElement("", "", "MessgeraeteSichtbar", null);
			String inhalt = Boolean.toString(an);
			hd.characters(inhalt.toCharArray(), 0, inhalt.length());
			hd.endElement("", "", "MessgeraeteSichtbar");
		}

		public void writeDividerLocation(int location) throws SAXException {
			if (hd == null)
				throw new SAXException(
						"[GHMXFile] Objekt ist nicht initialisiert.");
			hd.startElement("", "", "Trenner", null);
			String s = Integer.toString(location);
			hd.characters(s.toCharArray(), 0, s.length());
			hd.endElement("", "", "Trenner");
		}

		public void writeXmlEpilog() throws SAXException {
			if (hd == null)
				throw new SAXException(
						"[GHMXFile] Objekt ist nicht initialisiert.");
			hd.endElement("", "", "GHMessen");
			hd.endDocument();
		}

	}

	public static void createDTDHeader(PrintWriter out) {
		out.println("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
		out.println("<!ELEMENT GHMessen (Kommentar,Groessenliste,Messparameter,WerteTabelle,Darstellungsliste,");
		out.println("\tMessgeraeteSichtbar,Messgeraete,Trenner)>");
		out.println("<!ATTLIST GHMessen Version CDATA #REQUIRED>");
		out.println("<!ELEMENT Kommentar (#PCDATA)>");
		out.println("<!ELEMENT MessgeraeteSichtbar (BOOLEAN)>");
		out.println("<!ELEMENT Trenner (#PCDATA)>");
		out.println("<!ELEMENT INT (#PCDATA)>");
		out.println("<!ELEMENT DOUBLE (#PCDATA)>");
		out.println("<!ELEMENT BOOLEAN (true|false)>");
	}

	public static void createDTDFooter(PrintWriter out) {
		return;
	}

	public static void createDTD(PrintWriter out) {
		createDTDHeader(out);
		Groessenliste.createDTD(out);
		Messparameter.createDTD(out);
		WerteTabelle.createDTD(out);
		Darstellungsliste.createDTD(out);
		MessgeraeteToolBar.createDTD(out);
		createDTDFooter(out);
		out.close();
	}

	public static void main(String[] args) throws IOException {
		String dateiName = "src/ghmessen/ghmessen.dtd";
		if (args.length == 1)
			dateiName = args[0];
		File out = new File(dateiName);
		System.out.println("Neue DTD-Datei: " + out.getAbsolutePath());
		GHMXFile.createDTD(new PrintWriter(out));
	}
}
