/* 
 * GHMessen
 * --------
 * 
 * GHMessen is a measuring program to use in connection with a computer 
 * measuring interface. It's designed for school usage.
 * 
 * Copyright (C) 2010-2011  Frank Schütte - Gymnasium Himmelsthür
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms
 * of the GNU General Public License as published by the Free Software Foundation; either 
 * version 3 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with this program; 
 * if not, see <http://www.gnu.org/licenses/>.
 * 
 * Frank Schütte fschuett AT gymnasium-himmelsthuer DOT de
 * 
 */

package ghmessen;

import ghmessen.adapter.EAdapterException;
import ghmessen.components.GreekKeyEventDispatcher;
import ghmessen.einstellungen.PrefObjs;
import ghmessen.util.LocalHostSocketFactory;

import java.awt.Image;
import java.awt.KeyboardFocusManager;
import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.server.ExportException;
import java.rmi.server.RMISocketFactory;
import java.util.HashMap;
import java.util.Locale;
import java.util.Vector;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.prefs.BackingStoreException;
import java.util.prefs.InvalidPreferencesFormatException;
import java.util.prefs.Preferences;

import javax.help.HelpBroker;
import javax.help.HelpSet;
import javax.help.HelpSetException;
import javax.print.PrintService;
import javax.print.PrintServiceLookup;
import javax.print.attribute.HashPrintRequestAttributeSet;
import javax.print.attribute.PrintRequestAttributeSet;
import javax.print.attribute.standard.MediaSizeName;
import javax.swing.ImageIcon;
import javax.swing.UIManager;
import javax.swing.UIManager.LookAndFeelInfo;

import libGoIO.GoIOSensor;
import libGoIO.LibGoIO;

import org.jdesktop.application.Application;
import org.jdesktop.application.SingleFrameApplication;

/**
 * @author fschuett
 * 
 */
public class GHMessen extends SingleFrameApplication {
	private GHMessenView mainWindow = null;
	private String initialFile = null;
	private GreekKeyEventDispatcher gkDispatcher = null;
	private Vector<Process> subProcess;

	public double ghmMittelZeit = Konstanten.GROESSE_MITTELZEIT_STANDARD;

	private static final boolean DEBUG = false;
	private boolean rmiRegistryGeladen;

	// Gespeicherte Einstellungen
	public static Preferences prefs;
	public static Vector<String> recentFiles;
	public static String RECENTFILES = "ZuletztVerwendet";
	public static int RECENTFILESMAX = 10;
	
	// Datenstrukturen für den Ausdruck
	public PrintService druckDienst;
	public PrintRequestAttributeSet druckAttribut;
	
	// Eigenschaften des Programms <-> GHMessen.xml
	public HashMap<String, GoIOSensor> vernierSensoren;
	public static String VERNIERSENSOREN = "VernierSensoren";
	public static String EIGENSCHAFTEN = "Eigenschaften";

	// Programm-Hilfe
	public static HelpSet hSet;
	public static HelpBroker hilfe;

	// Bibliotheken für Messprogramme
	public transient LibGoIO goIO;
	
	public boolean istRmiRegistryGeladen() {
		return rmiRegistryGeladen;
	}

	public void ladeRmiRegistry() {
		if (rmiRegistryGeladen)
			return;
		// starte bei Bedarf die Registry
		try {
			if (System.getProperty("java.rmi.server.hostname") == null)
				System.setProperty("java.rmi.server.hostname", "localhost");
			RMISocketFactory factory = new LocalHostSocketFactory();
			RMISocketFactory.setSocketFactory(factory);
			LocateRegistry.createRegistry(Konstanten.RMI_PORT);
			if (DEBUG)
				System.out
						.println("[GHMessen] - ladeRmiRegistry: Registry wurde geladen.");
		} catch (ExportException e) {
			// registry already running, so connect
			if (DEBUG)
				e.printStackTrace();
		} catch (RemoteException e) {
			if (DEBUG)
				e.printStackTrace();
		} catch (IOException e) {
			if (DEBUG)
				e.printStackTrace();
		}
		rmiRegistryGeladen = true;
	}

	public String getInitialFile() {
		return initialFile;
	}

	@Override
	public void initialize(String[] args) {
		super.initialize(args);
		subProcess = new Vector<Process>();
		if (args.length > 0 && !"".equals(args[0])) {
			initialFile = args[0];
			System.out.println("[GHMessen] Argumentanzahl(" + args.length
					+ ") Argumente(" + args + ")");
		}
		gkDispatcher = new GreekKeyEventDispatcher();
		KeyboardFocusManager.getCurrentKeyboardFocusManager()
				.addKeyEventDispatcher(gkDispatcher);
	}

	@SuppressWarnings("unchecked")
	@Override
	protected void startup() {
		try {
		    for (LookAndFeelInfo info : UIManager.getInstalledLookAndFeels()) {
		        if ("Nimbus".equals(info.getName())) {
		            UIManager.setLookAndFeel(info.getClassName());
		            break;
		        }
		    }
		} catch (Exception e) {
		    // If Nimbus is not available, you can set the GUI to another look and feel.
		}
		try {
			goIO = new LibGoIO();
		} catch (EAdapterException e1) {
			if(DEBUG)
				e1.printStackTrace();
			goIO = null;
		}
		prefs = Preferences.userNodeForPackage(this.getClass());
		// Einstellungen aus XMl-Datei laden
		try {
			Eigenschaften.readConfig();
		} catch (IOException e) {
			if (DEBUG)
				e.printStackTrace();
		} catch (InvalidPreferencesFormatException e) {
			if (DEBUG)
				e.printStackTrace();
		}
		try {
			vernierSensoren = ((HashMap<String, GoIOSensor>) Eigenschaften
					.getObject(prefs, VERNIERSENSOREN));
		} catch (IOException e) {
			if (DEBUG)
				e.printStackTrace();
			vernierSensoren = new HashMap<String, GoIOSensor>();
		} catch (BackingStoreException e) {
			if (DEBUG)
				e.printStackTrace();
			vernierSensoren = new HashMap<String, GoIOSensor>();
		} catch (ClassNotFoundException e) {
			if (DEBUG)
				e.printStackTrace();
			vernierSensoren = new HashMap<String, GoIOSensor>();
		}

		// Druckereinstellungen
		try {
			druckDienst = PrintServiceLookup.lookupDefaultPrintService();
			String druckerName = prefs.get("Drucker", null);
			if (druckerName != null) {
				PrintService[] druckDienste = null;
				druckDienste = PrintServiceLookup.lookupPrintServices(null,
						null);
				for (PrintService psv : druckDienste) {
					if (druckerName.equals(psv.getName())) {
						druckDienst = psv;
						break;
					}
				}
			}
			druckAttribut = (PrintRequestAttributeSet) PrefObjs.getObject(
					prefs, "Druckattribute");
			if (druckAttribut == null) {
				druckAttribut = new HashPrintRequestAttributeSet();
				druckAttribut.add(MediaSizeName.ISO_A4);
			}
		} catch (BackingStoreException ex) {
			Logger.getLogger(GHMessen.class.getName()).log(Level.SEVERE, null,
					ex);
		} catch (IOException ex) {
			Logger.getLogger(GHMessen.class.getName()).log(Level.SEVERE, null,
					ex);
		} catch (ClassNotFoundException ex) {
			Logger.getLogger(GHMessen.class.getName()).log(Level.SEVERE, null,
					ex);
		}

		try {
			recentFiles = (Vector<String>) PrefObjs.getObject(prefs, RECENTFILES);
		} catch (BackingStoreException | IOException  | ClassNotFoundException ex) {
			Logger.getLogger(GHMessen.class.getName()).log(Level.SEVERE, null,
					ex);
			recentFiles = new Vector<>();
		}

		mainWindow = new GHMessenView(this);
		Image img = new ImageIcon(
				GHMessen.class.getResource("resources/ghmessen.png"))
				.getImage();
		mainWindow.getFrame().setIconImage(img);
		show(mainWindow);
		if (DEBUG)
			System.out.println("[GHMessen] Locale: " + Locale.getDefault());
		if (initialFile != null) {
			File f = new File(initialFile);
			mainWindow.dateiLaden(f);
		} else
			mainWindow.showEinstellungenBox(null);
	}

	public void registerProcess(Process p) {
		if (subProcess != null && p != null)
			subProcess.add(p);
	}

	public void unregisterProcess(Process p) {
		for (Process q : subProcess)
			if (q != null && q.equals(p)) {
				subProcess.remove(q);
			}
	}

	public static void appendToRecentFiles(String s) {
		if(recentFiles == null)
			recentFiles = new Vector<>();
		if(!recentFiles.contains(s))
			recentFiles.add(s);
		if(recentFiles.size() > RECENTFILESMAX)
			recentFiles.remove(0);
	}
	
	@Override
	protected void shutdown() {
		try {
			Eigenschaften.putObject(prefs, VERNIERSENSOREN, vernierSensoren);
		} catch (IOException e) {
			if (DEBUG)
				e.printStackTrace();
		} catch (BackingStoreException e) {
			if (DEBUG)
				e.printStackTrace();
		} catch (ClassNotFoundException e) {
			if (DEBUG)
				e.printStackTrace();
		}
		// Druckereinstellungen
		if (prefs != null) {
			try {
				if (druckDienst != null) {
					prefs.put("Drucker", druckDienst.getName());
				}
				PrefObjs.putObject(prefs, "Druckattribute", druckAttribut);
				PrefObjs.putObject(prefs, RECENTFILES, recentFiles);
			} catch (BackingStoreException ex) {
				Logger.getLogger(GHMessen.class.getName()).log(Level.SEVERE,
						null, ex);
			} catch (IOException ex) {
				Logger.getLogger(GHMessen.class.getName()).log(Level.SEVERE,
						null, ex);
			} catch (ClassNotFoundException ex) {
				Logger.getLogger(GHMessen.class.getName()).log(Level.SEVERE,
						null, ex);
			}
		}
		try {
			Eigenschaften.writeConfig(prefs);
			// Eigenschaften werden nicht unter den anderen Preferences gespeichert.
			prefs.node(EIGENSCHAFTEN).removeNode();
		} catch (IOException e) {
			if (DEBUG)
				e.printStackTrace();
		} catch (BackingStoreException e) {
			if (DEBUG)
				e.printStackTrace();
		}
		for (Process p : subProcess)
			if (p != null)
				p.destroy();
		if(goIO != null)
			goIO.uninit();
		goIO = null;
		if (mainWindow != null)
			mainWindow.close();
		super.shutdown();
	}

	/**
	 * This method is to initialize the specified window by injecting resources.
	 * Windows shown in our application come fully initialized from the GUI
	 * builder, so this additional configuration is not needed.
	 */
	@Override
	protected void configureWindow(java.awt.Window root) {
	}

	/**
	 * A convenient static getter for the application instance.
	 * 
	 * @return the instance of GHMessen
	 */
	public static GHMessen getApplication() {
		return Application.getInstance(GHMessen.class);
	}

	/**
	 * Main method launching the application.
	 */
	public static void main(String[] args) {
		launch(GHMessen.class, args);
	}

	@Override
	public GHMessenView getMainView() {
		return (GHMessenView) super.getMainView();
	}

	static {
		System.setProperty("java.rmi.server.hostname", "localhost");
		// Hilfeanbindung
		try {
			URL hsURL = HelpSet.findHelpSet(null, "ghmessen/docs/"
					+ Locale.getDefault().getLanguage() + "/GHMessen");
			if (hsURL == null)
				hsURL = HelpSet.findHelpSet(null, "ghmessen/docs/GHMessen.hs");
			hSet = new HelpSet(null, hsURL);
			hilfe = hSet.createHelpBroker();
		} catch (HelpSetException e) {
			if(DEBUG)
				e.printStackTrace();
			System.out.println("[GHMessen] - Hilfe-Datei nicht gefunden. Hilfe wird deaktiviert.");
			hSet = null;
			hilfe = null;
		}
	}
}
