/* 
 * GHMessen
 * --------
 * 
 * GHMessen is a measuring program to use in connection with a computer 
 * measuring interface. It's designed for school usage.
 * 
 * Copyright (C) 2010-2011  Frank Schütte - Gymnasium Himmelsthür
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms
 * of the GNU General Public License as published by the Free Software Foundation; either 
 * version 3 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with this program; 
 * if not, see <http://www.gnu.org/licenses/>.
 * 
 * Frank Schütte fschuett AT gymnasium-himmelsthuer DOT de
 * 
 */

package ghmessen;

import ghmessen.adapter.AdapterListe;
import ghmessen.adapter.IAdapter;
import ghmessen.filefilter.FileFormatException;
import ghmessen.groesse.Art;
import ghmessen.groesse.Einheit;
import ghmessen.groesse.Groesse;
import ghmessen.groesse.Groessenliste;
import ghmessen.groesse.ObservableObjectListListener;
import ghmessen.parser.FormelGenerator;
import ghmessen.util.Util;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.PrintWriter;
import java.io.Serializable;
import java.lang.reflect.Field;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.xml.transform.sax.TransformerHandler;

import ld.api.IXRay;
import ld.api.TriggerBits;
import ld.api.XRayData;
import ld.api.XRayModes;

import org.codehaus.janino.ExpressionEvaluator;
import org.jdesktop.observablecollections.ObservableList;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.AttributesImpl;

//TODO Messintervall von selektierten Größen abhängig ändern
public class Messparameter implements Serializable,
		ObservableObjectListListener {
	private static final long serialVersionUID = 2531635932762492340L;
	private static final boolean DEBUG = false;

	public static final int MODUSCASSY = 1;
	public static final int MODUSXRAY = 2;

	private int modus;
	private boolean einzelEin;
	private boolean signalEin;
	private boolean anhaengenEin;
	private boolean fortsetzenEin;
	private boolean ueberschreibenEin;
	private boolean wiederholungEin;

	/*
	 * Messmodus: einzelEin = true ueberschreibenEin = true : ersetzen des
	 * bisherigen Messwertes ueberschreibenEin = false: bisherige Messreihe
	 * fortsetzen
	 * 
	 * einzelEin = false ueberschreibenEin = true : ersetzen der letzten
	 * Messreihe ueberschreibenEin = false & anhaengenEin = true fortsetzenEin =
	 * true : bisherige Messreihe fortsetzen & anhaengenEin = true fortsetzenEin
	 * = false: mit null-Zeile als Trenner anhängen & anhaengenEin = false :
	 * neue Messreihe erzeugen
	 */
	
	/* Intervall, Wiederholung und Messzeit hängen zusammen.
	 * wiederholung != 0
	 *    Intervall und Wiederholung bestimmen gemeinsam den Wert von Messzeit.
	 *    
	 * wiederholung == 0
	 *    Intervall und Messzeit können unabhängig voneinander eingestellt werden.
	 *    Der Standardwert beträgt 2 Minuten.
	 */
	private double intervall; /* s */
	private long wiederholung;
	private double messzeit; /* s */
	
	
	private boolean timerBox;

	private boolean triggerEin;
	// Software-Trigger
	private String triggerTerm;
	private transient ExpressionEvaluator triggerEE;
	// Oszilloskop-Trigger
	private String triggerSymbol;
	private TriggerBits triggerFlanke;
	private double triggerWert;

	private boolean bedingungEin;
	private String bedingungTerm;
	private transient ExpressionEvaluator bedingungEE;
	private boolean stoppEin;
	private String stoppTerm;
	private transient ExpressionEvaluator stoppEE;

	/* XRay */
	public transient XRayData parameter;
	private transient XRayModes mode;

	protected Groesse xrg;

	/* Adapterliste */
	private AdapterListe al;

	/* manuelle Veränderung */
	protected boolean manuell;

	public static void createDTD(PrintWriter out) {
		out.println("<!ELEMENT Messparameter (XRayData)>");
		out.println("<!ATTLIST Messparameter ");
		out.println("\tmodus CDATA #REQUIRED");
		out.println("\teinzelEin (BOOLEAN) #REQUIRED");
		out.println("\tsignalEin (BOOLEAN) #REQUIRED");
		out.println("\tanhaengenEin (BOOLEAN) #REQUIRED");
		out.println("\tfortsetzenEin (BOOLEAN) #REQUIRED");
		out.println("\tueberschreibenEin (BOOLEAN) #REQUIRED");
		out.println("\twiederholungEin (BOOLEAN) #REQUIRED");
		out.println("\tintervall CDATA #REQUIRED");
		out.println("\twiederholung CDATA #REQUIRED");
		out.println("\tmesszeit CDATA #REQUIRED");
		out.println("\tmessEinheit CDATA #IMPLIED");
		out.println("\ttimerBox CDATA #REQUIRED");
		out.println("\ttriggerEin (BOOLEAN) #REQUIRED");
		out.println("\ttriggerTerm CDATA #REQUIRED");
		out.println("\ttriggerSymbol CDATA #REQUIRED");
		out.println("\ttriggerFlanke CDATA #REQUIRED");
		out.println("\ttriggerWert CDATA #REQUIRED");
		out.println("\tbedingungEin (BOOLEAN) #REQUIRED");
		out.println("\tbedingungTerm CDATA #REQUIRED");
		out.println("\tstoppEin (BOOLEAN) #REQUIRED");
		out.println("\tstoppTerm CDATA #REQUIRED");
		out.println("\txrg CDATA #REQUIRED");
		out.println("\tmanuell (BOOLEAN) #IMPLIED");
		out.println(">");
		// XRayData ist eine JNA-Klasse, daher modifiziere ich sie nicht
		out.println("<!ELEMENT XRayData (XRayPosition?)>");
		out.println("<!ATTLIST XRayData ");
		out.println("\tmode (INT) #REQUIRED");
		out.println("\tvoltage (DOUBLE) #REQUIRED");
		out.println("\tcurrent (DOUBLE) #REQUIRED");
		out.println("\tbeta (DOUBLE) #REQUIRED");
		out.println("\tbeta1 (DOUBLE) #REQUIRED");
		out.println("\tbeta2 (DOUBLE) #REQUIRED");
		out.println("\tdeltaBeta (DOUBLE) #REQUIRED");
		out.println("\tgateTime (DOUBLE) #REQUIRED");
		out.println("\trate (INT) #REQUIRED");
		out.println("\tcounts (INT) #REQUIRED");
		out.println("\tremainingTime (INT) #REQUIRED");
		out.println(">");
		// TXRayPosition ist eine JNA-Klasse, daher modifiziere ich sie nicht
		out.println("<!ELEMENT XRayPosition EMPTY>");
		out.println("<!ATTLIST XRayPosition bT (DOUBLE) #REQUIRED bS (DOUBLE) #REQUIRED>");
	}

	private static final Map<String, String> rename = new HashMap<String, String>() {
		/**
		 * 
		 */
		private static final long serialVersionUID = 1L;
		{
			put("Mode", "mode");
			put("voltage", "voltage");
			put("current", "current");
			put("b", "beta");
			put("b1", "beta1");
			put("b2", "beta2");
			put("db", "deltaBeta");
			put("dt", "gateTime");
			put("R", "rate");
			put("N", "counts");
			put("T", "remainingTime");
		}
	};

	public void writeToXml(TransformerHandler hd) throws SAXException {
		if (hd == null)
			throw new SAXException("[GHMXFile] Objekt ist nicht initialisiert.");
		AttributesImpl atts = new AttributesImpl();
		atts.addAttribute("", "", "modus", "INT", Integer.toString(modus));
		atts.addAttribute("", "", "einzelEin", "BOOLEAN",
				Boolean.toString(einzelEin));
		atts.addAttribute("", "", "signalEin", "BOOLEAN",
				Boolean.toString(signalEin));
		atts.addAttribute("", "", "anhaengenEin", "BOOLEAN",
				Boolean.toString(anhaengenEin));
		atts.addAttribute("", "", "fortsetzenEin", "BOOLEAN",
				Boolean.toString(fortsetzenEin));
		atts.addAttribute("", "", "ueberschreibenEin", "BOOLEAN",
				Boolean.toString(ueberschreibenEin));
		atts.addAttribute("", "", "wiederholungEin", "BOOLEAN",
				Boolean.toString(wiederholungEin));
		atts.addAttribute("", "", "intervall", "CDATA",
				Double.toString(intervall));
		atts.addAttribute("", "", "wiederholung", "CDATA",
				Long.toString(wiederholung));
		atts.addAttribute("", "", "messzeit", "CDATA",
				Double.toString(messzeit));
		atts.addAttribute("", "", "timerBox", "CDATA",
				Boolean.toString(timerBox));
		atts.addAttribute("", "", "triggerEin", "BOOLEAN",
				Boolean.toString(triggerEin));
		atts.addAttribute("", "", "triggerTerm", "CDATA", triggerTerm);
		atts.addAttribute("", "", "triggerSymbol", "CDATA", triggerSymbol);
		atts.addAttribute("", "", "triggerFlanke", "CDATA",
				triggerFlanke.toString());
		atts.addAttribute("", "", "triggerWert", "CDATA",
				Double.toString(triggerWert));
		atts.addAttribute("", "", "bedingungEin", "BOOLEAN",
				Boolean.toString(bedingungEin));
		atts.addAttribute("", "", "bedingungTerm", "CDATA", bedingungTerm);
		atts.addAttribute("", "", "stoppEin", "BOOLEAN",
				Boolean.toString(stoppEin));
		atts.addAttribute("", "", "stoppTerm", "CDATA", stoppTerm);
		atts.addAttribute("", "", "xrg", "CDATA",
				xrg == null ? "" : xrg.getSymbol());
		atts.addAttribute("", "", "manuell", "BOOLEAN",
				Boolean.toString(manuell));
		hd.startElement("", "", "Messparameter", atts);
		writeParameterToXml(hd);
		hd.endElement("", "", "Messparameter");
	}

	private void writeParameterToXml(TransformerHandler hd) throws SAXException {
		if (hd == null)
			throw new SAXException("[GHMXFile] Objekt ist nicht initialisiert.");
		AttributesImpl atts = new AttributesImpl();
		atts.addAttribute("", "", "mode", "INT",
				Integer.toString(parameter.getMode().getValue()));
		atts.addAttribute("", "", "voltage", "DOUBLE",
				Double.toString(parameter.getVoltage()));
		atts.addAttribute("", "", "current", "DOUBLE",
				Double.toString(parameter.getCurrent()));
		atts.addAttribute("", "", "beta", "DOUBLE",
				Double.toString(parameter.getBeta()));
		atts.addAttribute("", "", "beta1", "DOUBLE",
				Double.toString(parameter.getBeta1()));
		atts.addAttribute("", "", "beta2", "DOUBLE",
				Double.toString(parameter.getBeta2()));
		atts.addAttribute("", "", "deltaBeta", "DOUBLE",
				Double.toString(parameter.getDeltaBeta()));
		atts.addAttribute("", "", "gateTime", "DOUBLE",
				Double.toString(parameter.getGateTime()));
		atts.addAttribute("", "", "rate", "INT",
				Double.toString(parameter.getRate()));
		atts.addAttribute("", "", "counts", "INT",
				Double.toString(parameter.getCounts()));
		atts.addAttribute("", "", "remainingTime", "INT",
				Double.toString(parameter.getRemainingTime()));
		hd.startElement("", "", "XRayData", atts);
		writePositionToXml(hd);
		hd.endElement("", "", "XRayData");
	}

	private void writePositionToXml(TransformerHandler hd) throws SAXException {
		if (hd == null)
			throw new SAXException("[GHMXFile] Objekt ist nicht initialisiert.");
		AttributesImpl atts = new AttributesImpl();
		atts.addAttribute("", "", "bT", "DOUBLE",
				Double.toString(parameter.getPosition().getBetaTarget()));
		atts.addAttribute("", "", "bS", "DOUBLE",
				Double.toString(parameter.getPosition().getBetaSensor()));
		hd.startElement("", "", "TXRayPosition", atts);
		hd.endElement("", "", "TXRayPosition");
	}

	private static enum BEREICH {
		ANFANG, MESSPARAMETER, XRAYDATA, POSITION
	};

	private BEREICH bereich = BEREICH.ANFANG;

	public void startElement(String uri, String localName, String qName,
			Attributes attributes, Groessenliste gl) throws SAXException {
		switch (bereich) {
		case ANFANG:
			if (qName.equals("Messparameter")) {
				bereich = BEREICH.MESSPARAMETER;
				messzeit = 0.0;
				for (int index = 0; index < attributes.getLength(); index++) {
					try {
						String name = attributes.getLocalName(index);
						String inhalt = attributes.getValue(index);
						Field field = null;
						try {
							field = getClass().getDeclaredField(name);
						} catch (NoSuchFieldException ee) {
							// Das Feld messEinheit wurde entfernt, kann aber
							// noch in Dateien vorhanden sein.
							if (name.equals("messEinheit")) {
								try {
									if (messzeit == 0.0)
										messzeit = Einheit.toEinheit(inhalt)
												.getFaktor();
								} catch (IllegalArgumentException e) {
									if (DEBUG)
										System.out
												.println("[Messparameter] - startElement: messEinheit("
														+ e.getLocalizedMessage()
														+ ")");
								}
								continue;
							} else
								throw new NoSuchFieldException(
										ee.getLocalizedMessage());

						}
						if (DEBUG)
							System.out
									.println("[Messparameter] - endElement: Feld("
											+ name
											+ ") Typ("
											+ (field == null ? "" : field
													.getType())
											+ ") Inhalt("
											+ inhalt + ")");
						if (name.equals("xrg") && gl != null)
							xrg = Util.findeGroesse(gl, inhalt);
						else if (name.equals("messZeit")) {
							try {
								if (messzeit == 0.0)
									Util.setFieldValue(this, field, inhalt);
								else
									messzeit = messzeit
											* Double.valueOf(inhalt);
							} catch (IllegalArgumentException e) {
								if (DEBUG)
									System.out
											.println("[Messparameter] - startElement: messZeit("
													+ e.getLocalizedMessage()
													+ ")");
							}
						} else
							Util.setFieldValue(this, field, inhalt);
					} catch (SecurityException e) {
						if (DEBUG)
							e.printStackTrace();
						throw new SAXException(e.getMessage());
					} catch (NoSuchFieldException e) {
						if (DEBUG)
							e.printStackTrace();
						throw new SAXException(e.getMessage());
					} catch (IllegalArgumentException e) {
						if (DEBUG)
							e.printStackTrace();
						throw new SAXException(e.getMessage());
					}
				}
			} else
				throw new SAXException("[Messparameter] - Bereich(" + bereich
						+ ") Start(" + qName + ")");
			break;
		case MESSPARAMETER:
			if (qName.equals("XRayData"))
				bereich = BEREICH.XRAYDATA;
			break;
		case XRAYDATA:
			if (qName.equals("TXRayPosition"))
				bereich = BEREICH.POSITION;
			break;
		default:
			break;
		}
	}

	public void endElement(String uri, String localName, String qName,
			String inhalt) throws SAXException {
		switch (bereich) {
		default:
			throw new SAXException("[Messparameter] - Bereich(" + bereich
					+ ") Ende(" + qName + ")");
		case MESSPARAMETER:
			if (qName.equals("Messparameter")) {
				bereich = BEREICH.ANFANG;
				break;
			}
			break;
		case XRAYDATA:
			if (qName.equals("XRayData")) {
				bereich = BEREICH.MESSPARAMETER;
				break;
			}
			try {
				if (rename.containsKey(qName))
					qName = rename.get(qName);
				Field field = parameter.getClass().getField(qName);
				if (DEBUG)
					System.out.println("[Messparameter] - endElement: Feld("
							+ qName + ") Typ(" + field.getType() + ") Inhalt("
							+ inhalt + ")");
				if (Double.class.isAssignableFrom(field.getType()))
					field.setDouble(this, Double.parseDouble(inhalt));
			} catch (SecurityException e) {
				if (DEBUG)
					e.printStackTrace();
				throw new SAXException(e.getMessage());
			} catch (NoSuchFieldException e) {
				if (DEBUG)
					e.printStackTrace();
				throw new SAXException(e.getMessage());
			} catch (IllegalArgumentException e) {
				if (DEBUG)
					e.printStackTrace();
				throw new SAXException(e.getMessage());
			} catch (IllegalAccessException e) {
				if (DEBUG)
					e.printStackTrace();
				throw new SAXException(e.getMessage());
			}
			break;
		case POSITION:
			if (qName.equals("TXRayPosition")) {
				bereich = BEREICH.XRAYDATA;
				break;
			}
			try {
				Field field = parameter.getPosition().getClass()
						.getField(qName);
				if (DEBUG)
					System.out.println("[Messparameter] - endElement: Feld("
							+ qName + ") Typ(" + field.getType() + ") Inhalt("
							+ inhalt + ")");
				field.setDouble(this, Double.parseDouble(inhalt));
			} catch (SecurityException e) {
				if (DEBUG)
					e.printStackTrace();
				throw new SAXException(e.getMessage());
			} catch (NoSuchFieldException e) {
				if (DEBUG)
					e.printStackTrace();
				throw new SAXException(e.getMessage());
			} catch (IllegalArgumentException e) {
				if (DEBUG)
					e.printStackTrace();
				throw new SAXException(e.getMessage());
			} catch (IllegalAccessException e) {
				if (DEBUG)
					e.printStackTrace();
				throw new SAXException(e.getMessage());
			}
			break;

		}
	}

	public Messparameter(Groessenliste gl, AdapterListe al) {
		if (gl != null)
			gl.addObservableListListener(this);
		this.al = al;
		setModus(MODUSCASSY);
		setEinzelEin(false);
		setSignalEin(false);
		setAnhaengenEin(false);
		setFortsetzenEin(false);
		setUeberschreibenEin(true);
		setWiederholungEin(false);
		triggerEin = false;
		bedingungEin = false;
		stoppEin = false;
		setIntervall(Konstanten.MIN_ECHTZEIT_INTERVALL);
		setWiederholung(125);
		setMesszeit(intervall * wiederholung);
		setTimerBox(false);
		parameter = new XRayData();
		xrg = null;
		triggerTerm = "1";
		triggerEE = new FormelGenerator(triggerTerm).getAusdruck();
		triggerSymbol = "";
		triggerFlanke = TriggerBits.None;
		triggerWert = 0.0;
		bedingungTerm = "1";
		bedingungEE = new FormelGenerator(bedingungTerm).getAusdruck();
		stoppTerm = "0";
		stoppEE = new FormelGenerator(stoppTerm).getAusdruck();
		manuell = false;
	}

	public void reset() {
		setModus(MODUSCASSY);
		setEinzelEin(false);
		setSignalEin(false);
		setAnhaengenEin(false);
		setFortsetzenEin(false);
		setUeberschreibenEin(true);
		setWiederholungEin(false);
		setTriggerEin(false);
		setBedingungEin(false);
		setStoppEin(false);
		setIntervall(Konstanten.MIN_ECHTZEIT_INTERVALL);
		setWiederholung(125);
		setMesszeit(intervall * wiederholung);
		setTimerBox(false);
		parameter = new XRayData();
		xrg = null;
		triggerTerm = "1";
		triggerEE = new FormelGenerator(triggerTerm).getAusdruck();
		triggerSymbol = "";
		triggerFlanke = TriggerBits.None;
		triggerWert = 0.0;
		bedingungTerm = "1";
		bedingungEE = new FormelGenerator(bedingungTerm).getAusdruck();
		stoppTerm = "0";
		stoppEE = new FormelGenerator(stoppTerm).getAusdruck();
		manuell = false;
	}

	public void init(Messparameter mpneu) {
		setModus(mpneu.getModus());
		setEinzelEin(mpneu.isEinzelEin());
		setSignalEin(mpneu.isSignalEin());
		setAnhaengenEin(mpneu.isAnhaengenEin());
		setFortsetzenEin(mpneu.isFortsetzenEin());
		setUeberschreibenEin(mpneu.isUeberschreibenEin());
		setWiederholungEin(mpneu.isWiederholungEin());
		setTriggerEin(mpneu.isTriggerEin());
		setBedingungEin(mpneu.isBedingungEin());
		setStoppEin(mpneu.isStoppEin());
		setIntervall(mpneu.getIntervall());
		setWiederholung(mpneu.getWiederholung());
		setMesszeit(mpneu.getMesszeit());
		setTimerBox(mpneu.isTimerBox());
		parameter = mpneu.parameter;
		xrg = mpneu.xrg;
		setTriggerTerm(mpneu.getTriggerTerm());
		setTriggerEE(mpneu.getTriggerEE());
		setTriggerSymbol(mpneu.getTriggerSymbol());
		setTriggerFlanke(mpneu.getTriggerFlanke());
		setTriggerWert(mpneu.getTriggerWert());
		setBedingungTerm(mpneu.getBedingungTerm());
		setBedingungEE(mpneu.getBedingungEE());
		setStoppTerm(mpneu.getStoppTerm());
		setStoppEE(mpneu.getStoppEE());
		setManuell(mpneu.manuell);
	}

	public void setFromCL4(Groessenliste gl, String s)
			throws FileFormatException {
		int pos = s.lastIndexOf(" ");
		String term = s.substring(pos + 1);
		Double[] parameter = CL4File.leseDoubleListe(s.substring(0, pos), 12,
				0, false);
		if (1 == parameter[0].intValue())
			setMessparameterAutomatisch();
		else if (3 == parameter[0].intValue())
			setEinzelEin(true);
		setIntervall(parameter[1]);
		setWiederholung(parameter[2].longValue());
		if (wiederholung != 0)
			setMesszeit (wiederholung * intervall);
		setTriggerEin(1 == parameter[4].intValue());
		if (parameter[5].intValue() > -1) {
			setTriggerSymbol(gl.get(parameter[5].intValue()).getSymbol());
		}
		setTriggerFlanke(1 == parameter[6].intValue() ? TriggerBits.Rising
				: TriggerBits.None);
		setTriggerWert(parameter[7]);
		setWiederholungEin(1 == parameter[8].intValue());
		setAnhaengenEin(1 == parameter[9].intValue());
		setFortsetzenEin(false);
		setUeberschreibenEin(true);
		setSignalEin(1 == parameter[10].intValue());
		setBedingungEin(1 == parameter[11].intValue());
		setBedingungTerm(term);
		setManuell(false);
	}

	private transient PropertyChangeSupport support = new PropertyChangeSupport(
			this);

	private void readObject(ObjectInputStream in) throws IOException,
			ClassNotFoundException {
		in.defaultReadObject();
		support = new PropertyChangeSupport(this);
		parameter = new XRayData();
		FormelGenerator p = new FormelGenerator();
		Groessenliste gl = GHMessen.getApplication().getMainView().gl;
		if (gl != null)
			p.setKonstanten(gl.getSymbole());
		triggerEE = p.getAusdruck(triggerTerm);
		bedingungEE = p.getAusdruck(bedingungTerm);
		stoppEE = p.getAusdruck(stoppTerm);
	}

	private void writeObject(ObjectOutputStream out) throws IOException {
		out.defaultWriteObject();
	}

	public void addPropertyChangeListener(PropertyChangeListener listener) {
		this.support.addPropertyChangeListener(listener);
	}

	public void removePropertyChangeListener(PropertyChangeListener listener) {
		this.support.removePropertyChangeListener(listener);
	}

	public void addPropertyChangeListener(String property,
			PropertyChangeListener listener) {
		this.support.addPropertyChangeListener(property, listener);
	}

	public void removePropertyChangeListener(String property,
			PropertyChangeListener listener) {
		this.support.removePropertyChangeListener(property, listener);
	}

	private Groesse findeXrg() {
		if (xrg != null)
			return xrg;
		Groessenliste gl = GHMessen.getApplication().getMainView().gl;
		if (gl == null)
			return null;
		return gl.getXRayGroesse();
	}

	public void setModus(int modus) {
		int alt = this.modus;
		this.modus = modus;
		support.firePropertyChange("modus", alt, modus);
	}

	public int getModus() {
		return modus;
	}

	public void setManuell(boolean an) {
		boolean alt = manuell;
		manuell = an;
		support.firePropertyChange("manuell", alt, an);
	}

	public boolean isManuell() {
		return manuell;
	}

	public void setEinzelEin(boolean einzelEin) {
		boolean alt = this.einzelEin;
		this.einzelEin = einzelEin;
		if (einzelEin && !alt && modus == MODUSXRAY) {
			parameter.setBeta1(parameter.getBeta());
			parameter.setBeta2(parameter.getBeta());
			parameterToXRay();
		}
		support.firePropertyChange("einzelEin", alt, einzelEin);
	}

	public boolean isEinzelEin() {
		return einzelEin;
	}

	public void setTriggerEin(boolean triggerEin) {
		boolean alt = this.triggerEin;
		this.triggerEin = triggerEin;
		support.firePropertyChange("triggerEin", alt, triggerEin);
	}

	public boolean isTriggerEin() {
		return triggerEin;
	}

	public void setSignalEin(boolean signalEin) {
		boolean alt = this.signalEin;
		this.signalEin = signalEin;
		support.firePropertyChange("signalEin", alt, signalEin);
	}

	public boolean isSignalEin() {
		return signalEin;
	}

	public void setAnhaengenEin(boolean anhaengenEin) {
		boolean alt = this.anhaengenEin;
		this.anhaengenEin = anhaengenEin;
		support.firePropertyChange("anhaengenEin", alt, anhaengenEin);
	}

	public boolean isAnhaengenEin() {
		return anhaengenEin;
	}

	public void setFortsetzenEin(boolean fortsetzenEin) {
		boolean alt = this.fortsetzenEin;
		this.fortsetzenEin = fortsetzenEin;
		support.firePropertyChange("fortsetzenEin", alt, fortsetzenEin);
	}

	public boolean isFortsetzenEin() {
		return fortsetzenEin;
	}

	public boolean isUeberschreibenEin() {
		return ueberschreibenEin;
	}

	public void setUeberschreibenEin(boolean ueberschreibenEin) {
		boolean alt = this.ueberschreibenEin;
		this.ueberschreibenEin = ueberschreibenEin;
		support.firePropertyChange("ueberschreibenEin", alt, ueberschreibenEin);
	}

	public void setWiederholungEin(boolean wiederholungEin) {
		boolean alt = this.wiederholungEin;
		this.wiederholungEin = wiederholungEin;
		support.firePropertyChange("wiederholungEin", alt, wiederholungEin);
	}

	public boolean isWiederholungEin() {
		return wiederholungEin;
	}

	public void setIntervall(double intervall) {
		double alt = this.intervall;
		this.intervall = intervall;
		support.firePropertyChange("intervall", alt, intervall);
	}

	public double getIntervall() {
		return intervall;
	}

	public long getIntervallMS() {
		return Math.round(intervall * 1000);
	}

	public long getIntervallUS() {
		return Math.round(intervall * 1000000);
	}

	public void setWiederholung(long wiederholung) {
		long alt = this.wiederholung;
		this.wiederholung = wiederholung;
		support.firePropertyChange("wiederholung", alt, wiederholung);
	}

	public long getWiederholung() {
		return wiederholung;
	}

	public void setMesszeit(double messzeit) {
		double alt = this.messzeit;
		this.messzeit = messzeit;
		support.firePropertyChange("messzeit", alt, messzeit);
	}

	public double getMesszeit() {
		return messzeit;
	}

	public void setBedingungEin(boolean bedingungEin) {
		boolean alt = this.bedingungEin;
		this.bedingungEin = bedingungEin;
		support.firePropertyChange("bedingungEin", alt, bedingungEin);
	}

	public boolean isBedingungEin() {
		return bedingungEin;
	}

	public void setTimerBox(boolean timerBox) {
		boolean alt = this.timerBox;
		this.timerBox = timerBox;
		support.firePropertyChange("timerBox", alt, timerBox);
	}

	public boolean isTimerBox() {
		return timerBox;
	}

	public void setMessparameterAutomatisch() {
		setTimerBox(false);
		setEinzelEin(false);
		setIntervall(Konstanten.MIN_ECHTZEIT_INTERVALL);
		setWiederholung(0);
		setTriggerEin(false);
		setSignalEin(false);
		setBedingungEin(false);
	}

	public void setMessparameterTimerAutomatisch() {
		setMessparameterAutomatisch();
		setTimerBox(true);
		setMesszeit(2);
	}

	public void setMessparameterTimerManuell() {
		setTimerBox(true);
		setEinzelEin(true);
	}

	private void adjustTimerTyp(Groessenliste gl) {
		if (gl == null)
			return;
		if (gl.hatTimerBoxManuellGroesse()) {
			setMessparameterTimerManuell();
		} else if (gl.hatTimerBoxGroesse()) {
			setTimerBox(true);
		} else {
			setTimerBox(false);
		}
	}

	/*******************************************************************
	 * XRay - Methoden
	 *******************************************************************/
	public Groesse getXrg() {
		if (xrg == null)
			xrg = findeXrg();
		return xrg;
	}

	public void setXrg(Groesse g) {
		Groesse alt = this.xrg;
		this.xrg = g;
		support.firePropertyChange("xrg", alt, xrg);
	}

	protected void parameterToXRay() {
		if (xrg == null && (xrg = findeXrg()) == null)
			return;
		AdapterListe al = GHMessen.getApplication().getMainView().al;
		if(!(al.get(xrg.getNr()) instanceof IXRay))
			return;
		IXRay c = (IXRay) al.get(xrg.getNr());
		c.setData(parameter);
		c.setPosition(parameter.getPosition());
		c.setMode(mode);
	}

	protected void parameterFromXRay() {
		if (xrg == null && (xrg = findeXrg()) == null)
			return;
		AdapterListe al = GHMessen.getApplication().getMainView().al;
		if(!(al.get(xrg.getNr()) instanceof IXRay))
			return;
		IXRay c = (IXRay) al.get(xrg.getNr());
		parameter = c.getData();
		mode = parameter.getMode();
	}

	protected void modeToXRay(XRayModes mode) {
		if (xrg == null && (xrg = findeXrg()) == null)
			return;
		AdapterListe al = GHMessen.getApplication().getMainView().al;
		if(!(al.get(xrg.getNr()) instanceof IXRay))
			return;
		IXRay c = (IXRay) al.get(xrg.getNr());
		c.setMode(mode);
	}

	protected void modeFromXRay() {
		if (xrg == null && (xrg = findeXrg()) == null)
			return;
		AdapterListe al = GHMessen.getApplication().getMainView().al;
		if(!(al.get(xrg.getNr()) instanceof IXRay))
			return;
		IXRay c = (IXRay)al.get(xrg.getNr());
		mode = c.getMode();
	}

	public void setGoniometerMode(int mode) {
		setGoniometerMode(new XRayModes(mode));
	}

	public void setGoniometerMode(XRayModes mode) {
		if (this.parameter == null || this.mode == null) {
			if (DEBUG)
				System.out
						.println("[XRayMessparameter] - setGoniometerMode: Fehler parameter(null)|mode(null)");
			return;
		}
		mode.and(new XRayModes(XRayModes.ModeMask));
		if (mode.isEmpty()) {
			if (DEBUG)
				System.out
						.println("[XRayMessparameter] - setGoniometerMode falscher Modus("
								+ mode + ")");
			return;
		}
		this.mode.andNot(new XRayModes(XRayModes.ModeMask));
		this.mode.or(mode);
		modeToXRay(mode);
	}

	public XRayModes getGoniometerMode() {
		if (parameter == null)
			return new XRayModes(XRayModes.None);
		parameterFromXRay();
		return new XRayModes(mode.getValue()
				& XRayModes.ModeMask);
	}

	public void setU(double u) {
		if (parameter == null)
			return;
		parameter.setVoltage(u);
		parameterToXRay();
	}

	public double getU() {
		if (parameter == null)
			return 0.0;
		parameterFromXRay();
		return parameter.getVoltage();
	}

	public void setI(double i) {
		if (parameter == null)
			return;
		parameter.setCurrent(i);
		parameterToXRay();
	}

	public double getI() {
		if (parameter == null)
			return 0.0;
		parameterFromXRay();
		return parameter.getCurrent();
	}

	public void setB(double beta) {
		if (parameter == null)
			return;
		parameter.setBeta(beta);
		if (!mode.isAndEmpty(XRayModes.Sensor))
			parameter.getPosition().setBetaSensor(beta);
		else if (!mode.isAndEmpty(XRayModes.Target) || !mode.isAndEmpty(XRayModes.Coupled))
			parameter.getPosition().setBetaTarget(beta);
		parameterToXRay();
	}

	public double getBetaSensor() {
		if (parameter == null)
			return 0.0;
		parameterFromXRay();
		return parameter.getPosition().getBetaSensor();
	}
	
	public double getBetaTarget() {
		if (parameter == null)
			return 0.0;
		parameterFromXRay();
		return parameter.getPosition().getBetaTarget();
	}
	
	public double getB() {
		if (parameter == null)
			return 0.0;
		parameterFromXRay();
		return parameter.getBeta();
	}

	public void setBuG(double uG) {
		if (parameter == null)
			return;
		parameter.setBeta1(uG);
		parameterToXRay();
	}

	public double getBuG() {
		if (parameter == null)
			return 0.0;
		parameterFromXRay();
		return parameter.getBeta1();
	}

	public void setBoG(double oG) {
		if (parameter == null)
			return;
		parameter.setBeta2(oG);
		parameterToXRay();
	}

	public double getBoG() {
		if (parameter == null)
			return 0.0;
		parameterFromXRay();
		return parameter.getBeta2();
	}

	public void setDB(double deltaBeta) {
		if (parameter == null)
			return;
		parameter.setDeltaBeta(deltaBeta);
		parameterToXRay();
	}

	public double getDB() {
		if (parameter == null)
			return 0.0;
		parameterFromXRay();
		return parameter.getDeltaBeta();
	}

	public void setDT(double gateTime) {
		if (parameter == null)
			return;
		parameter.setGateTime(gateTime);
		parameterToXRay();
	}

	public double getDT() {
		if (parameter == null)
			return 0.0;
		parameterFromXRay();
		return parameter.getGateTime();
	}

	public void setTriggerTerm(String triggerTerm) {
		String alt = this.triggerTerm;
		this.triggerTerm = triggerTerm;
		support.firePropertyChange("triggerTerm", alt, triggerTerm);
	}

	public String getTriggerTerm() {
		return triggerTerm;
	}

	public void setTriggerEE(ExpressionEvaluator triggerEE) {
		this.triggerEE = triggerEE;
	}

	public ExpressionEvaluator getTriggerEE() {
		return triggerEE;
	}

	public void setTriggerSymbol(String triggerSymbol) {
		String alt = this.triggerSymbol;
		this.triggerSymbol = triggerSymbol;
		support.firePropertyChange("triggerSymbol", alt, triggerSymbol);
	}

	public String getTriggerSymbol() {
		return triggerSymbol;
	}

	public void setTriggerFlanke(TriggerBits triggerFlanke) {
		TriggerBits alt = this.triggerFlanke;
		this.triggerFlanke = triggerFlanke;
		support.firePropertyChange("triggerFlanke", alt, triggerFlanke);
	}

	public TriggerBits getTriggerFlanke() {
		return triggerFlanke;
	}

	public void setTriggerWert(double triggerWert) {
		double alt = this.triggerWert;
		this.triggerWert = triggerWert;
		support.firePropertyChange("triggerWert", alt, triggerWert);
	}

	public double getTriggerWert() {
		return triggerWert;
	}

	public void setBedingungTerm(String bedingungTerm) {
		String alt = this.bedingungTerm;
		this.bedingungTerm = bedingungTerm;
		support.firePropertyChange("bedingungTerm", alt, bedingungTerm);
	}

	public String getBedingungTerm() {
		return bedingungTerm;
	}

	public void setBedingungEE(ExpressionEvaluator bedingungEE) {
		this.bedingungEE = bedingungEE;
	}

	public ExpressionEvaluator getBedingungEE() {
		return bedingungEE;
	}

	public void setStoppEin(boolean ein) {
		boolean alt = stoppEin;
		stoppEin = ein;
		support.firePropertyChange("StoppEin", alt, ein);
	}

	public boolean isStoppEin() {
		return stoppEin;
	}

	public void setStoppTerm(String stoppTerm) {
		String alt = this.stoppTerm;
		this.stoppTerm = stoppTerm;
		support.firePropertyChange("stoppTerm", alt, stoppTerm);
	}

	public String getStoppTerm() {
		return stoppTerm;
	}

	public void setStoppEE(ExpressionEvaluator stoppEE) {
		this.stoppEE = stoppEE;
	}

	public ExpressionEvaluator getStoppEE() {
		return stoppEE;
	}

	/*******************************************************************
	 * ObservableList Listener
	 *******************************************************************/
	private void propertySymbolChanged(PropertyChangeEvent e) {
		String alt = (String) e.getOldValue();
		String neu = (String) e.getNewValue();
		if (triggerTerm != null && triggerTerm.contains(alt)) {
			setTriggerTerm(triggerTerm.replaceAll(alt, neu));
			triggerEE = new FormelGenerator(triggerTerm).getAusdruck();
		}
		if (bedingungTerm != null && bedingungTerm.contains(alt)) {
			setBedingungTerm(bedingungTerm.replaceAll(alt, neu));
			bedingungEE = new FormelGenerator(bedingungTerm).getAusdruck();
		}
		if (stoppTerm != null && stoppTerm.contains(alt)) {
			setStoppTerm(stoppTerm.replaceAll(alt, "1".equals(neu) ? "0" : neu));
			stoppEE = new FormelGenerator(stoppTerm).getAusdruck();
		}
		if (triggerSymbol != null && triggerSymbol.equals(alt))
			setTriggerSymbol("1".equals(neu) ? "" : neu);
	}

	private void propertyArtChanged(PropertyChangeEvent e) {
		if (al != null && e.getSource() instanceof Groesse) {
			Groesse g = (Groesse) e.getSource();
			Art neu = (Art) e.getNewValue();
			if (neu != Art.Messgroesse)
				return;
			if (!manuell) {
				IAdapter a = al.get(g.getNr());
				setIntervall(a.getIntervallStandard());
				setWiederholung(a.getWiederholungStandard());
				if (wiederholung != 0)
					setMesszeit(intervall * wiederholung);
			}
		}
	}

	@SuppressWarnings("rawtypes")
	@Override
	public void listElementPropertyChanged(ObservableList list, int index) {
		return;
	}

	@SuppressWarnings("rawtypes")
	@Override
	public void listElementReplaced(ObservableList list, int index,
			Object oldElement) {
		propertySymbolChanged(new PropertyChangeEvent(list, "symbol",
				((Groesse) oldElement).getSymbol(),
				((Groesse) list.get(index)).getSymbol()));
	}

	@SuppressWarnings("rawtypes")
	@Override
	public void listElementsAdded(ObservableList list, int index, int length) {
		return;
	}

	@SuppressWarnings("rawtypes")
	@Override
	public void listElementsRemoved(ObservableList list, int index,
			List oldElements) {
		for (Object o : oldElements)
			propertySymbolChanged(new PropertyChangeEvent(list, "symbol",
					((Groesse) o).getSymbol(), "1"));
	}

	@Override
	public void listElementPropertyChanged(
			@SuppressWarnings("rawtypes") ObservableList list, int index,
			PropertyChangeEvent e) {
		if ("symbol".equals(e.getPropertyName()))
			propertySymbolChanged(e);
		else if ("messGroesse".equals(e.getPropertyName())
				|| "art".equals(e.getPropertyName())) {
			if ("art".equals(e.getPropertyName()))
				propertyArtChanged(e);
			adjustTimerTyp(GHMessen.getApplication().getMainView().gl);
		}

	}

}
