/* 
 * GHMessen
 * --------
 * 
 * GHMessen is a measuring program to use in connection with a computer 
 * measuring interface. It's designed for school usage.
 * 
 * Copyright (C) 2010-2011  Frank Schütte - Gymnasium Himmelsthür
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms
 * of the GNU General Public License as published by the Free Software Foundation; either 
 * version 3 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with this program; 
 * if not, see <http://www.gnu.org/licenses/>.
 * 
 * Frank Schütte fschuett AT gymnasium-himmelsthuer DOT de
 * 
 */

package ghmessen.adapter;

import ghmessen.adapter.event.AdapterAddedEvent;
import ghmessen.adapter.event.AdapterChangedEvent;
import ghmessen.adapter.event.AdapterRemovedEvent;
import ghmessen.components.JImagePanel;
import ghmessen.components.JTransparentToggleButton;
import ghmessen.groesse.Art;
import ghmessen.groesse.Groesse;
import ghmessen.groesse.Groessenliste;
import ghmessen.util.HTML;
import ghmessen.util.Util;

import java.awt.Component;
import java.awt.Rectangle;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;

import javax.swing.ActionMap;
import javax.swing.ImageIcon;
import javax.swing.JLabel;
import javax.swing.border.BevelBorder;

import ld.api.CASSY;
import ld.api.CASSYs;
import ld.api.ITimer;

import org.jdesktop.observablecollections.ObservableList;

import udev.UdevDevice;

public abstract class Adapter extends JImagePanel implements IAdapter {
	/**
	 * 
	 */
	private static final long serialVersionUID = 8132222219420464032L;
	private static final boolean DEBUG = false;

	/*
	 * Die nächsten 4 Parameter sind für den Gerätemonitor gedacht. Jede
	 * implementierende Klasse muss diese Variablen neu setzen und in einem
	 * statischen Initialisierer initialisieren.
	 */
	public static final int vendorID;
	public static final HashSet<Integer> productID;
	public static final String[] subsystem;
	public static final String[] deviceGroup;
	public static final int ERSTER_SENSOR;

	/* Die Größenliste wird zur Verwaltung von aktiviere/deaktiviere benötigt. */
	protected Groessenliste liste;

	protected int nr;

	protected UdevDevice device; /* z.B. ldusb0 */
	protected JLabel schnittLabel; /* USB(1), COM(1) */

	/* Eingänge und ihre Darstellung */
	protected int eingaenge;
	protected JLabel[] inputLabel;
	protected JTransparentToggleButton[] inputButton;
	protected Rectangle[] rect;
	protected ActionMap buttonActions;
	protected Eingang[] sb;
	protected boolean[] aktiv;

	// Standard-Messparameter des Adapters
	protected double intervallStandard;
	protected double intervallMinimum;
	protected int wiederholungStandard;

	protected SymbolPropertyHandler symbolPH;
	protected ArtPropertyHandler artPH;

	public Adapter(ImageIcon img, int nr,
			UdevDevice device, Rectangle[] rect, Groessenliste liste)
			throws EAdapterException {
		super(img);
		this.intervallStandard = 0.1;
		this.intervallMinimum = 0.01;
		this.wiederholungStandard = 125;
		eingaenge = rect == null ? 0 : rect.length;
		this.nr = nr;
		this.device = device;
		this.rect = rect;
		this.sb = null;
		this.liste = liste;
		if (liste != null)
			this.liste.addObservableListListener(this);
		this.symbolPH = new SymbolPropertyHandler();
		this.artPH = new ArtPropertyHandler();
		setBorder(new BevelBorder(BevelBorder.RAISED, null, null, null, null));
		initEingaenge();
		initComponents();
		showAllLabels(false);
	}

	public Class<?> getKlasse() {
		return this.getClass();
	}
	
	public abstract void close();

	public int getVendorID() {
		return vendorID;
	}

	public HashSet<Integer> getProductID() {
		return productID;
	}

	public String[] getSubsystem() {
		return subsystem;
	}

	public String[] getDeviceGroup() {
		return deviceGroup;
	}

	public double getIntervallStandard() {
		if (eingaenge == 0 || sb == null)
			return this.intervallStandard;
		else {
			double intervall = this.intervallStandard;
			for (int i = 0; i < eingaenge; i++)
				if (sb[i] != null && aktiv[i])
					intervall = sb[i].getIntervallStandard();
			return intervall;
		}
	}

	public int getWiederholungStandard() {
		if (eingaenge == 0 || sb == null)
			return this.wiederholungStandard;
		else {
			int wiederholung = this.wiederholungStandard;
			for (int i = 0; i < eingaenge; i++)
				if (sb[i] != null && aktiv[i])
					wiederholung = sb[i].getWiederholungStandard();
			return wiederholung;
		}
	}

	public double getIntervallMinimum() {
		if (eingaenge == 0 || sb == null)
			return this.intervallMinimum;
		else {
			double intervall = this.intervallMinimum;
			for (int i = 0; i < eingaenge; i++)
				if (sb[i] != null && aktiv[i])
					intervall = Math.min(intervall, sb[i].getIntervallMinimum());
			return intervall;
		}
	}

	protected void initEingaenge() throws EAdapterException {
		if (eingaenge == 0) {
			inputLabel = null;
			inputButton = null;
			aktiv = null;
		} else {
			inputLabel = new JLabel[eingaenge];
			inputButton = new JTransparentToggleButton[eingaenge];
			buttonActions = null;
			aktiv = new boolean[eingaenge];
			Arrays.fill(aktiv, false);
			if (rect != null && rect.length != eingaenge) {
				rect = new Rectangle[eingaenge];
				Arrays.fill(rect, new Rectangle());
			}
		}
	}

	public IEingang getEingang(int ab) {
		if (ab >= eingaenge)
			return null;
		return sb[ab];
	}

	public IEingang getEingang(Groesse g) {
		return getEingang(g.getAB());
	}
	
	public abstract String getEingangString(int aB);
	public abstract int getEingangAnzahl();
	public abstract boolean istSensorEingang(int ab);
	public Component getComponent(){ return this; }

	protected void initButtonActions() {
		buttonActions = org.jdesktop.application.Application
				.getInstance(ghmessen.GHMessen.class).getContext()
				.getActionMap(this.getClass(), this);
	}

	protected void initInputLabels() {
		for (int i = 0; i < eingaenge; i++)
			inputLabel[i] = new JLabel("<html>ungültig</html>");
	}

	protected abstract void initButtons();

	public abstract void initGroessen();

	public String getDevice() {
		return device.getName();
	}

	public abstract String getName();

	public Groessenliste getGroessenliste(){
		return liste;
	}
	
	public int getNr() {
		return nr;
	}

	public void setNr(int i) {
		nr = i;
		if (schnittLabel != null)
			schnittLabel.setText(device.getPort() + "(" + nr + ")");
	}

	/*********
	 * spezielle CASSY-Methoden
	 */
	public void wechsleStellgroesse() { }

	public int getCassysNr() {
		return -1;
	}

	public void setCassysNr(int cassysNr) {	}

	public boolean istITimer() {
		return false;
	}
	
	public ITimer getITimer() {
		return null;
	}
	
	public void reset() {
		for (Groesse g : liste) {
			if (g.getNr() == getNr())
				deaktiviereGroesse(g);
		}
		for (int i = 0; i < eingaenge; i++) {
			inputButton[i].setAuswahl(0);
			inputButton[i].setSelected(false);
			aktiv[i] = false;
		}
		adjustImage();
	}

	protected abstract String erzeugeSymbol(int auswahl, int input, int kanal);

	protected void initComponents() {

		if (schnittLabel == null)
			schnittLabel = new JLabel(device.getPort() + "(" + nr + ")");

		initButtonActions();
		initInputLabels();
		initButtons();

	}

	public String getSchnittstelle() {
		return device.getPort();
	}

	/***************************************************************
	 * Methoden für den einfachen Zugriff auf die Cassys-Variablen
	 ***************************************************************/
	public CASSY getCassy() {
		return null;
	}

	public CASSY getCassy(int index) {
		return null;
	}

	public CASSYs getCassys() {
		return null;
	}

	public IEingang getSensorBox(int AB) {
		return null;
	}

	public boolean istOszilloskopVorhanden() {
		return false;
	}

	public void deaktiviereGroesse(Groesse g) {
		g.setArt(Art.Parameter);
		g.removePropertyChangeListener("symbol", symbolPH);
		g.removePropertyChangeListener("art", artPH);
	}

	public void aktiviereGroesse(Groesse g) {
		g.setArt(Art.Messgroesse);
		// Listener wird in listElementPropertyChanged automatisch hinzugefügt
	}

	public abstract void setzeStandardWerte(Groesse g, int AB, int ef, boolean neu);

	protected abstract void inputClicked(int aB, JTransparentToggleButton b);

	protected void deaktiviereInput(int abxy) {
		for (Groesse g : liste) {
			if (g.getNr() == nr && g.getAB() == abxy) {
				deaktiviereGroesse(g);
				setInputText(g.getAB(), g.getEF(), "");
			}
		}
		inputButton[abxy].setAuswahl(0);
		sb[abxy].setAuswahl(0);
		aktiv[abxy] = false;
	}

	protected Groesse aktiviereInput(int auswahl, int abxy, int ef,
			boolean unbedingt) {
		String name = erzeugeSymbol(auswahl, abxy, ef);
		Groesse g = Util.findeGroesse(liste, name, false);
		sb[abxy].setAuswahl(auswahl); // überflüssig
		if (g == null && !unbedingt)
			return null;
		else if (g == null) {
			g = new Groesse(name);
			setzeStandardWerte(g, abxy, ef, true);
			//FIXME: Standardwert x-Achse muss vor dem Hinzufügen gesetzt werden!?!
			liste.add(g);
		} else {
			setzeStandardWerte(g, abxy, ef, false);
		}
		aktiviereGroesse(g);
		aktiv[abxy] = true;
		setInputText(abxy, ef, name);
		return g;
	}

	protected void showAllLabels(boolean show) {
		if (eingaenge <= 0)
			return;
		for (JLabel lbl : inputLabel)
			lbl.setVisible(show);
	}

	protected void showLabel(int abxy, boolean show) {
		if (abxy >= eingaenge)
			return;
		inputLabel[abxy].setVisible(show);
	}

	protected void adjustImage() {
		resetImage();
	}

	protected boolean sind2Symbole(int abxy, int ef) {
		if (abxy >= eingaenge)
			return false;
		return inputLabel[abxy].getText().contains(" ");
	}

	protected void setInputText(int abxy, int ef, String text) {
		if (abxy >= eingaenge)
			return;
		if (ef == 0 && !sind2Symbole(abxy, ef))
			setInputText(abxy, text);
		String[] symbole = HTML.htmlZuSymbole(inputLabel[abxy].getText());
		if (symbole.length <= ef)
			symbole = Arrays.copyOf(symbole, ef + 1);
		for (int i = 0; i < symbole.length; i++)
			if (i == ef)
				symbole[i] = text;
			else if (symbole[i] == null)
				symbole[i] = "";
			else if (HTML.enthaeltHTML(symbole[i]))
				symbole[i] = HTML.htmlZuSymbol(symbole[i]);
		String label = "";
		for (int i = 0; i < symbole.length; i++)
			if (!symbole[i].equals("")) {
				if (!label.equals("")) {
					label += " <br>";
				}
				label += HTML.symbolZuHTML(symbole[i]);
			}
		label = "<html>" + label + "</html>";
		inputLabel[abxy].setText(label);
	}

	protected void setInputText(int abxy, String text) {
		if (abxy >= eingaenge)
			return;
		if (!HTML.enthaeltHTML(text))
			text = HTML.symbolZuHTML(text, true);
		inputLabel[abxy].setText(text);
	}

	/*
	 * PropertyChangeListeners
	 */
	protected class SymbolPropertyHandler implements PropertyChangeListener {
		@Override
		public void propertyChange(PropertyChangeEvent e) {
			handleSymbolEvent(e);
		}
	}

	protected class ArtPropertyHandler implements PropertyChangeListener {
		@Override
		public void propertyChange(PropertyChangeEvent e) {
			handleArtEvent(e);
		}
	}

	public void handleSymbolEvent(PropertyChangeEvent evt) {
		Groesse g = (Groesse) evt.getSource();
		inputButton[g.getAB()].setAuswahl(sb[g.getAB()].getAuswahl());
		setInputText(g.getAB(), g.getEF(), (String) evt.getNewValue());
		adjustImage();
	}

	public void handleArtEvent(PropertyChangeEvent evt) {
		Groesse g = (Groesse) evt.getSource();
		if (g.getEF() == 1 && evt.getNewValue() != Art.Messgroesse) {
			g.removePropertyChangeListener("symbol", symbolPH);
			g.removePropertyChangeListener("art", artPH);
			setInputText(g.getAB(), g.getEF(), "");
		}
	}

	protected void updateInputWith(Groesse g) {
		sb[g.getAB()].setAuswahl(g.getMessGroesse());
		inputButton[g.getAB()].setAuswahl(sb[g.getAB()].getAuswahl());
		inputButton[g.getAB()].setSelected(true);
		setInputText(g.getAB(), g.getEF(), g.getSymbol());
		aktiv[g.getAB()] = true;
		g.addPropertyChangeListener("symbol", symbolPH);
		g.addPropertyChangeListener("art", artPH);
	}

	public boolean isProductID(int id) {
		return getProductID() == null ? false : getProductID().contains(id);
	}

	/*
	 * AdapterEventListener
	 */
	@Override
	public void adapterAdded(AdapterAddedEvent evt) {
		if (evt.getKlasse() != null && evt.getNummer() == nr)
			initGroessen();
	}

	@Override
	public void adapterRemoved(AdapterRemovedEvent evt) {
		if (evt.getName().equals(device.getName()))
			for (int index = 0; index < eingaenge; index++)
				deaktiviereInput(index);
	}

	@Override
	public void adapterChanged(AdapterChangedEvent evt) {
		return;
	}

	/*
	 * ObservableObjectListListener
	 */
	@Override
	public void listElementPropertyChanged(
			@SuppressWarnings("rawtypes") ObservableList list, int index) {
		return;
	}

	@Override
	public void listElementPropertyChanged(
			@SuppressWarnings("rawtypes") ObservableList list, int index,
			PropertyChangeEvent e) {
		if (!"art".equals(e.getPropertyName()))
			return;
		if (!(list instanceof Groessenliste)
				|| !e.getNewValue().equals(Art.Messgroesse))
			return;
		Groesse g = ((Groessenliste) list).get(index);
		if (g.getNr() != nr)
			return;
		updateInputWith(g);
		if (DEBUG)
			System.out
					.println("[AbstractCassy] - listElementPropertyChanged: added to symbol("
							+ g.getSymbol() + ")");
		adjustImage();
	}

	@Override
	public void listElementReplaced(
			@SuppressWarnings("rawtypes") ObservableList list, int index,
			Object oldElement) {
		return;
	}

	@Override
	public void listElementsAdded(
			@SuppressWarnings("rawtypes") ObservableList list, int index,
			int length) {
		if (!(list instanceof Groessenliste))
			return;
		// nach dem Einlesen einer Groessenliste müssen die Interfaceobjekte
		// aktualisiert werden.
		Groesse g;
		boolean geaendert = false;
		for (int i = index; i < index + length; i++) {
			g = ((Groessenliste) list).get(i);
			if (g.getArt() != Art.Messgroesse || g.getNr() != nr)
				continue;
			updateInputWith(g);
			geaendert = true;
		}
		if (geaendert)
			adjustImage();
	}

	@Override
	public void listElementsRemoved(
			@SuppressWarnings("rawtypes") ObservableList list, int index,
			@SuppressWarnings("rawtypes") List oldElements) {
		return;
	}

	static {
		vendorID = 0;
		productID = new HashSet<Integer>();
		subsystem = null;
		deviceGroup = null;
		ERSTER_SENSOR = 1;
	}
}
