/* 
 * GHMessen
 * --------
 * 
 * GHMessen is a measuring program to use in connection with a computer 
 * measuring interface. It's designed for school usage.
 * 
 * Copyright (C) 2010-2011  Frank Schütte - Gymnasium Himmelsthür
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms
 * of the GNU General Public License as published by the Free Software Foundation; either 
 * version 3 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with this program; 
 * if not, see <http://www.gnu.org/licenses/>.
 * 
 * Frank Schütte fschuett AT gymnasium-himmelsthuer DOT de
 * 
 */

package ghmessen.adapter;

import ghmessen.GHMessen;
import ghmessen.Messparameter;
import ghmessen.adapter.arduino.AdapterArduino;
import ghmessen.adapter.cassy.AdapterCASSY;
import ghmessen.adapter.cassy.AdapterXRay;
import ghmessen.adapter.cassy.EingangCASSY;
import ghmessen.adapter.efm.AdapterEFM;
import ghmessen.adapter.event.AdapterAddedEvent;
import ghmessen.adapter.event.AdapterChangedEvent;
import ghmessen.adapter.event.AdapterEventListener;
import ghmessen.adapter.event.AdapterRemovedEvent;
import ghmessen.adapter.goIO.AdapterGoIO;
import ghmessen.adapter.goIO.EingangGoIO;
import ghmessen.groesse.Groessenliste;
import ghmessen.messtasks.CassyVector;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.lang.reflect.InvocationTargetException;
import java.util.AbstractMap;
import java.util.AbstractSet;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.ScheduledThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

import ld.api.CASSY;
import ld.api.CASSYs;
import ld.api.ICASSY;
import ld.api.IOscilloscope;
import ld.api.ios.CASSYIO;
import ld.api.ios.CASSYIOs;
import ld.api.ios.ICASSYIO;
import ld.api.ios.ScanMode;
import ld.components.events.CollectionChangedEvent;
import ld.components.events.CollectionChangedListener;

import org.jdesktop.observablecollections.ObservableMap;
import org.jdesktop.observablecollections.ObservableMapListener;

import udev.UdevDevice;

public class AdapterListe extends AbstractMap<Integer, IAdapter> implements PropertyChangeListener,
		ObservableMap<Integer, IAdapter>, Serializable, AdapterEventListener, CollectionChangedListener, AutoCloseable {

	/**
	 * 
	 */
	private static final long serialVersionUID = 4089687817965394424L;
	private static final boolean DEBUG = false;
	private final boolean supportsElementPropertyChanged;
	private Map<Integer, IAdapter> list;
	private transient List<ObservableMapListener> listeners;
	private transient Set<Map.Entry<Integer, IAdapter>> entrySet;
	private boolean cassyVorhanden;
	private transient CassyVector cassysv;
	private boolean goIOVorhanden;
	private AdapterMonitor monitor;
	private transient Groessenliste gl;
	private transient ScheduledThreadPoolExecutor zeitgeber;

	private static final int SENSOR_INTERVALL = 2;
	@SuppressWarnings("rawtypes")
	private transient ScheduledFuture sensorDetektor;
	private boolean pauseDetektor;

	private void writeObject(ObjectOutputStream out) throws IOException {
		out.defaultWriteObject();
	}

	private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {
		in.defaultReadObject();
		restoreListenings();
		gl = GHMessen.getApplication().getMainView().gl;
		zeitgeber = GHMessen.getApplication().getMainView().zeitgeber;
		cassysv = new CassyVector(zeitgeber);
		cassysv.getCASSYIOs().addCollectionChangedListener(this);
		monitor.addAdapterEventListener(this);
		listeners = new CopyOnWriteArrayList<ObservableMapListener>();
	}

	private void restoreListenings() {
		for (IAdapter a : list.values()) {
			a.addPropertyChangeListener(this);
		}
	}

	AdapterListe(Map<Integer, IAdapter> list, boolean supportsElementPropertyChanged, Groessenliste gl,
			ScheduledThreadPoolExecutor zeitgeber) {
		this.list = list;
		for (IAdapter a : list.values()) {
			a.addPropertyChangeListener(this);
		}
		this.gl = gl;
		this.zeitgeber = zeitgeber;
		pauseDetektor = false;
		cassyVorhanden = false;
		cassysv = new CassyVector(zeitgeber);
		cassysv.getCASSYIOs().addCollectionChangedListener(this);
		goIOVorhanden = false;
		listeners = new CopyOnWriteArrayList<ObservableMapListener>();
		this.supportsElementPropertyChanged = supportsElementPropertyChanged;
		starteMonitor();
	}

	public AdapterListe(Groessenliste gl, ScheduledThreadPoolExecutor zeitgeber) {
		this(new HashMap<Integer, IAdapter>(), true, gl, zeitgeber);
	}

	public void close() {
		stoppeSensorDetektor();
		stoppeMonitor();
	}

	private void starteMonitor() {
		monitor = new AdapterMonitor();
		monitor.addAdapterEventListener(this);
		monitor.registerAdapterClass(AdapterGoIO.class);
		monitor.registerAdapterClass(AdapterCASSY.class);
		monitor.registerAdapterClass(AdapterXRay.class);
		monitor.registerAdapterClass(AdapterEFM.class);
		monitor.registerAdapterClass(AdapterArduino.class);
		monitor.start();

	}

	private void stoppeMonitor() {
		if (monitor == null)
			return;
		monitor.stop();
		monitor.removeAdapterEventListener(this);
		monitor = null;
	}

	public AdapterMonitor getMonitor() {
		return monitor;
	}

	public boolean contains(Class<? extends IAdapter> aclass, int nr) {
		return list.containsKey(nr) && aclass.isAssignableFrom(list.get(nr).getClass());
	}

	public boolean contains(Class<? extends IAdapter> aclass) {
		for (IAdapter a : list.values())
			if (aclass.isAssignableFrom(a.getClass()))
				return true;
		return false;
	}

	public boolean containsCassy() {
		for (IAdapter a : list.values())
			if (a instanceof ICASSY)
				return true;
		return false;
	}

	public boolean istOszilloskopVorhanden() {
		boolean ret = false;
		for (IAdapter a : list.values())
			ret = ret || a instanceof IOscilloscope;
		return ret;
	}

	public boolean istSeriellerPortVorhanden() {
		boolean ret = false;
		for (IAdapter a : list.values())
			ret = ret || a.getDevice().matches(".*tty.*");
		return ret;
	}

	public boolean istSchnelleOszilloskopMessung(Messparameter mp) {
		boolean seriell = istSeriellerPortVorhanden();
		if (seriell && mp.getIntervallMS() < 100 || mp.getIntervallMS() < 10)
			return true;
		return false;
	}

	public IAdapter get(Class<? extends IAdapter> aclass, int nr) {
		if (list.containsKey(nr) && aclass.isAssignableFrom(list.get(nr).getClass()))
			return list.get(nr);
		return null;
	}

	public IAdapter getCassy(int nr) {
		IAdapter a = list.get(nr);
		if (a == null)
			return null;
		if (AdapterCASSY.class.isAssignableFrom(a.getClass()))
			return a;
		return null;
	}

	public CASSYs getCassys(UdevDevice device) {
		for (IAdapter a : list.values())
			if (a instanceof ICASSY && a.getDevice().equals(device.getName()))
				return ((ICASSY) a).getCASSYs();
		return null;
	}

	public CassyVector getCassys() {
		return cassysv;
	}

	public int getErsteFreieNummer() {
		int frei = 1;
		Set<Integer> belegt = keySet();
		while (belegt.contains(frei))
			frei++;
		return frei;
	}

	/*************************************
	 * Map - Interface
	 *************************************/

	public int getSize() {
		return list.size();
	}

	public int indexOf(IAdapter a) {
		for (int i = 0; i < list.size(); i++)
			if (list.get(i) == a)
				return i;
		return -1;
	}

	public IAdapter get(int index) {
		return list.get(index);
	}

	public Set<Map.Entry<Integer, IAdapter>> entrySet() {
		Set<Map.Entry<Integer, IAdapter>> es = entrySet;
		return es != null ? es : (entrySet = new EntrySet());
	}

	public int size() {
		return list.size();
	}

	public boolean containsKey(Integer key) {
		return list.containsKey(key);
	}

	public boolean containsValue(IAdapter value) {
		return list.containsValue(value);
	}

	public IAdapter get(Integer key) {
		return list.get(key);
	}

	public boolean isEmpty() {
		return list.isEmpty();
	}

	public IAdapter put(Integer key, IAdapter value) {
		IAdapter lastValue;
		if (containsKey(key)) {
			lastValue = list.put(key, value);
			for (ObservableMapListener listener : listeners) {
				listener.mapKeyValueChanged(this, key, lastValue);
			}
		} else {
			lastValue = list.put(key, value);
			for (ObservableMapListener listener : listeners) {
				listener.mapKeyAdded(this, key);
			}
		}
		return lastValue;
	}

	public IAdapter remove(Integer key) {
		if (containsKey(key)) {
			IAdapter value = list.remove(key);
			for (ObservableMapListener listener : listeners) {
				listener.mapKeyRemoved(this, key, value);
			}
			return value;
		}
		return null;
	}

	public void remove(IAdapter value) {
		if (containsValue(value)) {
			for (Integer index : list.keySet())
				if (list.get(index) == value) {
					remove(index);
					break;
				}
		}
	}

	public void removeAll(Collection<?> c) {
		if (c.isEmpty() || !(c.iterator().next() instanceof IAdapter))
			return;
	}

	public void putAll(Map<? extends Integer, ? extends IAdapter> m) {
		for (Integer key : m.keySet()) {
			put(key, m.get(key));
		}
	}

	public void clear() {
		Iterator<Integer> iterator = keySet().iterator();
		while (iterator.hasNext()) {
			iterator.next();
			iterator.remove();
		}
	}

	private class EntryIterator implements Iterator<Map.Entry<Integer, IAdapter>> {
		private Iterator<Map.Entry<Integer, IAdapter>> realIterator;
		private Map.Entry<Integer, IAdapter> last;

		EntryIterator() {
			realIterator = list.entrySet().iterator();
		}

		public boolean hasNext() {
			return realIterator.hasNext();
		}

		public Map.Entry<Integer, IAdapter> next() {
			last = realIterator.next();
			return last;
		}

		public void remove() {
			if (last == null) {
				throw new IllegalStateException();
			}
			Object toRemove = last.getKey();
			last = null;
			AdapterListe.this.remove(toRemove);
		}
	}

	private class EntrySet extends AbstractSet<Map.Entry<Integer, IAdapter>> {
		public Iterator<Map.Entry<Integer, IAdapter>> iterator() {
			return new EntryIterator();
		}

		@SuppressWarnings("unchecked")
		public boolean contains(Object o) {
			if (!(o instanceof Map.Entry)) {
				return false;
			}
			Map.Entry<Integer, IAdapter> e = (Map.Entry<Integer, IAdapter>) o;
			return containsKey(e.getKey());
		}

		@SuppressWarnings({ "unchecked", "unlikely-arg-type" })
		public boolean remove(Object o) {
			if (o instanceof Map.Entry) {
				Integer key = ((Map.Entry<Integer, IAdapter>) o).getKey();
				if (containsKey(key)) {
					remove(key);
					return true;
				}
			}
			return false;
		}

		public int size() {
			return AdapterListe.this.size();
		}

		public void clear() {
			AdapterListe.this.clear();
		}
	}

	public <T> T[] toArray(T[] a) {
		return list.values().toArray(a);
	}

	public Object[] toArray() {
		return list.values().toArray();
	}

	public void addObservableMapListener(ObservableMapListener listener) {
		listeners.add(listener);
	}

	public void removeObservableMapListener(ObservableMapListener listener) {
		listeners.remove(listener);
	}

	public boolean supportsElementPropertyChanged() {
		return supportsElementPropertyChanged;
	}

	@Override
	public void propertyChange(PropertyChangeEvent evt) {
		IAdapter a = (IAdapter) evt.getSource();
		int index = indexOf(a);
		if (index > -1)
			for (ObservableMapListener listener : listeners) {
				listener.mapKeyValueChanged(this, index, a);
			}
	}

	/*
	 * IAdapter Event Listener
	 */
	@Override
	public void adapterAdded(AdapterAddedEvent evt) {
		try {
			if (evt.getKlasse() == null)
				return;
			if (DEBUG)
				System.out.println("[AdapterListe] - adapterAdded: evt(" + evt + ")");
			int nr = evt.getNummer();
			IAdapter a = null;
			if (AdapterCASSY.class.isAssignableFrom(evt.getKlasse())) {
				// TODO Zumindest für seriell ist getNode(/dev/...) erforderlich, weil
				// getName(tty...) von ldapi nicht erkannt wird.
				int altesMaximum = cassysv.size() - 1;
				String devname = evt.getDevice().getName();
				if (devname.startsWith("tty")) {
					devname = evt.getDevice().getNode();
					cassysv.setCommPort(devname);
				}
				try {
					cassysv.scan();
				} catch (IOException e) {
					e.printStackTrace();
				}
				a = null;
				CASSY c = null;
				try {
					a = get(nr);
					if (evt.istKaskadiert())
						c = cassysv.get(evt.getKaskadeNr());
					else for(CASSY tmp : cassysv) {
						if(tmp.getDevicePath().endsWith(devname)) {
							c = tmp;
							if(c.getIsCascadeable())
								c = c.getFirstCASSY();
							}
					}
				} catch (Exception e) {
					if (DEBUG)
						System.out.println("[AdapterListe] - adapterAdded: " + e.getLocalizedMessage());
				}
				try {
					if (c != null && a == null) {
						nr = getErsteFreieNummer();
						evt.setNummer(nr);
						a = AdapterCASSY.create(c, nr, evt.getDevice(), gl);
						// Das funktioniert nicht, da die Proxy-Klasse als Klasse zurückgegeben wird.
						/*
						 * evt.setKlasse(a.getClass()); // Serielle Cassys wurden // erst nachträglich
						 * // erkannt.
						 * 
						 */
						monitor.addAdapterEventListener(a);
						put(nr, a);
						if (!evt.istKaskadiert() && c.getIsCascadeable()) {
							for (int cassysNr = 1; cassysNr < 8; cassysNr++)
								if (cassysv.size() >= evt.getNummer() + cassysNr)
									monitor.fireAdapterAdded(new AdapterAddedEvent(evt.getSource(), evt.getDevice(),
											evt.getKlasse(), evt.getNummer() + cassysNr, true, cassysNr));
						}
					}
				} catch (EAdapterException e) {
					if (DEBUG)
						e.printStackTrace();
				}
				cassyVorhanden = containsCassy();
			} else if (AdapterXRay.class.isAssignableFrom(evt.getKlasse())) {
				nr = getErsteFreieNummer();
				evt.setNummer(nr);
				a = evt.getKlasse().getConstructor(int.class, UdevDevice.class, Groessenliste.class).newInstance(nr,
						evt.getDevice(), gl);
				monitor.addAdapterEventListener(a);
				put(nr, a);
			} else {
				nr = getErsteFreieNummer();
				evt.setNummer(nr);
				a = evt.getKlasse().getConstructor(int.class, UdevDevice.class, Groessenliste.class).newInstance(nr,
						evt.getDevice(), gl);
				monitor.addAdapterEventListener(a);
				put(nr, a);
				goIOVorhanden = contains(AdapterGoIO.class);
			}
			if (goIOVorhanden || cassyVorhanden)
				starteSensorDetektor();
		} catch (

		InstantiationException e) {
			if (DEBUG)
				e.printStackTrace();
		} catch (IllegalAccessException e) {
			if (DEBUG)
				e.printStackTrace();
		} catch (IllegalArgumentException e) {
			if (DEBUG)
				e.printStackTrace();
		} catch (SecurityException e) {
			if (DEBUG)
				e.printStackTrace();
		} catch (InvocationTargetException e) {
			if (DEBUG)
				e.printStackTrace();
		} catch (NoSuchMethodException e) {
			if (DEBUG)
				e.printStackTrace();
		}
		if (DEBUG)
			System.out.println("[AdapterListe] - adapterAdded - complete:\n" + "\tliste(" + list + ")\n"
					+ "\tsensorDetektor(" + sensorDetektor + ") pause: " + pauseDetektor + ")\n" + "\tcassyVorhanden("
					+ cassyVorhanden + ") goioVorhanden(" + goIOVorhanden + ")");
	}

	@Override
	public void adapterRemoved(AdapterRemovedEvent evt) {
		suspendSensorDetektor();
		if (DEBUG)
			System.out.println("[AdapterListe] - adapterRemoved: evt(" + evt + ")");
		// alle anderen IAdapter überprüfen
		// Bei kaskadierbaren CASSY interfaces können mehrere Adapter 
		// gleichzeitig verschwinden.
		Set<Integer> removals = new HashSet<>();
		for (IAdapter a : list.values())
			if (a.getDevice().equals(evt.getName())) {
				if (a instanceof ICASSY) {
					try {
						cassysv.scan();
					} catch (IOException e) {
						e.printStackTrace();
					}
				}
				a.close();
				removals.add(a.getNr());
				monitor.removeAdapterEventListener(a);
			}
		// Adapter einzeln entfernen, damit Listeners benachrichtig werden.
		for (int key : removals)
			remove(key);
		resumeSensorDetektor();
		cassyVorhanden = containsCassy();
		goIOVorhanden = contains(AdapterGoIO.class);
		if (!goIOVorhanden)
			stoppeSensorDetektor();
		if (DEBUG)
			System.out.println("[AdapterListe] - adapterRemoved - complete:\n" + "\tliste(" + list + ")\n"
					+ "\tsensorDetektor(" + sensorDetektor + ") pause: " + pauseDetektor + ")\n" + "\tcassyVorhanden("
					+ cassyVorhanden + ") goioVorhanden(" + goIOVorhanden + ")");
	}

	@Override
	public void adapterChanged(AdapterChangedEvent evt) {
		if (DEBUG)
			System.out.println("[AdapterListe] - adapterChanged: evt(" + evt + ")");
		for (IAdapter a : list.values())
			a.adapterChanged(evt);
		if (DEBUG)
			System.out.println("[AdapterListe] - adapterChanged - complete:\n" + "\tliste(" + list + ")\n"
					+ "\tsensorDetektor(" + sensorDetektor + ") pause: " + pauseDetektor + ")\n" + "\tcassyVorhanden("
					+ cassyVorhanden + ") goioVorhanden(" + goIOVorhanden + ")");
	}

	/*
	 * Sensor-Detektor
	 */

	private class SensorDetektor implements Runnable {

		@Override
		public void run() {
			if (monitor == null || pauseDetektor)
				return;
			try {
				if (goIOVorhanden || cassyVorhanden) {
					for (IAdapter a : list.values()) {
						if (a instanceof ICASSY) {
							cassysv.getCASSYIOs().scan(ScanMode.ConvertAll);
						} else if (a instanceof AdapterGoIO) {
							EingangGoIO goIO = (EingangGoIO) a.getEingang(0);
							int alt = goIO.getSensor() == null ? 0 : goIO.getSensor().getSensorNumber();
							int neu = goIO.getCurrentSensorNumber();
							if (alt != neu) {
								monitor.fireAdapterChanged(
										new AdapterChangedEvent(this, a, a.getNr(), 0, goIO, goIO, alt, neu));
							}
						}
					}
				}
			} catch (EAdapterException e) {
				if (DEBUG)
					System.err.println(
							"[SensorDetektor] - run: Die Suche nach einem neuen Sensor hat eine EAdapterException ausgelöst.");
			} catch (Exception e) {
				e.printStackTrace();
			}

		}
	}

	private void starteSensorDetektor() {
		if (zeitgeber != null && sensorDetektor == null) {
			sensorDetektor = zeitgeber.scheduleWithFixedDelay(new SensorDetektor(), SENSOR_INTERVALL, SENSOR_INTERVALL,
					TimeUnit.SECONDS);
		}
	}

	private void stoppeSensorDetektor() {
		if (zeitgeber != null && sensorDetektor != null) {
			sensorDetektor.cancel(true);
		}
		sensorDetektor = null;
	}

	public void suspendSensorDetektor() {
		pauseDetektor = true;
	}

	public void resumeSensorDetektor() {
		pauseDetektor = false;
	}

	@Override
	public void collectionChanged(CollectionChangedEvent e) {
		// Ein CASSYIO hat sich geändert.
		for (IAdapter a : list.values())
			if (a instanceof ICASSY) {
				ICASSY c = ((AdapterCASSY.CASSYComponent) a.getComponent()).getCASSY();
				if (cassysv.getCASSYIOs() == e.getSource()) {
					CASSYIOs cassyios = (CASSYIOs) e.getSource();
					for (int ab = 0; ab < a.getEingangAnzahl(); ab++) {
						if (!a.istSensorEingang(ab))
							continue;
						IEingang alt = a.getEingang(ab);
						int altNr = ((ICASSYIO) alt).getSensorBox();
						CASSYIO neuIO = cassyios.getItem(c, ab);
						int neuNr = neuIO.getSensorBox();
						if (altNr != neuNr) {
							IEingang neu = EingangCASSY.create(neuIO, a);
							monitor.fireAdapterChanged(
									new AdapterChangedEvent(this, a, a.getNr(), ab, alt, neu, altNr, neuNr));
						}
					}
				}
			}
	}

}
