/* 
 * GHMessen
 * --------
 * 
 * GHMessen is a measuring program to use in connection with a computer 
 * measuring interface. It's designed for school usage.
 * 
 * Copyright (C) 2010-2011  Frank Schütte - Gymnasium Himmelsthür
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms
 * of the GNU General Public License as published by the Free Software Foundation; either 
 * version 3 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with this program; 
 * if not, see <http://www.gnu.org/licenses/>.
 * 
 * Frank Schütte fschuett AT gymnasium-himmelsthuer DOT de
 * 
 */

package ghmessen.adapter;

import ghmessen.GHMessen;
import ghmessen.adapter.event.AdapterAddedEvent;
import ghmessen.adapter.event.AdapterChangedEvent;
import ghmessen.adapter.event.AdapterEventListener;
import ghmessen.adapter.event.AdapterInterface;
import ghmessen.adapter.event.AdapterRemovedEvent;
import ghmessen.adapter.goIO.AdapterGoIO;
import ghmessen.util.CursorTk;
import gnu.io.CommPortIdentifier;
import gnu.io.NoSuchPortException;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;
import java.util.Vector;

import javax.swing.event.EventListenerList;

import udev.DeviceAddedEvent;
import udev.DeviceEventListener;
import udev.DeviceRemovedEvent;
import udev.Udev;
import udev.UdevDevice;

public class AdapterMonitor implements Serializable, DeviceEventListener, AutoCloseable,
		AdapterInterface {
	private static final long serialVersionUID = -4090458856267098899L;
	private static final boolean DEBUG = false;

	// Private declarations
	private transient EventListenerList listener;

	// USB Devices
	protected HashMap<Integer, HashSet<Integer>> productID;
	protected HashMap<String, Class<? extends IAdapter>> adapterKlasse;
	protected HashMap<String, UdevDevice> device;
	// Serial Devices
	protected HashMap<String, CommPortIdentifier> commPort;

	protected HashMap<String, Integer> nr;
	protected Vector<String> subsystem;
	protected Vector<String> devGroup;
	protected boolean active;
	protected transient Udev udev;
	protected transient Thread udevMonitor;

	@Override
	public void addAdapterEventListener(AdapterEventListener listener) {
		this.listener.add(AdapterEventListener.class, listener);
	}

	@Override
	public void removeAdapterEventListener(AdapterEventListener listener) {
		this.listener.remove(AdapterEventListener.class, listener);
	}

	public void fireAdapterAdded(AdapterAddedEvent e) {
		if (DEBUG)
			System.out.println("[AdapterMonitor] - fireAdapterAdded: e(" + e
					+ ")");
		Object[] listeners = listener.getListenerList();
		for (int i = listeners.length - 2; i >= 0; i -= 2) {
			if (listeners[i] == AdapterEventListener.class) {
				((AdapterEventListener) listeners[i + 1]).adapterAdded(e);
			}
		}
	}

	private void fireAdapterRemoved(AdapterRemovedEvent e) {
		if (DEBUG)
			System.out.println("[AdapterMonitor] - fireAdapterRemoved: e(" + e
					+ ")");
		Object[] listeners = listener.getListenerList();
		for (int i = listeners.length - 2; i >= 0; i -= 2) {
			if (listeners[i] == AdapterEventListener.class) {
				((AdapterEventListener) listeners[i + 1]).adapterRemoved(e);
			}
		}
	}

	public void fireAdapterChanged(AdapterChangedEvent e) {
		if (DEBUG)
			System.out.println("[AdapterMonitor] - fireAdapterChanged: e(" + e
					+ ")");
		Object[] listeners = listener.getListenerList();
		for (int i = listeners.length - 2; i >= 0; i -= 2) {
			if (listeners[i] == AdapterEventListener.class) {
				((AdapterEventListener) listeners[i + 1]).adapterChanged(e);
			}
		}
	}

	/*
	 * Serializable Interface
	 */
	private void writeObject(ObjectOutputStream out) throws IOException {
		out.defaultWriteObject();
	}

	private void readObject(ObjectInputStream in) throws IOException,
			ClassNotFoundException {
		in.defaultReadObject();
		listener = new EventListenerList();
		if (active)
			start();
		removeObsoleteDevices();
		findExistingPorts();
	}

	/*
	 * Udev Device Event Listener
	 */
	public synchronized void deviceAdded(DeviceAddedEvent e) {
		UdevDevice d = e.getDevice();
		int vendor = d.getVendorID();
		int product = d.getProductID();
		if (d.getName().startsWith("ttyUSB")) {
			try {
				device.put(d.getName(), d);
				CommPortIdentifier port = CommPortIdentifier
						.getPortIdentifier(d.getNode());
				commPort.put(d.getName(), port);
				fireAdapterAdded(new AdapterAddedEvent(this, d, null));
			} catch (NoSuchPortException ee) {
				if (DEBUG)
					System.out.println("[AdapterMonitor] - deviceAdded: "
							+ ee.getLocalizedMessage());
			}
			return;
		} else if (!productID.containsKey(vendor)
				|| !productID.get(vendor).contains(product))
			return;
		device.put(d.getName(), d);
		if (DEBUG)
			System.out.println("[AdapterMonitor] AdapterAdded [" + e
					+ "] vendor(" + vendor + ") product(" + product + ")");
		try {
			CursorTk.startWaitCursor(GHMessen.getApplication().getMainView().einstellungenBox);
		} catch (IllegalStateException ex) {
			if (DEBUG)
				System.out
						.println("[AdapterMonitor] - GHMessen ist nicht initialisiert.");
		}
		fireAdapterAdded(new AdapterAddedEvent(this, d,
				adapterKlasse.get(berechneSchluessel(d.getName(), vendor, product))));
		try {
			CursorTk.stopWaitCursor(GHMessen.getApplication().getMainView().einstellungenBox);
		} catch (IllegalStateException ex) {
			if (DEBUG)
				System.out
						.println("[AdapterMonitor] - GHMessen ist nicht initialisiert.");
		}
	}

	public synchronized void deviceAdapterChosen(String deviceName, int vendor,
			int product) {
		UdevDevice d = null;
		if (device.containsKey(deviceName)) {
			d = device.get(deviceName);
			d.setVendorID(Integer.toString(vendor));
			d.setProductID(Integer.toString(product));
		} else if (commPort.containsKey(deviceName)) {
			d = new UdevDevice(deviceName, Integer.toString(vendor), "",
					Integer.toString(product), "", "/dev/" + deviceName, "", "", "");
		}
		if (d == null) {
			if (DEBUG)
				System.out
						.println("[AdapterMonitor] - deviceAdapterChosen: deviceName("
								+ deviceName
								+ ") vendor("
								+ vendor
								+ ") product("
								+ product
								+ ") wurde nicht gefunden.");
			return;
		}
		if (vendor == -1)
			fireAdapterRemoved(new AdapterRemovedEvent(this, d, false));
		else
			fireAdapterAdded(new AdapterAddedEvent(this, d,
					adapterKlasse.get(berechneSchluessel(d.getName(), vendor, product))));
	}

	public synchronized void deviceRemoved(DeviceRemovedEvent e) {
		UdevDevice d = null;
		if (device.containsKey(e.getName()))
			d = device.remove(e.getName());
		else if (commPort.containsKey(e.getName())) {
			commPort.remove(e.getName());
			d = new UdevDevice(e.getName(), "", "", "", "", "/dev/"
					+ e.getName(), "", "", "");
		}
		if (d == null) {
			if (DEBUG)
				System.out.println("[AdapterMonitor] - deviceRemoved: device("
						+ e.getName() + ") wurde nicht gefunden.");
			return;
		}
		if (DEBUG)
			System.out.println("[AdapterMonitor] AdapterRemoved [" + e
					+ "] vendor(" + d.getVendorID() + ") product("
					+ d.getProductID() + ")");
		fireAdapterRemoved(new AdapterRemovedEvent(this, d));
	}

	/*
	 * Konstruktoren
	 */
	public AdapterMonitor() {
		super();
		active = false;
		device = new HashMap<String, UdevDevice>();
		nr = new HashMap<String, Integer>();
		adapterKlasse = new HashMap<String, Class<? extends IAdapter>>();
		listener = new EventListenerList();
		productID = new HashMap<Integer, HashSet<Integer>>();
		this.subsystem = new Vector<String>();
		subsystem.add("tty");
		this.devGroup = new Vector<String>();
		devGroup.add("ttyUSB*");
		commPort = new HashMap<String, CommPortIdentifier>();
		// findExistingPorts
		@SuppressWarnings("rawtypes")
		Enumeration portListe = CommPortIdentifier.getPortIdentifiers();
		CommPortIdentifier port;
		while (portListe.hasMoreElements()) {
			port = (CommPortIdentifier) portListe.nextElement();
			if (port.getPortType() == CommPortIdentifier.PORT_SERIAL) {
				commPort.put(
						port.getName().substring(
								port.getName().lastIndexOf("/") + 1), port);
			}
		}

	}

	public void close() {
		stop();
	}
	
	public synchronized void start() {
		if(DEBUG)
			System.out.println("[AdapterMonitor] - start: udevMonitor("+udevMonitor+") udev("+udev+")");
		if (udevMonitor == null && udev == null) {
			startUdevMonitor();
			findExistingDevices();
			active = true;
		}
	}

	public synchronized void stop() {
		if(DEBUG)
			System.out.println("[AdapterMonitor] - start: udevMonitor("+udevMonitor+") udev("+udev+")");
		if (udevMonitor != null && udev != null) {
			stopUdevMonitor();
			active = false;
		}
	}

	private synchronized void stopUdevMonitor() {
		if(DEBUG)
			System.out.println("[AdapterMonitor] - stopUdevMonitor: udev("+udev+") udevMonitor("+udevMonitor+")");
		udevMonitor.interrupt();
		udevMonitor = null;
		udev.removeDeviceEventListener(this);
		udev = null;
	}

	private synchronized void startUdevMonitor() {
		udev = new Udev(subsystem.toArray(new String[] {}),
				devGroup.toArray(new String[] {}));
		if(DEBUG)
			System.out.println("[AdapterMonitor] - startUdevMonitor: udev("+udev+")");
		udevMonitor = new Thread(udev);
		udevMonitor.start();
		if (udev != null)
			udev.addDeviceEventListener(this);
	}

	protected synchronized void findExistingPorts() {
		@SuppressWarnings("rawtypes")
		Enumeration portListe = CommPortIdentifier.getPortIdentifiers();
		CommPortIdentifier port;
		while (portListe.hasMoreElements()) {
			port = (CommPortIdentifier) portListe.nextElement();
			if (commPort.containsKey(port.getName()))
				continue;
			if (port.getPortType() == CommPortIdentifier.PORT_SERIAL) {
				commPort.put(
						port.getName().substring(
								port.getName().lastIndexOf("/") + 1), port);
			}
		}
	}

	protected synchronized void findExistingDevices() {
		Vector<UdevDevice> devices = udev.getDevices();
		if(DEBUG)
			System.out.println("[AdapterMonitor] - findExistingDevices devices("+devices+")");
		if (devices == null)
			return;
		for (UdevDevice d : devices) {
			if (d.getNode() == null || "".equals(d.getNode()))
				continue;
			if (device.containsKey(d.getName()))
				continue;
			if (d.getName().startsWith("ttyUSB")) {
				try {
					device.put(d.getName(), d);
					CommPortIdentifier port = CommPortIdentifier
							.getPortIdentifier(d.getNode());
					commPort.put(d.getName(), port);
					fireAdapterAdded(new AdapterAddedEvent(this, d, null));
				} catch (NoSuchPortException ee) {
					if (DEBUG)
						System.out
								.println("[AdapterMonitor] - findExistingDevices: "
										+ ee.getLocalizedMessage());
				}
			} else {
				int vendor = d.getVendorID();
				int product = d.getProductID();
				if (!productID.containsKey(vendor)
						|| !productID.get(vendor).contains(product))
					continue;
				Class <? extends IAdapter> klasse = adapterKlasse.get(berechneSchluessel(d.getName(), vendor, product));
				device.put(d.getName(), d);
				fireAdapterAdded(new AdapterAddedEvent(this, d, klasse));
			}
		}
	}

	protected synchronized void removeObsoleteDevices() {
		Vector<UdevDevice> devices = udev.getDevices();
		Vector<String> obsolete = new Vector<String>();
		for (UdevDevice d : device.values()) {
			if (!devices.contains(d))
				obsolete.add(d.getName());
		}
		for (String s : obsolete) {
			UdevDevice old = device.remove(s);
			fireAdapterRemoved(new AdapterRemovedEvent(this, old));
		}
	}

	public Set<String> getPortListe() {
		return commPort.keySet();
	}

	/*
	 * Register / Unregister Devices
	 */
	public synchronized void registerAdapterClass(String[] subsystem,
			String[] deviceGroup, int vendor, HashSet<Integer> products,
			Class<? extends IAdapter> klasse) {
		boolean saveActive = active;
		if (active)
			stop();
		if (subsystem != null && subsystem.length > 0)
			for (int i = 0; i < subsystem.length; i++)
				addDevGroup(subsystem[i], deviceGroup[i]);
		if (productID.containsKey(vendor))
			productID.get(vendor).addAll(products);
		else
			productID.put(vendor, products);
		for (int i = 0; i < deviceGroup.length; i++)
			for (Integer p : products)
				adapterKlasse.put(berechneSchluessel(deviceGroup[i], vendor, p), klasse);
		if (saveActive)
			start();
	}

	public synchronized void registerAdapterClass(
			Class<? extends IAdapter> klasse) {
		boolean saveActive = active;
		if (active)
			stop();
		try {
			String[] subsystem = (String[]) klasse.getField("subsystem").get(
					null);
			String[] deviceGroup = (String[]) klasse.getField("deviceGroup")
					.get(null);
			if(subsystem != null && subsystem.length > 0)
				for (int i=0;i<subsystem.length;i++)
				addDevGroup(subsystem[i], deviceGroup[i]);
			int vendor = klasse.getField("vendorID").getInt(null);
			Object o = klasse.getField("productID").get(null);
			if (!(o instanceof HashSet<?>))
				throw new IllegalArgumentException(
						"productID is not an instance of HashSet");
			@SuppressWarnings("unchecked")
			HashSet<Integer> products = (HashSet<Integer>) o;
			if (productID.containsKey(vendor))
				productID.get(vendor).addAll(products);
			else
				productID.put(vendor, products);
			for (int i = 0; i < deviceGroup.length; i++)
				for (Integer p : products)
					adapterKlasse.put(berechneSchluessel(deviceGroup[i], vendor, p), klasse);
		} catch (IllegalAccessException e) {
			if (DEBUG)
				e.printStackTrace();
		} catch (IllegalArgumentException e) {
			if (DEBUG)
				e.printStackTrace();
		} catch (SecurityException e) {
			if (DEBUG)
				e.printStackTrace();
		} catch (NoSuchFieldException e) {
			if (DEBUG)
				e.printStackTrace();
		}
		if (saveActive)
			start();
	}

	private String berechneSchluessel(String devname, int vendor, int product) {
		return devname.replaceAll("\\d", "*") + "|" + (vendor << 16) + product;
	}

	public synchronized void unregisterAdapterClass(int vendorID,
			HashSet<Integer> productID) {

	}

	public synchronized void unregisterAdapterClass(int vendorID) {

	}

	public synchronized void unregisterAdapterClass(
			Class<? extends IAdapter> klasse) {

	}

	public synchronized void addDevGroup(String subsystem, String devGroup) {
		this.subsystem.add(subsystem);
		this.devGroup.add(devGroup);
	}

	public synchronized void removeDevGroup(String subsystem, String devGroup) {
		int si = -1,di = -1;
		while((si = this.subsystem.indexOf(subsystem,si)) > -1 && (di = this.devGroup.indexOf(devGroup,di)) > -1){
			si = di = Math.max(si,di);
			if(this.subsystem.get(si).equals(subsystem) && this.devGroup.get(di).equals(devGroup)){
				this.subsystem.remove(si);
				this.devGroup.remove(di);
				return;
			}
		}
	}

	public static void main(String[] args) {
		try (AdapterMonitor monitor = new AdapterMonitor()) {
			monitor.registerAdapterClass(AdapterGoIO.class);
			monitor.start();
			while (true)
				try {
					Thread.sleep(1000);
				} catch (InterruptedException e) {
					if (DEBUG)
						e.printStackTrace();
				}
			}
	}
}
