/* 
 * GHMessen
 * --------
 * 
 * GHMessen is a measuring program to use in connection with a computer 
 * measuring interface. It's designed for school usage.
 * 
 * Copyright (C) 2010-2011  Frank Schütte - Gymnasium Himmelsthür
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms
 * of the GNU General Public License as published by the Free Software Foundation; either 
 * version 3 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with this program; 
 * if not, see <http://www.gnu.org/licenses/>.
 * 
 * Frank Schütte fschuett AT gymnasium-himmelsthuer DOT de
 * 
 */

package ghmessen.adapter.cassy;

import java.lang.reflect.InvocationHandler;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.util.ArrayList;
import java.util.concurrent.ScheduledThreadPoolExecutor;

import javax.swing.JFrame;

import ghmessen.adapter.EAdapterException;
import ghmessen.adapter.IEingang;
import ghmessen.adapter.IMesskanal;
import ghmessen.adapter.MesskanalException;
import ghmessen.groesse.Groesse;
import ghmessen.groesse.Messbereich;
import ghmessen.instrumente.Messgeraet;
import ghmessen.instrumente.MessgeraetCASSY;
import ghmessen.instrumente.MessgeraetRS;
import ghmessen.instrumente.MessgeraetStellgroesse;
import ghmessen.instrumente.SensorPanel.CASSYBoxPanel;
import ghmessen.instrumente.SensorPanel.SensorBoxPanel;
import ghmessen.instrumente.SensorPanel.StellgroessePanel;
import ghmessen.instrumente.SensorPanel.TimerBoxPanel;
import ghmessen.util.ProxyUtil;

import static ghmessen.groesse.Messgroesse.Pegel;
import ld.api.events.SelectedChangedEvent;
import ld.api.events.SelectedChangedListener;
import ld.api.ios.CASSYRange;
import ld.api.ios.CASSYTimerQuantity;
import ld.api.ios.CASSYTimerTime;
import ld.api.ios.IBmwTime;
import ld.api.ios.ICASSYQuantity;
import ld.api.ios.ICASSYTimerQuantity;
import ld.api.ios.IMirror;
import ld.api.ios.ITare;
import ld.api.ios.IToggle;
import ld.api.ios.ITriggered;
import ld.api.ios.IUltrasonic;
import ld.api.ios.IWaveform;
import ld.api.ios.MeasurementMode;

public class MesskanalCASSY implements IMesskanal, InvocationHandler,
		SelectedChangedListener {

	private static final boolean DEBUG = false;

	private final EingangCASSY eingang;
	private final ICASSYQuantity q;
	private final Groesse g;
	private Boolean istOszilloskopZeitKanal = null;
	private Boolean istLangsamerZeitKanal = null;

	public MesskanalCASSY(EingangCASSY eingang, Groesse g, ICASSYQuantity q) {
		this.q = q;
		this.eingang = eingang;
		this.g = g;
		q.addSelectedChangedListener(this);
		selectedChanged(new SelectedChangedEvent(q));
	}

	public static IMesskanal create(EingangCASSY eingang, Groesse g,
			ICASSYQuantity q) {
		ArrayList<Class<?>> interfaces = new ArrayList<Class<?>>();
		Class<?> dieseKlasse = q.getClass();
		ProxyUtil.sammleIF(interfaces, dieseKlasse);
		ProxyUtil.addIF(interfaces, IMesskanal.class);
		ProxyUtil.addIF(interfaces, SelectedChangedListener.class);
		IMesskanal result = (IMesskanal) Proxy.newProxyInstance(q.getClass()
				.getClassLoader(), interfaces.toArray(new Class<?>[] {}),
				new MesskanalCASSY(eingang, g, q));
		return result;
	}

	@Override
	public Object invoke(Object proxy, Method method, Object[] args)
			throws Throwable {
		Object result = null;
		try {
			if (method.getDeclaringClass() == IMesskanal.class
					|| method.getDeclaringClass() == SelectedChangedListener.class)
				result = method.invoke(this, args);
			else
				result = method.invoke(q, args);
		} catch (InvocationTargetException e) {
			throw e.getTargetException();
		} catch (Exception e) {
			throw new RuntimeException("unexpected invocation exception "
					+ e.getMessage());
		}
		return result;
	}

	@Override
	public String[] getMessbereicheStringArray() {
		return Messbereich.getBereicheStringList(getMessbereiche());
	}

	@Override
	public int getMessbereichIndex() {
		return q.getRangeIndex();
	}

	@Override
	public void setMessbereich(int index) {
		q.setRangeIndex(index);
	}

	@Override
	public void setMessbereich(Messbereich messbereich) {
		Messbereich[] bereiche = getMessbereiche();
		for (int index = 0; index < bereiche.length; index++)
			if (bereiche[index].equals(messbereich)) {
				setMessbereich(index);
				return;
			}
	}

	@Override
	public void selectedChanged(SelectedChangedEvent e) {
		if(DEBUG)
		System.out.println("[" + this.getClass().toString()
				+ "] SelectedChangedEvent " + q);
	}

	@Override
	public double einlesen() throws MesskanalException {
		if (q.getOscilloscopeMode() != MeasurementMode.FastOscilloscope 
				|| q instanceof IToggle) {
//	FIXME		if (!q.getValueValid())
//				throw new MesskanalException("Der Wert ist nicht gültig.");
//			if (DEBUG)
//				System.out.println("[MesskanalCASSY] - einlesen: Einzelwert("+q.getValue()+")");
			return q.getValue();
		} else if(q.getValues() > 0){
			if (DEBUG)
				System.out.println("[MesskanalCASSY] - einlesen: Oszilloskop("+q.getValues()+" gültige Werte)");
			return q.getValues(q.getValues()-1);
		} else
			throw new MesskanalException("Es sind noch keine Messwerte vorhanden.");
	}

	@Override
	public double einlesenZeit() throws MesskanalException {
		if(!istOszilloskopZeitKanal())
			throw new MesskanalException("Der Messkanal ist kein Zeitkanal.");
		if(q instanceof IBmwTime)
			return ((IBmwTime)q).getBmwTime();
		else {
			CASSYTimerTime t = (CASSYTimerTime)q.getDefaultXQuantity();
			if(t == null)
				throw new MesskanalException("Der Zeitkanal hat keine Zeit-Quantity.");
			if(q.getOscilloscopeMode() != MeasurementMode.FastOscilloscope)
				return t.getValue();
			else if(t.getValues() > 0)
				return t.getValues(t.getValues()-1);
			else
				throw new MesskanalException("Es sind noch keine Zeitwerte vorhanden.");
		}
	}
	
	@Override
	public int fuellePuffer(int bisher, int neu) throws InterruptedException,
			EAdapterException {
		if (!istOszilloskopKanal())
			throw new EAdapterException("Der Messkanal hat kein Oszilloskop.");
		int n = q.getValues();
		n = Math.max(n-bisher,0);
		n = Math.min(n, neu); // ev. sind mehr Werte vorhanden als erwartet!
		if (DEBUG)
			System.out.println("[werteEinlesen] Werte im Oszilloskop-Array("
					+ n + ")");
		return n;
	}

	public double lesePufferWert(int index) throws InterruptedException,
			EAdapterException {
		if (!istOszilloskopKanal())
			throw new EAdapterException("Der Messkanal hat kein Oszilloskop.");
		try {
			return q.getValues(index);
		} catch (IndexOutOfBoundsException e) {
			throw new EAdapterException("Der Indexwert " + index
					+ " ist ungültig.");
		}
	}

	public double lesePufferZeit(int index) throws InterruptedException,
			EAdapterException {
		if(!istOszilloskopZeitKanal() || !istOszilloskopKanal())
			throw new EAdapterException("Der Messkanal ist kein Oszilloskop- oder kein Zeitkanal.");
		try {
			if(q instanceof IBmwTime)
				throw new EAdapterException("TODO: BMWQuantity Zeit");
			else {
				CASSYTimerTime t = (CASSYTimerTime)q.getDefaultXQuantity();
				return t.getValues(index);
			}
		} catch (IndexOutOfBoundsException e) {
			throw new EAdapterException("Der Zeitwert " + index + " ist ungültig.");
		}
	}

	@Override
	public boolean istKanalOffen() {
		return true;
	}

	@Override
	public boolean istOszilloskopZeitKanal() {
		if(istOszilloskopZeitKanal != null)
			return istOszilloskopZeitKanal;
		return istOszilloskopZeitKanal = q instanceof IBmwTime
				|| (q instanceof ICASSYTimerQuantity
				    && q.getDefaultXQuantity() instanceof CASSYTimerTime);
	}

	@Override
	public boolean istLangsamerZeitKanal() {
		if(istLangsamerZeitKanal != null)
			return istLangsamerZeitKanal;
		return istLangsamerZeitKanal = q instanceof IBmwTime
				|| (q instanceof ICASSYTimerQuantity
						&& q.getDefaultXQuantity() instanceof CASSYTimerTime
						&& g.getMessGroesse() != Pegel);
	}
	
	@Override
	public boolean istTriggerKanal() {
		return this instanceof ITriggered;
	}
	
	@Override
	public double leseLetztenWert() throws MesskanalException {
		return einlesen();
	}

	@Override
	public Groesse getGroesse() {
		return g;
	}

	@Override
	public boolean istGueltig() {
		return q.getValueValid();
	}

	@Override
	public boolean istUeberschritten() {
		return q.getCASSY().getOverRange();
	}

	@Override
	public boolean istUnterschritten() {
		return q.getCASSY().getUnderRange();
	}

	@Override
	public boolean istOszilloskopKanal() {
		return q.getIsFastOscilloscopeChannel();
	}

	@Override
	public IEingang getEingang() {
		return eingang;
	}

	@Override
	public boolean hatKorrektur() {
		return q instanceof ITare;
	}

	@Override
	public Messbereich[] getMessbereiche() {
		ArrayList<Messbereich> result = new ArrayList<Messbereich>();
		for (CASSYRange r : q.getRanges())
			result.add(new Messbereich(r));
		return result.toArray(new Messbereich[] {});

	}

	@Override
	public Messbereich getMessbereich() {
		return new Messbereich(q.getRange());
	}

	@Override
	public void korrigiereNullpunkt() {
		if (q instanceof ITare)
			((ITare) q).doTare();
	}

	@Override
	public boolean hatVorzeichenWechsel() {
		return q instanceof IMirror;
	}

	@Override
	public void wechseleVorzeichen() {
		if (q instanceof IMirror)
			((IMirror) q).doMirror();
	}

	@Override
	public void start() throws MesskanalException {
	}

	@Override
	public void stop() throws MesskanalException {
	}

	@Override
	public void setAbstand(double abstand) throws MesskanalException {
	}

	@Override
	public double getAbstand() throws MesskanalException {
		return 0.1;
	}

	@Override
	public Messgeraet createMessgeraet(JFrame parent, IMesskanal messkanal,
			ScheduledThreadPoolExecutor timer) {
		if (q instanceof IToggle)
			return new MessgeraetRS(parent, messkanal,
					eingang.getAdapter(), g, timer);
		else if (q instanceof IWaveform)
			return new MessgeraetStellgroesse(parent, messkanal,
					eingang.getAdapter(), g, timer);
		else
			return new MessgeraetCASSY(parent, messkanal, eingang.getAdapter(),
					g, timer);
	}

	@Override
	public SensorBoxPanel getPanel(IMesskanal messkanal) {
		if (q instanceof IWaveform)
			return new StellgroessePanel(g, eingang.getAdapter(), messkanal);
		else if (q instanceof IUltrasonic)
			return new CASSYBoxPanel(g, eingang.getAdapter(), messkanal);
		else if (q instanceof CASSYTimerQuantity)
			return new TimerBoxPanel(g, eingang.getAdapter(), messkanal);
		else
			return new CASSYBoxPanel(g, eingang.getAdapter(), messkanal);
	}

}
