/* 
 * GHMessen
 * --------
 * 
 * GHMessen is a measuring program to use in connection with a computer 
 * measuring interface. It's designed for school usage.
 * 
 * Copyright (C) 2010-2012  Frank Schütte - Gymnasium Himmelsthür
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms
 * of the GNU General Public License as published by the Free Software Foundation; either 
 * version 3 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with this program; 
 * if not, see <http://www.gnu.org/licenses/>.
 * 
 * Frank Schütte fschuett AT gymnasium-himmelsthuer DOT de
 * 
 */

package ghmessen.adapter.goIO;

import ghmessen.adapter.EAdapterException;
import ghmessen.adapter.IEingang;
import ghmessen.adapter.IMesskanal;
import ghmessen.adapter.Messkanal;
import ghmessen.adapter.MesskanalException;
import ghmessen.groesse.Groesse;
import ghmessen.groesse.Messbereich;
import ghmessen.instrumente.Messgeraet;
import ghmessen.instrumente.MessgeraetGoIO;
import ghmessen.instrumente.SensorPanel.GoIOBoxPanel;
import ghmessen.instrumente.SensorPanel.SensorBoxPanel;

import java.util.Vector;
import java.util.concurrent.ScheduledThreadPoolExecutor;

import javax.swing.JFrame;

import libGoIO.GoIOCalibration;
import libGoIO.GoIOSensor;
import libGoIO.LibGoIO;

import org.bridj.Pointer;

public class MesskanalGoIO extends Messkanal {

	private static final boolean DEBUG = false;

	private static final int PUFFERSIZE = 20 * 6;
	private LibGoIO lib;
	private GoIOSensor sensor;
	private Vector<Messbereich> messbereich;
	private int mbIndex;
	private GoIOCalibration cal;
	private boolean istKalibriert;
	private Pointer<?> handle;
	private double[] puffer;
	private int laenge;

	public MesskanalGoIO(IEingang eingang, Groesse g, GoIOCalibration cal,
			LibGoIO lib, Pointer<?> handle, boolean istKalibriert)
			throws MesskanalException {
		this(eingang, g, cal, lib, handle, istKalibriert, true);
	}

	public MesskanalGoIO(IEingang eingang, Groesse g, GoIOCalibration cal,
			LibGoIO lib, Pointer<?> handle, boolean istKalibriert,
			boolean autostart) throws MesskanalException {
		super(eingang, g);
		this.cal = cal;
		this.lib = lib;
		this.sensor = ((EingangGoIO) eingang).getSensor();
		this.messbereich = sensor.getMessbereicheForSensorNumber();
		this.mbIndex = 0;
		this.handle = handle;
		this.istKalibriert = istKalibriert;
		this.puffer = new double[PUFFERSIZE];
		this.laenge = 0;
		if (autostart)
			start();
	}

	public void setCalibration(GoIOCalibration cal) {
		this.cal = cal;
	}

	@Override
	public double einlesen() throws MesskanalException {
		if (istKalibriert) {
			d = lib.getLatestMeasurement(handle);
		} else {
			d = lib.getLatestVoltageMeasurement(handle);
			d = cal.calibrateData(d);
		}
		d = g.getFaktor() * (d + g.getOffset() + korrekturWert);
		if (DEBUG)
			System.out.println("[MesskanalGoIO] - einlesen: d(" + d + ")");
		gueltig = true;
		return d;
	}

	@Override
	public int fuellePuffer(int bisher, int neu) throws InterruptedException,
			EAdapterException {
		laenge = lib.getNumMeasurementsAvailable(handle);
		laenge = Math.min(laenge, PUFFERSIZE);
		laenge = Math.min(laenge, neu);
		try {
			if (istKalibriert)
				laenge = lib.getCalibratedMeasurements(handle, puffer, laenge);
			else
				laenge = lib.getVoltageMeasurements(handle, puffer, laenge);
		} catch (EAdapterException e) {
			throw new EAdapterException(e.getLocalizedMessage());
		}
		return laenge;
	}

	@Override
	public double lesePufferWert(int index) throws InterruptedException,
			EAdapterException {
		try {
			if (istKalibriert)
				d = puffer[index];
			else
				d = cal.calibrateData(puffer[index]);
			return g.getFaktor()
					* (puffer[index] + g.getOffset() + korrekturWert);
		} catch (ArrayIndexOutOfBoundsException e) {
			throw new EAdapterException(e.getLocalizedMessage());
		}
	}

	@Override
	public void setAbstand(double abstand) throws MesskanalException {
		if (gestartet)
			stop();
		try {
			lib.setAbstandMS(handle, abstand);
		} catch (EAdapterException e) {
			throw new MesskanalException(e.getLocalizedMessage());
		}
		start();
	}

	@Override
	public double getAbstand() throws MesskanalException {
		try {
			return lib.getAbstand(handle);
		} catch (EAdapterException e) {
			throw new MesskanalException(e.getLocalizedMessage());
		}
	}

	@Override
	public void start() throws MesskanalException {
		if (!gestartet)
			try {
				lib.startMeasurements(handle);
				gestartet = true;
			} catch (EAdapterException e) {
				throw new MesskanalException(e.getLocalizedMessage());
			}
	}

	@Override
	public void stop() throws MesskanalException {
		if (gestartet) {
			gueltig = false;
			try {
				lib.stopMeasurements(handle);
				gestartet = false;
			} catch (EAdapterException e) {
				throw new MesskanalException(e.getLocalizedMessage());
			}
		}
	}

	@Override
	public boolean istKanalOffen() {
		return gestartet;
	}

	@Override
	public Messbereich getMessbereich() {
		return messbereich.get(mbIndex);
	}

	@Override
	public String[] getMessbereicheStringArray() {
		return Messbereich.getBereicheStringList(messbereich);
	}

	@Override
	public int getMessbereichIndex() {
		return mbIndex;
	}

	@Override
	public void setMessbereich(int index) {
		if (messbereich != null && index < messbereich.size())
			mbIndex = index;
	}

	@Override
	public void setMessbereich(Messbereich mbneu) {
		if (messbereich != null)
			for (int index = 0; index < messbereich.size(); index++)
				if (messbereich.get(index).equals(mbneu)) {
					mbIndex = index;
					return;
				}
	}

	@Override
	public Messgeraet createMessgeraet(JFrame parent, IMesskanal messkanal,
			ScheduledThreadPoolExecutor timer) {
		return new MessgeraetGoIO(parent, this, (AdapterGoIO) messkanal
				.getEingang().getAdapter(), g, timer);
	}

	@Override
	public SensorBoxPanel getPanel(IMesskanal messkanal) {
		return new GoIOBoxPanel(g, (AdapterGoIO) eingang.getAdapter(), (MesskanalGoIO) messkanal);
	}
}
