/* 
 * GHMessen
 * --------
 * 
 * GHMessen is a measuring program to use in connection with a computer 
 * measuring interface. It's designed for school usage.
 * 
 * Copyright (C) 2010-2011  Frank Schütte - Gymnasium Himmelsthür
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms
 * of the GNU General Public License as published by the Free Software Foundation; either 
 * version 3 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with this program; 
 * if not, see <http://www.gnu.org/licenses/>.
 * 
 * Frank Schütte fschuett AT gymnasium-himmelsthuer DOT de
 * 
 */

package ghmessen.adapter.seriell;

/**
 * Synchronized ring buffer. 
 * Suitable to hand over data from one thread to another.
 **/
public class RingBuffer {

    /** internal buffer to hold the data **/
    protected byte buffer[];

    /** size of the buffer **/
    protected int size;

    /** current start of data area **/
    protected int start;

    /** current end of data area **/
    protected int end;


    /**
     * Construct a RingBuffer with a default buffer size of 1k.
     */
    public RingBuffer() {
         this(32000);
    }

    /**
     * Construct a RingBuffer with a certain buffer size.
     * @param size   Buffer size in bytes
     */
    public RingBuffer(int size) {
         this.size = size;
         buffer = new byte[size];
         clear();
    }

    /**
     * Clear the buffer contents. All data still in the buffer is lost.
     */
    public synchronized void clear() {
        // Just reset the pointers. The remaining data fragments, if any,
        // will be overwritten during normal operation.
        start = end = 0;
    }

    /**
     * Return used space in buffer. This is the size of the
     * data currently in the buffer.
     * <p>
     * Note: While the value is correct upon returning, it
     * is not necessarily valid when data is read from the 
     * buffer or written to the buffer. Another thread might
     * have filled the buffer or emptied it in the mean time.
     *
     * @return currently amount of data available in buffer
     */
    public synchronized int data() {
         return start <= end
                     ? end - start
                     : end - start + size;
    }

    /**
     * Return unused space in buffer. Note: While the value is
     * correct upon returning, it is not necessarily valid when
     * data is written to the buffer or read from the buffer.
     * Another thread might have filled the buffer or emptied
     * it in the mean time.
     *
     * @return currently available free space
     */
    public synchronized int free() {
         return start <= end
                     ? size + start - end
                     : start - end;
    }

    /**
     * Write as much data as possible to the buffer.
     * @param data   Data to be written
     * @return       Amount of data actually written
     */
    synchronized int write(byte data[]) {
        return write(data, 0, data.length);  
    }

    /**
     * Write as much data as possible to the buffer.
     * @param data   Array holding data to be written
     * @param off    Offset of data in array
     * @param n      Amount of data to write, starting from off.
     * @return       Amount of data actually written
     */
    synchronized int write(byte data[], int off, int n) {
        if(n <= 0) return 0;
        int remain = n;
        // @todo check if off is valid: 0= <= off < data.length; throw exception if not

        int i = Math.min(remain, (end < start ? start : buffer.length) - end);
        if(i > 0) {
             System.arraycopy(data, off, buffer, end, i);
             off    += i;
             remain -= i;
             end    += i;
        }

        i = Math.min(remain, end >= start ? start : 0);
        if(i > 0 ) {
             System.arraycopy(data, off, buffer, 0, i);
             remain -= i;
             end = i;
        }
        return n - remain;
    }


    /**
     * Read as much data as possible from the buffer.
     * @param data   Where to store the data
     * @return       Amount of data read
     */
    synchronized int read(byte data[]) {
        return read(data, 0, data.length);  
    }

    /**
     * Read as much data as possible from the buffer.
     * @param data   Where to store the read data
     * @param off    Offset of data in array
     * @param n      Amount of data to read
     * @return       Amount of data actually read
     */
    synchronized int read(byte data[], int off, int n) {
        if(n <= 0) return 0;
        int remain = n;
        // @todo check if off is valid: 0= <= off < data.length; throw exception if not

        int i = Math.min(remain, (end < start ? buffer.length : end) - start);
        if(i > 0) {
             System.arraycopy(buffer, start, data, off, i);
             off    += i;
             remain -= i;
             start  += i;
             if(start >= buffer.length) start = 0;
        }

        i = Math.min(remain, end >= start ? 0 : end);
        if(i > 0 ) {
             System.arraycopy(buffer, 0, data, off, i);
             remain -= i;
             start = i;
        }
        return n - remain;
    }
}
