/* 
 * GHMessen
 * --------
 * 
 * GHMessen is a measuring program to use in connection with a computer 
 * measuring interface. It's designed for school usage.
 * 
 * Copyright (C) 2010-2011  Frank Schütte - Gymnasium Himmelsthür
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms
 * of the GNU General Public License as published by the Free Software Foundation; either 
 * version 3 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with this program; 
 * if not, see <http://www.gnu.org/licenses/>.
 * 
 * Frank Schütte fschuett AT gymnasium-himmelsthuer DOT de
 * 
 */

package ghmessen.adapter.seriell;

import gnu.io.CommPortIdentifier;
import gnu.io.NoSuchPortException;
import gnu.io.PortInUseException;
import gnu.io.SerialPort;
import gnu.io.UnsupportedCommOperationException;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.Serializable;
import java.util.Arrays;

import javax.swing.JOptionPane;

import udev.UdevDevice;

public class SerialAdapter implements Serializable, AutoCloseable {
	private static final long serialVersionUID = -4090458856267098899L;
	private static final boolean DEBUG = false;

	public static final File INVALID_HANDLE_VALUE = null;
	public static final int INFINITE = 0xFFFFFFFF;
	public static final int WAIT_TIMEOUT = 0;
	public static final int WAIT_OBJECT_0 = 1;
	public static final int BINARY = 0x0001; // binary mode, no EOF check
	public static final int DTRCONTROLENABLE = 0x0010; // DTR flow control type
	public static final int RTSCONTROLENABLE = 0x1000; // RTS flow control

	// device declarations
	protected File hFile = INVALID_HANDLE_VALUE;
	protected UdevDevice device;
	protected SerialPort commPort = null;

	// Serial comm streams
	protected OutputStream comWrite = null;
	protected InputStream comRead = null;

	protected SerialReadWriteThread serialReadWriteThread = null;

	// Communications control parameters
	protected int version = 0;
	protected boolean fullDuplex = false, useThreads = true;
	protected int inReportSize, inReportStart;
	protected int txDDelay = 0;
	protected int resyncTimeout = 0;

	protected int flowControl = SerialPort.FLOWCONTROL_NONE;
	protected int baudrate = 38400;
	protected int parity = SerialPort.PARITY_NONE;
	protected int stopBits = SerialPort.STOPBITS_1;
	protected int dataBits = SerialPort.DATABITS_8;
	
	private int pid = -1;

	public static final int MAXREPORTSIZE = 64;
	protected byte[] inReport = new byte[MAXREPORTSIZE + 1];

	// Public declarations

	public SerialAdapter(UdevDevice device, int baudrate) {
		this(device, baudrate, SerialPort.FLOWCONTROL_NONE);
	}
	
	public SerialAdapter(UdevDevice device, int baudrate, int flowControl) {
		super();
		useThreads = true;
		this.device = device;
		this.baudrate = baudrate;
		this.flowControl = flowControl;
		if (!isOpen()) // start tty adapter
			start();
	}
// FIXME: überflüssig ?
	public void setRTS(boolean status) {
		if(commPort != null)
			commPort.setRTS(status);
	}
	
	public void setDTR(boolean status) {
		if(commPort != null)
			commPort.setDTR(status);
	}
	
	public void close() {
		stop();
	}

	public void finalize() {
		stop();
	}

	public File getHandle() {
		return hFile;
	}

	public String getDevice() {
		return device == null ? "" : device.getName();
	}

	public void setDevice(UdevDevice value) {
		if (device != value) {
			stop();
			device = value;
			start();
		}
	}

	public void setBaudrate(int value) {
		if (baudrate != value) {
			stop();
			baudrate = value;
			start();
		}
	}

	public int getConnectedPID() {
		return 0;
	}

	public void setPID(int value) {
		if (pid != value) {
			stop();
			pid = value;
			start();
		}
	}

	public int getPID() {
		return pid;
	}

	public void setFTxDDelay(int value) {
		txDDelay = value;
	}

	public boolean isUseThreads() {
		return useThreads;
	}

	public void setUseThreads(boolean value) {
		if (useThreads != value) {
			stop();
			useThreads = value;
			start();
		}
	}

	public boolean isOpen() {
		return hFile != INVALID_HANDLE_VALUE;
	}

	public boolean alreadyOpen() {
		return false;
	}

	public boolean getTimeout(long startTime, long timeout) {
		return System.currentTimeMillis() - startTime > timeout;
	}

	public void start() {
		File handle;
		if (!isOpen()) {
			handle = INVALID_HANDLE_VALUE;
			version = 0;
			inReportSize = 0;
			if (device == null)
				return;
			CommPortIdentifier portId = null;
			SerialPort portHandle = null;
			try {
				portId = CommPortIdentifier.getPortIdentifier(device
						.getNode());
				if (portId == null) {
					System.err
							.println("[SerialAdapter] - start: Could not find serial port("
									+ device.getNode() + ")");
					handle = INVALID_HANDLE_VALUE;
					return;
				}
				portHandle = (SerialPort) portId.open("GHMessen", 10000);
				portHandle.setSerialPortParams(baudrate, dataBits,
						stopBits, parity);
				portHandle.setFlowControlMode(flowControl);
			} catch (UnsatisfiedLinkError e) {
				if (DEBUG)
					System.out
							.println("[SerialAdapter] - start: Geräte("
									+ device.getNode()
									+ ") "
									+ e.getLocalizedMessage());
			} catch (NoSuchPortException e) {
				if (DEBUG)
					System.out
							.println("[SerialAdapter] - start: Port("
									+ device.getNode() + ") "
									+ e.getLocalizedMessage());
			} catch (PortInUseException e) {
				System.err
						.println("[SerialAdapter] - start: Port already in use("
								+ e.getMessage() + ")");
				hFile = INVALID_HANDLE_VALUE;
				JOptionPane
						.showMessageDialog(
								null,
								"<html><p>Die serielle Schnittstelle <b>"
										+ portId.getName()
										+ "</b> konnte nicht geöffnet werden.</p>"
										+ "<p>Hinweis: Schnittstellen abgezogener USB-Seriell-Konverter werden häufig nicht korrekt wieder frei gegeben.</p>"
										+ "<p>Beenden Sie das Programm und starten Sie es wieder neu, um das Problem zu beheben.</p></html>",
								"Serielle Schnittstelle öffnen",
								JOptionPane.ERROR_MESSAGE);
				return;
			} catch (RuntimeException e) {
				System.err
						.println("[SerialAdapter] - start: Port is not existing("
								+ e.getMessage() + ")");
				hFile = INVALID_HANDLE_VALUE;
				return;
			} catch (UnsupportedCommOperationException e) {
				if (DEBUG)
					e.printStackTrace();
				hFile = INVALID_HANDLE_VALUE;
				System.err.println("[SerialAdapter] - start: Fehler("
						+ e.getMessage() + ")");
				return;
			}
			if (useThreads) {
				serialReadWriteThread = new SerialReadWriteThread(
						portHandle, 8, 8);
				serialReadWriteThread.start();
			} else {
				try {
					comRead = portHandle.getInputStream();
					comWrite = portHandle.getOutputStream();
				} catch (IOException e) {
					hFile = INVALID_HANDLE_VALUE;
					System.out.println("[SerialAdapter] - start:"
							+ e.getMessage() + ")");
					return;
				}
			}
			handle = new File(device.getNode());
			commPort = portHandle;
			hFile = handle;
		}
	}

	public synchronized void stop() {
		boolean usedThreads;
		if (isOpen()) {
			usedThreads = useThreads;
			useThreads = false;
			if (usedThreads) {
				serialReadWriteThread.terminate = true;
				serialReadWriteThread = null;
			} else {
				try {
					comRead.close();
					comRead = null;
					comWrite.close();
					comWrite = null;
				} catch (IOException e) {
					if (DEBUG)
						e.printStackTrace();
				}
			}
			if (DEBUG)
				System.out
						.println("[SerialAdapter] - stop: close port("
								+ commPort + ")");
			commPort.close();
			commPort = null;
			hFile = INVALID_HANDLE_VALUE;
			useThreads = usedThreads;
		}
	}

	public boolean autoTestDevice() {
		return isOpen();
	}

	public synchronized void write(byte[] buffer, int length)
			throws InterruptedException {
		if (isOpen()) {
			if (txDDelay > 0 && length > 1) {
				for (int i = 0; i < length; i++) {
					byte[] buffer1 = Arrays.copyOfRange(buffer, i, 1);
					if (i > 0)
						Thread.sleep(txDDelay);
					write(buffer1, 1);
				}
			} else {
				if (!fullDuplex) {
					if (useThreads) {
						inReportSize = 0;
						serialReadWriteThread.resetIn();
					} else {
						serialFlush(comRead);
					}
				}
				if (length > 0) {
					byte[] comReport = new byte[length];
					System.arraycopy(buffer, 0, comReport, 0, length);
					if (useThreads)
						serialReadWriteThread.writeOutReport(comReport);
					else {
						try {
							comWrite.write(comReport);
						} catch (IOException e) {
							if (DEBUG)
								e.printStackTrace();
						}
					}
				}
			}
		}
	}

	protected void serialFlush(InputStream is) {
		byte[] buffer = new byte[MAXREPORTSIZE];
		try {
			while (is.read(buffer) > 0)
				;
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	public int read(byte[] buffer, int offset, int maxLength)
			throws InterruptedException {
		byte[] buf = new byte[maxLength];
		int bRead = read(buf, maxLength);
		System.arraycopy(buf, 0, buffer, offset, bRead);
		return bRead;
	}

	public synchronized int read(byte[] buffer, int maxLength)
			throws InterruptedException {
		int result = 0;
		if (isOpen()) {
			if (inReportSize == 0) { 
				// no old data available --> get new data
				if (useThreads) {
					try {
						if (!serialReadWriteThread.inAvailable())
							Thread.sleep(10);
					} catch (InterruptedException e) {
						if (DEBUG)
							e.printStackTrace();
						throw new InterruptedException(
								"[SerialAdapter] COM: interrupted while waiting for incoming data.");
					} // reduce system load
					if (serialReadWriteThread.inAvailable()) {
						inReport = serialReadWriteThread
								.readInReport(maxLength);
						inReportSize = inReport.length;
						inReportStart = 0;
					}
				} else {
					byte[] bufferIn = new byte[inReport.length];
					try {
						if (comRead.read(bufferIn, 1, 8) == 8) {
							inReportSize = inReport[1]; 
							// ignore Report Type, get length
							if (inReportSize >= 8) { 
								// no valid length in InReport[1]
								inReportSize = 8;
								inReportStart = 1; 
								// data begins at InReport[1]
							} else
								inReportStart = 2; 
							// data begins at InReport[2]

						}
					} catch (IOException e) {
						if (DEBUG)
							e.printStackTrace();
					}
				}
			}
			if (inReportSize > maxLength)
				result = maxLength;
			else
				result = inReportSize;
			if (result > 0) {
				if (DEBUG)
					System.out
							.println("[SerialAdapter] - read: result("
									+ result + ")");
				System.arraycopy(inReport, inReportStart, buffer, 0, result);
				inReportSize -= result;
				inReportStart += result;
			}
		}
		return result;
	}

	protected void setfullDuplex(boolean fullDuplex) {
		this.fullDuplex = fullDuplex;
	}

	protected boolean isfullDuplex() {
		return fullDuplex;
	}

	public int getBaudrate() {
		return baudrate;
	}

	public void setTxDDelay(int txDDelay) {
		this.txDDelay = txDDelay;
	}

	public int getTxDDelay() {
		return txDDelay;
	}
}
