/* 
 * GHMessen
 * --------
 * 
 * GHMessen is a measuring program to use in connection with a computer 
 * measuring interface. It's designed for school usage.
 * 
 * Copyright (C) 2010-2011  Frank Schütte - Gymnasium Himmelsthür
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms
 * of the GNU General Public License as published by the Free Software Foundation; either 
 * version 3 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with this program; 
 * if not, see <http://www.gnu.org/licenses/>.
 * 
 * Frank Schütte fschuett AT gymnasium-himmelsthuer DOT de
 * 
 */

package ghmessen.adapter.seriell;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.TooManyListenersException;

import gnu.io.SerialPort;
import gnu.io.SerialPortEvent;
import gnu.io.SerialPortEventListener;

public class SerialReadWriteThread extends Thread implements
		SerialPortEventListener {
	private static final boolean DEBUG = false;
	public boolean terminate = false;
	private SerialPort port = null;
	private InputStream in = null;
	private OutputStream out = null;
	private RingBuffer outReportBuffer = null;
	private RingBuffer inReportBuffer = null;

	public SerialReadWriteThread(SerialPort hPort, int usbInReportSize,
			int usbOutReportSize) {
		super();
		this.setName("serialReadWrite");
		this.setDaemon(true);
		outReportBuffer = new RingBuffer();
		inReportBuffer = new RingBuffer();
		port = hPort;
		setPriority(MAX_PRIORITY); // highest Priority
		if (DEBUG)
			System.out.println("[SerialReadWriteThread] - init: ");
	}

	public void finalize() {
		terminate = true;
		if (DEBUG)
			System.out.println("[SerialReadWriteThread] - finalized: ");
	}

	/****************
	 * Es ist ein Report zur Dateiausgabe im Puffer
	 * 
	 * @return
	 * @throws InterruptedException
	 */
	private boolean outAvailable() throws InterruptedException {
		synchronized (outReportBuffer) {
			while (outReportBuffer.data() <= 0) {
				if (DEBUG)
					System.out
							.println("[SerialReadWriteThread] - outAvailable: --> wait");
				outReportBuffer.wait();
			}
		}
		if (DEBUG)
			System.out
					.println("[SerialReadWriteThread] - outAvailable: --> true");
		return true;
	}

	/****************
	 * Es ist ein Report zum Lesen im Dateieingabepuffer vorhanden.
	 * 
	 * @return
	 */
	public boolean inAvailable() {
		synchronized (inReportBuffer) {
			if (DEBUG)
				System.out.println("[SerialReadWriteThread] - inAvailable: n("
						+ inReportBuffer.data() + ")");
			return inReportBuffer.data() > 0;
		}
	}

	private byte[] readOutBuffer() {
		synchronized (outReportBuffer) {
			byte[] result = new byte[outReportBuffer.data()];
			outReportBuffer.read(result);
			if (DEBUG)
				System.out
						.println("[SerialReadWriteThread] - readOutBuffer: n("
								+ result + ")");
			return result;
		}
	}

	/*********************
	 * Schreibe einen Report in den Puffer für die Dateiausgabe.
	 * 
	 * @param report
	 */
	public void writeOutReport(byte[] report) {
		synchronized (outReportBuffer) {
			outReportBuffer.write(report);
			if (DEBUG)
				System.out
						.println("[SerialReadWriteThread] - writeOutReport: n("
								+ report.length + ") --> notifyAll");
			outReportBuffer.notifyAll();
		}
	}

	public void writeOutReport(byte[] report, int offset, int length) {
		synchronized (outReportBuffer) {
			outReportBuffer.write(report, offset, length);
			if (DEBUG)
				System.out
						.println("[SerialReadWriteThread] - writeOutReport: ofs("
								+ offset
								+ ") length("
								+ length
								+ ") --> notifyAll");
			outReportBuffer.notifyAll();
		}
	}

	/*
	 * Lese maximal max Zeichen aus dem Eingabepuffer.
	 */
	public byte[] readInReport(int max) {
		synchronized (inReportBuffer) {
			int length = Math.min(max, inReportBuffer.data());
			byte[] result = new byte[length];
			inReportBuffer.read(result);
			if (DEBUG)
				System.out
						.println("[SerialReadWriteThread] - readInReport: max("
								+ max + ") length(" + length
								+ ") --> notifyAll");
			inReportBuffer.notifyAll();
			return result;
		}
	}

	/*
	 * Schreibe einen Report aus der Dateieingabe in den Einlesepuffer.
	 */
	private void writeInBuffer(byte[] report) {
		synchronized (inReportBuffer) {
			inReportBuffer.write(report);
			if (DEBUG)
				System.out
						.println("[SerialReadWriteThread] - writeInBuffer: n("
								+ report.length + ") --> notifyAll");
			inReportBuffer.notifyAll();
		}
	}

	/*
	 * Leere den Einlesepuffer und verwirf alle gelesenen Reports.
	 */
	public void resetIn() {
		synchronized (inReportBuffer) {
			inReportBuffer.clear();
			if (DEBUG)
				System.out.println("[SerialReadWriteThread] - resetIn:");
		}
	}

	public void run() {
		int available;
		byte[] outReport;
		try {
			in = port.getInputStream();
			out = port.getOutputStream();
			port.addEventListener(this);
			port.notifyOnDataAvailable(true);
			if (DEBUG)
				System.out
						.println("[SerialReadWriteThread] - run: Thread gestartet");
			while (!terminate) {
				// Zeichen ausgeben, falls Zeichen im Puffer sind.
				if (outAvailable()) {
					synchronized (outReportBuffer) {
						available = outReportBuffer.data();
						outReport = readOutBuffer();
					}
					out.write(outReport);
					if (DEBUG) {
						System.out.print("[SerialReadWriteThread] - write:n("
								+ available + ") bytes(");
						for (byte b : outReport)
							System.out.print(b + ",");
						System.out.println(")");
					}
				}
			}
		} catch (IOException e) {
			if (DEBUG) {
				e.printStackTrace();
				System.err
						.println("[SerialReadWriteThread] Dateiein-/ausgabefehler.");
			}
		} catch (NullPointerException e) {
			if(DEBUG) {
				e.printStackTrace();
				System.err.println("[SerialReadWriteThread] port ist null!");
			}
		} catch (TooManyListenersException e) {
			if (DEBUG)
				e.printStackTrace();
		} catch (InterruptedException e) {
			if (DEBUG)
				e.printStackTrace();
		} finally {
			try {
				in.close();
				out.close();
				port.removeEventListener();
			} catch (IOException|NullPointerException e) {
				if (DEBUG)
					e.printStackTrace();
			}
		}
		if (DEBUG)
			System.err
					.println("[SerialReadWriteThread] - run: Thread gestoppt");

	}

	@Override
	public void serialEvent(SerialPortEvent event) {
		switch (event.getEventType()) {
		case SerialPortEvent.DATA_AVAILABLE:
			dataAvailable(event);
			break;
		}
	}

	protected void dataAvailable(SerialPortEvent event) {
		byte[] inReport = null;
		try {
			int available = in.available();
			inReport = new byte[available];
			in.read(inReport);
			synchronized (inReportBuffer) {
				writeInBuffer(inReport);
			}
			if (DEBUG) {
				System.out.print("[SerialReadWriteThread] - read:n("
						+ available + ") bytes(");
				for (byte b : inReport)
					System.out.print(b + ",");
				System.out.println(")");
			}
		} catch (IOException e) {
			if (DEBUG)
				e.printStackTrace();
		}

	}
}
