/* 
 * GHMessen
 * --------
 * 
 * GHMessen is a measuring program to use in connection with a computer 
 * measuring interface. It's designed for school usage.
 * 
 * Copyright (C) 2010-2011  Frank Schütte - Gymnasium Himmelsthür
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms
 * of the GNU General Public License as published by the Free Software Foundation; either 
 * version 3 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with this program; 
 * if not, see <http://www.gnu.org/licenses/>.
 * 
 * Frank Schütte fschuett AT gymnasium-himmelsthuer DOT de
 * 
 */

package ghmessen.components;

import ghmessen.groesse.Messbereich;
import ghmessen.groesse.Skalierung;

import java.awt.geom.AffineTransform;
import java.awt.geom.NoninvertibleTransformException;
import java.awt.geom.Point2D;

import static ghmessen.util.MathUtil.*;

public class Skala {

	private static final boolean DEBUG = false;
	private double minimum, maximum;
	private double unterteilung;
	private double laenge;
	private Skalierung skalierung;
	private int[] stricheKlein;
	private Skalentext[] skalenText;
	private boolean logarithmisch = false;
	private boolean autoBerechnen = true;
	private boolean fehler = false;
	private String fehlerText;
	private boolean vertikal = false;
	private AffineTransform transform = null;
	private AffineTransform inversTransform = null;

	// Zeigeranzeige verwendet Weltkoordinaten:
	// Alle Zahlen zwischen minimum und maximum landen zwischen 0 und 1(laenge).

	public Skala(double minimum, double maximum, double unterteilung,
			Skalierung skalierung, boolean vertikal, double laenge) {
		this.minimum = minimum;
		this.maximum = maximum;
		this.unterteilung = unterteilung;
		this.skalierung = skalierung;
		this.logarithmisch = skalierung.istLogarithmisch();
		this.laenge = laenge;
		this.vertikal = vertikal;
		if (DEBUG)
			System.out.println("[Skala] - Konstruktor:\n"
					+"\tmin-max("+minimum+"-"+maximum+") u("+unterteilung+")\n"
					+"\tskalierung("+skalierung+"|log("+logarithmisch+")\n"
					+"\tlaenge("+laenge+") vertikal("+vertikal+")");
		ueberpruefeWerte();
		if (autoBerechnen)
			aktualisieren();
	}

	public Skala(double minimum, double maximum, double unterteilung,
			boolean vertikal) {
		this(minimum, maximum, unterteilung, Skalierung.y, vertikal, 1.0);
	}

	public Skala(Double minimum, Double maximum, double unterteilung,
			Skalierung skalierung, boolean vertikal, double laenge) {
		this(minimum == null ? -5 : minimum.doubleValue(), maximum == null ? 5
				: maximum.doubleValue(), unterteilung, skalierung, vertikal,
				laenge);
	}

	public Skala(Double minimum, Double maximum, double unterteilung, Skalierung skalierung,
			boolean vertikal) {
		this(minimum, maximum, unterteilung, skalierung, vertikal, 1.0);
	}

	/****************************************************************
	 * Berechnungen
	 ****************************************************************/

	private void ueberpruefeWerte() {
		if (minimum > maximum) {
			fehler = true;
			fehlerText = "Das Minimum " + minimum
					+ " ist größer als das Maximum " + maximum
					+ ". Das ist nicht möglich.";
		} else if (minimum == maximum) {
			fehler = true;
			fehlerText = "Das Minimum des Intervalls ist gleich dem Maximum. Das ist nicht möglich.";
		} else
			fehler = false;
	}

	private void aktualisieren() {
		if (DEBUG)
			System.out.println("[Skala] - aktualisieren:\n"
					+"\tfehler("+fehler+")\n"
					+"\tmin-max("+minimum+"-"+maximum+") u("+unterteilung+")\n"
					+"\tskalierung("+skalierung+"|log("+logarithmisch+")\n"
					+"\tlaenge("+laenge+") vertikal("+vertikal+")");
		berechneAffineTransform();
		berechneInverseAffineTransform();
		berechneSkala();
	}

	private void berechneSkala() {
		stricheKlein = berechneStricheKlein();
		skalenText = berechneSkalentextPos();
	}

	private void berechneAffineTransform() {
		if (vertikal) {
			transform = AffineTransform.getTranslateInstance(0, laenge);
			transform.scale(1.0, -laenge / (maximum - minimum));
			transform.translate(0, -minimum);
		} else {
			transform = AffineTransform.getScaleInstance(laenge
					/ (maximum - minimum), 1.0);
			transform.translate(-minimum, 0);
		}
	}

	private void berechneInverseAffineTransform() {
		if (transform == null)
			berechneAffineTransform();
		try {
			inversTransform = transform.createInverse();
			fehler = false;
		} catch (NoninvertibleTransformException e) {
			if (DEBUG)
				System.err.println("[Skala] - berechneInverseAffineTransform: ["+minimum+" - "+maximum+"/"+unterteilung+"] -> "+laenge+(logarithmisch?"(log)":""));
			inversTransform = AffineTransform.getTranslateInstance(0, 0);
			fehler = true;
			fehlerText = "Die verwendete Skala ist nicht bijektiv und kann daher nicht invertiert werden.";
		}
	}

	public String skaliereText(String s) {
		return skalierung.toString(s);
	}

	public AffineTransform getAffineTransform() {
		if (transform == null)
			berechneAffineTransform();
		return transform;
	}

	public AffineTransform getInverseAffineTransform() {
		if (transform == null)
			berechneAffineTransform();
		if (inversTransform == null)
			berechneInverseAffineTransform();
		return inversTransform;
	}

	private double transform2D(double l) {
		if (vertikal)
			return transform.transform(new Point2D.Double(0.0, l), null).getY();
		else
			return transform.transform(new Point2D.Double(l, 0.0), null).getX();
	}

	public int transform(double l) {
		return (int) Math.round(transform2D(l));
	}

	private double inversTransform2D(double k) {
		if (vertikal)
			return inversTransform.transform(new Point2D.Double(0.0, k), null)
					.getY();
		else
			return inversTransform.transform(new Point2D.Double(k, 0.0), null)
					.getX();
	}

	public double inversTransform(int l) {
		return inversTransform2D(l);
	}

	private int[] berechneStricheKlein() {
		int[] result;
		if (logarithmisch) {
			double unlogmin = Math.pow(10.0,minimum);
			double unlogmax = Math.pow(10.0,maximum);
			double minex = Math.floor(flog10(unlogmin));
			double maxex = Math.floor(flog10(unlogmax));
			double minf = Math.ceil(unlogmin * Math.pow(10.0, -minex));
			double maxf = Math.floor(unlogmax * Math.pow(10.0, -maxex));
			result = new int[(int) (maxex > minex ? 10 - minf + maxf + 9
					* (maxex - minex - 1) : maxf - minf + 1)];
			int index = 0;
			for (int ex = (int) minex; ex <= maxex; ex++) {
				int links = (int) (ex == minex ? minf : 1);
				int rechts = (int) (ex == maxex ? maxf : 9);
				for (int f = links; f <= rechts; f++)
					result[index++] = Math.round(transform(Math.log10(f
							* Math.pow(10.0, ex))));
			}
		} else {
			double dy = findeFeineUnterteilung(minimum, maximum);
			double y0 = moduloGrenze(minimum, dy, false);
			double yn = moduloGrenze(maximum, dy, false);
			result = new int[(int) Math.round(Math.abs(yn - y0) / dy) + 1];
			for (int i = 0; i < result.length; i++)
				result[i] = Math.round(transform(y0 + i * dy));
		}
		return result;
	}

	private Skalentext[] berechneSkalentextPos() {
		Skalentext[] result;
		if (logarithmisch) {
			double unlogmin = Math.pow(10.0,minimum);
			double unlogmax = Math.pow(10.0,maximum);
			double minex = Math.ceil(flog10(unlogmin));
			double maxex = Math.floor(flog10(unlogmax));
			result = new Skalentext[(int)Math.max(0, maxex-minex+1)];
			int index = 0;
			for (int ex = (int) minex; ex <= maxex; ex++) {
				result[index] = new Skalentext();
				result[index].pos = transform(Math.log10(Math.pow(10.0, ex)));
				result[index].text = formatNum(Math.pow(10.0, ex),Math.max(0, -ex));
				result[index].ofs = 0;
				index++;
			}
		} else {
			double dy = unterteilung == -1.0 ? findeGrobeUnterteilung(minimum,
					maximum) : unterteilung;

			double ds = findeMittlereUnterteilung(minimum, maximum);
			String[] text = berechneSkalaListe(minimum, maximum, ds);
			double y0 = moduloGrenze(minimum, ds, false);
			result = new Skalentext[text.length];
			if (DEBUG)
				System.out.println("[Skala] - berechneSkalentextPos: min("
						+ minimum + ") max(" + maximum + ") dy(" + dy
						+ ") text(" + text.length + ")");
			int y;
			for (int i = 0; i < result.length; i++) {
				result[i] = new Skalentext();
				y = transform(y0 + i * ds);
				result[i].pos = y;
				result[i].text = (Math.abs(Math.IEEEremainder(y0 + i * ds, dy)) < 0.01 * dy) ? text[i]
						: null;
				result[i].ofs = 0;
			}
		}
		return result;
	}

	/*********************************************************
	 * get / set
	 *********************************************************/

	public void setMinimum(double minimum) {
		this.minimum = minimum;
		ueberpruefeWerte();
		if (autoBerechnen)
			aktualisieren();
	}

	public void setMinimum(Double minimum) {
		setMinimum(minimum == null ? -5 : minimum.doubleValue());
	}

	public double getMinimum() {
		return minimum;
	}

	public void setMaximum(double maximum) {
		this.maximum = maximum;
		ueberpruefeWerte();
		if (autoBerechnen)
			aktualisieren();
	}

	public void setMaximum(Double maximum) {
		setMaximum(maximum == null ? 5 : maximum.doubleValue());
	}

	public double getMaximum() {
		return maximum;
	}

	public void setIntervall(Intervall intervall, Double unterteilung) {
		if (intervall == null)
			return;
		setMinMax(intervall.getMin(), intervall.getMax(), unterteilung);
	}

	public void setMinMax(Double minimum, Double maximum, Double unterteilung) {
		if (unterteilung == null && minimum != null && maximum != null)
			setUnterteilung(findeGrobeUnterteilung(minimum, maximum));
		else if (unterteilung == null)
			setUnterteilung(-1.0);
		else
			setUnterteilung(unterteilung);
		setMinMax(minimum, maximum);
	}

	private void setMinMax(Double minimum, Double maximum) {
		// Minimum und Maximum müssen verschieden sein!
		// Minimum / Maximum können auch einzeln gesetzt werden.
		boolean korrigieren = false;
		if (minimum != null && maximum != null) {
			korrigieren = minimum == maximum;
		}
		autoBerechnen = maximum == null;
		if (minimum != null)
			setMinimum(korrigieren ? minimum - 5 : minimum);
		autoBerechnen = true;
		if (maximum != null)
			setMaximum(korrigieren ? maximum + 5 : maximum);
	}

	public void setMessbereich(Messbereich m) {
		autoBerechnen = false;
		minimum = m.getMinimum();
		maximum = m.getMaximum();
		unterteilung = m.getUnterteilung();
		autoBerechnen = true;
		ueberpruefeWerte();
		aktualisieren();
	}

	public void setUnterteilung(double unterteilung) {
		this.unterteilung = unterteilung;
	}

	public double getUnterteilung() {
		return unterteilung;
	}

	public void setLaenge(double laenge) {
		if (this.laenge == laenge)
			return;
		this.laenge = laenge;
		if (autoBerechnen)
			aktualisieren();
	}

	public double getLaenge() {
		return laenge;
	}

	public int[] getStricheKlein() {
		return stricheKlein;
	}

	public int[] getStricheGross() {
		int[] result = new int[skalenText.length];
		for (int i = 0; i < skalenText.length; i++)
			result[i] = skalenText[i].pos;
		return result;
	}

	public String[] getZahlen() {
		String[] result = new String[skalenText.length];
		for (int i = 0; i < skalenText.length; i++)
			result[i] = skalenText[i].text;
		return result;
	}

	public Skalentext[] getSkalenText() {
		return skalenText;
	}

	public void setSkalierung(Skalierung skalierung) {
		this.skalierung = skalierung;
		this.logarithmisch = skalierung.istLogarithmisch();
		if (logarithmisch && autoBerechnen)
			aktualisieren();
	}

	public Skalierung getSkalierung() {
		return skalierung;
	}

	public boolean istFehlerhaft() {
		return fehler;
	}

	public String getFehlerText() {
		return fehlerText;
	}

}
