/* 
 * GHMessen
 * --------
 * 
 * GHMessen is a measuring program to use in connection with a computer 
 * measuring interface. It's designed for school usage.
 * 
 * Copyright (C) 2010-2011  Frank Schütte - Gymnasium Himmelsthür
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms
 * of the GNU General Public License as published by the Free Software Foundation; either 
 * version 3 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with this program; 
 * if not, see <http://www.gnu.org/licenses/>.
 * 
 * Frank Schütte fschuett AT gymnasium-himmelsthuer DOT de
 * 
 */

package ghmessen.groesse;

import ghmessen.CL4File;
import ghmessen.GHMessen;
import ghmessen.Konstanten;
import ghmessen.Messparameter;
import ghmessen.util.Util;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.PrintWriter;
import java.io.Serializable;
import java.util.AbstractList;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Locale;
import java.util.Scanner;
import java.util.Vector;
import java.util.concurrent.CopyOnWriteArrayList;

import javax.xml.transform.sax.TransformerHandler;

import org.jdesktop.observablecollections.ObservableList;
import org.jdesktop.observablecollections.ObservableListListener;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.AttributesImpl;

public class Darstellungsliste extends AbstractList<Darstellung>
		implements ObservableList<Darstellung>, Serializable, PropertyChangeListener {
	/**
	 * 
	 */
	private static final long serialVersionUID = -5605020260724329033L;
	private static final boolean DEBUG = false;
	private final boolean supportsElementPropertyChanged;
	private List<Darstellung> list;
	private transient List<ObservableListListener> listeners;
	private Darstellung histogramm;
	private Darstellung spektrum;

	public static void createDTD(PrintWriter out) {
		out.println("<!ELEMENT Darstellungsliste (Darstellung*)>");
		out.println("<!ATTLIST Darstellungsliste Histogramm (INT) #IMPLIED Spektrum (INT) #IMPLIED>");
		Darstellung.createDTD(out);
	}

	public void writeToXml(TransformerHandler hd) throws SAXException {
		if (hd == null)
			throw new SAXException("[Darstellungsliste] Objekt ist nicht initialisiert.");
		AttributesImpl atts = new AttributesImpl();
		if (histogramm != null)
			atts.addAttribute("", "", "Histogramm", "INT", Integer.toString(list.indexOf(histogramm)));
		if (spektrum != null)
			atts.addAttribute("", "", "Spektrum", "INT", Integer.toString(list.indexOf(spektrum)));
		hd.startElement("", "", "Darstellungsliste", null);
		for (int index = 0; index < list.size(); index++) {
			list.get(index).writeToXml(hd);
		}
		hd.endElement("", "", "Darstellungsliste");
	}

	private transient int histogrammIndex;
	private transient int spektrumIndex;

	public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException {
		if (qName.equals("Darstellungsliste")) {
			histogrammIndex = -1;
			spektrumIndex = -1;
			for (int index = 0; index < attributes.getLength(); index++) {
				try {
					String name = attributes.getLocalName(index);
					String inhalt = attributes.getValue(index);
					if (DEBUG)
						System.out.println("[Messparameter] - endElement: Feld(" + name + ") Inhalt(" + inhalt + ")");
					if ("Histogramm".equals(name))
						histogrammIndex = Integer.parseInt(inhalt);
					else if ("Spektrum".equals(name))
						spektrumIndex = Integer.parseInt(inhalt);
				} catch (SecurityException e) {
					if (DEBUG)
						e.printStackTrace();
					throw new SAXException(e.getMessage());
				} catch (IllegalArgumentException e) {
					if (DEBUG)
						e.printStackTrace();
					throw new SAXException(e.getMessage());
				}
			}
			return;
		} else if (qName.equals("Darstellung"))
			list.add(new Darstellung());
		list.get(list.size() - 1).startElement(uri, localName, qName, attributes);
	}

	public void endElement(String uri, String localName, String qName, String inhalt) throws SAXException {
		if (qName.equals("Darstellungsliste"))
			return;
		else if (qName.equals("Darstellung")) {
			int lastIndex = list.size() - 1;
			list.get(lastIndex).endElement(uri, localName, qName, inhalt);
			if (histogrammIndex == lastIndex)
				histogramm = list.get(lastIndex);
			if (spektrumIndex == lastIndex)
				spektrum = list.get(lastIndex);
		} else
			list.get(list.size() - 1).endElement(uri, localName, qName, inhalt);
	}

	Darstellungsliste(List<Darstellung> list, boolean supportsElementPropertyChanged) {
		this.list = list;
		for (Darstellung d : list)
			d.addPropertyChangeListener(this);
		listeners = new CopyOnWriteArrayList<ObservableListListener>();
		this.supportsElementPropertyChanged = supportsElementPropertyChanged;
		this.histogramm = null;
		this.spektrum = null;
	}

	public Darstellungsliste() {
		this(new Vector<Darstellung>(), true);
	}

	public Darstellungsliste(Vector<Darstellung> v) {
		this(v, true);
	}

	private void writeObject(ObjectOutputStream out) throws IOException {
		out.defaultWriteObject();
	}

	private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {
		in.defaultReadObject();
		listeners = new CopyOnWriteArrayList<ObservableListListener>();
	}

	public void setFromCL4(String cl4) {
		try (Scanner a0 = new Scanner(cl4)) {
			a0.useLocale(Locale.US);
			CL4File.AUSWERTUNG art = CL4File.AUSWERTUNG.toAuswertung(a0.nextInt());
			if (art == null) {
				if (DEBUG)
					System.out.println(
							"[Darstellungsliste] - setFromCL4: Ich ignoriere die unbekannte Auswertung(" + cl4 + ")");
				return;
			}
			int achse = a0.nextInt();
			Darstellung ds = achseDivDarstellungCL4(achse);
			int yA = achseModDarstellungCL4(achse);
			if (ds == null || yA < 0 || yA > ds.getColumnCount() - 1) {
				if (DEBUG)
					System.out.println("[Darstellungsliste] - setFromCL4: Fehler ds(" + ds + ") yA(" + yA
							+ ") bei der Auswertung(" + cl4 + ")");
				return;
			}
			int farbe = a0.next().charAt(1) - '0';
			a0.nextInt(); // ignoriere 2x -1
			a0.nextInt();
			String s;
			do {
				s = a0.next(); // ignoriere die Darstellungsparameter
			} while (s != null && s.charAt(0) >= 'A' && s.charAt(0) <= 'Z');
			Double[] parameter = new Double[art.getDoubleAnzahl()];
			if (s != null && parameter.length > 0) // lese die Parameterliste
				parameter[0] = Double.parseDouble(s);
			for (int i = 1; i < parameter.length; i++)
				parameter[i] = a0.nextDouble();
			String text = art.hatText() ? a0.nextLine() : "";
			ds.neueAuswertung(art.toAuswertung(ds.getColumnArray(Darstellung.XACHSE, 0, ds.getRowCount() - 1),
					ds.getColumnArray(yA, 0, ds.getRowCount() - 1), text, ds.getGraphEigenschaften().getSchrift(),
					Konstanten.FARBE_GRAPHEN.get(farbe), ds.getGraphEigenschaften().getLinie(), parameter));
		}
	}

	private Darstellung achseDivDarstellungCL4(int achse) {
		Darstellung ds = null;
		for (int i = 0; i < list.size(); i++) {
			ds = list.get(i);
			if (ds.getColumnCount() - 2 < achse)
				achse -= ds.getColumnCount() - 1;
			else
				return ds;
		}
		return null;
	}

	private int achseModDarstellungCL4(int achse) {
		Darstellung ds = null;
		for (int i = 0; i < list.size(); i++) {
			ds = list.get(i);
			if (ds.getColumnCount() - 2 < achse)
				achse -= ds.getColumnCount() - 1;
			else
				return achse + 1;
		}
		return -1;
	}

	public String[] getNamen() {
		if (size() == 0)
			return new String[] {};
		String[] result = new String[size()];
		for (int i = 0; i < size(); i++)
			result[i] = ((Darstellung) get(i)).getName();
		return result;
	}

	public void setzeMessmodus(boolean an) {
		if (size() == 0)
			return;
		for (Darstellung d : list)
			d.setzeMessmodus(an);
	}

	public void removeElementMitNamen(String name) {
		int index = indexOfName(name);
		if (index != -1)
			remove(index);
	}

	public int indexOfName(String name) {
		for (int i = 0; i < size(); i++) {
			Darstellung d = (Darstellung) get(i);
			if (d.getName().equals(name))
				return i;
		}
		return -1;
	}

	public int getSize() {
		return list.size();
	}

	public int indexOf(Darstellung d) {
		for (int i = 0; i < list.size(); i++)
			if (list.get(i) == d)
				return i;
		return -1;
	}

	public Darstellung get(int index) {
		return list.get(index);
	}

	public int size() {
		return list.size();
	}

	public Darstellung set(int index, Darstellung element) {
		Darstellung oldValue = list.set(index, element);
		oldValue.removePropertyChangeListener(this);
		element.addPropertyChangeListener(this);
		for (ObservableListListener listener : listeners) {
			listener.listElementReplaced(this, index, oldValue);
		}
		return oldValue;
	}

	public Darstellung neueDarstellung() {
		String neu = "Standard";
		neu = Util.findUniqueName(this, neu);
		Darstellung ds = new Darstellung(GHMessen.getApplication().getMainView().gt);
		ds.setName(neu);
		add(ds);
		return ds;
	}

	public void add(int index, Darstellung element) {
		list.add(index, element);
		element.addPropertyChangeListener(this);
		modCount++;
		for (ObservableListListener listener : listeners) {
			listener.listElementsAdded(this, index, 1);
		}
	}

	public void addHistogramm(Darstellung element) {
		add(element);
		histogramm = element;
	}

	public void addSpektrum(Darstellung element) {
		add(element);
		spektrum = element;
	}

	public void removeHistogramm() {
		if (histogramm == null)
			return;
		remove(histogramm);
		histogramm = null;
	}

	public void removeSpektrum() {
		if (spektrum == null)
			return;
		remove(spektrum);
		spektrum = null;
	}

	public Darstellung remove(int index) {
		Darstellung oldValue = list.remove(index);
		oldValue.removePropertyChangeListener(this);
		modCount++;
		for (ObservableListListener listener : listeners) {
			listener.listElementsRemoved(this, index, java.util.Collections.singletonList(oldValue));
		}
		if (oldValue == histogramm)
			histogramm = null;
		if (oldValue == spektrum)
			spektrum = null;
		return oldValue;
	}

	public boolean addAll(Collection<? extends Darstellung> c) {
		return addAll(size(), c);
	}

	public boolean addAll(Collection<? extends Darstellung> c, WerteTabelle gt) {
		for (Darstellung d : c)
			d.setWerteTabelle(gt);
		return addAll(c);
	}

	public boolean addAll(int index, Collection<? extends Darstellung> c) {
		if (list.addAll(index, c)) {
			for (int i = index; i < list.size(); i++)
				list.get(i).addPropertyChangeListener(this);
			modCount++;
			for (ObservableListListener listener : listeners) {
				listener.listElementsAdded(this, index, c.size());
			}
		}
		if (c instanceof Darstellungsliste) {
			if (histogramm == null)
				histogramm = ((Darstellungsliste) c).getHistogramm();
			if (spektrum == null)
				spektrum = ((Darstellungsliste) c).getSpektrum();
		}
		return false;
	}

	public void clear() {
		List<Darstellung> dup = new ArrayList<Darstellung>(list);
		for (Darstellung g : list)
			g.removePropertyChangeListener(this);
		list.clear();
		modCount++;
		if (dup.size() != 0) {
			for (ObservableListListener listener : listeners) {
				listener.listElementsRemoved(this, 0, dup);
			}
		}
		histogramm = null;
		spektrum = null;
	}

	public boolean containsAll(Collection<?> c) {
		return list.containsAll(c);
	}

	public <T> T[] toArray(T[] a) {
		return list.toArray(a);
	}

	public Object[] toArray() {
		return list.toArray();
	}

	private void fireElementChanged(int index) {
		for (ObservableListListener listener : listeners) {
			listener.listElementPropertyChanged(this, index);
		}
	}

	public void addObservableListListener(ObservableListListener listener) {
		listeners.add(listener);
	}

	public void removeObservableListListener(ObservableListListener listener) {
		listeners.remove(listener);
	}

	public boolean supportsElementPropertyChanged() {
		return supportsElementPropertyChanged;
	}

	@Override
	public void propertyChange(PropertyChangeEvent evt) {
		if ("einzelEin".equals(evt.getPropertyName())) {
			for (Darstellung d : this) {
				if ("Standard".equals(d.getName())) {
					if ((Boolean)evt.getNewValue())
						d.setManuelleMessung();
					else
						d.setZeitmessung();
					return;
				}
			}
		} else {
			int index = indexOf(evt.getSource());
			if (index > -1)
				fireElementChanged(index);
		}
	}

	public Darstellung getHistogramm() {
		return histogramm;
	}

	public Darstellung getSpektrum() {
		return spektrum;
	}

}
