/* 
 * GHMessen
 * --------
 * 
 * GHMessen is a measuring program to use in connection with a computer 
 * measuring interface. It's designed for school usage.
 * 
 * Copyright (C) 2010-2011  Frank Schütte - Gymnasium Himmelsthür
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms
 * of the GNU General Public License as published by the Free Software Foundation; either 
 * version 3 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with this program; 
 * if not, see <http://www.gnu.org/licenses/>.
 * 
 * Frank Schütte fschuett AT gymnasium-himmelsthuer DOT de
 * 
 */

package ghmessen.groesse;

import static ghmessen.groesse.Messgroesse.Spannung;
import static ghmessen.groesse.wertetabelle.XAchseTyp.T;
import static ghmessen.groesse.wertetabelle.XAchseTyp.UNABHAENGIG;

import ghmessen.CL4File;
import ghmessen.GHMessen;
import ghmessen.adapter.AdapterListe;
import ghmessen.filefilter.FileFormatException;
import ghmessen.groesse.wertetabelle.XAchseTyp;
import ghmessen.instrumente.komponenten.Nullpunkt;
import ghmessen.parser.FormelGenerator;
import ghmessen.util.DTD;
import ghmessen.util.Util;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.PrintWriter;
import java.io.Serializable;
import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Locale;
import java.util.Scanner;
import java.util.Set;

import javax.xml.transform.sax.TransformerHandler;

import ld.api.ICASSY;
import ld.api.IXRay;
import ld.api.Waveforms;
import ld.api.ios.ICASSYQuantity;
import ld.components.MathString;
import util.StringUtil;

import org.codehaus.janino.ExpressionEvaluator;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.AttributesImpl;

public class Groesse implements Serializable {
	/**
	 * 
	 */
	private static final long serialVersionUID = 1341769800258321284L;
	private static boolean DEBUG = false;

	private transient FormelGenerator p = new FormelGenerator();

	/* eine aktive Größe wird bei der nächsten Messung berücksichtigt */
	private boolean aktiv;

	private String name;
	private boolean manuellerName;
	private Art art;
	private String symbol;
	private Einheit einheit;
	private double von;
	private double bis;
	private int stellen;

	private XAchseTyp xAchse;

	private int nr; // Adapternummer in der Adapterliste
	// Für Cassys-Adapter:
	private int AB; // Power-CASSY: 0 = Stellgröße, 1 = Messgröße
	private int EF; // Kanal: erforderlich für z.B. die Timer-Box
	/*
	 * Ableitung, Integral, Mittelwert, FFT und Histogramm erfordern eine
	 * Bezugsgröße
	 */
	/* Leistung, Cos-Phi, Phi benötigen zwei Bezugsgrößen */
	private Groesse bezug;
	private Groesse bezug2;
	/* Formel, R und S und die Wellenform "USER" erfordern eine Formel */
	private String formel;
	private transient ExpressionEvaluator formelterm;
	/* R und S sind synchronisiert oder unsynchronisiert */
	private boolean synchron;
	/* Mittelwert erfordert ein Intervall und eine Einheit */
	private double intervall;
	private Einheit mweinheit;
	/* Histogramm erfordert ein delta */
	private double delta;
	/* Messgroesse erfordert eine Werteanzeige vom Typ Messgroesse */
	private Messgroesse messGroesse;
	/* Messgroesse erfordert Korrekturwerte und ein Messgeraet */
	/* U,I,s,N,R */
	private double wert;
	private double offset;
	private double faktor;
	private Messbereich messbereich;
	private Nullpunkt nullpunkt;
	private Aufnahme aufnahme;
	private double zeitraum; /* Counter-Torzeit: 0 = unendlich */
	/* US-Box */
	private boolean nichtImmerAktiv;
	private boolean ausloeserTriggert;
	private boolean schmalesSichtfeld;
	/* Timer-Box */
	private boolean start1Flanke;
	private boolean stoppFlanke;
	private int flanke;
	private boolean invertieren;
	private boolean pendel;
	private boolean ueberschreiben;
	private double fahne;
	private double radius;
	/* Linearer Stoß und Drehstoß */
	private int anordnung;

	/* XRay erfordert Gitterabstand und Energie */
	private double d_pm;
	private double e_keV;

	/* Power-CASSY */
	private Waveforms form;
	private boolean formSynchron;
	private double amplitude; // zwischen -10 und 10
	private double formOffset;
	private double anstieg; // Prozentzahl zwischen 0 und 100
	private double frequenz; // Frequenz

	/* eine abhängige Größe ist die n-te Messreihe (n=1,2,...) einer Größe */
	private boolean abhaengig;
	private int messreihe;

	private transient PropertyChangeSupport support = new PropertyChangeSupport(this);

	private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {
		in.defaultReadObject();
		support = new PropertyChangeSupport(this);
		p = new FormelGenerator();
		formelterm = null;
		if (art == Art.Messgroesse) {
			AdapterListe al = GHMessen.getApplication().getMainView().al;
			if (al != null && (al.size() <= nr || !(al.get(nr) instanceof ICASSY)))
				art = Art.Parameter;
		}
	}

	public void addPropertyChangeListener(PropertyChangeListener listener) {
		this.support.addPropertyChangeListener(listener);
	}

	public void addPropertyChangeListener(String property, PropertyChangeListener listener) {
		this.support.addPropertyChangeListener(property, listener);
	}

	public void removePropertyChangeListener(PropertyChangeListener listener) {
		this.support.removePropertyChangeListener(listener);
	}

	public void removePropertyChangeListener(String property, PropertyChangeListener listener) {
		this.support.removePropertyChangeListener(property, listener);
	}

	public boolean istPropertyChangeListener(PropertyChangeListener listener) {
		for (PropertyChangeListener pcl : this.support.getPropertyChangeListeners())
			if (pcl.equals(listener))
				return true;
		return false;
	}

	public static void createDTD(PrintWriter out) {
		Art.createDTD(out);
		XAchseTyp.createDTD(out);
		Messbereich.createDTD(out);
		Messgroesse.createDTD(out);
		Nullpunkt.createDTD(out);
		Aufnahme.createDTD(out);
		DTD.create(out, Waveforms.class);
		out.println("<!ELEMENT Groesse EMPTY>");
		out.println("<!ATTLIST Groesse ");
		out.println("\taktiv (BOOLEAN) #REQUIRED");
		out.println("\tmanuellerName (BOOLEAN) #REQUIRED");
		out.println("\tart (Art) #REQUIRED");
		out.println("\tvon (DOUBLE) #REQUIRED");
		out.println("\tbis (DOUBLE) #REQUIRED");
		out.println("\tstellen (INT) #REQUIRED");
		out.println("\txAchse (XAchseTyp) #REQUIRED");
		out.println("\ttyp (AdapterTyp) #REQUIRED");
		out.println("\tnr (INT) #REQUIRED");
		out.println("\tAB (INT) #REQUIRED");
		out.println("\tEF (INT) #REQUIRED");
		out.println("\tbezug CDATA #REQUIRED");
		out.println("\tbezug2 CDATA #REQUIRED");
		out.println("\tformel CDATA #REQUIRED");
		out.println("\tsynchron (BOOLEAN) #REQUIRED");
		out.println("\tintervall (DOUBLE) #REQUIRED");
		out.println("\tmweinheit (CDATA) #REQUIRED");
		out.println("\tdelta (DOUBLE) #REQUIRED");
		out.println("\twert (DOUBLE) #REQUIRED");
		out.println("\toffset (DOUBLE) #REQUIRED");
		out.println("\tfaktor (DOUBLE) #REQUIRED");
		out.println("\tzeitraum CDATA #REQUIRED");
		out.println("\tnichtImmerAktiv (BOOLEAN) #REQUIRED");
		out.println("\tausloeserTriggert (BOOLEAN) #REQUIRED");
		out.println("\tschmalesSichtfeld (BOOLEAN) #REQUIRED");
		out.println("\tstart1Flanke (BOOLEAN) #REQUIRED");
		out.println("\tstoppFlanke (BOOLEAN) #REQUIRED");
		out.println("\tflanke (INT) #REQUIRED");
		out.println("\tinvertieren (BOOLEAN) #REQUIRED");
		out.println("\tpendel (BOOLEAN) #REQUIRED");
		out.println("\tueberschreiben (BOOLEAN) #REQUIRED");
		out.println("\tfahne (DOUBLE) #REQUIRED");
		out.println("\tradius (DOUBLE) #REQUIRED");
		out.println("\tanordnung (INT) #REQUIRED");
		out.println("\td_pm (DOUBLE) #REQUIRED");
		out.println("\te_keV (DOUBLE) #REQUIRED");
		out.println("\tformSynchron (BOOLEAN) #REQUIRED");
		out.println("\tamplitude (DOUBLE) #REQUIRED");
		out.println("\tformoffset (DOUBLE) #REQUIRED");
		out.println("\tanstieg (DOUBLE) #REQUIRED");
		out.println("\tfrequenz (DOUBLE) #REQUIRED");
		out.println("\tmessreihe (INT) #REQUIRED");
		out.println("\tname CDATA #REQUIRED");
		out.println("\tsymbol CDATA #REQUIRED");
		out.println("\teinheit (CDATA) #REQUIRED");
		out.println("\tmessGroesse (Messgroesse) #REQUIRED");
		out.println("\tnullpunkt (Nullpunkt) #REQUIRED");
		out.println("\taufnahme (Aufnahme) #REQUIRED");
		out.println("\twaveForm (WaveForm) #REQUIRED");
		out.println("\tabhaengig (BOOLEAN) #REQUIRED>");
	}

	public void writeToXml(TransformerHandler hd) throws SAXException {
		if (hd == null)
			throw new SAXException("[Groesse] Objekt ist nicht initialisiert.");
		AttributesImpl atts = new AttributesImpl();
		String inhalt = "";
		for (Field f : this.getClass().getDeclaredFields())
			try {
				if (Modifier.isStatic(f.getModifiers()) || Modifier.isTransient(f.getModifiers()))
					continue;
				if (f.getName().equals("messbereich"))
					continue;
				if (f.get(this) != null && (f.getName().equals("mweinheit") || f.getName().equals("einheit"))) {
					Einheit e = (Einheit) f.get(this);
					inhalt = e.toString();
				} else if (f.get(this) != null && f.getName().equals("messGroesse")) {
					Messgroesse m = (Messgroesse) f.get(this);
					inhalt = m.toString();
				} else if (f.get(this) != null && f.getName().equals("art")) {
					Art a = (Art) f.get(this);
					inhalt = a.toString();
				} else if (f.get(this) != null && (f.getName().equals("bezug") || f.getName().equals("bezug2"))) {
					inhalt = ((Groesse) f.get(this)).getSymbol();
				} else
					inhalt = f.get(this) == null ? "" : f.get(this).toString();
				atts.addAttribute("", "", f.getName(), "CDATA", inhalt);
			} catch (NullPointerException e) {
				System.err.println("[Groesse] - writeToXml: Feld(" + f.getName() + ") Inhalt(" + inhalt
						+ ") konnte nicht geschrieben werden.");
			} catch (IllegalArgumentException e) {
				if (DEBUG)
					System.out.println(
							"[Groesse] - writeToXml: Feld(" + f.getName() + ") konnte nicht geschrieben werden.");
			} catch (IllegalAccessException e) {
				if (DEBUG)
					System.out.println(
							"[Groesse] - writeToXml: Feld(" + f.getName() + ") konnte nicht geschrieben werden.");
			}
		hd.startElement("", "", "Groesse", atts);
		messbereich.writeToXml(hd);
		hd.endElement("", "", "Groesse");
	}

	private static enum BEREICH {
		ANFANG, GROESSE, MESSBEREICH, MESSGROESSE
	};

	private transient BEREICH bereich;

	public void startElement(String uri, String localName, String qName, Attributes attributes, Groessenliste gl)
			throws SAXException {
		switch (bereich) {
		case ANFANG:
			if (qName.equals("Groesse")) {
				bereich = BEREICH.GROESSE;
				for (int index = 0; index < attributes.getLength(); index++) {
					try {
						String name = attributes.getLocalName(index);
						String inhalt = attributes.getValue(index);
						Field field = getClass().getDeclaredField(name);
						if (DEBUG)
							System.out.println("[Groesse] - startElement: Feld(" + name + ") Typ(" + field.getType()
									+ ") Inhalt(" + inhalt + ")");
						if (name.equals("bezug") && gl != null)
							bezug = Util.findeGroesse(gl, inhalt);
						else if (name.equals("bezug2") && gl != null)
							bezug = Util.findeGroesse(gl, inhalt);
						else
							Util.setFieldValue(this, field, inhalt);
					} catch (SecurityException e) {
						if (DEBUG)
							e.printStackTrace();
						throw new SAXException(e.getMessage());
					} catch (NoSuchFieldException e) {
						if (DEBUG)
							e.printStackTrace();
						throw new SAXException(e.getMessage());
					} catch (IllegalArgumentException e) {
						if (DEBUG)
							e.printStackTrace();
						throw new SAXException(e.getMessage());
					}
				}
			} else
				throw new SAXException("[Messparameter] - Bereich(" + bereich + ") Start(" + qName + ")");
			break;
		case GROESSE:
			if (qName.equals("Messbereich")) {
				bereich = BEREICH.MESSBEREICH;
				messbereich = new Messbereich(0);
				messbereich = messbereich.startElement(uri, localName, qName, attributes);
			} else if (qName.equals("Messgroesse")) {
				bereich = BEREICH.MESSGROESSE;
				messGroesse = new Messgroesse();
				messGroesse = messGroesse.startElement(uri, localName, qName, attributes);
			}
			break;
		case MESSBEREICH:
			messbereich = messbereich.startElement(uri, localName, qName, attributes);
			break;
		case MESSGROESSE:
			messGroesse = messGroesse.startElement(uri, localName, qName, attributes);
			break;
		}
	}

	public void endElement(String uri, String localName, String qName, String inhalt) throws SAXException {
		switch (bereich) {
		default:
			throw new SAXException("[Groesse] - Bereich(" + bereich + ") Ende(" + qName + ")");
		case GROESSE:
			if (qName.equals("Groesse")) {
				if (art == Art.Messgroesse) {
					AdapterListe al = GHMessen.getApplication().getMainView().al;
					if (al == null || !(al.containsKey(nr)) || !(al.get(nr) instanceof ICASSY))
						art = Art.Parameter;
				}
				bereich = BEREICH.ANFANG;
				break;
			}
			break;
		case MESSBEREICH:
			messbereich.endElement(uri, localName, qName, inhalt);
			if (qName.equals("Messbereich")) {
				bereich = BEREICH.GROESSE;
			}
			break;
		case MESSGROESSE:
			messGroesse.endElement(uri, localName, qName, inhalt);
			if (qName.equals("Messgroesse"))
				bereich = BEREICH.GROESSE;
			break;
		}
	}

	public Groesse(String name, String symbol, Einheit einheit, boolean abhaengig) {
		super();
		aktiv = true;
		manuellerName = false;
		art = Art.Parameter;
		von = -10;
		bis = 10;
		stellen = 2;
		xAchse = T;
		nr = 0;
		AB = -1;
		EF = -1;
		bezug = null;
		bezug2 = null;
		formel = "0";
		formelterm = null;
		synchron = false;
		intervall = GHMessen.getApplication().ghmMittelZeit;
		mweinheit = null;
		delta = 1;
		messGroesse = Spannung;
		wert = 0;
		offset = 0;
		faktor = 1;
		messbereich = Messbereich.U10V;
		nullpunkt = Nullpunkt.mittig;
		aufnahme = Aufnahme.momentan;
		zeitraum = 0;
		nichtImmerAktiv = false;
		ausloeserTriggert = false;
		schmalesSichtfeld = false;
		start1Flanke = false;
		stoppFlanke = false;
		flanke = 0;
		invertieren = false;
		pendel = false;
		ueberschreiben = true;
		fahne = 1.0;
		radius = 1.0;
		anordnung = 0;
		d_pm = 564.0; /* NaCl: 2d */
		e_keV = 1239.8; /* E in keV pro pm */
		form = Waveforms.DC;
		formSynchron = false;
		amplitude = 0;
		formOffset = 0;
		anstieg = 50;
		frequenz = 1.0;
		messreihe = 0;
		this.name = name;
		this.symbol = symbol;
		this.einheit = einheit;
		this.abhaengig = abhaengig;
	}

	public Groesse(String name, String symbol) {
		this(name, symbol, Einheit.m, false);
	}

	public Groesse(String symbol) {
		this("neue Größe", symbol, Einheit.m, false);
	}

	public Groesse(int nr, int ab, int ef) {
		this();
		this.nr = nr;
		this.AB = ab;
		this.EF = ef;
	}

	public Groesse(int nr, ICASSYQuantity q, XAchseTyp xAchse) {
		this();
		this.nr = nr;
		this.AB = q.getAB();
		this.name = q.getToolTipText();
		this.manuellerName = true;
		this.symbol = MathString.replaceAllGreekLetters(q.getSymbol());
		String unit = q.getUnit();
		if (unit == "")
			this.einheit = Einheit.Keine;
		else
			this.einheit = Einheit.toEinheit(q.getUnit());
		String s = q.getSymbol();
		if (s.indexOf("_") > -1)
			s = s.substring(0, s.indexOf("_"));
		this.messGroesse = Messgroesse.toMessgroesse(q.getName(), q.getUnit(), s);
		this.messbereich = new Messbereich(q.getRange());
		this.von = messbereich.getMinimum();
		this.bis = messbereich.getMaximum();
		this.stellen = messbereich.getStellen();
		if (xAchse == null)
			this.xAchse = T;
		else
			this.xAchse = xAchse;
		this.abhaengig = false;
	}

	public Groesse() {
		this("neue Größe", "f_1", Einheit.m, false);
		bereich = BEREICH.ANFANG;
	}

	public Groesse(Groesse g) {
		super();
		aktiv = g.aktiv;
		name = g.name;
		manuellerName = g.manuellerName;
		art = g.art;
		symbol = g.symbol;
		einheit = g.einheit;
		von = g.von;
		bis = g.bis;
		stellen = g.stellen;
		xAchse = g.xAchse;
		nr = g.nr;
		AB = g.AB;
		EF = g.EF;
		bezug = g.bezug;
		bezug2 = g.bezug2;
		formel = g.formel;
		formelterm = g.formelterm;
		synchron = g.synchron;
		intervall = g.intervall;
		mweinheit = g.mweinheit;
		delta = g.delta;
		messGroesse = g.messGroesse;
		wert = g.wert;
		offset = g.offset;
		faktor = g.faktor;
		messbereich = g.messbereich;
		nullpunkt = g.nullpunkt;
		aufnahme = g.aufnahme;
		zeitraum = g.zeitraum;
		nichtImmerAktiv = g.nichtImmerAktiv;
		ausloeserTriggert = g.ausloeserTriggert;
		schmalesSichtfeld = g.schmalesSichtfeld;
		start1Flanke = g.start1Flanke;
		stoppFlanke = g.stoppFlanke;
		flanke = g.flanke;
		invertieren = g.invertieren;
		pendel = g.pendel;
		ueberschreiben = g.ueberschreiben;
		fahne = g.fahne;
		radius = g.radius;
		anordnung = g.anordnung;
		d_pm = g.d_pm;
		e_keV = g.e_keV;
		form = g.form;
		formSynchron = g.formSynchron;
		amplitude = g.amplitude;
		formOffset = g.formOffset;
		anstieg = g.anstieg;
		frequenz = g.frequenz;
		abhaengig = g.abhaengig;
		messreihe = g.messreihe;

	}

	public static Groesse createFromCL4(Groessenliste gl, String[] cl4, int messReihe) throws FileFormatException {
		Groesse ret = new Groesse();
		Scanner a0 = new Scanner(cl4[0]);
		a0.useLocale(Locale.US);
		int artInt = a0.nextInt();
		double typ = a0.nextDouble();
		ret.setArt(Art.fromCL4(artInt, typ));
		ret.setName(Util.findUniqueName(gl, cl4[2], true));
		Messgroesse m = null;
		ret.setManuellerName(true);
		if (ret.art == Art.Messgroesse) {
			try {
				m = Messgroesse.toMessgroesse(cl4[2]);
				ret.setSymbol(CL4File.getSymbolCL4(cl4[3], m) + (messReihe == 0 ? "" : "#" + messReihe));
			} catch (Exception e) {
				ret.setSymbol(CL4File.getSymbolCL4(cl4[3]) + (messReihe == 0 ? "" : "#" + messReihe));
			}
		} else {
			ret.setSymbol(CL4File.getSymbolCL4(cl4[3]) + (messReihe == 0 ? "" : "#" + messReihe));
		}
		ret.abhaengig = ret.symbol.contains("#");
		ret.messreihe = messReihe;
		int bezug = -1;
		switch (ret.art) {
		case Klasse:
			bezug = (int) typ;
			break;
		case Messgroesse:
			ret.setFaktor(typ);
			ret.setOffset(a0.nextDouble());
			if (a0.hasNext())
				ret.setFormel(CL4File.getFormelCL4(gl, ret.symbol, ret.getSymbolCL4(), a0.nextLine()));
			break;
		default:
			bezug = a0.nextInt();
			if (a0.hasNext())
				ret.setFormel(CL4File.getFormelCL4(gl, ret.symbol, ret.getSymbolCL4(), a0.nextLine()));
			break;
		}
		if (ret.art == Art.Messgroesse && m != null)
			m.setzeStandardWerte(ret);
		if (ret.art == Art.Messgroesse)
			ret.setArt(Art.Parameter);
		if (bezug > -1 && bezug < gl.size()) {
			if (messReihe == 0)
				ret.setBezug(gl.get(bezug));
			else
				ret.setBezug(gl, gl.get(bezug).symbol, messReihe);
		}
		try {
			ret.setEinheit(Einheit.toEinheit(cl4[4]));
		} catch (IllegalArgumentException e) {
			if (DEBUG)
				System.out.println("Groesse.setFromCL4: Die Einheit(" + cl4[4] + ") existiert nicht.");
			ret.setEinheit(Einheit.Keine);
		}
		CL4File.leseDoubleListe(cl4[1], 29, 0, true);
		Double[] parameter16 = CL4File.leseDoubleListe(cl4[5], 16, 0, true);
		ret.setVon(parameter16[0]);
		ret.setBis(parameter16[1]);
		ret.setStellen(parameter16[4].intValue());
		ret.setWert(parameter16[13] != null ? parameter16[13] : 0.0);
		a0.close();
		return ret;
	}

	public Groesse copy(int nr) {
		Groesse g = new Groesse(this);
		g.symbol += "#" + nr;
		g.abhaengig = true;
		g.messreihe = nr;
		return g;
	}

	public void korrigiereWerte(Groessenliste gl) {
		String alt = symbol.substring(0, symbol.indexOf("#"));
		if (name.contains(alt))
			setName(name.replaceAll(alt, symbol));
		if (bezug != null)
			setBezug(Util.findeGroesse(gl, bezug.symbol + "#" + messreihe));
		if (bezug2 != null)
			setBezug2(Util.findeGroesse(gl, bezug2.symbol + "#" + messreihe));
		if (art == Art.Formel && formel != null) {
			String[] find = gl.getTextSymbole(0);
			String[] replace = new String[find.length];
			for (int i = 0; i < find.length; i++)
				replace[i] = find[i] + "#" + messreihe;
			String s = formel;
			for (int i = 0; i < find.length; i++)
				s = s.replaceAll(find[i], replace[i]);
			setFormel(s);
			if (p == null)
				p = new FormelGenerator();
			p.setName(symbol);
			p.setKonstanten(Arrays.copyOfRange(gl.getTextSymbole(), 0, gl.indexOf(this)));
			if (p.parsen(formel))
				formelterm = p.getAusdruck(formel);
		}

	}

	public int getMessreihe() {
		if (abhaengig)
			try {
				return Integer.parseInt(symbol.substring(symbol.indexOf("#") + 1));
			} catch (NumberFormatException e) {
				return -1;
			}
		else
			return 0;
	}

	public void setMessreihe(int messreihe) {
		int alt = this.messreihe;
		this.messreihe = messreihe;
		support.firePropertyChange("messreihe", alt, messreihe);
	}

	@Override
	public String toString() {
		return name;
	}

	public void setName(String name) {
		String alt = this.name;
		this.name = name;
		support.firePropertyChange("name", alt, name);
	}

	public boolean isManuellerName() {
		return manuellerName;
	}

	public void setManuellerName(boolean manuellerName) {
		this.manuellerName = manuellerName;
	}

	public String getName() {
		return name;
	}

	public void setArt(Art art) {
		Art alt = this.art;
		this.art = art;
		if (!manuellerName) {
			if (art == Art.Messgroesse)
				setName((messGroesse.getName() == null ? "" : messGroesse.getName() + " ") + symbol);
			else
				setName(art.toString() + " " + symbol);
		}
		support.firePropertyChange("art", alt, art);
	}

	public Art getArt() {
		return art;
	}

	public void setSymbol(String symbol) {
		String alt = this.symbol;
		this.symbol = symbol;
		if (!manuellerName) {
			if (art == Art.Messgroesse)
				setName((messGroesse.getName() == null ? "" : messGroesse.getName() + " ") + symbol);
			else
				setName(art.toString() + " " + symbol);
		}
		support.firePropertyChange("symbol", alt, symbol);
	}

	public String getSymbol() {
		return symbol;
	}

	public String getTextSymbol() {
		return StringUtil.toSimpleText(symbol);
	}

	public String getSymbolCL4() {
		return abhaengig ? symbol.substring(0, symbol.indexOf("#")).replace("_", "") : symbol.replaceAll("_", "");
	}

	public SymbolNameEinheit getSymbolNameEinheit() {
		return new SymbolNameEinheit(symbol, einheit, stellen, von, bis);
	}

	public void setEinheit(Einheit einheit) {
		Einheit alt = this.einheit;
		this.einheit = einheit;
		support.firePropertyChange("einheit", alt, einheit);
	}

	public Einheit getEinheit() {
		return einheit;
	}

	public void setEinheitString(String einheit) {
		Einheit e = Einheit.toEinheit(einheit);
		setEinheit(e);
	}

	public String getEinheitString() {
		return einheit == null ? "" : einheit.toString();
	}

	public void setVon(double von) {
		double alt = this.von;
		this.von = von;
		support.firePropertyChange("von", alt, von);
	}

	public double getVon() {
		return von;
	}

	public boolean isAktiv() {
		return aktiv;
	}

	public void setAktiv(boolean an) {
		if (DEBUG)
			System.out.println("[Groesse] - setAktiv: groesse(" + symbol + ") an(" + an + ")");
		boolean alt = this.aktiv;
		aktiv = an;
		support.firePropertyChange("aktiv", alt, an);
	}

	public void setBis(double bis) {
		double alt = this.bis;
		this.bis = bis;
		support.firePropertyChange("bis", alt, bis);
	}

	public double getBis() {
		return bis;
	}

	public void setStellen(int stellen) {
		int alt = this.stellen;
		this.stellen = stellen;
		support.firePropertyChange("stellen", alt, stellen);
	}

	public int getStellen() {
		return stellen;
	}

	public XAchseTyp getXAchse() {
		return xAchse;
	}

	public void setXAchse(XAchseTyp xachse) {
		XAchseTyp alt = this.xAchse;
		this.xAchse = xachse;
		support.firePropertyChange("xAchse", alt, xachse);
	}

	public void setNr(int nr) {
		int alt = this.nr;
		this.nr = nr;
		support.firePropertyChange("nr", alt, nr);
	}

	public int getNr() {
		return nr;
	}

	public void setAB(int aB) {
		int alt = this.AB;
		this.AB = aB;
		support.firePropertyChange("aB", alt, AB);
	}

	public int getAB() {
		return AB;
	}

	public void setEF(int eF) {
		int alt = this.EF;
		EF = eF;
		support.firePropertyChange("eF", alt, eF);
	}

	public int getEF() {
		return EF;
	}

	public void setBezug(Groessenliste gl, String symbol, int messReihe) {
		symbol = symbol + "#" + messReihe;
		setBezug(Util.findeGroesse(gl, symbol));
	}

	public void setBezug(Groesse bezug) {
		Groesse alt = this.bezug;
		this.bezug = bezug;
		if (art == Art.Histogramm && bezug2 != null)
			bezug2.setBezug(bezug);
		if (art == Art.Klasse && bezug != null)
			setSymbol(bezug.getSymbol() + "_^^'");
		support.firePropertyChange("bezug", alt, bezug);
	}

	public Groesse getBezug() {
		return bezug;
	}

	public void setBezug2(Groesse bezug2) {
		Groesse alt = this.bezug2;
		this.bezug2 = bezug2;
		support.firePropertyChange("bezug2", alt, bezug2);
	}

	public Groesse getBezug2() {
		return bezug2;
	}

	public void setFormel(String formel) {
		String alt = this.formel;
		this.formel = formel;
		if (!alt.equals(formel))
			setFormelterm(null);
		support.firePropertyChange("formel", alt, formel);
	}

	public String getFormel() {
		return formel;
	}

	public boolean isFormelValid(String[] konstanten) {
		if (p == null)
			return false;
		if (konstanten != null)
			p.setKonstanten(konstanten);
		return p.parsen(formel, false);
	}

	public void setIntervall(double intervall) {
		double alt = this.intervall;
		this.intervall = intervall;
		support.firePropertyChange("intervall", alt, intervall);
	}

	public double getIntervall() {
		return intervall;
	}

	public void setMweinheit(Einheit mweinheit) {
		Einheit alt = this.mweinheit;
		this.mweinheit = mweinheit;
		support.firePropertyChange("mweinheit", alt, mweinheit);
	}

	public Einheit getMweinheit() {
		return mweinheit;
	}

	public void setDelta(double delta) {
		double alt = this.delta;
		this.delta = delta;
		if (art == Art.Histogramm && bezug2 != null)
			bezug2.setDelta(delta);
		support.firePropertyChange("delta", alt, delta);
	}

	public double getDelta() {
		return delta;
	}

	public void setAbhaengig(boolean abhaengig) {
		boolean alt = this.abhaengig;
		this.abhaengig = abhaengig;
		support.firePropertyChange("abhaengig", alt, abhaengig);
	}

	public boolean isAbhaengig() {
		return abhaengig;
	}

	public void setOffset(double offset) {
		double alt = this.offset;
		this.offset = offset;
		support.firePropertyChange("offset", alt, offset);
	}

	public double getOffset() {
		return offset;
	}

	public void setFaktor(double faktor) {
		double alt = this.faktor;
		this.faktor = faktor;
		support.firePropertyChange("faktor", alt, faktor);
	}

	public double getFaktor() {
		return faktor;
	}

	public void setMessbereich(Messbereich messbereich) {
		Messbereich alt = this.messbereich;
		this.messbereich = messbereich;
		support.firePropertyChange("messbereich", alt, messbereich);
		setEinheit(messbereich.getEinheit());
		if (alt.getMaximum() < messbereich.getMinimum()) {
			setBis(messbereich.getMaximum());
			setVon(messbereich.getMinimum());
		} else {
			setVon(messbereich.getMinimum());
			setBis(messbereich.getMaximum());
		}
		setStellen(messbereich.getStellen());
	}

	public Messbereich getMessbereich() {
		return messbereich;
	}

	public void setNullpunkt(Nullpunkt nullpunkt) {
		Nullpunkt alt = this.nullpunkt;
		this.nullpunkt = nullpunkt;
		support.firePropertyChange("nullpunkt", alt, nullpunkt);
	}

	public Nullpunkt getNullpunkt() {
		return nullpunkt;
	}

	public void setAufnahme(Aufnahme aufnahme) {
		Aufnahme alt = this.aufnahme;
		this.aufnahme = aufnahme;
		support.firePropertyChange("aufnahme", alt, aufnahme);
	}

	public Aufnahme getAufnahme() {
		return aufnahme;
	}

	public void setZeitraum(double zeitraum) {
		double alt = this.zeitraum;
		this.zeitraum = zeitraum;
		support.firePropertyChange("zeitraum", alt, zeitraum);
	}

	public void setZeitraumString(String zeitraum) {
		if (zeitraum.endsWith("  μs"))
			setZeitraum(Double.parseDouble(zeitraum.substring(0, zeitraum.length() - 3)) / 1000000.0);
		else if (zeitraum.endsWith(" ms"))
			setZeitraum(Double.parseDouble(zeitraum.substring(0, zeitraum.length() - 3)) / 1000.0);
		else
			setZeitraum(Double.parseDouble(zeitraum));
	}

	public double getZeitraum() {
		return zeitraum;
	}

	public String getZeitraumString() {
		if (zeitraum < 0.001)
			return zeitraum * 1000000 + " μs";
		else if (zeitraum < 1)
			return zeitraum * 1000 + " ms";
		else
			return zeitraum + " s";
	}

	public void setWert(double wert) {
		double alt = this.wert;
		this.wert = wert;
		support.firePropertyChange("wert", alt, wert);
	}

	public double getWert() {
		return wert;
	}

	public void setFormelterm(ExpressionEvaluator formelterm) {
		ExpressionEvaluator alt = this.formelterm;
		this.formelterm = formelterm;
		support.firePropertyChange("formelterm", alt, formelterm);
	}

	public void createFormelterm(String[] abhaengig) {
		if (formelterm == null) {
			p.setName(symbol);
			p.setKonstanten(abhaengig);
			if (p.parsen(formel)) {
				setFormelterm(p.getAusdruck(formel));
			} else {
				setFormelterm(FormelGenerator.NULLAUSDRUCK);
			}
		}
	}

	public ExpressionEvaluator getFormelterm() {
		return formelterm;
	}

	public void setNichtImmerAktiv(boolean nichtImmerAktiv) {
		boolean alt = this.nichtImmerAktiv;
		this.nichtImmerAktiv = nichtImmerAktiv;
		support.firePropertyChange("nichtImmerAktiv", alt, nichtImmerAktiv);
	}

	public boolean isNichtImmerAktiv() {
		return nichtImmerAktiv;
	}

	public void setAusloeserTriggert(boolean ausloeserTriggert) {
		boolean alt = this.ausloeserTriggert;
		this.ausloeserTriggert = ausloeserTriggert;
		support.firePropertyChange("ausloeserTriggert", alt, ausloeserTriggert);
	}

	public boolean isAusloeserTriggert() {
		return ausloeserTriggert;
	}

	public void setSchmalesSichtfeld(boolean schmalesSichtfeld) {
		boolean alt = this.schmalesSichtfeld;
		this.schmalesSichtfeld = schmalesSichtfeld;
		support.firePropertyChange("schmalesSichtfeld", alt, schmalesSichtfeld);
	}

	public boolean isSchmalesSichtfeld() {
		return schmalesSichtfeld;
	}

	public void setMessGroesse(Messgroesse messGroesse) {
		Messgroesse alt = this.messGroesse;
		this.messGroesse = messGroesse;
		if (!manuellerName) {
			if (art == Art.Messgroesse)
				setName((messGroesse.getName() == null ? "" : messGroesse.getName() + " ") + symbol);
			else
				setName(art.toString() + " " + symbol);
		}
		support.firePropertyChange("messGroesse", alt, messGroesse);
	}

	public Messgroesse getMessGroesse() {
		return messGroesse;
	}

	public void setStart1Flanke(boolean start1Flanke) {
		boolean alt = this.start1Flanke;
		this.start1Flanke = start1Flanke;
		support.firePropertyChange("start1Flanke", alt, start1Flanke);
	}

	public boolean isStart1Flanke() {
		return start1Flanke;
	}

	public void setStoppFlanke(boolean stoppFlanke) {
		boolean alt = this.stoppFlanke;
		this.stoppFlanke = stoppFlanke;
		support.firePropertyChange("stoppFlanke", alt, stoppFlanke);
	}

	public boolean isStoppFlanke() {
		return stoppFlanke;
	}

	public void setFlanke(int flanke) {
		int alt = this.flanke;
		this.flanke = flanke;
		support.firePropertyChange("flanke", alt, flanke);
	}

	public int getFlanke() {
		return flanke;
	}

	public void setInvertieren(boolean invertieren) {
		boolean alt = this.invertieren;
		this.invertieren = invertieren;
		support.firePropertyChange("invertieren", alt, invertieren);
	}

	public boolean isInvertieren() {
		return invertieren;
	}

	public void setPendel(boolean pendel) {
		boolean alt = this.pendel;
		this.pendel = pendel;
		support.firePropertyChange("pendel", alt, pendel);
	}

	public boolean isPendel() {
		return pendel;
	}

	public void setUeberschreiben(boolean ueberschreiben) {
		boolean alt = this.ueberschreiben;
		this.ueberschreiben = ueberschreiben;
		support.firePropertyChange("ueberschreiben", alt, ueberschreiben);
	}

	public boolean isUeberschreiben() {
		return ueberschreiben;
	}

	public void setAnordnung(int anordnung) {
		int alt = this.anordnung;
		this.anordnung = anordnung;
		support.firePropertyChange("anordnung", alt, anordnung);
	}

	public int getAnordnung() {
		return anordnung;
	}

	public void setSynchron(boolean synchron) {
		boolean alt = this.synchron;
		this.synchron = synchron;
		support.firePropertyChange("synchron", alt, synchron);
	}

	public boolean isSynchron() {
		return synchron;
	}

	public void setFahne(double fahne) {
		double alt = this.fahne;
		this.fahne = fahne;
		support.firePropertyChange("fahne", alt, fahne);
	}

	public double getFahne() {
		return fahne;
	}

	public void setRadius(double radius) {
		double alt = this.radius;
		this.radius = radius;
		support.firePropertyChange("radius", alt, radius);
	}

	public double getRadius() {
		return radius;
	}

	public void setD_pm(double d) {
		double alt = this.d_pm;
		this.d_pm = d;
		support.firePropertyChange("d_pm", alt, d_pm);
	}

	public double getD_pm() {
		return d_pm;
	}

	public void setE_keV(double e) {
		double alt = this.e_keV;
		this.e_keV = e;
		support.firePropertyChange("e_keV", alt, e_keV);
	}

	public double getE_keV() {
		return e_keV;
	}

	public void setForm(Waveforms form) {
		Waveforms alt = this.form;
		this.form = form;
		support.firePropertyChange("form", alt, form);
	}

	public Waveforms getForm() {
		return form;
	}

	public void setFormSynchron(boolean formSynchron) {
		boolean alt = this.formSynchron;
		this.formSynchron = formSynchron;
		support.firePropertyChange("formSynchron", alt, formSynchron);
	}

	public boolean isFormSynchron() {
		return formSynchron;
	}

	public void setAmplitude(double amplitude) {
		double alt = this.amplitude;
		this.amplitude = amplitude;
		support.firePropertyChange("amplitude", alt, amplitude);
	}

	public double getAmplitude() {
		return amplitude;
	}

	public void setFormOffset(double formOffset) {
		double alt = this.formOffset;
		this.formOffset = formOffset;
		support.firePropertyChange("formoffset", alt, formOffset);
	}

	public double getFormOffset() {
		return formOffset;
	}

	public void setAnstieg(double anstieg) {
		double alt = this.anstieg;
		this.anstieg = anstieg;
		support.firePropertyChange("anstieg", alt, anstieg);
	}

	public double getAnstieg() {
		return anstieg;
	}

	public void setFrequenz(double frequenz) {
		double alt = this.frequenz;
		this.frequenz = frequenz;
		support.firePropertyChange("frequenz", alt, frequenz);
	}

	public double getFrequenz() {
		return frequenz;
	}

	public boolean istXRayGroesse() {
		AdapterListe al = GHMessen.getApplication().getMainView().al;
		if (al == null)
			return false;
		return al.get(nr) instanceof IXRay;
	}
}
