/* 
 * GHMessen
 * --------
 * 
 * GHMessen is a measuring program to use in connection with a computer 
 * measuring interface. It's designed for school usage.
 * 
 * Copyright (C) 2010-2011  Frank Schütte - Gymnasium Himmelsthür
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms
 * of the GNU General Public License as published by the Free Software Foundation; either 
 * version 3 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with this program; 
 * if not, see <http://www.gnu.org/licenses/>.
 * 
 * Frank Schütte fschuett AT gymnasium-himmelsthuer DOT de
 * 
 */

/**
 * Die Klasse Groessenliste enthält eine Liste von Größen, die als DataModel für
 * eine ComboBox verwendet werden können.
 * Außerdem enthält sie das Interface TableModel, um auf die Daten jeder enthaltenen
 * Größe zugreifen zu können.
 */

package ghmessen.groesse;

import static ghmessen.groesse.Messgroesse.Drehstoß;
import static ghmessen.groesse.Messgroesse.Dunkelzeit;
import static ghmessen.groesse.Messgroesse.Laufzeit;
import static ghmessen.groesse.Messgroesse.LinearerStoß;

import ghmessen.GHMessen;
import ghmessen.adapter.AdapterListe;
import ghmessen.adapter.IMesskanal;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.PrintWriter;
import java.io.Serializable;
import java.util.AbstractList;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.Set;
import java.util.Vector;
import java.util.concurrent.CopyOnWriteArrayList;

import javax.xml.transform.sax.TransformerHandler;

import ld.api.ITimer;
import ld.api.IXRay;
import util.StringUtil;

import org.jdesktop.observablecollections.ObservableListListener;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;

public class Groessenliste extends AbstractList<Groesse>
		implements PropertyChangeListener, ObservableObjectList<Groesse>, Serializable {
	private static final long serialVersionUID = 8530533245187684853L;
	private static final boolean DEBUG = false;

	private class SymbolPropertyHandler implements Serializable, PropertyChangeListener {
		/**
		 * 
		 */
		private static final long serialVersionUID = 1L;

		@Override
		public void propertyChange(PropertyChangeEvent evt) {
			handleSymbolEvent(evt);
		}
	}

	private class NamePropertyHandler implements Serializable, PropertyChangeListener {
		/**
		 * 
		 */
		private static final long serialVersionUID = 1L;

		@Override
		public void propertyChange(PropertyChangeEvent evt) {
			handleNameEvent(evt);
		}
	}

	private class ArtPropertyHandler implements Serializable, PropertyChangeListener {
		/**
		 * 
		 */
		private static final long serialVersionUID = 1L;

		@Override
		public void propertyChange(PropertyChangeEvent evt) {
			handleArtEvent(evt);
		}
	}

	private final boolean supportsElementPropertyChanged;
	private List<Groesse> list;
	private List<String> symbole = new ArrayList<String>();
	private List<String> namen = new ArrayList<String>();
	private transient List<ObservableListListener> listeners;

	protected SymbolPropertyHandler symbolPH;
	protected NamePropertyHandler namePH;
	protected ArtPropertyHandler artPH;

	public static void createDTD(PrintWriter out) {
		out.println("<!ELEMENT Groessenliste (GlGroesse*)>");
		out.println("<!ELEMENT GlGroesse (Groesse,GlSymbol,GlName)>");
		Groesse.createDTD(out);
		out.println("<!ELEMENT GlSymbol (#PCDATA)>");
		out.println("<!ELEMENT GlName (#PCDATA)>");
	}

	public void writeToXml(TransformerHandler hd) throws SAXException {
		if (hd == null)
			throw new SAXException("[Groessenliste] Objekt ist nicht initialisiert.");
		hd.startElement("", "", "Groessenliste", null);
		for (int index = 0; index < list.size(); index++) {
			hd.startElement("", "", "GlGroesse", null);
			list.get(index).writeToXml(hd);
			hd.startElement("", "", "GlSymbol", null);
			hd.characters(symbole.get(index).toCharArray(), 0, symbole.get(index).length());
			hd.endElement("", "", "GlSymbol");
			hd.startElement("", "", "GlName", null);
			hd.characters(namen.get(index).toCharArray(), 0, namen.get(index).length());
			hd.endElement("", "", "GlName");
			hd.endElement("", "", "GlGroesse");
		}
		hd.endElement("", "", "Groessenliste");
	}

	private static enum BEREICH {
		ANFANG, GROESSENLISTE, GLGROESSE, GLSYMBOL, GLNAME
	};

	private transient BEREICH bereich;

	public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException {
		switch (bereich) {
		default:
		case ANFANG:
			if (qName.equals("Groessenliste")) {
				bereich = BEREICH.GROESSENLISTE;
			} else
				throw new SAXException("[Messparameter] - Bereich(" + bereich + ") Start(" + qName + ")");
			break;
		case GROESSENLISTE:
			if (qName.equals("GlGroesse")) {
				bereich = BEREICH.GLGROESSE;
				add(new Groesse());
			}
			break;
		case GLGROESSE:
			if (qName.equals("GlSymbol"))
				bereich = BEREICH.GLSYMBOL;
			else if (qName.equals("GlName"))
				bereich = BEREICH.GLNAME;
			else
				list.get(list.size() - 1).startElement(uri, localName, qName, attributes, this);
			break;
		}
	}

	public void endElement(String uri, String localName, String qName, String inhalt) throws SAXException {
		switch (bereich) {
		default:
			throw new SAXException("[Groessenliste] - Bereich(" + bereich + ") Ende(" + qName + ")");
		case GROESSENLISTE:
			if (qName.equals("Groessenliste")) {
				for (Groesse g : this)
					if (g.getArt() == Art.Formel)
						g.createFormelterm(this.getTextSymbole());
				bereich = BEREICH.ANFANG;
			}
			break;
		case GLGROESSE:
			if (qName.equals("GlGroesse")) {
				bereich = BEREICH.GROESSENLISTE;
			} else
				list.get(list.size() - 1).endElement(uri, localName, qName, inhalt);
			break;
		case GLSYMBOL:
			if (qName.equals("GlSymbol")) {
				bereich = BEREICH.GLGROESSE;
			} else
				throw new SAXException("[Groessenliste] - Bereich(" + bereich + ") Ende(" + qName + ")");
			break;
		case GLNAME:
			if (qName.equals("GlName")) {
				bereich = BEREICH.GLGROESSE;
			} else
				throw new SAXException("[Groessenliste] - Bereich(" + bereich + ") Ende(" + qName + ")");
			break;
		}
	}

	Groessenliste(List<Groesse> list, boolean supportsElementPropertyChanged) {
		this.list = list;
		this.bereich = BEREICH.ANFANG;
		symbolPH = new SymbolPropertyHandler();
		namePH = new NamePropertyHandler();
		artPH = new ArtPropertyHandler();
		for (Groesse g : list) {
			g.addPropertyChangeListener("symbol", symbolPH);
			g.addPropertyChangeListener("name", namePH);
			g.addPropertyChangeListener("art", artPH);
			g.addPropertyChangeListener(this);
		}
		listeners = new CopyOnWriteArrayList<ObservableListListener>();
		this.supportsElementPropertyChanged = supportsElementPropertyChanged;
		if (DEBUG)
			System.out.println("[Groessenliste] - Konstruktor(list,boolean)...");
	}

	private void writeObject(ObjectOutputStream out) throws IOException {
		out.defaultWriteObject();
	}

	private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {
		in.defaultReadObject();
		restoreListenings();
		listeners = new CopyOnWriteArrayList<ObservableListListener>();
	}

	private void restoreListenings() {
		for (Groesse g : list) {
			g.addPropertyChangeListener("symbol", symbolPH);
			g.addPropertyChangeListener("name", namePH);
			g.addPropertyChangeListener("art", artPH);
			g.addPropertyChangeListener(this);
		}
	}

	public Groessenliste() {
		this(new Vector<Groesse>(), true);
	}

	public Groessenliste(Vector<Groesse> v) {
		this(v, true);
	}

	public String[] getNamen() {
		return namen.toArray(new String[] {});
	}

	public int indexOfName(String name) {
		return namen.indexOf(name);
	}

	public String[] getSymbole() {
		return symbole.toArray(new String[] {});
	}

	public String[] getSymbole(int nr) {
		Vector<String> mrSymbole = new Vector<String>();
		for (Groesse g : this)
			if (g.getMessreihe() == nr)
				mrSymbole.add(g.getSymbol());
		return mrSymbole.toArray(new String[] {});
	}

	public String[] getTextSymbole() {
		String[] ret = getSymbole();
		for(int i=0;i<ret.length;i++)
			ret[i] = StringUtil.toSimpleText(ret[i]);
		return ret;
	}
	
	public String[] getTextSymbole(int nr) {
		Vector<String> mrSymbole = new Vector<String>();
		for (Groesse g : this)
			if (g.getMessreihe() == nr)
				mrSymbole.add(g.getTextSymbol());
		return mrSymbole.toArray(new String[] {});
	}
	
	public boolean enthaeltSymbol(String symbol) {
		return symbole.contains(symbol);
	}

	public int indexOfSymbol(String symbol) {
		return symbole.indexOf(symbol);
	}

	public boolean hatOszilloskopGroesse() {
		AdapterListe al = GHMessen.getApplication().getMainView().al;
		if (al == null)
			return false;
		IMesskanal messkanal = null;
		for (Groesse g : this)
			try {
				if (g.getArt() == Art.Messgroesse) {
					messkanal = al.get(g.getNr()).getMesskanal(g);
					if (messkanal != null && messkanal.istOszilloskopKanal())
						return true;
				}
			} catch (ArrayIndexOutOfBoundsException e) {
			}
		return false;
	}

	public String[] getOszilloskopSymbole() {
		AdapterListe al = GHMessen.getApplication().getMainView().al;
		Vector<String> result = new Vector<String>();
		if (al == null)
			return new String[] {};
		IMesskanal messkanal = null;
		for (Groesse g : this)
			try {
				if (g.getArt() == Art.Messgroesse) {
					messkanal = al.get(g.getNr()).getMesskanal(g);
					if (messkanal != null && messkanal.istOszilloskopKanal())
						result.add(g.getSymbol());
				}
			} catch (ArrayIndexOutOfBoundsException e) {
			}
		return result.toArray(new String[] {});
	}

	public String[] getTriggerSymbole() {
		AdapterListe al = GHMessen.getApplication().getMainView().al;
		Vector<String> result = new Vector<String>();
		if (al == null)
			return new String[] {};
		IMesskanal messkanal = null;
		for (Groesse g : this)
			try {
				if (g.getArt() == Art.Messgroesse) {
					messkanal = al.get(g.getNr()).getMesskanal(g);
					if (messkanal != null && messkanal.istTriggerKanal())
						result.add(g.getSymbol());
				}
			} catch (ArrayIndexOutOfBoundsException e) {
			}
		return result.toArray(new String[] {});
	}

	public boolean hatXRayGroesse() {
		AdapterListe al = GHMessen.getApplication().getMainView().al;
		if (al == null)
			return false;
		for (Groesse g : this)
			if (g.getArt() == Art.Messgroesse && al.get(g.getNr()) instanceof IXRay)
				return true;
		return false;
	}

	public Groesse getXRayGroesse() {
		AdapterListe al = GHMessen.getApplication().getMainView().al;
		if (al == null)
			return null;
		for (Groesse g : this)
			if (al.get(g.getNr()) instanceof IXRay)
				return g;
		return null;
	}

	public boolean hatTimerGroesse() {
		AdapterListe al = GHMessen.getApplication().getMainView().al;
		if (al == null)
			return false;
		for (Groesse g : this)
			try {
				if (g.getArt() == Art.Messgroesse && al.get(g.getNr()) instanceof ITimer)
					return true;
			} catch (ArrayIndexOutOfBoundsException e) {
			}
		return false;
	}

	public boolean hatTimerBoxGroesse() {
		AdapterListe al = GHMessen.getApplication().getMainView().al;
		if (al == null)
			return false;
		for (Groesse g : this)
			try {
				if (g.getArt() == Art.Messgroesse && al.get(g.getNr()) instanceof ITimer) {
					if (g.getMessGroesse() == Laufzeit || g.getMessGroesse() == Dunkelzeit
							|| g.getMessGroesse() == LinearerStoß || g.getMessGroesse() == Drehstoß)
						return true;
				}
			} catch (ArrayIndexOutOfBoundsException e) {
			}
		return false;
	}

	public boolean hatTimerBoxManuellGroesse() {
		AdapterListe al = GHMessen.getApplication().getMainView().al;
		if (al == null)
			return false;
		for (Groesse g : this)
			try {
				if (g.getArt() == Art.Messgroesse && al.get(g.getNr()) instanceof ITimer) {
					if (g.getMessGroesse() == LinearerStoß || g.getMessGroesse() == Drehstoß)
						return true;
				}
			} catch (ArrayIndexOutOfBoundsException e) {
			}
		return false;
	}

	public boolean hatHistogrammGroesse() {
		for (Groesse g : this)
			try {
				if (g.getArt() == Art.Histogramm)
					return true;
			} catch (ArrayIndexOutOfBoundsException e) {
			}
		return false;
	}

	public boolean hatSpektrumGroesse() {
		for (Groesse g : this)
			try {
				if (g.getArt() == Art.FFT)
					return true;
			} catch (ArrayIndexOutOfBoundsException e) {
			}
		return false;
	}

	public boolean hatStart1Flanke() {
		for (Groesse g : this)
			if (g.isStart1Flanke())
				return true;
		return false;
	}

	public boolean hatStoppFlanke() {
		for (Groesse g : this)
			if (g.isStoppFlanke())
				return true;
		return false;
	}

	public int stoppFlanke() {
		for (Groesse g : this)
			if (g.isStoppFlanke())
				return g.getFlanke();
		return -1;
	}

	public boolean hatArrayZugriff() {
		for (Groesse g : this)
			if (g.isAktiv() && g.getArt() == Art.Formel && g.getFormel().contains("["))
				return true;
		return false;
	}

	/*************************************
	 * Messreihenverwaltung
	 *************************************/
	public int getMaxMessreihe() {
		int maxMessreihe = -1;
		for (Groesse g : this)
			if (g.isAbhaengig())
				maxMessreihe = Math.max(maxMessreihe, g.getMessreihe());
			else
				maxMessreihe = Math.max(maxMessreihe, 0);

		return maxMessreihe;
	}

	public int addMessreihe() {
		int maxMessreihe = getMaxMessreihe();
		Vector<Groesse> v = new Vector<Groesse>();
		for (Groesse g : this)
			if (!g.isAbhaengig()) {
				Groesse gneu = g.copy(maxMessreihe + 1);
				v.add(gneu);
			}
		addAll(v);
		for (Groesse g : v)
			g.korrigiereWerte(this);
		return maxMessreihe + 1;
	}

	public void removeLetzteMessreihe() {
		int maxMessreihe = getMaxMessreihe();
		Vector<Groesse> v = new Vector<Groesse>();
		if (maxMessreihe > 0)
			for (Groesse g : this)
				if (g.isAbhaengig() && g.getMessreihe() == maxMessreihe)
					v.add(g);
		removeAll(v);
	}

	public void removeAlleMessreihen() {
		Vector<Groesse> v = new Vector<Groesse>();
		for (Groesse g : this)
			if (g.isAbhaengig())
				v.add(g);
		removeAll(v);
	}

	public void removeMessreihe(int nr) {
		Vector<Groesse> v = new Vector<Groesse>();
		for (Groesse g : this)
			if (g.isAbhaengig() && g.getMessreihe() == nr)
				v.add(g);
		removeAll(v);
	}

	/*************************************
	 * Liste - Interface
	 *************************************/

	public int getSize() {
		return list.size();
	}

	public int indexOf(Groesse g) {
		for (int i = 0; i < list.size(); i++)
			if (list.get(i) == g)
				return i;
		return -1;
	}

	public Groesse get(int index) {
		return list.get(index);
	}

	public int size() {
		return list.size();
	}

	public Groesse set(int index, Groesse element) {
		Groesse oldValue = list.set(index, element);
		oldValue.removePropertyChangeListener("symbol", symbolPH);
		oldValue.removePropertyChangeListener("name", namePH);
		oldValue.removePropertyChangeListener("art", artPH);
		oldValue.removePropertyChangeListener(this);
		element.addPropertyChangeListener("symbol", symbolPH);
		element.addPropertyChangeListener("name", namePH);
		element.addPropertyChangeListener("art", artPH);
		element.addPropertyChangeListener(this);
		symbole.set(index, element.getSymbol());
		namen.set(index, element.getName());
		for (ObservableListListener listener : listeners) {
			listener.listElementReplaced(this, index, oldValue);
		}
		return oldValue;
	}

	public void add(int index, Groesse element) {
		list.add(index, element);
		element.addPropertyChangeListener("symbol", symbolPH);
		element.addPropertyChangeListener("name", namePH);
		element.addPropertyChangeListener("art", artPH);
		element.addPropertyChangeListener(this);
		symbole.add(index, element.getSymbol());
		namen.add(index, element.getName());
		modCount++;
		for (ObservableListListener listener : listeners) {
			listener.listElementsAdded(this, index, 1);
		}
		if (hatHistogrammGroesse() && element.getArt() == Art.Histogramm) {
			Darstellungsliste ansicht = GHMessen.getApplication().getMainView().ansicht;
			if (ansicht != null && ansicht.getHistogramm() == null) {
				Darstellung d = new Darstellung(GHMessen.getApplication().getMainView().gt, "Häufigkeitsverteilung",
						(element.getBezug2() != null ? element.getBezug2().getSymbol() : element.getSymbol()),
						element.getSymbol());
				ansicht.addHistogramm(d);
			}
		}
		if (hatSpektrumGroesse() && element.getArt() == Art.FFT) {
			Darstellungsliste ansicht = GHMessen.getApplication().getMainView().ansicht;
			if (ansicht != null && ansicht.getSpektrum() == null) {
				Darstellung d = new Darstellung(GHMessen.getApplication().getMainView().gt, "Frequenzspektrum", "f",
						element.getSymbol());
				ansicht.addSpektrum(d);
			}
		}
	}

	@Override
	public boolean remove(Object o) {
		int index = list.indexOf(o);
		if (index < 0)
			return false;
		boolean ret = list.remove(o);
		if (!(o instanceof Groesse))
			return ret;
		Groesse oldValue = (Groesse) o;
		oldValue.removePropertyChangeListener("symbol", symbolPH);
		oldValue.removePropertyChangeListener("name", namePH);
		oldValue.removePropertyChangeListener("art", artPH);
		oldValue.removePropertyChangeListener(this);
		symbole.remove(index);
		namen.remove(index);
		for (Groesse g : list) {
			if (g.getBezug() == oldValue)
				g.setBezug(null);
			if (g.getBezug2() == oldValue)
				g.setBezug2(null);
		}
		modCount++;
		for (ObservableListListener listener : listeners) {
			listener.listElementsRemoved(this, index, java.util.Collections.singletonList(oldValue));
		}
		if (!hatHistogrammGroesse()) {
			Darstellungsliste ansicht = GHMessen.getApplication().getMainView().ansicht;
			if (ansicht != null && ansicht.getHistogramm() != null)
				ansicht.removeHistogramm();
		}
		if (!hatSpektrumGroesse()) {
			Darstellungsliste ansicht = GHMessen.getApplication().getMainView().ansicht;
			if (ansicht != null && ansicht.getSpektrum() != null)
				ansicht.removeSpektrum();
		}
		return ret;
	}

	public Groesse remove(int index) {
		Groesse oldValue = list.remove(index);
		oldValue.removePropertyChangeListener("symbol", symbolPH);
		oldValue.removePropertyChangeListener("name", namePH);
		oldValue.removePropertyChangeListener("art", artPH);
		oldValue.removePropertyChangeListener(this);
		symbole.remove(index);
		namen.remove(index);
		for (Groesse g : list) {
			if (g.getBezug() == oldValue)
				g.setBezug(null);
			if (g.getBezug2() == oldValue)
				g.setBezug2(null);
		}
		modCount++;
		for (ObservableListListener listener : listeners) {
			listener.listElementsRemoved(this, index, java.util.Collections.singletonList(oldValue));
		}
		if (!hatHistogrammGroesse()) {
			Darstellungsliste ansicht = GHMessen.getApplication().getMainView().ansicht;
			if (ansicht != null && ansicht.getHistogramm() != null)
				ansicht.removeHistogramm();
		}
		if (!hatSpektrumGroesse()) {
			Darstellungsliste ansicht = GHMessen.getApplication().getMainView().ansicht;
			if (ansicht != null && ansicht.getSpektrum() != null)
				ansicht.removeSpektrum();
		}
		return oldValue;
	}

	public boolean removeAll(Collection<?> c) {
		if (c.isEmpty() || !(c.iterator().next() instanceof Groesse))
			return true;
		@SuppressWarnings("unchecked")
		Collection<? extends Groesse> c1 = (Collection<? extends Groesse>) c;
		int index = list.indexOf(c.iterator().next());
		boolean ret = list.removeAll(c);
		for (Groesse g : c1) {
			g.removePropertyChangeListener("symbol", symbolPH);
			g.removePropertyChangeListener("name", namePH);
			g.removePropertyChangeListener("art", artPH);
			g.removePropertyChangeListener(this);
			symbole.remove(g.getSymbol());
			namen.remove(g.getName());
		}
		for (Groesse g : list) {
			if (c1.contains(g.getBezug()))
				g.setBezug(null);
			if (c1.contains(g.getBezug2()))
				g.setBezug2(null);
		}
		modCount++;
		for (ObservableListListener listener : listeners) {
			listener.listElementsRemoved(this, index, (List<? extends Groesse>) c1);
		}
		if (!hatHistogrammGroesse()) {
			Darstellungsliste ansicht = GHMessen.getApplication().getMainView().ansicht;
			if (ansicht != null && ansicht.getHistogramm() != null)
				ansicht.removeHistogramm();
		}
		if (!hatSpektrumGroesse()) {
			Darstellungsliste ansicht = GHMessen.getApplication().getMainView().ansicht;
			if (ansicht != null && ansicht.getSpektrum() != null)
				ansicht.removeSpektrum();
		}
		return ret;
	}

	public boolean addAll(Collection<? extends Groesse> c) {
		return addAll(size(), c);
	}

	public boolean addAll(int index, Collection<? extends Groesse> c) {
		if (list.addAll(index, c)) {
			for (int i = index; i < list.size(); i++) {
				list.get(i).addPropertyChangeListener("symbol", symbolPH);
				list.get(i).addPropertyChangeListener("name", namePH);
				list.get(i).addPropertyChangeListener("art", artPH);
				list.get(i).addPropertyChangeListener(this);
				symbole.add(list.get(i).getSymbol());
				namen.add(list.get(i).getName());
			}
			modCount++;
			for (ObservableListListener listener : listeners) {
				listener.listElementsAdded(this, index, c.size());
			}
		}
		return false;
	}

	public void clear() {
		List<Groesse> dup = new ArrayList<Groesse>(list);
		for (Groesse g : list) {
			g.removePropertyChangeListener("symbol", symbolPH);
			g.removePropertyChangeListener("name", namePH);
			g.removePropertyChangeListener("art", artPH);
			g.removePropertyChangeListener(this);
		}
		list.clear();
		symbole.clear();
		namen.clear();
		modCount++;
		if (dup.size() != 0) {
			for (ObservableListListener listener : listeners) {
				listener.listElementsRemoved(this, 0, dup);
			}
		}
	}

	public boolean containsAll(Collection<?> c) {
		return list.containsAll(c);
	}

	public <T> T[] toArray(T[] a) {
		return list.toArray(a);
	}

	public Object[] toArray() {
		return list.toArray();
	}

	private void fireElementChanged(int index, PropertyChangeEvent evt) {
		for (ObservableListListener listener : listeners) {
			if (listener instanceof ObservableObjectListListener)
				((ObservableObjectListListener) listener).listElementPropertyChanged(this, index, evt);
			else
				listener.listElementPropertyChanged(this, index);
		}
	}

	public void addObservableListListener(ObservableListListener listener) {
		listeners.add(listener);
	}

	public void removeObservableListListener(ObservableListListener listener) {
		listeners.remove(listener);
	}

	public boolean supportsElementPropertyChanged() {
		return supportsElementPropertyChanged;
	}

	protected void handleSymbolEvent(PropertyChangeEvent evt) {
		int index = indexOf(evt.getSource());
		symbole.set(index, (String) evt.getNewValue());
	}

	protected void handleNameEvent(PropertyChangeEvent evt) {
		int index = indexOf(evt.getSource());
		namen.set(index, (String) evt.getNewValue());
	}

	protected void handleArtEvent(PropertyChangeEvent evt) {
		Groesse s = (Groesse) evt.getSource();
		int change = ((Art) evt.getOldValue() == Art.Histogramm ? 1 : 0) + ((Art) evt.getOldValue() == Art.FFT ? 2 : 0)
				+ ((Art) evt.getNewValue() == Art.Histogramm ? 4 : 0) + ((Art) evt.getNewValue() == Art.FFT ? 8 : 0);
		if (change > 0) {
			switch (change) {
			case 1: {
				if (!hatHistogrammGroesse()) {
					Darstellungsliste ansicht = GHMessen.getApplication().getMainView().ansicht;
					if (ansicht != null && ansicht.getHistogramm() != null)
						ansicht.removeHistogramm();
				}
				break;
			}
			case 4: {
				Darstellungsliste ansicht = GHMessen.getApplication().getMainView().ansicht;
				if (ansicht != null && ansicht.getHistogramm() == null) {
					Darstellung d = new Darstellung(GHMessen.getApplication().getMainView().gt, "Häufigkeitsverteilung",
							(s.getBezug2() != null ? s.getBezug2().getSymbol() : s.getSymbol()), s.getSymbol());
					ansicht.addHistogramm(d);
				}
				break;
			}
			case 2: {
				if (!hatSpektrumGroesse()) {
					Darstellungsliste ansicht = GHMessen.getApplication().getMainView().ansicht;
					if (ansicht != null && ansicht.getSpektrum() != null)
						ansicht.removeSpektrum();
				}
				break;
			}
			case 8: {
				Darstellungsliste ansicht = GHMessen.getApplication().getMainView().ansicht;
				if (ansicht != null && ansicht.getSpektrum() == null) {
					Darstellung d = new Darstellung(GHMessen.getApplication().getMainView().gt, "Frequenzspektrum", "f",
							s.getSymbol());
					ansicht.addSpektrum(d);
				}
				break;
			}
			}
		}
	}

	/**
	 * getAbsoluteDataVector alle aktuellen Werte aller Groessen werden als
	 * Double-Vektor von absoluten Werten zurückgegeben.
	 * 
	 * @return
	 */
	public Double[] getAbsoluteDataVector() {
		Double[] ret = new Double[list.size()];
		for (int i = 0; i < list.size(); i++) {
			ret[i] = list.get(i).getWert() * list.get(i).getEinheit().getFaktor();
		}
		return ret;
	}

	@Override
	public void propertyChange(PropertyChangeEvent evt) {
		int index = indexOf(evt.getSource());
		if (index > -1)
			fireElementChanged(index, evt);
	}

}
