/* 
 * GHMessen
 * --------
 * 
 * GHMessen is a measuring program to use in connection with a computer 
 * measuring interface. It's designed for school usage.
 * 
 * Copyright (C) 2010-2011  Frank Schütte - Gymnasium Himmelsthür
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms
 * of the GNU General Public License as published by the Free Software Foundation; either 
 * version 3 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with this program; 
 * if not, see <http://www.gnu.org/licenses/>.
 * 
 * Frank Schütte fschuett AT gymnasium-himmelsthuer DOT de
 * 
 */

package ghmessen.groesse;

import static ghmessen.util.MathUtil.findeGrobeUnterteilung;
import ghmessen.components.Intervall;
import ghmessen.instrumente.komponenten.Nullpunkt;
import ghmessen.util.Util;

import java.io.PrintWriter;
import java.io.Serializable;
import java.util.Vector;

import javax.xml.transform.sax.TransformerHandler;

import ld.api.CASSYRanges;
import ld.api.ios.CASSYRange;
import ld.api.ios.Measurement;
import ld.api.ios.Origin;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.AttributesImpl;

public class Messbereich extends CASSYRange implements Serializable {
	/**
	 * 
	 */
	private static final long serialVersionUID = -3323759699907976107L;

	public static final Messbereich Invalid;

	public static final Messbereich U100V;
	public static final Messbereich U30V;
	public static final Messbereich U10V;
	public static final Messbereich U3V;
	public static final Messbereich U1V;
	public static final Messbereich U10Vb;
	public static final Messbereich U3Vb;
	public static final Messbereich U0V3;

	public static final Messbereich I3A;
	public static final Messbereich I1A;
	public static final Messbereich I0A3;
	public static final Messbereich I0A1;

	public static final Messbereich P0W03;
	public static final Messbereich P0W1;
	public static final Messbereich P1W;
	public static final Messbereich P3W;
	public static final Messbereich P10W;
	public static final Messbereich P100W;
	public static final Messbereich P300W;

	public static final Messbereich PHI;
	public static final Messbereich COSPHI;

	public static final Vector<Messbereich> spannungsBereich;
	public static final Vector<Messbereich> stromBereich;
	public static final Vector<Messbereich> leistungsBereich;

	private static final boolean DEBUG = false;

	private String name;
	private double unterteilung;
	private Einheit einheit;
	private String bereichName;
	private String[] skala;

	public static void createDTD(PrintWriter out) {
		out.println("<!ELEMENT Messbereich (MbSkala)>");
		out.println("<!ATTLIST Messbereich ");
		out.println("\tname CDATA #REQUIRED");
		out.println("\twert (INT) #REQUIRED");
		out.println("\talternative (INT) #REQUIRED");
		out.println("\tfaktor (DOUBLE) #REQUIRED");
		out.println("\tminimum (DOUBLE) #REQUIRED");
		out.println("\tmaximum (DOUBLE) #REQUIRED");
		out.println("\tunterteilung (DOUBLE) #REQUIRED");
		out.println("\tstellen (INT) #REQUIRED");
		out.println("\teinheit (Einheit) #REQUIRED");
		out.println("\tbereichName CDATA #REQUIRED");
		out.println(">");
		out.println("<!ELEMENT MbSkala (#PCDATA)*>");
	}

	public void writeToXml(TransformerHandler hd) throws SAXException {
		if (hd == null)
			throw new SAXException(
					"[Messbereich] Objekt ist nicht initialisiert.");
		AttributesImpl atts = new AttributesImpl();
		if(name != null && !"".equals(name))
			atts.addAttribute("","","name","CDATA",name);
		atts.addAttribute("","","wert","INT", Integer.toString(getWert()));
		atts.addAttribute("","","alternative","INT", Integer.toString(getAlternative()));
		atts.addAttribute("","","faktor","DOUBLE", Double.toString(getFaktor()));
		atts.addAttribute("","","minimum","DOUBLE", Double.toString(getMinimum()));
		atts.addAttribute("","","maximum", "DOUBLE", Double.toString(getMaximum()));
		atts.addAttribute("","","unterteilung", "DOUBLE", Double.toString(getUnterteilung()));
		atts.addAttribute("","","stellen", "INT", Integer.toString(getStellen()));
		atts.addAttribute("","","einheit", "Einheit", getEinheit().toString());
		atts.addAttribute("","","bereichName","CDATA", bereichName);
		hd.startElement("", "", "Messbereich", atts);
		hd.startElement("", "", "MbSkala", null);
		String out = "";
		for (String s : skala) {
			out = out + "\"" + s + "\",";
		}
		if (!"".equals(out))
			out = out.substring(0, out.length() - 1);
		hd.characters(out.toCharArray(), 0, out.length());
		hd.endElement("", "", "MbSkala");
		hd.endElement("", "", "Messbereich");
	}

	public Messbereich startElement(String uri, String localName, String qName,
			Attributes attributes) throws SAXException {
		if (qName.equals("Messbereich")) {
			for (int index = 0; index < attributes.getLength(); index++) {
				String lname = attributes.getLocalName(index);
				if("name".equals(lname)) {
					name = attributes.getValue(index);
				} else if("wert".equals(lname)) {
					setRange1(CASSYRanges.forValue(Integer.parseInt(attributes.getValue(index))));
				} else if("alternative".equals(lname)) {
					setRange2(CASSYRanges.forValue(Integer.parseInt(attributes.getValue(index))));					
				} else if("faktor".equals(lname)) {
					setBoxScale(Double.parseDouble(attributes.getValue(index)));
				} else if("minimum".equals(lname)) {
					setMinimum(Double.parseDouble(attributes.getValue(index)));
				} else if("maximum".equals(lname)) {
					setMaximum(Double.parseDouble(attributes.getValue(index)));					
				} else if("unterteilung".equals(lname)) {
					unterteilung = Double.parseDouble(attributes.getValue(index));
				} else if("stellen".equals(lname)) {
					setStellen(Integer.parseInt(attributes.getValue(index)));
				} else if("einheit".equals(lname)) {
					einheit = Einheit.toEinheit(attributes.getValue(index));
				} else if("bereichName".equals(lname)) {
					bereichName = attributes.getValue(index);
				}
			}
			if (name == null)
				return this;
			else
				try {
					return (Messbereich) getClass().getField(name).get(this);
				} catch (IllegalArgumentException e) {
					if (DEBUG)
						System.out
								.println("[Messbereich] - startElement: Statisches Feld("
										+ name + ") nicht gefunden.");
					e.printStackTrace();
				} catch (SecurityException e) {
					if (DEBUG)
						System.out
								.println("[Messbereich] - startElement: Statisches Feld("
										+ name + ") nicht gefunden.");
					e.printStackTrace();
				} catch (IllegalAccessException e) {
					if (DEBUG)
						System.out
								.println("[Messbereich] - startElement: Statisches Feld("
										+ name + ") nicht gefunden.");
					e.printStackTrace();
				} catch (NoSuchFieldException e) {
					if (DEBUG)
						System.out
								.println("[Messbereich] - startElement: Statisches Feld("
										+ name + ") nicht gefunden.");
					e.printStackTrace();
				}
			return this;
		} else if (qName.equals("MbSkala")) {
			return this;
		} else
			throw new SAXException("[Messbereich] - Start(" + qName + ")");
	}

	public void endElement(String uri, String localName, String qName,
			String inhalt) throws SAXException {
		if (qName.equals("MbSkala")) {
			if (DEBUG)
				System.out.println("[MbSkala] - endElement: qName(" + qName
						+ ") inhalt(" + inhalt + ")");
			skala = Util.aStringFromXml(inhalt);
		} else if (qName.equals("Messbereich"))
			return;
		else
			throw new SAXException("[Messbereich] - endElement(" + qName + ")");
	}

	public Messbereich(int wert) {
		this(wert, 0, 1.0, 0.0, 1.0, 0.5, 0, Einheit.Keine, "");
	}

	public Messbereich(int wert, int alternative) {
		this(wert, alternative, 1.0, 0.0, 1.0, 0.5, 0, Einheit.Keine, "");
	}

	public Messbereich(int wert, double faktor, double min, double max,
			double unterteilung, int stellen, Einheit einheit, String bereich) {
		this(null, wert, 0, faktor, min, max, unterteilung, stellen,
				einheit, bereich);
	}

	public Messbereich(int wert, int alternative, double faktor,
			double min, double max, double unterteilung, int stellen,
			Einheit einheit, String bereich) {
		this(null, wert, alternative, faktor, min, max, unterteilung,
				stellen, einheit, bereich);
	}

	public Messbereich(CASSYRange r) {
		this(null, r.getRange1().getValue(), r.getRange2().getValue(), r.getBoxScale(),
				r.getMin(), r.getMax(), -1, r.getDec(), 
				r.getUnit() == "" ? Einheit.Keine : Einheit.toEinheit(r.getUnit()),r.toString(Measurement.Normal, Origin.Center));
	}
	
	private Messbereich(String name, int wert, double faktor, double min,
			double max, double unterteilung, int stellen, Einheit einheit,
			String bereich) {
		this(name, wert, wert, faktor, min, max, unterteilung, stellen,
				einheit, bereich);
	}

	private Messbereich(String name, int wert, int alternative, double faktor,
			double min, double max, double unterteilung, int stellen,
			Einheit einheit, String bereich) {
		super(wert, alternative, faktor, min, max, 1, stellen, einheit == null ? "" : einheit.toString());
		this.name = name;
		this.unterteilung = unterteilung;
		this.einheit = einheit;
		this.bereichName = bereich;
		this.skala = berechneSkalaListe();
	}

	private Messbereich(String name, CASSYRanges r1, double faktor, double min, double max,
			double unterteilung, int stellen, Einheit einheit, String bereich) {
		super(r1, faktor, min, max, 1, stellen, einheit == null ? "" : einheit.toString());
		this.name = name;
		this.unterteilung = unterteilung;
		this.einheit = einheit;
		this.bereichName = bereich;
		this.skala = berechneSkalaListe();
	}
	
	public void setzeStandardWerte(Groesse g) {
		if (g == null)
			return;
		g.setVon(getMin());
		g.setBis(getMax());
		g.setStellen(getDec());
		g.setNullpunkt(getStandardNullpunkt());
		g.setEinheit(einheit);
	}

	public int getWert() {
		return getRange1().getValue();
	}

	public int getAlternative() {
		return getRange2().getValue();
	}

	public double getFaktor() {
		return getBoxScale();
	}

	public void setEinheit(Einheit einheit) {
		this.einheit = einheit;
	}

	public Einheit getEinheit() {
		return einheit;
	}

	public Nullpunkt getStandardNullpunkt() {
		if (getMin() < 0) {
			if (getMax() > 0)
				return Nullpunkt.mittig;
			else
				return Nullpunkt.rechts;
		} else if (getMin() == 0)
			return Nullpunkt.links;
		else {
			return Nullpunkt.mittig;
		}
	}

	public void setName(String name) {
		this.name = name;
	}

	public String getName() {
		return name;
	}

	public String getBereichName() {
		return bereichName;
	}

	public void setBereichName(String name) {
		bereichName = name;
	}

	public static Messbereich toMessbereich(String aName, Messbereich[] werte) {
		if (DEBUG)
			System.out.println(aName);
		for (Messbereich m : werte)
			if (m.getBereichName().equals(aName))
				return m;
		return null;
	}

	public static String[] getBereicheStringList(Vector<Messbereich> bereiche) {
		if (bereiche == null)
			return new String[] {};
		String[] result = new String[bereiche.size()];
		for (int i = 0; i < bereiche.size(); i++)
			result[i] = bereiche.get(i).getBereichName();
		return result;
	}

	public static String[] getBereicheStringList(Messbereich[] bereiche) {
		if (bereiche == null)
			return new String[] {};
		String[] result = new String[bereiche.length];
		for (int i = 0; i < bereiche.length; i++)
			result[i] = bereiche[i].getBereichName();
		return result;
	}

	public String[] getSkalaListe() {
		return skala;
	}

	private String[] berechneSkalaListe() {
		return ghmessen.util.MathUtil.berechneSkalaListe(getMin(), getMax(),
				unterteilung);
	}

	public void setMinimum(double min) {
		this.setMin(min);
	}

	public double getMinimum() {
		return getMin();
	}

	public void setStellen(int stellen) {
		this.setDec(stellen);
	}

	public int getStellen() {
		return getDec();
	}

	public void setMaximum(double max) {
		this.setMax(max);
	}

	public double getMaximum() {
		return getMax();
	}

	public void setUnterteilung(double unterteilung) {
		this.unterteilung = unterteilung;
	}

	public double getUnterteilung() {
		return unterteilung;
	}

	public Intervall getIntervall() {
		return new Intervall(getMin(), getMax());
	}

	public void setIntervall(Intervall intervall) {
		this.setMin(intervall.getMin());
		this.setMax(intervall.getMax());
		unterteilung = findeGrobeUnterteilung(getMin(), getMax());
		skala = berechneSkalaListe();
	}

	public String[] getBereicheStringList(Messbereich mb) {
		if (spannungsBereich.contains(mb))
			return new String[] { U100V.getBereichName(),
					U30V.getBereichName(), U10V.getBereichName(),
					U3V.getBereichName(), U1V.getBereichName(),
					U0V3.getBereichName() };
		else if (stromBereich.contains(mb))
			return new String[] { I3A.getBereichName(), I1A.getBereichName(),
					I0A3.getBereichName(), I0A1.getBereichName() };
		else if (leistungsBereich.contains(mb))
			return new String[] { P0W03.getBereichName(),
					P0W1.getBereichName(), P1W.getBereichName(),
					P3W.getBereichName(), P10W.getBereichName(),
					P100W.getBereichName(), P300W.getBereichName() };
		else
			return new String[] { Invalid.getBereichName() };
	}

	public Messbereich[] getBereicheList(Messbereich mb) {
		if (spannungsBereich.contains(mb))
			return new Messbereich[] { U100V, U30V, U10V, U3V, U1V, U0V3 };
		else if (stromBereich.contains(mb))
			return new Messbereich[] { I3A, I1A, I0A3, I0A1 };
		else if (leistungsBereich.contains(mb))
			return new Messbereich[] { P0W03, P0W1, P1W, P3W, P10W, P100W,
					P300W };
		else
			return new Messbereich[] { Invalid };
	}

	public String toString() {
		return this.bereichName;
	}
	
	static {
		Invalid = new Messbereich("Invalid", -1, 1, 0, 1, 0.5, 0,
				Einheit.Keine, "Ungültig");
		U100V = new Messbereich("U100V", CASSYRanges.U100V, 1, -100, 100, 50, 1,
				Einheit.V, "-100V .. 100V");
		U30V = new Messbereich("U30V", CASSYRanges.U30V, 1, -30, 30, 10, 2, Einheit.V,
				"-30V .. 30V");
		U10V = new Messbereich("U10V", CASSYRanges.U10V, 1, -10, 10, 5, 2, Einheit.V,
				"-10V .. 10V");
		U3V = new Messbereich("U3V", CASSYRanges.U3V, 1, -3, 3, 1, 2, Einheit.V,
				"-3V .. 3V");
		U10Vb = new Messbereich("U10Vb", CASSYRanges.U10VB, 1, -10, 10, 5, 2, Einheit.V,
				"-10V .. 10V");
		U3Vb = new Messbereich("U3Vb", CASSYRanges.U3VB, 1, -3, 3, 50, 2, Einheit.V,
				"-3V .. 3V");
		U1V = new Messbereich("U1V", CASSYRanges.U1V, 1, -1, 1, 0.5, 3, Einheit.V,
				"-1V .. 1V");
		U0V3 = new Messbereich("U0V3", CASSYRanges.U0V3, 1, -0.3, 0.3, 0.1, 3, Einheit.V,
				"-0,3V .. 0,3V");
		spannungsBereich = new Vector<Messbereich>();
		spannungsBereich.add(U100V);
		spannungsBereich.add(U30V);
		spannungsBereich.add(U10V);
		spannungsBereich.add(U3V);
		spannungsBereich.add(U1V);
		spannungsBereich.add(U0V3);
		I3A = new Messbereich("I3A", CASSYRanges.I3A, 1, -3, 3, 1, 2, Einheit.A,
				"-3A .. 3A");
		I1A = new Messbereich("I1A", CASSYRanges.I1A, 1, -1, 1, 0.5, 3, Einheit.A,
				"-1A .. 1A");
		I0A3 = new Messbereich("I0A3", CASSYRanges.I0A3, 1, -0.3, 0.3, 0.1, 3, Einheit.A,
				"-0,3A .. 0,3A");
		I0A1 = new Messbereich("I0A1", CASSYRanges.I0A1, 1, -0.1, 0.1, 0.05, 3, Einheit.A,
				"-0,1A .. 0,1A");
		stromBereich = new Vector<Messbereich>();
		stromBereich.add(I3A);
		stromBereich.add(I1A);
		stromBereich.add(I0A3);
		stromBereich.add(I0A1);
		P0W03 = new Messbereich("P0W03", CASSYRanges.None, 1, -0.03, 0.03, 0.01, 3,
				Einheit.W, "-0.03W .. 0.03W");
		P0W1 = new Messbereich("P0W1", CASSYRanges.None, 1, -0.1, 0.1, 0.05, 3,
				Einheit.W, "-0.1W .. 0.1W");
		P1W = new Messbereich("P1W", CASSYRanges.None, 1, -1.0, 1.0, 0.5, 2,
				Einheit.W, "-1W .. 1W");
		P3W = new Messbereich("P3W", CASSYRanges.None, 1, -3.0, 3.0, 1.0, 2,
				Einheit.W, "-3W .. 3W");
		P10W = new Messbereich("P10W", CASSYRanges.None, 1, -10, 10, 5, 1, Einheit.W,
				"-10W .. 10W");
		P100W = new Messbereich("P100W", CASSYRanges.None, 1, -100, 100, 50, 1,
				Einheit.W, "-100W .. 100W");
		P300W = new Messbereich("P300W", CASSYRanges.None, 1, -300, 300, 100, 0,
				Einheit.W, "-300W .. 300W");
		leistungsBereich = new Vector<Messbereich>();
		leistungsBereich.add(P0W03);
		leistungsBereich.add(P0W1);
		leistungsBereich.add(P1W);
		leistungsBereich.add(P3W);
		leistungsBereich.add(P10W);
		leistungsBereich.add(P100W);
		leistungsBereich.add(P300W);
		PHI = new Messbereich("PHI", CASSYRanges.None, 1, -180, 180, 60, 1,
				Einheit._g, "-180° .. 180°");
		COSPHI = new Messbereich("COSPHI", CASSYRanges.None, 1, -1, 1, 0.5, 2,
				Einheit.Keine, "-1 .. 1");
	}

}
