/* 
 * GHMessen
 * --------
 * 
 * GHMessen is a measuring program to use in connection with a computer 
 * measuring interface. It's designed for school usage.
 * 
 * Copyright (C) 2010-2011  Frank Schütte - Gymnasium Himmelsthür
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms
 * of the GNU General Public License as published by the Free Software Foundation; either 
 * version 3 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with this program; 
 * if not, see <http://www.gnu.org/licenses/>.
 * 
 * Frank Schütte fschuett AT gymnasium-himmelsthuer DOT de
 * 
 */

package ghmessen.groesse.darstellung;

import static ghmessen.groesse.darstellung.AnpassungEnum.Ausgleichsgerade;
import flanagan.analysis.Stat;
import ghmessen.components.plotcanvas.GraphBereich;
import ghmessen.parser.FormelGenerator;
import ghmessen.util.Util;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Component;
import java.awt.Font;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionListener;
import java.awt.geom.Path2D;
import java.awt.geom.Point2D;
import java.io.PrintWriter;
import java.io.Serializable;

import javax.xml.transform.sax.TransformerHandler;

import org.xml.sax.SAXException;
import org.xml.sax.helpers.AttributesImpl;

/*
 * Diese Klasse enthält eine Auswertung, die sich auf den Abschnitt einer
 * Reihe von Messwerten innerhalb einer Darstellung bezieht.
 * 
 * Der Bezug wird nur für externe Abfragen mit angegeben und nicht intern
 * benutzt.
 * 
 * typ : Typ der Anpassung
 * x : Erste Spalte (0) Darstellung.XACHSE
 * bezug:
 * col : Spalte (>=1)
 * row1 : Erste verwendete Zeile
 * row2 : Letzte Verwendete Zeile
 * 
 * ausdruck : JANINO-Rechenausdruck
 * term : unformatierter String des Rechenausdrucks
 * htmlTerm : in HTML formatierter String zur Ausgabe
 * rsqr : Mittlere quadratische Abweichung
 * 
 * zeichnung : Zeichnung der zugehörigen Grafik
 * 
 */

public abstract class Anpassung implements Auswertung, Serializable {

	/**
	 * 
	 */
	private static final long serialVersionUID = 6797140082415421098L;
	private static final boolean DEBUG = false;
	/* Vorbedingungen */
	protected AnpassungEnum typ;
	protected double[] x;
	protected double[] y;
	protected double von, bis;
	protected int n;

	/* Bezug */
	protected GraphBereich bezug;
	String symbolX, symbolY;
	int stellenX, stellenY;
	String einheitX, einheitY;
	
	/* Auswertungen */
	protected String term;
	protected String htmlTerm;
	protected Color farbe;

	protected double[] y1;
	protected double rsqr;
	protected Path2D zeichnung;
	protected Path2D zeichnungVonBis;

	/* Ablaufsteuerung */
	protected boolean autoZeichnung;
	protected boolean autoAnpassung;
	protected boolean autoZeichnungVonBis;

	/* Darstellung im Canvas */
	protected transient TextAuswertung benennung;
	protected boolean manuell;
	protected transient FormelGenerator parser;

	public static void createDTD(PrintWriter out) {
		AnpassungEnum.createDTD(out);
		GraphBereich.createDTD(out);
		out.println("<!ELEMENT Anpassung (APX,APY,APY1,GraphBereich,TextAuswertung)>");
		out.println("<!ATTLIST Anpassung typ (AnpassungEnum) #REQUIRED von (DOUBLE) #REQUIRED bis (DOUBLE) #REQUIRED n (INT) #REQUIRED");
		out.println("\tterm CDATA #REQUIRED htmlTerm CDATA #REQUIRED farbe CDATA #REQUIRED rsqr (DOUBLE) #REQUIRED autoZeichnung (BOOLEAN) #REQUIRED");
		out.println("\tautoAnpassung (BOOLEAN) #REQUIRED autoZeichnungVonBis (BOOLEAN) #REQUIRED manuell (BOOLEAN) #IMPLIED");
		out.println("\tsymbolX CDATA #REQUIRED symbolY CDATA #REQUIRED");
		out.println("\tstellenX (INT) #REQUIRED stellenY (INT) #REQUIRED einheitX CDATA #REQUIRED einheitY CDATA #REQUIRED>");
		out.println("<!ELEMENT APX (DOUBLE*)>");
		out.println("<!ELEMENT APY (DOUBLE*)>");
		out.println("<!ELEMENT APY1 (DOUBLE*)>");
	}

	public void writeToXml(TransformerHandler hd) throws SAXException {
		if (hd == null)
			throw new SAXException(
					"[Anpassung] Objekt ist nicht initialisiert.");
		AttributesImpl atts = new AttributesImpl();
		atts.addAttribute("", "", "typ", "AnpassungEnum", typ.toString());
		atts.addAttribute("", "", "von", "DOUBLE", Double.toString(von));
		atts.addAttribute("", "", "bis", "DOUBLE", Double.toString(bis));
		atts.addAttribute("", "", "symbolX", "CDATA", symbolX);
		atts.addAttribute("", "", "symbolY", "CDATA", symbolY);
		atts.addAttribute("", "", "stellenX", "INT", Integer.toString(stellenX));
		atts.addAttribute("", "", "stellenY", "INT", Integer.toString(stellenY));
		atts.addAttribute("", "", "einheitX", "CDATA", einheitX);
		atts.addAttribute("", "", "einheitY", "CDATA", einheitY);
		atts.addAttribute("", "", "n", "INT", Integer.toString(n));

		atts.addAttribute("", "", "term", "CDATA", term);
		atts.addAttribute("", "", "htmlTerm", "CDATA", htmlTerm);
		atts.addAttribute("", "", "farbe", "CDATA", Integer.toHexString(farbe.getRGB()));
		atts.addAttribute("", "", "rsqr", "DOUBLE", Double.toString(rsqr));
		atts.addAttribute("", "", "autoZeichnung", "BOOLEAN",
				Boolean.toString(autoZeichnung));
		atts.addAttribute("", "", "autoAnpassung", "BOOLEAN",
				Boolean.toString(autoAnpassung));
		atts.addAttribute("", "", "autoZeichnungVonBis", "BOOLEAN",
				Boolean.toString(autoZeichnungVonBis));
		atts.addAttribute("", "", "manuell", "BOOLEAN", Boolean.toString(manuell));
		hd.startElement("", "", "Anpassung", atts);
		String inhalt;
		hd.startElement("", "", "APX", null);
		inhalt = Util.writeDoublesToXml(x);
		hd.characters(inhalt.toCharArray(), 0, inhalt.length());
		hd.endElement("", "", "APX");
		hd.startElement("", "", "APY", null);
		inhalt = Util.writeDoublesToXml(y);
		hd.characters(inhalt.toCharArray(), 0, inhalt.length());
		hd.endElement("", "", "APY");
		hd.startElement("", "", "APY1", null);
		inhalt = Util.writeDoublesToXml(y1);
		hd.characters(inhalt.toCharArray(), 0, inhalt.length());
		hd.endElement("", "", "APY1");
		bezug.writeToXml(hd);
		if(benennung != null)
			benennung.writeToXml(hd);
		hd.endElement("", "", "Anpassung");
	}

	public Anpassung(AnpassungEnum typ, Double[] x, Double[] y, double von,
			double bis, GraphBereich bezug, String symbolX, String symbolY, int stellenX, int stellenY,
			String einheitX, String einheitY,
			String term, String htmlTerm, double rsqr, Path2D zeichnung,
			boolean autoZeichnung, boolean autoAnpassung,
			boolean autoZeichnungVonBis, Color farbe) {
		this.typ = typ;
		nullEntfernenUndEintragen(x, y);
		n = Math.min(this.x == null ? 0 : this.x.length, this.y == null ? 0
				: this.y.length);
		this.von = von;
		this.bis = bis;
		this.bezug = bezug;
		this.symbolX = symbolX;
		this.symbolY = symbolY;
		this.stellenX = stellenX;
		this.stellenY = stellenY;
		this.einheitX = einheitX;
		this.einheitY = einheitY;
		this.term = term;
		this.htmlTerm = htmlTerm;
		this.rsqr = rsqr;
		this.zeichnung = zeichnung;
		this.zeichnungVonBis = zeichnung;
		this.autoZeichnung = autoZeichnung;
		this.autoAnpassung = autoAnpassung;
		this.autoZeichnungVonBis = autoZeichnungVonBis;
		this.y1 = null;
		this.parser = new FormelGenerator();
		this.benennung = new TextAuswertung(null, htmlTerm, farbe,
				berechnePosition(), bezug == null ? -1 : bezug.getCol());
		this.farbe = farbe;
		this.manuell = false;
		this.benennung.addMouseMotionListener(new MouseMotionListener(){
			@Override
			public void mouseDragged(MouseEvent e) { manuell = true; }
			@Override
			public void mouseMoved(MouseEvent e) { }
		});
		aktualisieren();
	}

	public Anpassung() {
		this(Ausgleichsgerade, null, null, 0, 0, null, "t", "y", 2, 2, null, null, "", "", 0.0, null,
				false, false, false, null);
	}

	/********************************************************
	 * Berechnungen
	 * 
	 ********************************************************/
	protected abstract void anpassen() throws AnpassungException;

	protected void htmlAusTermErzeugen() {
		if (parser != null) {
			htmlTerm = "<html>" + term + "</html>";
		} else
			htmlTerm = term;
	}

	protected abstract void zeichnen() throws AnpassungException;

	protected abstract double value(double x) throws AnpassungException;

	protected abstract void zeichnen(double von, double bis)
			throws AnpassungException;

	protected void aktualisieren() {
		if (autoAnpassung) {
			try {
				zeichnung = null;
				anpassen();
				if (autoZeichnung)
					zeichnen();
				if (autoZeichnungVonBis)
					zeichnen(von, bis);
				if (benennung != null) {
					benennung.setText(htmlTerm);
					if(!manuell)
						benennung.setPos(berechnePosition());
				}
			} catch (AnpassungException e) {
				if (DEBUG)
					e.printStackTrace();
			}
		} else {
			rsqr = 0.0;
			zeichnung = null;
			zeichnungVonBis = null;
			if (term == null)
				term = "";
			if (htmlTerm == null)
				htmlTerm = "";
		}
	}

	protected void nullEntfernenUndEintragen(Double[] x1, Double[] y1) {
		if (x1 == null || x1.length == 0 || y1 == null || y1.length == 0
				|| x1.length != y1.length)
			return;
		int nullWerte = nullWerteZaehlen(x1, y1);
		if (nullWerte >= x1.length)
			x = null;
		x = new double[x1.length - nullWerte];
		y = new double[y1.length - nullWerte];
		int j = 0;
		for (int i = 0; i < x1.length; i++)
			if (x1[i] != null && y1[i] != null) {
				x[j] = x1[i].doubleValue();
				y[j++] = y1[i].doubleValue();
			}
	}

	private int nullWerteZaehlen(Double[] x1, Double[] y1) {
		if (x1 == null || y1 == null || x1.length != y1.length)
			return 0;
		int result = 0;
		for (int i = 0; i < x1.length; i++)
			if (x1[i] == null || y1[i] == null)
				result++;
		return result;
	}

	public void setTyp(AnpassungEnum typ) {
		this.typ = typ;
		aktualisieren();
	}

	public AnpassungEnum getTyp() {
		return typ;
	}

	public double[] getX() {
		return x;
	}

	public double[] getY() {
		return y;
	}

	public String getSymbolX() {
		return symbolX;
	}
	
	public void setSymbolX(String symbolX) {
		this.symbolX = symbolX;
	}
	
	public String getSymbolY() {
		return symbolY;
	}
	
	public void setSymbolY(String symbolY) {
		this.symbolY = symbolY;
	}
	
	public int getStellenX() {
		return stellenX;
	}
	
	public void setStellenX(int stellenX) {
		this.stellenX = stellenX;
	}
	
	public int getStellenY() {
		return stellenY;
	}
	
	public void setStellenY(int stellenY) {
		this.stellenY = stellenY;
	}
	
	public String getEinheitX() {
		return einheitX;
	}
	
	public void setEinheitX(String einheitX) {
		this.einheitX = einheitX;
	}
	
	public String getEinheitY() {
		return einheitY;
	}
	
	public void setEinheitY(String einheitY) {
		this.einheitY = einheitY;
	}
	
	public void setXY(double[] x, double[] y) {
		this.x = x;
		this.y = y;
		n = Math.min(this.x == null ? 0 : this.x.length, this.y == null ? 0
				: this.y.length);
		aktualisieren();
	}

	public void setXY(Double[] x, Double[] y) {
		nullEntfernenUndEintragen(x, y);
		n = Math.min(this.x == null ? 0 : this.x.length, this.y == null ? 0
				: this.y.length);
		aktualisieren();
	}

	public void setBezug(int col, int zeile1, int zeile2, String symbolY, int stellenY, String einheitY) {
		bezug = new GraphBereich(col, zeile1, zeile2);
		this.symbolY = symbolY;
		this.stellenY = stellenY;
		this.einheitY = einheitY;
	}

	public void setBezug(GraphBereich bezug, String symbolY, int stellenY, String einheitY) {
		if (bezug == null)
			return;
		this.bezug = bezug;
		this.symbolY = symbolY;
		this.stellenY = stellenY;
		this.einheitY = einheitY;
	}

	public void setAnfang(int zeile1) {
		bezug.setRow1(zeile1);
	}

	public void setEnde(int zeile2) {
		bezug.setRow2(zeile2);
	}

	/******************************************************************
	 * Interface Auswertung
	 *****************************************************************/

	@Override
	public Point2D getPos() {
		return null;
	}

	@Override
	public void setPos(Point2D pos) {
		return;
	}

	@Override
	public void setLinienbreite(BasicStroke breite) {
		return;
	}

	@Override
	public BasicStroke getLinienbreite() {
		return null;
	}

	@Override
	public boolean isManuell() {
		return manuell;
	}
	
	@Override
	public void setManuell(boolean an) {
		manuell = an;
	}
	
	@Override
	public boolean enthaeltZeile(int zeile) {
		return bezug != null && bezug.contains(zeile);
	}

	@Override
	public int getEndeZeile() {
		if (bezug == null)
			return -1;
		else
			return bezug.getRow2();
	}

	@Override
	public int getSpalte() {
		if (bezug == null)
			return -1;
		else
			return bezug.getCol();
	}

	@Override
	public void setTextgroesse(Font font) {
		return;
	}

	@Override
	public Font getTextgroesse() {
		return null;
	}

	@Override
	public void setFarbe(Color farbe) {
		this.farbe = farbe;
	}

	@Override
	public Color getFarbe() {
		return farbe;
	}

	@Override
	public void updateXY() {
		return;
	}

	@Override
	public void updateLocation() {
		return;
	}

	public void setTerm(String term) {
		this.term = term;
		htmlAusTermErzeugen();
	}

	public String getTerm() {
		return term;
	}

	public void setHtmlTerm(String htmlTerm){
		this.htmlTerm = htmlTerm;
	}
	
	public String getHtmlTerm() {
		return htmlTerm;
	}

	public void setRsqr(double rsqr) {
		this.rsqr = rsqr;
	}

	public double getRsqr() {
		return rsqr;
	}

	public void setZeichnung(Path2D zeichnung) {
		this.zeichnung = zeichnung;
		this.zeichnungVonBis = zeichnung;
		aktualisieren();
	}

	public void setZeichnungVonBis(Path2D zeichnungVonBis) {
		this.zeichnungVonBis = zeichnungVonBis;
		aktualisieren();
	}

	public Path2D getZeichnung() {
		return zeichnung;
	}

	public double[] getWerte() {
		return y1;
	}

	public void setVonBis(double von, double bis) {
		this.von = von;
		this.bis = bis;
		autoZeichnungVonBis = true;
		aktualisieren();
	}

	public void setVon(double von){
		this.von = von;
	}
	
	public void setBis(double bis){
		this.bis = bis;
	}
	
	public void setN(int n){
		this.n = n;
	}
	
	public double getVon() {
		return von;
	}

	public double getBis() {
		return bis;
	}

	public void setAutoZeichnung(boolean an){
		this.autoZeichnung = an;
	}
	
	public void setAutoAnpassung(boolean an){
		this.autoAnpassung = an;
	}
	
	public void setAutoZeichnungVonBis(boolean an){
		this.autoZeichnungVonBis = an;
	}
	
	public boolean hatZeichnungVonBis() {
		return autoZeichnungVonBis;
	}

	public Path2D getZeichnungVonBis() {
		return zeichnungVonBis;
	}

	public void setFont(Font f) {
		if (benennung != null)
			benennung.setFont(f);
	}

	@Override
	public boolean hatKomponente() {
		return benennung != null;
	}

	@Override
	public TextAuswertung getKomponente() {
		return benennung;
	}

	@Override
	public void setKomponente(Component b) {
		if (b instanceof TextAuswertung) {
			benennung = (TextAuswertung) b;
			if(!manuell)
				benennung.setPos(berechnePosition());
		}
	}

	protected Point2D berechnePosition() {
		if (benennung == null || x == null || y == null || x.length < 1
				|| y.length < 1)
			return new Point2D.Double();
		return new Point2D.Double(Stat.mean(x), Stat.mean(y));
	}
}
