/* 
 * GHMessen
 * --------
 * 
 * GHMessen is a measuring program to use in connection with a computer 
 * measuring interface. It's designed for school usage.
 * 
 * Copyright (C) 2010-2011  Frank Schütte - Gymnasium Himmelsthür
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms
 * of the GNU General Public License as published by the Free Software Foundation; either 
 * version 3 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with this program; 
 * if not, see <http://www.gnu.org/licenses/>.
 * 
 * Frank Schütte fschuett AT gymnasium-himmelsthuer DOT de
 * 
 */

package ghmessen.instrumente;

import static ghmessen.Konstanten.FARBE_MESSGROESSE;
import ghmessen.adapter.EAdapterException;
import ghmessen.adapter.IAdapter;
import ghmessen.adapter.IMesskanal;
import ghmessen.adapter.MesskanalException;
import ghmessen.groesse.Aufnahme;
import ghmessen.groesse.Groesse;
import ghmessen.groesse.Messbereich;
import ghmessen.instrumente.komponenten.Nullpunkt;
import ghmessen.instrumente.komponenten.NullpunktZeigeranzeige;
import ghmessen.instrumente.komponenten.WerteanzeigeCASSY;

import java.awt.Color;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.Serializable;
import java.util.concurrent.ScheduledThreadPoolExecutor;

import javax.swing.JFrame;

import ld.api.events.ValueChangedEvent;
import ld.api.events.ValueChangedListener;
import ld.api.ios.ICASSYQuantity;
import ld.api.ios.IRangeMax;
import ld.api.ios.ITare;
import ld.api.ios.IUltrasonic;
import ld.components.MathString;

public class MessgeraetCASSY extends Messgeraet implements ValueChangedListener {
	/**
	 * 
	 */
	private static final long serialVersionUID = -3501529264432959901L;
	private static final boolean DEBUG = false;

	private class ZeitraumPropertyHandler implements Serializable,
			PropertyChangeListener {
		/**
		 * 
		 */
		private static final long serialVersionUID = 1L;

		@Override
		public void propertyChange(PropertyChangeEvent evt) {
			handleZeitraumEvent(evt);
		}
	}

	private class SchmalesSichtfeldHandler implements Serializable,
			PropertyChangeListener {
		private static final long serialVersionUID = 1L;

		@Override
		public void propertyChange(PropertyChangeEvent evt) {
			handleSchmalesSichtfeldEvent(evt);
		}
	}

	private class NichtImmerAktivHandler implements Serializable,
			PropertyChangeListener {
		private static final long serialVersionUID = 1L;

		@Override
		public void propertyChange(PropertyChangeEvent evt) {
			handleNichtImmerAktivEvent(evt);
		}
	}

	private SchmalesSichtfeldHandler schmalesSichtfeldPH;
	private NichtImmerAktivHandler nichtImmerAktivPH;
	private ZeitraumPropertyHandler zeitraumPH;

	public MessgeraetCASSY(JFrame parent, IMesskanal messkanal, IAdapter a,
			Groesse g, ScheduledThreadPoolExecutor hauptTimer) {
		super(parent, messkanal, a, g, hauptTimer);
		schmalesSichtfeldPH = new SchmalesSichtfeldHandler();
		nichtImmerAktivPH = new NichtImmerAktivHandler();
		zeitraumPH = new ZeitraumPropertyHandler();
		g.addPropertyChangeListener("schmalesSichtfeld", schmalesSichtfeldPH);
		g.addPropertyChangeListener("nichtImmerAktiv", nichtImmerAktivPH);
		g.addPropertyChangeListener("zeitraum", zeitraumPH);
		((ICASSYQuantity) messkanal).addValueChangedListener(this);
		setTitle(MathString.convertToToolTipText("Messgerät " + "$"
				+ ((ICASSYQuantity) messkanal).getSymbol() + "$"));
		if(messkanal instanceof ITare)
			((ITare)messkanal).doTare();
	}

	@Override
	protected void initSkala() {
		Messbereich m = g.getMessbereich();
		skala = new NullpunktZeigeranzeige(g.getWert(), g.getNullpunkt(), m,
				g.getAufnahme(), FARBE_MESSGROESSE);
	}

	@Override
	protected void initWert() {
		wert = new WerteanzeigeCASSY("");
	}

	@Override
	protected void startTask() {
		executeTask();
	}

	@Override
	protected void stopTask() {
	}

	@Override
	protected boolean executeTask() {
		Double d = 0.0;
		try {
			d = messkanal.einlesen();
			try {
				setWert(d, messkanal.istUeberschritten(),
						messkanal.istUnterschritten());
			} catch (NullPointerException e) {
				if (DEBUG) // Der Messkanal wurde geschlossen, bevor das
							// Messgerät davon Kenntnis erhielt.
					e.printStackTrace();
			} catch (EAdapterException e) {
				if (DEBUG) {
					e.printStackTrace();
					System.out.println("[" + getClass().toString()
							+ "] - executeTask: d(" + d
							+ ") EAdapterException(" + e.getMessage() + ")");
				}
				d = Double.NaN;

			}

		} catch (MesskanalException e1) {
			if (DEBUG)
				e1.printStackTrace();
		}
		return true;
	}

	@Override
	protected void setWert(Double zahl, boolean overRange, boolean underRange)
			throws EAdapterException {
		if (DEBUG)
			System.out.println("[" + getClass().toString()
					+ "] - setWert: Symbol(" + g.getSymbol() + ") Messbereich("
					+ g.getMessbereich() + ") Wert(" + zahl + ") overRange("
					+ overRange + ") underRange(" + underRange + ")");
		if (overRange) {
			wert.setBackground(Color.ORANGE);
			throw new EAdapterException(
					"Messbereichsüberschreitung - Messwert verwerfen Wert("
							+ zahl + ")");
		} else
			wert.setBackground(wertHintergrund);
		g.setWert(zahl);
		skala.setWert(zahl);
		((WerteanzeigeCASSY) wert).setWert(((ICASSYQuantity) messkanal)
				.getSymbolValueUnitString());
	}

	@Override
	public void handleNullpunktEvent(PropertyChangeEvent evt) {
		skala.setNullpunkt((Nullpunkt) evt.getNewValue());
		if (DEBUG)
			System.out.println("[" + getClass().toString()
					+ "] - handleNullpunktEvent: Nullpunkt("
					+ (Nullpunkt) evt.getNewValue() + ")");
	}

	@Override
	public void handleSymbolEvent(PropertyChangeEvent evt) {
		setTitle("Messgerät " + evt.getNewValue());
		((WerteanzeigeCASSY) wert).setWert(((ICASSYQuantity) messkanal)
				.getSymbolValueUnitString());
	}

	@Override
	public void handleAufnahmeEvent(PropertyChangeEvent evt) {
		skala.setAufnahme((Aufnahme) evt.getNewValue());
		restartTask();
	}

	@Override
	public void handleEinheitEvent(PropertyChangeEvent evt) {
		((WerteanzeigeCASSY) wert).setWert(((ICASSYQuantity) messkanal)
				.getSymbolValueUnitString());
	}

	@Override
	public void handleMessbereichEvent(PropertyChangeEvent evt) {
		if (DEBUG)
			System.out.println("[" + getClass().toString()
					+ "] - handleMessbereichEvent: Messbereich alt("
					+ evt.getOldValue() + ") neu(" + evt.getNewValue() + ")");
		Messbereich m = (Messbereich) evt.getNewValue();
		skala.setMessbereich(m);
		((WerteanzeigeCASSY) wert).setWert(((ICASSYQuantity) messkanal)
				.getSymbolValueUnitString());
		if (DEBUG)
			System.out.println("[" + getClass().toString()
					+ "] - handleMessbereichEvent: minimum(" + m.getMinimum()
					+ ") maximum(" + m.getMaximum() + ") Zahlen("
					+ m.getSkalaListe() + ") stellen(" + m.getStellen()
					+ ") einheit(" + m.getEinheit() + ")");
	}

	@Override
	public void handleBisEvent(PropertyChangeEvent evt) {
		if(messkanal instanceof IRangeMax) {
			if(DEBUG)
				System.out.println("["+getClass().getCanonicalName()+"]"
						+ " - handleBisEvent");
			skala.setMaximum(((IRangeMax) messkanal).getRangeMax());
		}
	}
	
	public void handleSchmalesSichtfeldEvent(PropertyChangeEvent evt) {
		if (messkanal instanceof IUltrasonic)
			((IUltrasonic) messkanal).setSmallBeam((Boolean) evt.getNewValue());
	}

	public void handleNichtImmerAktivEvent(PropertyChangeEvent evt) {
		if (messkanal instanceof IUltrasonic) {
			((IUltrasonic) messkanal).setDuringOscilloscopeOnly((Boolean) evt
					.getNewValue());
			if ((Boolean) evt.getNewValue())
				stopTask();
			else
				startTask();
		}
	}

	public void handleZeitraumEvent(PropertyChangeEvent evt) {
		restartTask();
	}

	@Override
	public void valueChanged(ValueChangedEvent e) {
		if (!isVisible())
			return;
		executeTask();
	}
}
