/* 
 * GHMessen
 * --------
 * 
 * GHMessen is a measuring program to use in connection with a computer 
 * measuring interface. It's designed for school usage.
 * 
 * Copyright (C) 2010-2011  Frank Schütte - Gymnasium Himmelsthür
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms
 * of the GNU General Public License as published by the Free Software Foundation; either 
 * version 3 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with this program; 
 * if not, see <http://www.gnu.org/licenses/>.
 * 
 * Frank Schütte fschuett AT gymnasium-himmelsthuer DOT de
 * 
 */

package ghmessen.instrumente;

import static ghmessen.Konstanten.FARBE_FORMEL;

import java.beans.PropertyChangeEvent;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.ScheduledThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

import javax.swing.JFrame;

import org.codehaus.janino.ExpressionEvaluator;

import ghmessen.GHMessen;
import ghmessen.Konstanten;
import ghmessen.adapter.EAdapterException;
import ghmessen.adapter.MesskanalException;
import ghmessen.components.Intervall;
import ghmessen.groesse.Einheit;
import ghmessen.groesse.Groesse;
import ghmessen.groesse.Groessenliste;
import ghmessen.groesse.WerteTabelle;
import ghmessen.groesse.wertetabelle.XAchseTyp;
import ghmessen.instrumente.komponenten.Werteanzeige;
import ghmessen.instrumente.komponenten.Zeigeranzeige;

public class MessgeraetFormel extends Messgeraet {

	private static final long serialVersionUID = 4564764541107776275L;
	private static final boolean DEBUG = false;

	private Groessenliste gl = null;
	private WerteTabelle gt = null;

	public MessgeraetFormel(JFrame parent, Groessenliste gl, Groesse g, ScheduledThreadPoolExecutor hauptTimer) {
		super(parent, null, null, g, hauptTimer);
		if (DEBUG)
			System.out.println("[MessgeraetFormel] - Konstruktor: g(" + g + ")");
	}

	private Double formelBerechnen() throws InvocationTargetException {
		Double intervall;
		Double[] d;
		Double result = 0.0;
		Calendar jetzt = Calendar.getInstance();
		int datum = ghmessen.util.Math.datum(jetzt);
		int uhrzeit = ghmessen.util.Math.uhrzeit(jetzt);
		intervall = Konstanten.GROESSE_MITTELZEIT_STANDARD;
		if (gl == null)
			gl = GHMessen.getApplication().getMainView().gl;
		if (gt == null)
			gt = GHMessen.getApplication().getMainView().gt;
		d = gl.getAbsoluteDataVector();
		ExpressionEvaluator ee = g.getFormelterm();
		if (ee == null)
			return Double.NaN;
		result = (Double) ee.evaluate(new Object[] { 0.0, 1.0, 1.0, 0.0, 0.0, 0.0, intervall, datum, uhrzeit, d, gt });
		return result;
	}

	@Override
	protected void startTask() {
		try {
			super.startTask();
			scheduleTask();
		} catch (MesskanalException e) {
			if (DEBUG)
				e.printStackTrace();
		}
	}

	protected void scheduleTask() throws MesskanalException {
		if (task != null) {
			runningTask = (ScheduledFuture<?>) hauptTimer.scheduleWithFixedDelay(task, 0, delay, TimeUnit.MILLISECONDS);
		}
	}

	@Override
	protected boolean executeTask() {
		Double d = 0.0;
		try {
			d = formelBerechnen();
			setWert(d, false, false);
		} catch (InvocationTargetException e) {
			if (DEBUG) {
				System.err.println("[MessgeraetFormel]:executeTask - formel could not be evaluated.");
				e.printStackTrace();
			}
		} catch (EAdapterException e) {
			System.err.println("[MessgeraetFormel]:executeTask - this exception is not meant to be thrown!");
			if (DEBUG)
				e.printStackTrace();
		}
		return true;
	}

	@Override
	protected void setWert(Double zahl, boolean overRange, boolean underRange) throws EAdapterException {
		if (DEBUG)
			System.out.println("[MessgeraetFormel] - setWert: Symbol(" + g.getSymbol() + ") Messbereich("
					+ g.getMessbereich() + ") Wert(" + zahl + ")");
		g.setWert(zahl);
		skala.setWert(zahl);
		wert.setWert(zahl);
	}

	protected void initSkala() {
		if (g != null) {
			skala = new Zeigeranzeige(g.getWert(), g.getNullpunkt(), g.getVon(), g.getBis(), -1, FARBE_FORMEL);
		} else
			skala = new Zeigeranzeige();
	}

	protected void initWert() {
		if (g != null) {
			wert = new Werteanzeige(g.getEinheit().toString(), "%2.2f", g.getSymbol(), g.getWert());
			wert.setDezimalStellen(g.getStellen());
		} else
			wert = new Werteanzeige();
	}

	public void showEinstellungenBox() {
		return;
	}

	@Override
	public void handleSymbolEvent(PropertyChangeEvent evt) {
		setTitle("Messgerät " + evt.getNewValue());
		wert.setGroesse((String) evt.getNewValue());
	}

	@Override
	public void handleEinheitEvent(PropertyChangeEvent evt) {
		Einheit e = (Einheit) evt.getNewValue();
		wert.setEinheit(e.toString());
	}

	@Override
	public void handleVonEvent(PropertyChangeEvent evt) {
		if (DEBUG)
			System.out.println("[MessgeraetKonstante] - propertyChange: Messbereich von - bis geändert.");
		skala.setIntervall(new Intervall(g.getVon(), g.getBis()), -1.0);
	}

	@Override
	public void handleBisEvent(PropertyChangeEvent evt) {
		if (DEBUG)
			System.out.println("[MessgeraetKonstante] - propertyChange: Messbereich von - bis geändert.");
		skala.setIntervall(new Intervall(g.getVon(), g.getBis()), -1.0);
	}

	@Override
	public void handleStellenEvent(PropertyChangeEvent evt) {
		wert.setDezimalStellen((Integer) evt.getNewValue());
	}
}
