/* 
 * GHMessen
 * --------
 * 
 * GHMessen is a measuring program to use in connection with a computer 
 * measuring interface. It's designed for school usage.
 * 
 * Copyright (C) 2010-2011  Frank Schütte - Gymnasium Himmelsthür
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms
 * of the GNU General Public License as published by the Free Software Foundation; either 
 * version 3 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with this program; 
 * if not, see <http://www.gnu.org/licenses/>.
 * 
 * Frank Schütte fschuett AT gymnasium-himmelsthuer DOT de
 * 
 */

package ghmessen.instrumente;

import static ghmessen.Konstanten.FARBE_MESSGROESSE;
import ghmessen.adapter.EAdapterException;
import ghmessen.adapter.MesskanalException;
import ghmessen.adapter.event.AdapterEventListener;
import ghmessen.adapter.goIO.AdapterGoIO;
import ghmessen.adapter.goIO.MesskanalGoIO;
import ghmessen.groesse.Aufnahme;
import ghmessen.groesse.Einheit;
import ghmessen.groesse.Groesse;
import ghmessen.groesse.Messbereich;
import ghmessen.instrumente.komponenten.Nullpunkt;
import ghmessen.instrumente.komponenten.NullpunktZeigeranzeige;
import ghmessen.instrumente.komponenten.Werteanzeige;
import java.awt.Color;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.Serializable;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.ScheduledThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

import javax.swing.JFrame;

public class MessgeraetGoIO extends Messgeraet implements AdapterEventListener {
	/**
	 * 
	 */
	private static final long serialVersionUID = -3501529264432959901L;
	private static final boolean DEBUG = false;

	private class ZeitraumPropertyHandler implements Serializable,
			PropertyChangeListener {
		/**
		 * 
		 */
		private static final long serialVersionUID = 1L;

		@Override
		public void propertyChange(PropertyChangeEvent evt) {
			handleZeitraumEvent(evt);
		}
	}

	private ZeitraumPropertyHandler zeitraumPH;

	public MessgeraetGoIO(JFrame parent, MesskanalGoIO messkanal, AdapterGoIO a, Groesse g,
			ScheduledThreadPoolExecutor hauptTimer) {
		super(parent, messkanal, a, g, hauptTimer);
		zeitraumPH = new ZeitraumPropertyHandler();
		if (g != null)
			g.addPropertyChangeListener("zeitraum", zeitraumPH);
		if(DEBUG)
			System.out.println("[MessgeraetGoIO] - Konstruktor aufgerufen...");
	}

	@Override
	protected void initSkala() {
		Messbereich m = g.getMessbereich();
		skala = new NullpunktZeigeranzeige(g.getWert(), g.getNullpunkt(), m,
				g.getAufnahme(), FARBE_MESSGROESSE);
	}

	@Override
	protected void initWert() {
		wert = new Werteanzeige(g.getEinheit().toString(), "%2.2f",
				g.getSymbol(), g.getWert());
		wert.setDezimalStellen(g.getStellen());
	}

	@Override
	protected void startTask() {
		try {
			super.startTask();
			messkanal.setAbstand(delay);
			scheduleTask();
		} catch (MesskanalException e) {
			if (DEBUG)
				e.printStackTrace();
		}
	}

	protected void scheduleTask() throws MesskanalException {
		if (task != null) {
			runningTask = (ScheduledFuture<?>) hauptTimer
					.scheduleWithFixedDelay(task, 0, delay,
							TimeUnit.MILLISECONDS);
		}
	}

	@Override
	protected boolean executeTask() {
		if (!super.executeTask())
			return false;
		Double d = 0.0;
		try {
			if(einlesen)
				d = messkanal.einlesen();
			else
				d = messkanal.leseLetztenWert();
			try {
				setWert(d, messkanal.istUeberschritten(),
						messkanal.istUnterschritten());
			} catch (NullPointerException e) {
				if (DEBUG) // Der Messkanal wurde geschlossen, bevor das
							// Messgerät davon Kenntnis erhielt.
					e.printStackTrace();
			} catch (EAdapterException e) {
				if (DEBUG) {
					e.printStackTrace();
					System.out.println("[MessgeraetGoIO] - executeTask: d(" + d
							+ ") EAdapterException(" + e.getMessage() + ")");
				}
				d = Double.NaN;

			}

		} catch (MesskanalException e1) {
			if (DEBUG)
				e1.printStackTrace();
		}
		return true;
	}

	@Override
	protected void setWert(Double zahl, boolean overRange, boolean underRange)
			throws EAdapterException {
		if (DEBUG)
			System.out.println("[MessgeraetGoIO] - setWert: Symbol("
					+ g.getSymbol() + ") Messbereich(" + g.getMessbereich()
					+ ") Wert(" + zahl + ") overRange(" + overRange
					+ ") underRange(" + underRange + ")");
		if (overRange) {
			wert.setBackground(Color.ORANGE);
			throw new EAdapterException(
					"Messbereichsüberschreitung - Messwert verwerfen Wert("
							+ zahl + ")");
		} else
			wert.setBackground(wertHintergrund);
		g.setWert(zahl);
		skala.setWert(zahl);
		wert.setWert(zahl);
	}

	@Override
	public void handleNullpunktEvent(PropertyChangeEvent evt) {
		skala.setNullpunkt((Nullpunkt) evt.getNewValue());
		if (DEBUG)
			System.out
					.println("[MessgeraetGoIO] - handleNullpunktEvent: Nullpunkt("
							+ (Nullpunkt) evt.getNewValue() + ")");
	}

	@Override
	public void handleSymbolEvent(PropertyChangeEvent evt) {
		setTitle("Messgerät " + evt.getNewValue());
		wert.setGroesse((String) evt.getNewValue());
	}

	@Override
	public void handleAufnahmeEvent(PropertyChangeEvent evt) {
		skala.setAufnahme((Aufnahme) evt.getNewValue());
		restartTask();
	}

	@Override
	public void handleEinheitEvent(PropertyChangeEvent evt) {
		Einheit e = (Einheit) evt.getNewValue();
		wert.setEinheit(e.toString());
	}

	@Override
	public void handleMessbereichEvent(PropertyChangeEvent evt) {
		if (DEBUG)
			System.out
					.println("[MessgeraetGoIO] - handleMessbereichEvent: Messbereich alt("
							+ evt.getOldValue()
							+ ") neu("
							+ evt.getNewValue()
							+ ")");
		Messbereich m = (Messbereich) evt.getNewValue();
		skala.setMessbereich(m);
		wert.setDezimalStellen(m.getStellen());
		wert.setEinheit(m.getEinheit().toString());
		if (DEBUG)
			System.out
					.println("[MessgeraetGoIO] - handleMessbereichEvent: minimum("
							+ m.getMinimum()
							+ ") maximum("
							+ m.getMaximum()
							+ ") Zahlen("
							+ m.getSkalaListe()
							+ ") stellen("
							+ m.getStellen()
							+ ") einheit("
							+ m.getEinheit()
							+ ")");
	}

	public void handleZeitraumEvent(PropertyChangeEvent evt) {
		restartTask();
	}
}
