/* 
 * GHMessen
 * --------
 * 
 * GHMessen is a measuring program to use in connection with a computer 
 * measuring interface. It's designed for school usage.
 * 
 * Copyright (C) 2010-2011  Frank Schütte - Gymnasium Himmelsthür
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms
 * of the GNU General Public License as published by the Free Software Foundation; either 
 * version 3 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with this program; 
 * if not, see <http://www.gnu.org/licenses/>.
 * 
 * Frank Schütte fschuett AT gymnasium-himmelsthuer DOT de
 * 
 */

package ghmessen.instrumente;

import static ghmessen.Konstanten.MESSGERAETE_OFFSET;
import static ghmessen.Konstanten.MESSGERAETE_VERSCHIEBUNG_HORIZONTAL;
import static ghmessen.Konstanten.MESSGERAETE_VERSCHIEBUNG_VERTIKAL;
import static ghmessen.Konstanten.MESSGERAET_BREITE;
import static ghmessen.Konstanten.MESSGERAET_HOEHE;
import static ghmessen.groesse.Art.Formel;
import static ghmessen.groesse.Art.Konstante;
import static ghmessen.groesse.Art.Messgroesse;
import ghmessen.GHMessen;
import ghmessen.adapter.AdapterListe;
import ghmessen.adapter.EAdapterException;
import ghmessen.adapter.IAdapter;
import ghmessen.adapter.IMesskanal;
import ghmessen.adapter.event.AdapterAddedEvent;
import ghmessen.adapter.event.AdapterChangedEvent;
import ghmessen.adapter.event.AdapterEventListener;
import ghmessen.adapter.event.AdapterRemovedEvent;
import ghmessen.groesse.Art;
import ghmessen.groesse.Groesse;
import ghmessen.groesse.Groessenliste;
import ghmessen.util.CursorTk;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.KeyboardFocusManager;
import java.awt.Toolkit;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Vector;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.ScheduledThreadPoolExecutor;

import javax.swing.JComponent;
import javax.swing.JFrame;

import org.jdesktop.observablecollections.ObservableList;
import org.jdesktop.observablecollections.ObservableListListener;
import org.jdesktop.observablecollections.ObservableMap;
import org.jdesktop.observablecollections.ObservableMapListener;

public class MessgeraeteMap extends HashMap<String, Messgeraet> implements
		ObservableMap<String, Messgeraet>, Serializable, ObservableListListener {

	/**
	 * 
	 */
	private static final long serialVersionUID = 8530533245187684853L;
	private static final boolean DEBUG = false;

	private class SymbolPropertyHandler implements Serializable,
			PropertyChangeListener {
		/**
		 * 
		 */
		private static final long serialVersionUID = 1L;

		@Override
		public void propertyChange(PropertyChangeEvent evt) {
			handleSymbolEvent(evt);
		}
	}

	private class ArtPropertyHandler implements Serializable,
			PropertyChangeListener {
		/**
		 * 
		 */
		private static final long serialVersionUID = 1L;

		@Override
		public void propertyChange(PropertyChangeEvent evt) {
			handleArtEvent(evt);
		}
	}

	private class AdapterEventHandler implements AdapterEventListener {
		@Override
		public void adapterAdded(AdapterAddedEvent evt) {
			return;

		}

		@Override
		public void adapterRemoved(AdapterRemovedEvent evt) {
			for (Messgeraet m : map.values())
				if (m.getAdapter().getDevice().equals(evt.getName()))
					m.entferneMessgeraet();
		}

		@Override
		public void adapterChanged(AdapterChangedEvent evt) {
			return;
		}
	}

	private static final HashSet<Art> messgeraete = new HashSet<Art>() {
		private static final long serialVersionUID = 1L;

		{
			add(Messgroesse);
			add(Konstante);
			add(Formel);

		}
	};
	private final boolean supportsElementPropertyChanged;
	private HashMap<String, Messgeraet> map;
	private Groessenliste gl;
	private AdapterListe al;
	private transient List<ObservableMapListener> listeners;

	private SymbolPropertyHandler symbolPH;
	private ArtPropertyHandler artPH;
	private AdapterEventHandler adapterEH;

	public MessgeraeteMap(HashMap<String, Messgeraet> map,
			boolean supportsElementPropertyChanged, Groessenliste gl,
			AdapterListe al) {
		this.map = map;
		this.gl = gl;
		this.al = al;
		adapterEH = new AdapterEventHandler();
		if (this.al != null)
			this.al.getMonitor().addAdapterEventListener(adapterEH);
		symbolPH = new SymbolPropertyHandler();
		artPH = new ArtPropertyHandler();
		if (gl != null) {
			gl.addObservableListListener(this);
			for (int i = 0; i < gl.size(); i++) {
				Groesse g = gl.get(i);
				if (g.isAbhaengig())
					continue;
				g.addPropertyChangeListener("symbol", symbolPH);
				g.addPropertyChangeListener("art", artPH);
				if (!messgeraete.contains(g.getArt()))
					continue;
				Messgeraet m = get(g.getSymbol());
				if (m == null) {
					trageNeuesMessgeraetEin(g);
				}
			}
		}
		listeners = new CopyOnWriteArrayList<ObservableMapListener>();
		this.supportsElementPropertyChanged = supportsElementPropertyChanged;
	}

	public MessgeraeteMap() {
		this(new HashMap<String, Messgeraet>(), false, null, null);
	}

	public MessgeraeteMap(HashMap<String, Messgeraet> map) {
		this(map, false, null, null);
	}

	public MessgeraeteMap(Groessenliste gl, AdapterListe al) {
		this(new HashMap<String, Messgeraet>(), false, gl, al);
	}

	private void writeObject(ObjectOutputStream out) throws IOException {
		out.defaultWriteObject();
	}

	private void readObject(ObjectInputStream in) throws IOException,
			ClassNotFoundException {
		in.defaultReadObject();
		listeners = new CopyOnWriteArrayList<ObservableMapListener>();
		if (gl != null) {
			gl.addObservableListListener(this);
			for (int i = 0; i < gl.size(); i++) {
				Groesse g = gl.get(i);
				Messgeraet m = map.get(g.getSymbol());
				if (m != null) {
					g.addPropertyChangeListener("symbol", symbolPH);
					g.addPropertyChangeListener("art", artPH);
				}
			}
		}
	}

	private Messgeraet erzeugeMessgeraet(Groesse g)
			throws InterruptedException, EAdapterException {
		JFrame mainFrame = GHMessen.getApplication().getMainFrame();
		ScheduledThreadPoolExecutor hauptTimer = GHMessen.getApplication()
				.getMainView().zeitgeber;
		Component focussed = KeyboardFocusManager.getCurrentKeyboardFocusManager().getFocusOwner();
		focussed = focussed instanceof JComponent? ((JComponent)focussed).getTopLevelAncestor():focussed;
		CursorTk.startWaitCursor(focussed);
		Messgeraet result = null;
		if (DEBUG)
			System.out.println("[MessgeraeteMap] - erzeugeMessgeraet: g("+g.getSymbol()+": Nr("+g.getNr()+")|AB("+g.getAB()+")|"+g.getArt()+")");
		if (g.getArt() == Konstante)
			result = new MessgeraetKonstante(mainFrame, g);
		else if (g.getArt() == Formel)
			result = new MessgeraetFormel(mainFrame, gl, g, hauptTimer);
		else {
			IAdapter a = al.get(g.getNr());
			IMesskanal messkanal = a.getMesskanal(g);
			result = messkanal.createMessgeraet(mainFrame, messkanal, hauptTimer);
		}
		CursorTk.stopWaitCursor(focussed);
		return result;
	}

	private void trageNeuesMessgeraetEin(Groesse g) {
		Messgeraet m;
		try {
			m = erzeugeMessgeraet(g);
			if (m == null)
				return;
			put(g.getSymbol(), m);
			int nummer = size();
			Dimension linksoben = Toolkit.getDefaultToolkit().getScreenSize();
			linksoben = new Dimension(linksoben.width - MESSGERAET_BREITE,
					linksoben.height - MESSGERAET_HOEHE);
			int x = (MESSGERAETE_OFFSET + nummer
					* MESSGERAETE_VERSCHIEBUNG_HORIZONTAL)
					% linksoben.width;
			int y = (MESSGERAETE_OFFSET + nummer
					* MESSGERAETE_VERSCHIEBUNG_VERTIKAL)
					% linksoben.height;
			m.setLocation(x, y);
		} catch (InterruptedException e) {
			if (DEBUG)
				e.printStackTrace();
		} catch (EAdapterException e) {
			if (DEBUG)
				e.printStackTrace();
		}

	}

	public void fireMapKeyValueChanged(ObservableMap<?, ?> map, Object key,
			Object lastValue) {
		for (ObservableMapListener listener : listeners)
			listener.mapKeyValueChanged(map, key, lastValue);
	}

	public void fireMapKeyAdded(ObservableMap<?, ?> map, Object key) {
		for (ObservableMapListener listener : listeners)
			listener.mapKeyAdded(map, key);
	}

	public void fireMapKeyRemoved(ObservableMap<?, ?> map, Object key,
			Object value) {
		for (ObservableMapListener listener : listeners)
			listener.mapKeyRemoved(map, key, value);
	}

	public void addObservableMapListener(ObservableMapListener listener) {
		listeners.add(listener);
	}

	public void removeObservableMapListener(ObservableMapListener listener) {
		listeners.remove(listener);
	}

	public boolean supportsElementPropertyChanged() {
		return supportsElementPropertyChanged;
	}

	@Override
	public int size() {
		return map.size();
	}

	@Override
	public boolean isEmpty() {
		return map.isEmpty();
	}

	@Override
	public Messgeraet get(Object key) {
		return map.get(key);
	}

	@Override
	public boolean containsKey(Object key) {
		return map.containsKey(key);
	}

	@Override
	public Messgeraet put(String key, Messgeraet value) {
		Messgeraet alt = map.put(key, value);
		fireMapKeyAdded(this, key);
		return alt;
	}

	@Override
	public void putAll(Map<? extends String, ? extends Messgeraet> m) {
		for (Iterator<? extends Map.Entry<? extends String, ? extends Messgeraet>> i = m
				.entrySet().iterator(); i.hasNext();) {
			Map.Entry<? extends String, ? extends Messgeraet> e = i.next();
			put(e.getKey(), e.getValue());
		}
	}

	@Override
	public Messgeraet remove(Object key) {
		Messgeraet alt = map.remove(key);
		fireMapKeyRemoved(this, key, alt);
		return alt;
	}

	@Override
	public void clear() {
		Vector<String> alle = new Vector<String>();
		for (String s : map.keySet())
			alle.add(s);
		for (String s : alle) {
			Messgeraet alt = map.remove(s);
			fireMapKeyRemoved(this, s, alt);
		}
	}

	@Override
	public boolean containsValue(Object value) {
		return map.containsValue(value);
	}

	@Override
	public Object clone() {
		return map.clone();
	}

	@Override
	public Set<String> keySet() {
		return map.keySet();
	}

	@Override
	public Collection<Messgeraet> values() {
		return map.values();
	}

	@Override
	public Set<java.util.Map.Entry<String, Messgeraet>> entrySet() {
		return map.entrySet();
	}

	@Override
	public void listElementsAdded(
			@SuppressWarnings("rawtypes") ObservableList list, int index,
			int length) {
		if (list instanceof Groessenliste) {
			Groessenliste gl = (Groessenliste) list;
			for (int i = index; i < index + length; i++) {
				Groesse g = gl.get(i);
				if (g.isAbhaengig())
					continue;
				g.addPropertyChangeListener("symbol", symbolPH);
				g.addPropertyChangeListener("art", artPH);
				if (!messgeraete.contains(g.getArt()))
					continue;
				Messgeraet m = get(g.getSymbol());
				if (m == null) {
					trageNeuesMessgeraetEin(g);
				}
			}
		}
	}

	@Override
	public void listElementsRemoved(
			@SuppressWarnings("rawtypes") ObservableList list, int index,
			@SuppressWarnings("rawtypes") List oldElements) {
		if (list instanceof Groessenliste) {
			for (Object o : oldElements) {
				if (!(o instanceof Groesse))
					continue;
				Groesse g = (Groesse) o;
				if (g.isAbhaengig())
					continue;
				g.removePropertyChangeListener("symbol", symbolPH);
				g.removePropertyChangeListener("art", artPH);
				if (!messgeraete.contains(g.getArt()))
					continue;
				Messgeraet m = get(g.getSymbol());
				if (m != null) {
					m.entferneMessgeraet();
					remove(g.getSymbol());
				}
			}
		} else if (list instanceof Groessenliste
				&& oldElements instanceof ArrayList<?>) {
			for (int i = 0; i < oldElements.size(); i++) {
				if (oldElements.get(i) instanceof Groesse) {
					Groesse g = (Groesse) oldElements.get(i);
					if (g.isAbhaengig())
						continue;
					g.removePropertyChangeListener("symbol", symbolPH);
					g.removePropertyChangeListener("art", artPH);
					if (!messgeraete.contains(g.getArt()))
						continue;
					Messgeraet m = get(g.getSymbol());
					if (m != null) {
						m.entferneMessgeraet();
						remove(g.getSymbol());
					}
				}
			}

		}
	}

	@Override
	public void listElementReplaced(
			@SuppressWarnings("rawtypes") ObservableList list, int index,
			Object oldElement) {
		if (DEBUG)
			System.out.println("[MessgeraeteMap] - listElementReplaced: index("
					+ index + ") alte Groesse(" + oldElement + ")");
	}

	@Override
	public void listElementPropertyChanged(
			@SuppressWarnings("rawtypes") ObservableList list, int index) {
		return;
	}

	public void handleSymbolEvent(PropertyChangeEvent evt) {
		Groesse g = (Groesse) evt.getSource();
		if (g.isAbhaengig())
			return;
		Messgeraet m = remove(evt.getOldValue());
		if (m != null) {
			m.entferneMessgeraet();
			trageNeuesMessgeraetEin(g);
		}
	}

	public void handleArtEvent(PropertyChangeEvent evt) {
		Groesse g = (Groesse) evt.getSource();
		if (g.isAbhaengig())
			return;
		Art alt = (Art) evt.getOldValue();
		Art neu = (Art) evt.getNewValue();
		if (messgeraete.contains(alt)) {
			Messgeraet m = remove(g.getSymbol());
			if (m != null) {
				m.entferneMessgeraet();
			}
		}
		if (messgeraete.contains(neu)) {
			Messgeraet m = get(g.getSymbol());
			if (m == null) {
				trageNeuesMessgeraetEin(g);
			}
		}
	}

}
