/* 
 * GHMessen
 * --------
 * 
 * GHMessen is a measuring program to use in connection with a computer 
 * measuring interface. It's designed for school usage.
 * 
 * Copyright (C) 2010-2011  Frank Schütte - Gymnasium Himmelsthür
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms
 * of the GNU General Public License as published by the Free Software Foundation; either 
 * version 3 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with this program; 
 * if not, see <http://www.gnu.org/licenses/>.
 * 
 * Frank Schütte fschuett AT gymnasium-himmelsthuer DOT de
 * 
 */

package ghmessen.messtasks;

import ghmessen.Messparameter;
import ghmessen.adapter.AdapterListe;
import ghmessen.adapter.EAdapterException;
import ghmessen.adapter.MesskanalException;
import ghmessen.groesse.Groesse;
import ghmessen.groesse.Groessenliste;
import ghmessen.groesse.WerteTabelle;

import static ghmessen.groesse.wertetabelle.XAchseTyp.F;
import static ghmessen.groesse.wertetabelle.XAchseTyp.T;
import static ghmessen.groesse.wertetabelle.XAchseTyp.UNABHAENGIG;
import static ghmessen.groesse.Art.Messgroesse;

import java.util.concurrent.ScheduledThreadPoolExecutor;

/* 
 * gemessen wird mit Hilfe der Array-Funktionen der Timer,
 * dabei wird im Abstand einer Sekunde der jeweilige Array ausgelesen, alle anderen
 * Werte werden berechnet und das ganze dargestellt, bis alle Arrays voll sind oder
 * bis die Messzeit abgelaufen ist.
 * In diesem Fall wird die Messtask selbstständig abgebrochen.
 */
public class AnzahlTimerTask extends ZeitTimerTask {
	public static final boolean DEBUG = false;

	protected boolean initialized;
	protected boolean started;
	protected boolean stopped;
	protected double zeitOffset;
	protected boolean start1Flanke;
	
	/*
	 * AnzahlTimerTask ist von der Messzeit abhängig, nicht von der Anzahl der
	 * Wiederholungen. Daher muss die Anzahl der Wiederholungen passend gesetzt 
	 * werden. Außerdem muss das Intervall vom CASSY gelesen werden, da die TimerBox
	 * zwingend intervall <= 0.05 vorschreibt.
	 */
	public AnzahlTimerTask(WerteTabelle gt, Groessenliste gl, Messparameter mp,
			ScheduledThreadPoolExecutor timer, AdapterListe al)
			throws EAdapterException {
		super(gt, gl, mp, timer, al);
		double intervall = Math.min(0.05, mp.getIntervall());
		wiederholung = (int)(mp.getMesszeit()/intervall);
		cassys.setOszilloskopeWiederholung((int)wiederholung);
		cassys.setOszilloskopeIntervall(intervall);
		initialized = false;
		started = false;
		stopped = false;
		zeitOffset = 0.0;
		start1Flanke = false;
	}

	/*
	 * wiederholung = 1 for TimerBoxManuell
	 * @see ghmessen.messtasks.ZeitTimerTask#init()
	 */
	@Override
	public void init() throws EAdapterException {
		super.init();
		if (mp.isEinzelEin())
			wiederholung = 1;
		initialized = true;
	}

	@Override
	public void restart() {
		initialized = false;
		super.restart();
	}

	/*
	 * Notiert die Startzeit in Sekunden.
	 * @see ghmessen.messtasks.ZeitTimerTask#start()
	 */
	@Override
	public void start() throws EAdapterException {
		if (!initialized)
			throw new EAdapterException(
					"[AnzahlTimerTask] - start: Timer ist nicht initialisiert und kann daher nicht gestartet werden.");
		super.start();
		started = true;
	}

	@Override
	public void initGroesse(int i, Groesse g) {
		super.initGroesse(i, g);
		start1Flanke = start1Flanke || g.isStart1Flanke();
	}

	/* Beim TimerArray muss sichergestellt werden, dass der richtige Wert
	 * und die richtige Zeit ausgelesen werden.
	 * @see ghmessen.messtasks.ZeitTimerTask#messeGroesse(int, ghmessen.groesse.Groesse, java.lang.Double[])
	 */
	@Override
	protected void messeGroesse(int i, Groesse g, Double[] d)
			throws MesskanalException, InterruptedException, EAdapterException {
		super.messeGroesse(i, g, d);
		if(g.getArt() == Messgroesse && g.isAktiv() && messkanal[i] != null && messkanal[i].istOszilloskopZeitKanal()) {
			d[i + UNABHAENGIG] = messkanal[i].lesePufferWert((int) anzahl);
			d[T.getIndex()] = messkanal[i].lesePufferZeit((int) anzahl);
			d[F.getIndex()] = 0.0;			
		}
	}

	@Override
	public void einzelMessung() {
		if(cassys.oszilloskopStarted != null) {
			if(cassys.oszilloskopStarted.getCount() > 0)
				return;
			cassys.oszilloskopStarted = null;
		}
		int messwerte = cassys.getTimes();
		// Es liegen soviele Messwerte vor, wie die Anzahl der Zeitwerte angibt. Dabei wir immer die steigende und fallende
		// Flanke registriert.
		if (messwerte <= anzahl)
			return;

		//if (anzahl == 0 && start1Flanke)
		//	zeitOffset = cassys.getTimes(0);
		synchronized (this) {
			while (messwerte > anzahl && anzahl < wiederholung) {
				super.einzelMessung();
				anzahl++;
			}
		}
	}

	@Override
	protected double getTime() {
		return super.getTime() - zeitOffset;
	}
	
	public int minimum(int[] feld) {
		int result;
		if (feld == null || feld.length < 1)
			return -1;
		result = feld[0];
		for (int i = 1; i < feld.length; i++)
			result = Math.min(result, feld[i]);
		return result;
	}

	@Override
	public int prozentFertig() {
		return (int) Math.round((System.currentTimeMillis() - messBeginn)
				/ mp.getMesszeit() / 10.0);	
	}
	
	public boolean fertig() {
		boolean fertig = false;
		if (!mp.isEinzelEin())
			fertig = fertig
					|| (System.currentTimeMillis() - messBeginn) / 1000.0 > mp
							.getMesszeit();
		fertig = fertig || wiederholung > 0 && anzahl >= wiederholung;
		return fertig;
	}

	public boolean hatTrue(boolean[] feld) {
		if (feld == null)
			return false;
		for (int i = 0; i < feld.length; i++)
			if (feld[i])
				return true;
		return false;
	}

	public void setWerteTabelle(WerteTabelle w) {
		this.gt = w;
	}

	public void setGroessenliste(Groessenliste g) {
		this.gl = g;
	}

	@Override
	public void run() {
		if (gt != null && gl != null && mp != null && al != null) {
			if (!started)
				try {
					start();
				} catch (EAdapterException e) {
					if (DEBUG)
						e.printStackTrace();
				}
			einzelMessung();
			if (fertig()) {
				if (mp.isWiederholungEin())
					restart();
				else
					this.cancel();
			}
		}
	}
}
