package ghmessen.messtasks;

import ghmessen.adapter.EAdapterException;

import java.util.ArrayList;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.ScheduledThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

import ld.api.CASSY;
import ld.api.CASSYs;
import ld.api.ICASSY;
import ld.api.ITimer;
import ld.api.OscilloscopeStates;
import ld.api.TriggerBits;
import ld.api.ios.CASSYIO;
import ld.api.ios.CASSYIOs;
import ld.api.ios.CASSYQuantity;
import ld.api.ios.ICASSYQuantity;
import ld.api.ios.MeasurementMode;
import ld.api.ios.OscilloscopeError;

public class CassyVector extends CASSYs {
	private static final long serialVersionUID = 1L;
	private static final boolean DEBUG = false;

	private boolean quantitiesSelected;
	private ScheduledFuture<?> einzelMessung;
	private CountDownLatch einzelMessungComplete;
	private Boolean restartEinzelMessung;
	private OszilloskopMessung oszilloskopMessung;
	public CountDownLatch oszilloskopStarted;
	private CountDownLatch oszilloskopMessungComplete;
	private Boolean oszilloskopWerteVorhanden;
	private ScheduledThreadPoolExecutor timer;

	public CassyVector(ScheduledThreadPoolExecutor timer) {
		super();
		quantitiesSelected = false;
		getCASSYIOs().setAutoDefaults(true);
		einzelMessung = null;
		oszilloskopStarted = null;
		einzelMessungComplete = null;
		restartEinzelMessung = false;
		oszilloskopMessung = null;
		oszilloskopMessungComplete = null;
		oszilloskopWerteVorhanden = false;
		this.timer = timer;
		if (DEBUG)
			System.out.println("[" + this.getClass().toString() + "] Konstruktor ausgeführt.");
	}

	// CASSYs verwendet die Methode add(int i, CASSY c) zum Hinzufügen
	@Override
	public void add(int i, CASSY c) {
		super.add(i, c);
		starteMessung();
	}

	public int getNr(ICASSYQuantity quantity) {
		ICASSY cassy = quantity.getInput().getCASSY();
		for (int index = 0; index < size(); index++)
			if (get(index) == cassy)
				return index;
		return -1;
	}

	public boolean isQuantitiesSelected() {
		return quantitiesSelected;
	}

	public int getTimes() {
		return getCASSYIOs().getTimes();
	}

	public Double getTimes(int i) {
		return getCASSYIOs().getTimes(i);
	}

	// CASSYs verwendet remove(int i) um nicht mehr vorhandene CASSY interfaces zu entfernen
	@Override
	public CASSY remove(int i) {
		stoppeMessung();
		CASSY result = super.remove(i);
		if (size() >= 1) {
			starteMessung();
		}
		return result;
	}

	private boolean istOszilloskopMessung(double intervall) {
		// Falls beide Möglichkeiten erlaubt sind, entscheiden die Intervalle.
		boolean seriell = false;
		boolean einzelAllowed = getCASSYIOs().getSlowOscilloscopeAllowed();
		boolean osziAllowed = getCASSYIOs().getFastOscilloscopeAllowed();
		if (getCommPort().matches(".*tty.*"))
			seriell = true;
		if (osziAllowed && (seriell && intervall < 0.1 || intervall < 0.01))
			return true;
		else if (einzelAllowed)
			return false;
		else if (osziAllowed)
			return true;
		else
			return false;
	}

	public void init(double intervall) {
		quantitiesSelected = false;
		for (CASSYIO io : getCASSYIOs()) {
			if (io.size() > 0) {
				quantitiesSelected = true;
				break;
			}
		}
		if (istOszilloskopMessung(intervall))
			getCASSYIOs().setMeasurementMode(MeasurementMode.FastOscilloscope);
		else
			getCASSYIOs().setMeasurementMode(MeasurementMode.SlowOscilloscope);
	}

	//
	// Messungen
	//
	public boolean starteMessung() {
		if (DEBUG)
			System.out.println("[CassyVector] - starteMessung: " + System.currentTimeMillis());
		if (size() > 0 && timer != null) {
			getCASSYIOs().setMeasurementMode(MeasurementMode.SingleMeasurement);
			if (einzelMessung == null)
				einzelMessung = (ScheduledFuture<?>) timer.scheduleWithFixedDelay(new EinzelMessung(), 0, 100,
						TimeUnit.MILLISECONDS);
			return true;
		} else
			return false;
	}

	public boolean stoppeMessung() {
		if (DEBUG)
			System.out.println("[CassyVector] - stoppeMessung");
		if (timer != null && einzelMessung != null) {
			einzelMessung.cancel(false);
			if (einzelMessungComplete != null)
				try {
					einzelMessungComplete.await(100, TimeUnit.MILLISECONDS);
				} catch (InterruptedException e) {
				}
			einzelMessung = null;
		}
		return true;
	}

	public boolean verwendetTimer() {
		for (CASSY c : this)
			if (c instanceof ITimer)
				return true;
		return false;
	}

	public void setTriggerQuantity(String symbol) {
		for (CASSYIO io : getCASSYIOs())
			for (CASSYQuantity q : io.getSelectedQuantities())
				if (q.getSymbol().equals(symbol)) {
					getCASSYIOs().setTriggerQuantity(q);
					return;
				}
	}

	public void setTrigger(boolean ein) {
		getCASSYIOs().setTrigger(ein);
	}

	public void setTriggerFlanke(TriggerBits trigger) {
		getCASSYIOs().setTriggerRising(!trigger.isAndEmpty(TriggerBits.Rising));
	}

	public void setTriggerValue(double value) {
		getCASSYIOs().setTriggerValue(value);
	}

	public void setOszilloskopeWiederholung(int wiederholung) {
		getCASSYIOs().setOscilloscopeCount(wiederholung);
	}

	public void setOszilloskopeIntervall(double interval) {
		getCASSYIOs().setOscilloscopeInterval(interval);
	}

	public void starteOszilloskope() throws EAdapterException {
		if (DEBUG)
			System.out.println("[CassyVector] - starteOszilloskope");
		if (size() > 0 && timer != null) {
			synchronized (oszilloskopWerteVorhanden) {
				oszilloskopWerteVorhanden = false;
			}
			synchronized (restartEinzelMessung) {
				restartEinzelMessung = einzelMessung != null;
				stoppeMessung();
			}
			if (oszilloskopMessung == null) {
				if (DEBUG)
					System.out.println("[CassyVector] - starteOszilloskope: Oszilloskopmessung (100ms delay).");
				oszilloskopStarted = new CountDownLatch(1);
				oszilloskopMessung = new OszilloskopMessung();
				timer.schedule(oszilloskopMessung, 0, TimeUnit.MILLISECONDS);
			}
		}
	}

	public String getOszilloskopError() {
		OscilloscopeError resultat = getCASSYIOs().getOscilloscopeError();
		if (resultat == OscilloscopeError.None)
			return null;
		if (DEBUG)
			System.err.println("[CassyVector] - getOszilloskopError(" + resultat + ")");
		return "Es ist der Oszilloskopfehler " + resultat.getValue() + " aufgetreten.";
	}

	public void stoppeOszilloskope() throws EAdapterException {
		if (DEBUG)
			System.out.println("[CassyVector] - stoppeOszilloskope");
		if (timer != null && oszilloskopMessung != null) {
			synchronized (oszilloskopMessung.doTerminate) {
				oszilloskopMessung.doTerminate = true;
			}
			if (oszilloskopMessungComplete != null) {
				try {
					oszilloskopMessungComplete.await(100, TimeUnit.MILLISECONDS);
				} catch (InterruptedException e) {
				}
			}
			oszilloskopMessung = null;
		}
		if (einzelMessung == null && restartEinzelMessung) {
			starteMessung();
			synchronized (restartEinzelMessung) {
				restartEinzelMessung = false;
			}
		}
	}

	public boolean istMesswertVorhanden() {
		if (einzelMessung != null)
			return true;
		return oszilloskopWerteVorhanden;
	}

	private class EinzelMessung implements Runnable {
		private static final boolean DEBUG = false;

		@Override
		public void run() {
			if (DEBUG)
				System.out.print("E");
			if (einzelMessungComplete != null && einzelMessungComplete.getCount() > 0)
				return;
			// es darf nur eine Messung zur Zeit laufen, abwarten ist nicht sinnvoll,
			// andernfalls
			// können sich Messungen "aufstauen".
			einzelMessungComplete = new CountDownLatch(1);
			getCASSYIOs().doSingleMeasurement();
			einzelMessungComplete.countDown();
			// verhindert eine Unterbrechung der Einzelmessung, sondern stoppt nach einer
			// Messung
			if (DEBUG)
				System.out.print(".");
		}
	}

	private class OszilloskopMessung implements Runnable {
		private static final boolean DEBUG = false;
		private OscilloscopeStates status = OscilloscopeStates.Stopped;
		public Boolean doTerminate = false;

		@Override
		public void run() {
			if (oszilloskopMessungComplete != null && oszilloskopMessungComplete.getCount() > 0) {
				System.err.println("[OszilloskopMessung] - run: Es läuft bereits eine Messung!");
				return;
			}
			oszilloskopMessungComplete = new CountDownLatch(1);
			while (!doTerminate && !getCASSYIOs().doOscilloscope()) {
				if (oszilloskopStarted != null && oszilloskopStarted.getCount() > 0)
					oszilloskopStarted.countDown();
				if (DEBUG)
					System.out.print("O");
				status = getCASSYIOs().getOscilloscopeStatus();
				synchronized (oszilloskopWerteVorhanden) {
					oszilloskopWerteVorhanden = status != OscilloscopeStates.TriggerWaiting;
				}
				if (DEBUG)
					System.out.print(".");
			}
			getCASSYIOs().stopOscilloscope();
			oszilloskopMessungComplete.countDown();
			if (DEBUG)
				System.out.println("O!");
		}
	}
}
