/* 
 * GHMessen
 * --------
 * 
 * GHMessen is a measuring program to use in connection with a computer 
 * measuring interface. It's designed for school usage.
 * 
 * Copyright (C) 2010-2011  Frank Schütte - Gymnasium Himmelsthür
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms
 * of the GNU General Public License as published by the Free Software Foundation; either 
 * version 3 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with this program; 
 * if not, see <http://www.gnu.org/licenses/>.
 * 
 * Frank Schütte fschuett AT gymnasium-himmelsthuer DOT de
 * 
 */

package ghmessen.messtasks;

import static ghmessen.groesse.wertetabelle.XAchseTyp.BETA;
import static ghmessen.groesse.wertetabelle.XAchseTyp.ENERGIE;
import static ghmessen.groesse.wertetabelle.XAchseTyp.F;
import static ghmessen.groesse.wertetabelle.XAchseTyp.LAMBDA;
import static ghmessen.groesse.wertetabelle.XAchseTyp.N;
import static ghmessen.groesse.wertetabelle.XAchseTyp.T;
import static ghmessen.groesse.wertetabelle.XAchseTyp.UNABHAENGIG;
import ghmessen.Messparameter;
import ghmessen.adapter.AdapterListe;
import ghmessen.adapter.EAdapterException;
import ghmessen.adapter.IEingang;
import ghmessen.adapter.IMesskanal;
import ghmessen.adapter.MesskanalException;
import ghmessen.groesse.Groesse;
import ghmessen.groesse.Groessenliste;
import ghmessen.groesse.WerteTabelle;
import ghmessen.groesse.wertetabelle.XAchseTyp;
import ghmessen.parser.FormelGenerator;
import ghmessen.util.Util;

import java.lang.reflect.InvocationTargetException;
import java.util.Arrays;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.Set;

public abstract class MessTimerTask implements Runnable {
	private static final boolean DEBUG = false;

	protected long wiederholung;
	protected long anzahl;
	protected WerteTabelle gt;
	protected int gtOffset;
	protected Groessenliste gl;
	protected int mrNr;
	protected Messparameter mp;
	protected CassyVector cassys;
	protected AdapterListe al;
	protected IMesskanal[] messkanal;
	protected IEingang[] eingang;
	protected Double[] d;
	protected Double[] dAlt;
	protected Calendar jetzt;
	protected Calendar messStart;
	protected boolean ersterMesswert;
	protected int datum;
	protected int uhrzeit;
	protected boolean warteAufSoftTrigger;

	public MessTimerTask(WerteTabelle gt, Groessenliste gl, Messparameter mp, AdapterListe al) {
		super();
		this.gt = gt;
		this.gl = gl;
		this.mp = mp;
		this.al = al;
		this.cassys = al.getCassys();

		anzahl = 0;
		gtOffset = 0;
		mrNr = 0;
		wiederholung = mp.getWiederholung();
		if (mp.isEinzelEin()) {
			wiederholung = 0;
			if (mp.isUeberschreibenEin()) {
				// Überschreiben
				gt.setNumRows(0);
			} else {
				// isEinzelEin == true && isUeberschreibenEin == false
				// An vorhandene Messreihe anhängen
				if (gt.getRowCount() > 0) {
					dAlt = gt.getRow(gt.getRowCount() - 1);
					mrNr = gl.getMaxMessreihe();
				}
			}

		} else {
			// isEinzelEin == false
			// bei kurzer Messzeit diese als Begrenzung verwenden
			if (mp.getMesszeit() < mp.getIntervall() * mp.getWiederholung())
				wiederholung = 0;
			if (wiederholung == 0) {
				// Anzahl über die Messzeit gesteuert
				wiederholung = (long) (mp.getMesszeit() / mp.getIntervall());
			}
			if (mp.isUeberschreibenEin()) {
				// Überschreiben
				gt.setNumRows(0);
			} else {
				// isEinzelEin == false && isUeberschreibenEin == false
				if (mp.isAnhaengenEin()) {
					if (mp.isFortsetzenEin()) {
						// An vorhandene Messreihe anhängen
						if (gt.getRowCount() > 0) {
							dAlt = gt.getRow(gt.getRowCount() - 1);
							mrNr = gl.getMaxMessreihe();
						}
					} else {
						// anhängen, aber nicht fortsetzen
						dAlt = null;
						mrNr = gl.getMaxMessreihe();
					}
				} else {
					// isEinzelEin == false && isUeberschreibenEin == false &&
					// isAnhaengenEin == false
					// Neue Messreihe erzeugen
					if (gt.getRowCount() > 0) {
						mrNr = gl.addMessreihe();
					} else {
						mrNr = 0;
					}
				}
			}
		}
	}

	@Override
	public void run() {
	}

	/*
	 * initialisiert alle benötigten Datenstrukturen und initialisiert für
	 * Formelgrößen die Berechnungsobjekte
	 */
	public void init() throws EAdapterException {
		if (DEBUG)
			System.out.println("[MessTimerTask] - init...");
		if (gl == null)
			throw new EAdapterException("Es sind keine Messgrößen vorhanden.");
		messkanal = new IMesskanal[gl.getSize()];
		eingang = new IEingang[gl.getSize()];
		Arrays.fill(eingang, null);
		for (int i = 0; i < gl.getSize(); i++) {
			Groesse g = gl.get(i);
			if (g.getMessreihe() != mrNr)
				continue;
			initGroesse(i, g);
		}
		if (mp.isEinzelEin()) {
			if (mp.isUeberschreibenEin()) {
				dAlt = null;
				messStart = null;
				ersterMesswert = true;
			} else {
				// isEinzelEin == true && isUeberschreibenEin == false
				if (dAlt != null) {
					messStart = new GregorianCalendar();
					messStart.setTimeInMillis((long) (messStart.getTimeInMillis()
							- dAlt[T.getIndex()] * (1000.0 + 1000.0 / dAlt[N.getIndex()])));
					ersterMesswert = false;
				} else {
					ersterMesswert = true;
				}
			}
		} else {
			// isEinzelEin == false
			if (mp.isUeberschreibenEin()) {
				dAlt = null;
				messStart = null;
				ersterMesswert = true;
			} else {
				// isEinzelEin == false && isUeberschreibenEin == false
				if (mp.isAnhaengenEin()) {
					if (mp.isFortsetzenEin()) {
						// An vorhandene Messreihe anhängen
						if (dAlt != null) {
							messStart = new GregorianCalendar();
							messStart.setTimeInMillis((long) (messStart.getTimeInMillis()
									- dAlt[T.getIndex()] * (1000.0 + 1000.0 / dAlt[N.getIndex()])));
							ersterMesswert = false;
						} else
							ersterMesswert = true;
					} else {
						// anhängen, aber nicht fortsetzen
						dAlt = null;
						messStart = null;
						ersterMesswert = true;
						if (gt.getRowCount() > 0) {
							gt.addEmptyRow();
							gtOffset = gt.getRowCount();
						}
					}
				} else {
					// isEinzelEin == false && isUeberschreibenEin == false &&
					dAlt = null;
					messStart = null;
					ersterMesswert = true;
					if (gt.getRowCount() > 0)
						dAlt = gt.getRow(0);
				}
			}
		}
	}

	protected void initGroesse(int i, Groesse g) {
		switch (g.getArt()) {
		case Formel:
			if (g.getFormelterm() == null)
				g.createFormelterm(gl.getTextSymbole());
			break;
		default:
			break;
		}
	}

	protected double getTime() {
		return (jetzt.getTimeInMillis() - messStart.getTimeInMillis()) / 1000.0;
	}

	protected void einzelMessung() {
		if (gl.getSize() > 0) { // Messgröße vorhanden
			Double[] d = new Double[gt.getColumnCount()];
			jetzt = new GregorianCalendar();
			if (ersterMesswert)
				messStart = jetzt;
			datum = ghmessen.util.Math.datum(jetzt);
			uhrzeit = ghmessen.util.Math.uhrzeit(jetzt);
			if (mrNr == 0 || ersterMesswert && gt.getRowCount() < 1
					|| !ersterMesswert && gt.getRowCount() < gtOffset + dAlt[N.getIndex()] + 1) {
				d[T.getIndex()] = getTime();
				d[F.getIndex()] = 0.0;
				d[N.getIndex()] = ersterMesswert ? 1 : dAlt[N.getIndex()] + 1;
				d[BETA.getIndex()] = 0.0;
				d[LAMBDA.getIndex()] = 0.0;
				d[ENERGIE.getIndex()] = 0.0;
			} else
				d = gt.getRow(gtOffset + dAlt[N.getIndex()].intValue());
			for (int i = 0; i < gl.getSize(); i++) {
				Groesse g = gl.get(i);
				try {
					if (g.getMessreihe() != mrNr)
						continue;
					messeGroesse(i, g, d);
				} catch (NullPointerException e) {
					d[i + UNABHAENGIG] = null;
				} catch (MesskanalException e) {
					if (DEBUG)
						e.printStackTrace();
					d[i + UNABHAENGIG] = null;
				} catch (InterruptedException e) {
					if (DEBUG)
						e.printStackTrace();
					d[i + UNABHAENGIG] = null;
				} catch (EAdapterException e) {
					if (DEBUG)
						e.printStackTrace();
					d[i + UNABHAENGIG] = null;
				}
			}
			schreiben(d);
		}
	}

	public void schreiben(Double[] d) {
		if (d != null) {
			gt.replaceRow(gtOffset + d[N.getIndex()].intValue() - 1, d, false, true);
			dAlt = d;
			ersterMesswert = false;
		}
	}

	protected void messeGroesse(int i, Groesse g, Double[] d)
			throws MesskanalException, InterruptedException, EAdapterException {
		switch (g.getArt()) {
		case Parameter:
		case Konstante:
			d[i + UNABHAENGIG] = g.getWert();
			break;
		case Ableitung:
			int bezug = gl.indexOf(g.getBezug()) + UNABHAENGIG;
			if (dAlt != null && dAlt[bezug] != null && dAlt[T.getIndex()] != null && d[bezug] != null)
				d[i + UNABHAENGIG] = (d[bezug] - dAlt[bezug]) / (d[T.getIndex()] - dAlt[T.getIndex()]); // Differenzenquotient
			else
				d[i + UNABHAENGIG] = null;
			break;
		case Integral:
			bezug = gl.indexOf(g.getBezug()) + UNABHAENGIG;
			if (dAlt != null && dAlt[bezug] != null && dAlt[T.getIndex()] != null && d[bezug] != null)
				d[i + UNABHAENGIG] = dAlt[i + UNABHAENGIG]
						+ (d[T.getIndex()] - dAlt[T.getIndex()]) * (d[bezug] + dAlt[bezug]) / 2.0; // Trapezformel
			else
				d[i + UNABHAENGIG] = 0.0;
			break;
		case Formel:
			/* Formelparameter in absolute Zahlen umrechnen */
			Double[] params = Util.dataAbsoluteFromRel(d, UNABHAENGIG, d.length, gl);
			try {
				d[i + UNABHAENGIG] = (Double) g.getFormelterm()
						.evaluate(new Object[] { d[T.getIndex()], d[N.getIndex()], d[F.getIndex()], d[BETA.getIndex()],
								d[LAMBDA.getIndex()], d[ENERGIE.getIndex()],
								Double.valueOf(dAlt != null ? d[T.getIndex()] - dAlt[T.getIndex()] : mp.getIntervall()),
								datum, uhrzeit, params, gt });
			} catch (InvocationTargetException e) {
				if (DEBUG)
					e.printStackTrace();
				d[i + UNABHAENGIG] = null;
			}
			break;
		default:
			d[i + UNABHAENGIG] = 0.0;
			break;
		}
	}

	public void start() throws EAdapterException {
		if (DEBUG)
			System.out.println("[MessTimerTask] - start...");
		if (mp.isTriggerEin())
			warteAufSoftTrigger = true;
	}

	private void ergaenzeBerechnungenAmEnde() {
		/* Berechnung der Histogrammwerte/FFT-Werte im Anschluss an die Messung */
		for (int i = 0; i < gl.getSize(); i++) {
			Groesse g = gl.get(i);
			if (g.getMessreihe() != mrNr)
				continue;
			switch (g.getArt()) {
			case Histogramm:
				if (g.getBezug() == null)
					break;
				gt.spalteBerechnen(i + UNABHAENGIG, g);
				break;
			case FFT: {
				if (g.getBezug() == null)
					break;
				gt.spalteBerechnen(i + UNABHAENGIG, g);
				break;
			}
			default:
				break;
			}
		}
	}

	public void stop() {
		if (DEBUG)
			System.out.println("[MessTimerTask] - stop...");
		cancel();
	}

	public boolean cancel() {
		ergaenzeBerechnungenAmEnde();
		return true;
	}

	public void restart() {
		if (DEBUG)
			System.out.println("[MessTimerTask] - restart...");
		try {
			if (!mp.isUeberschreibenEin() && !mp.isAnhaengenEin()) {
				// Neue Messreihe erzeugen
				if (gt.getRowCount() > 0) {
					mrNr = gl.addMessreihe();
				} else
					mrNr = 0;
			} else if ((mp.isEinzelEin() || (mp.isAnhaengenEin() && mp.isFortsetzenEin())) && gt.getRowCount() > 0) {
				// letzte Messwerte verwenden
				dAlt = gt.getRow(gt.getRowCount() - 1);
				mrNr = gl.getMaxMessreihe();
			} else if (!mp.isEinzelEin() && mp.isAnhaengenEin() && !mp.isFortsetzenEin()) {
				// anhängen, aber nicht fortsetzen
				mrNr = gl.getMaxMessreihe();
				dAlt = null;
			} else {
				gt.setNumRows(0);
				mrNr = 0;
			}
			anzahl = 0;
			init();
			start();
		} catch (EAdapterException e) {
			if (DEBUG)
				e.printStackTrace();
			this.cancel();
		}
	}

	public void setWerteTabelle(WerteTabelle gt) {
		this.gt = gt;
	}

	public WerteTabelle getWerteTabelle() {
		return gt;
	}

	public void setGroessenliste(Groessenliste gl) {
		this.gl = gl;
	}

	public Groessenliste getGroessenliste() {
		return gl;
	}

	public void setMessparameter(Messparameter mp) {
		this.mp = mp;
	}

	public Messparameter getMessparameter() {
		return mp;
	}

	public void setCassys(CassyVector cassys) {
		this.cassys = cassys;
	}

	public CassyVector getCassys() {
		return cassys;
	}

	public void setWiederhoung(long wiederholung) {
		this.wiederholung = wiederholung;
		anzahl = 0;
	}

	public long getWiederholung() {
		return wiederholung;
	}

	public long getAnzahl() {
		return anzahl;
	}

	public void setAnzahl(long anzahl) {
		this.anzahl = (anzahl > wiederholung) ? wiederholung : anzahl;
	}

	public abstract int prozentFertig();

	public abstract boolean messBedingungErfuellt();

	public abstract boolean stoppBedingungErfuellt();

	public abstract boolean triggerBedingungErfuellt();
}
