/* 
 * GHMessen
 * --------
 * 
 * GHMessen is a measuring program to use in connection with a computer 
 * measuring interface. It's designed for school usage.
 * 
 * Copyright (C) 2010-2011  Frank Schütte - Gymnasium Himmelsthür
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms
 * of the GNU General Public License as published by the Free Software Foundation; either 
 * version 3 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with this program; 
 * if not, see <http://www.gnu.org/licenses/>.
 * 
 * Frank Schütte fschuett AT gymnasium-himmelsthuer DOT de
 * 
 */

package ghmessen.messtasks;

import static ghmessen.groesse.wertetabelle.XAchseTyp.BETA;
import static ghmessen.groesse.wertetabelle.XAchseTyp.ENERGIE;
import static ghmessen.groesse.wertetabelle.XAchseTyp.F;
import static ghmessen.groesse.wertetabelle.XAchseTyp.LAMBDA;
import static ghmessen.groesse.wertetabelle.XAchseTyp.N;
import static ghmessen.groesse.wertetabelle.XAchseTyp.T;
import static ghmessen.groesse.wertetabelle.XAchseTyp.UNABHAENGIG;
import ghmessen.Messparameter;
import ghmessen.adapter.AdapterListe;
import ghmessen.adapter.EAdapterException;
import ghmessen.adapter.IAdapter;
import ghmessen.adapter.IMesskanal;
import ghmessen.adapter.MesskanalException;
import ghmessen.groesse.Groesse;
import ghmessen.groesse.Groessenliste;
import ghmessen.groesse.WerteTabelle;
import ghmessen.util.Util;

import java.lang.reflect.InvocationTargetException;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Vector;
import java.util.concurrent.ScheduledThreadPoolExecutor;

/* 
 * gemessen wird mit Hilfe der Array-Funktionen Oscilloscope,
 * dabei wird im Abstand einer Sekunde der jeweilige Array ausgelesen, alle anderen
 * Werte werden berechnet und das ganze dargestellt, bis alle Arrays voll sind.
 * In diesem Fall wird die Messtask selbstständig abgebrochen.
 */
public class OszilloskopTimerTask extends MessTimerTask {
	public static final boolean DEBUG = false;

	protected Double[][] daten; 
	// Spalten : Größen+t,n , Zeilen : Messwerte einer Größe
	protected int[] gelesen; // Anzahl gelesener Werte
	protected int grZahl; // Anzahl der abhängigen Größen
	protected int geschrieben; // bisher in die Tabelle geschriebene Zeilen
	protected int gesamt; // Anzahl zu lesender Werte
	protected boolean initialized;
	protected boolean started;
	protected boolean startFG;
	protected double messBeginn;

	public OszilloskopTimerTask(WerteTabelle gt, Groessenliste gl,
			Messparameter mp, ScheduledThreadPoolExecutor timer, AdapterListe al)
			throws EAdapterException {
		super(gt, gl, mp, al);
		initialized = false;
		startFG = false;
		messBeginn = 0;
		started = false;
		grZahl = this.gl.getSize();
		gesamt = (int) mp.getWiederholung();
		geschrieben = 0;
	}

	private void fuelleUnabhaengigeSpalten() throws EAdapterException {
		if (daten == null || daten.length < gesamt)
			throw new EAdapterException(
					"[OszilloskopTimerTask] Variable daten zu klein(" + daten
							+ ")");
		// t, n,f Spalten füllen
		for (int i = 0; i < gesamt; i++) {
			daten[i][T.getIndex()] = i * mp.getIntervall();
			daten[i][N.getIndex()] = (double) (i + 1);
			daten[i][F.getIndex()] = i == 0 ? 1.0 : 1.0 / daten[i][T.getIndex()];
			daten[i][BETA.getIndex()] = 0.0;
			daten[i][LAMBDA.getIndex()] = 0.0;
			daten[i][ENERGIE.getIndex()] = 0.0;
		}
	}

	private void fuelleAbhaengigeSpalten() throws EAdapterException {
		for (int i = 0; i < grZahl; i++) {
			Groesse g = gl.get(i);
			switch (g.getArt()) {
			case Integral:
				daten[0][i + UNABHAENGIG] = 0.0;
				gelesen[i] = 1;
				break;
			case Parameter:
			case Konstante:
				for (int j = 0; j < gesamt; j++)
					daten[j][i + UNABHAENGIG] = g.getWert();
				gelesen[i] = gesamt;
				break;
			default:
				break;
			}
		}
	}

	@Override
	public void init() throws EAdapterException {
		cassys.init(mp.getIntervall());
		cassys.setOszilloskopeWiederholung((int) wiederholung);
		cassys.setOszilloskopeIntervall(mp.getIntervall());
		super.init();
		daten = new Double[(int) mp.getWiederholung()][grZahl + UNABHAENGIG];
		gelesen = new int[grZahl];
		Arrays.fill(gelesen, 0);
		fuelleUnabhaengigeSpalten();
		fuelleAbhaengigeSpalten();
		/* Benötigte Oszilloskope initialisieren */
		cassys.setTrigger(mp.isTriggerEin());
		if(mp.isTriggerEin()){
			cassys.setTriggerQuantity(mp.getTriggerSymbol());
			cassys.setTriggerFlanke(mp.getTriggerFlanke());
			cassys.setTriggerValue(mp.getTriggerWert());
		}
		if (mp.isBedingungEin())
			mp.setBedingungEin(false);
		if (mp.isStoppEin())
			mp.setStoppEin(false);
		warteAufSoftTrigger = false;
		initialized = true;
	}

	@Override
	protected void initGroesse(int i, Groesse g) {
		switch (g.getArt()) {
		case Messgroesse:
			if (!g.isAktiv())
				return;
			IAdapter a = al.get(g.getNr());
			if(a != null) {
				eingang[i] = a.getEingang(g);
				messkanal[i] = a.getMesskanal(g);
			}
			if (messkanal[i] != null) {
				try {
					messkanal[i].setAbstand(0.5 * mp.getIntervallMS());
				} catch (MesskanalException e) {
					if (DEBUG)
						e.printStackTrace();
				}
				if(!mp.isTriggerEin() && g.isAusloeserTriggert() 
						&& messkanal[i].istTriggerKanal()) {
					cassys.setTriggerQuantity(g.getSymbol());
				}
			}
			break;
		default:
			super.initGroesse(i, g);
			break;
		}
	}

	@Override
	public void restart() {
		initialized = false;
		geschrieben = 0;
		messBeginn = 0;
		super.restart();
	}

	@Override
	public void start() throws EAdapterException {
		super.start();
		if (!initialized)
			throw new EAdapterException(
					"[OszilloskopTimerTask] - start: Oszilloskop nicht initialisiert, soll aber gestartet werden.");
		else if(cassys.getOszilloskopError() != null)
			throw new EAdapterException("[OszilloskopTimerTask] - start: Oszilloskopfehler("+cassys.getOszilloskopError()+")");
		try {
			for (IMesskanal m : messkanal)
				if (m != null)
					m.start();
			cassys.starteOszilloskope();
		} catch (MesskanalException e) {
			if (DEBUG)
				e.printStackTrace();
		}
		messBeginn = System.currentTimeMillis();
		started = true;
	}

	@Override
	public void stop() {
		for (IMesskanal m : messkanal)
			try {
				if (m != null)
					m.stop();
				cassys.stoppeOszilloskope();
			} catch (MesskanalException e) {
				if (DEBUG)
					e.printStackTrace();
			} catch (EAdapterException e) {
				if (DEBUG)
					e.printStackTrace();
			}
		cancel();
	}

	@Override
	public boolean cancel() {
		try {
			cassys.stoppeOszilloskope();
		} catch (EAdapterException e) {
			if (DEBUG)
				e.printStackTrace();
		}
		if (DEBUG)
			System.out.println("[OszilloskopTimerTask] - Messzeit("
					+ ((System.currentTimeMillis() - messBeginn) / 1000.0)
					+ ")");
		return super.cancel();
	}

	@Override
	public void einzelMessung() {
		if(cassys.oszilloskopStarted != null) {
			if( cassys.oszilloskopStarted.getCount() > 0)
				return;
			else
				cassys.oszilloskopStarted = null;
		}
		jetzt = Calendar.getInstance();
		datum = ghmessen.util.Math.datum(jetzt);
		uhrzeit = ghmessen.util.Math.uhrzeit(jetzt);
		for (int i = 0; i < grZahl; i++) {
			Groesse g = gl.get(i);
			switch (g.getArt()) {
			case Messgroesse:
				if (messkanal[i].istOszilloskopKanal()) {
					if (g.isAktiv()) {
						try {
							int n = messkanal[i].fuellePuffer(gelesen[i],
									gesamt - gelesen[i]);
							System.out
									.println("[OszilloskopTimerTask] - Größe("
											+ i + ") neu(" + n + ") von("+gesamt+")");
							n += gelesen[i];
							for (int n1 = gelesen[i]; n1 < n; n1++)
								try {
									daten[gelesen[i]++][i + UNABHAENGIG] = messkanal[i]
											.lesePufferWert(n1);
								} catch (InterruptedException e) {
									if (DEBUG)
										e.printStackTrace();
									daten[gelesen[i]++][i + UNABHAENGIG] = null;
								} catch (EAdapterException e) {
									if (DEBUG)
										e.printStackTrace();
									daten[gelesen[i]++][i + UNABHAENGIG] = null;
								}
						} catch (InterruptedException e) {
							if (DEBUG)
								e.printStackTrace();
						} catch (EAdapterException e) {
							if (DEBUG)
								e.printStackTrace();
							// Dieser Kanal ist kein Oszilloskopkanal und wird
							// daher mit null gefüllt.
							for (int ii = gelesen[i]; ii < gesamt; ii++)
								daten[ii][i + UNABHAENGIG] = null;
							gelesen[i] = gesamt;
						}
					} else if (!g.isAktiv() && gelesen[i] < gesamt) {
						for (int ii = gelesen[i]; ii < gesamt; ii++)
							daten[ii][i + UNABHAENGIG] = null;
						gelesen[i] = gesamt;
					}
				}
				break;
			case Ableitung:
				int bezug = gl.indexOf(g.getBezug()) + UNABHAENGIG;
				if (bezug >= UNABHAENGIG) {
					while (gelesen[i] < gelesen[bezug])
						daten[++gelesen[i]][i + UNABHAENGIG] = (daten[gelesen[i]][bezug] - daten[gelesen[i] - 1][bezug])
								/ (daten[gelesen[i]][T.getIndex()] - daten[gelesen[i] - 1][T.getIndex()]); // Differenzenquotient
				}
				break;
			case Integral:
				bezug = gl.indexOf(g.getBezug()) + UNABHAENGIG;
				if (bezug >= UNABHAENGIG) {
					while (gelesen[i] < gelesen[bezug])
						daten[++gelesen[i]][i + UNABHAENGIG] = daten[gelesen[i] - 1][i
								+ UNABHAENGIG]
								+ (daten[gelesen[i]][bezug] + daten[gelesen[i] - 1][bezug])
								* (daten[gelesen[i]][T.getIndex()] - daten[gelesen[i] - 1][T.getIndex()])
								/ 2.0; // Trapezformel
				}
				break;
			case Formel:
				int max = minimumBisIndex(gelesen, i);
				while (gelesen[i] < max)
					try {
						Double[] params = Util.dataAbsoluteFromRel(daten[gelesen[i]], UNABHAENGIG, daten[gelesen[i]].length, gl);
						daten[gelesen[i]][i + UNABHAENGIG] = (Double) g
								.getFormelterm()
								.evaluate(
										new Object[] {
												daten[gelesen[i]][T.getIndex()],
												Double.valueOf(gelesen[i]),
												daten[gelesen[i]][F.getIndex()],
												daten[gelesen[i]][BETA.getIndex()],
												daten[gelesen[i]][LAMBDA.getIndex()],
												daten[gelesen[i]][ENERGIE.getIndex()],
												Double.valueOf(mp.getIntervall()),
												datum,
												uhrzeit,
												params,
												gt });
						gelesen[i]++;
					} catch (InvocationTargetException e) {
						if (DEBUG)
							e.printStackTrace();
						daten[gelesen[i]++][i + UNABHAENGIG] = null;
					} catch (Exception e) {
						if (DEBUG)
							e.printStackTrace();
						System.err
								.println("[OszilloskopTimerTask] - einzelMessung: Exception("
										+ e.getLocalizedMessage() + ")");
					}
				break;
			default:
				break;
			}
		}
		werteSchreiben();
	}

	public void werteSchreiben() {
		if (daten == null || gt == null)
			return;
		int minGelesen = minimum(gelesen);
		if (minGelesen <= geschrieben)
			return;
		@SuppressWarnings("rawtypes")
		Vector<Vector> v = new Vector<Vector>(minGelesen
				- geschrieben);
		for (int i = geschrieben; i < minGelesen; i++) {
			Vector<Double> row = new Vector<Double>(grZahl + UNABHAENGIG);
			for (int j = 0; j < grZahl + UNABHAENGIG; j++)
				row.add(daten[i][j]);
			v.add(row);
		}
		gt.addRows(v);
		geschrieben = minGelesen;
	}

	private int minimum(int[] feld) {
		return minimumBisIndex(feld, feld.length);
	}

	private int minimumBisIndex(int[] feld, int index) {
		int result;
		if (feld == null || feld.length < 1 || feld.length < index)
			return -1;
		result = feld[0];
		for (int i = 1; i < index; i++)
			result = Math.min(result, feld[i]);
		return result;

	}

	@Override
	public int prozentFertig() {
		return (int) Math.round(geschrieben * 100.0 / gesamt);
	}

	@Override
	public void run() {
		super.run();
		try {
			if (!started)
				start();
			if (gt != null && gl != null && mp != null && al != null) {
				einzelMessung();
				if (prozentFertig() >= 100) {
					if (mp.isWiederholungEin())
						restart();
					else
						this.cancel();
				}
			}
		} catch (EAdapterException e) {
			if (DEBUG)
				e.printStackTrace();
		}

	}

	@Override
	public boolean messBedingungErfuellt() {
		return true;
	}

	@Override
	public boolean stoppBedingungErfuellt() {
		return false;
	}

	@Override
	public boolean triggerBedingungErfuellt() {
		return true;
	}

}
