/* 
 * GHMessen
 * --------
 * 
 * GHMessen is a measuring program to use in connection with a computer 
 * measuring interface. It's designed for school usage.
 * 
 * Copyright (C) 2010-2011  Frank Schütte - Gymnasium Himmelsthür
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms
 * of the GNU General Public License as published by the Free Software Foundation; either 
 * version 3 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with this program; 
 * if not, see <http://www.gnu.org/licenses/>.
 * 
 * Frank Schütte fschuett AT gymnasium-himmelsthuer DOT de
 * 
 */

package ghmessen.messtasks;

import static ghmessen.groesse.wertetabelle.XAchseTyp.BETA;
import static ghmessen.groesse.wertetabelle.XAchseTyp.ENERGIE;
import static ghmessen.groesse.wertetabelle.XAchseTyp.F;
import static ghmessen.groesse.wertetabelle.XAchseTyp.LAMBDA;
import static ghmessen.groesse.wertetabelle.XAchseTyp.N;
import static ghmessen.groesse.wertetabelle.XAchseTyp.T;
import static ghmessen.groesse.wertetabelle.XAchseTyp.UNABHAENGIG;
import ghmessen.Messparameter;
import ghmessen.adapter.AdapterListe;
import ghmessen.adapter.EAdapterException;
import ghmessen.adapter.IAdapter;
import ghmessen.groesse.Groesse;
import ghmessen.groesse.Groessenliste;
import ghmessen.groesse.WerteTabelle;

import java.util.Arrays;
import java.util.Vector;
import java.util.concurrent.ScheduledThreadPoolExecutor;

import ld.api.IXRay;
import ld.api.XRayData;
import ld.api.XRayModes;
import ld.api.ios.IWaveform;

/* 
 * gemessen wird winkelabhängig mit Hilfe der Daten aus dem Röntgengerät
 * 
 *  Die Messparameter bestimmen die Größen und Intervalle für die Winkel.
 *  
 */
public class XRayTimerTask extends MessTimerTask {
	public static final boolean DEBUG = false;

	protected Double[][] daten; // Spalten : Größen+t,n , Zeilen : Messwerte
								// einer
								// Größe
	protected int[] gelesen; // Anzahl gelesener Werte
	protected int grZahl; // Anzahl der abhängigen Größen
	protected int geschrieben; // bisher in die Tabelle geschriebene Zeilen
	protected int gesamt; // Anzahl zu lesender Werte
	protected boolean initialized = false;
	protected boolean started = false;
	protected boolean startFG = false;
	protected ScheduledThreadPoolExecutor timer;
	protected XRayData xrp;
	protected IXRay xray;

	public XRayTimerTask(WerteTabelle gt, Groessenliste gl, Messparameter mp,
			ScheduledThreadPoolExecutor timer, AdapterListe al)
			throws EAdapterException {
		super(gt, gl, mp, al);
		this.timer = timer;
		this.xrp = mp.parameter;
		if (xrp == null) {
			if (DEBUG)
				System.out
						.println("[XRayTimerTask] - XRayTimerTask() Fehler: xrp == null");
			xrp = new XRayData();
		}
		grZahl = this.gl.getSize();
		xray = null;
		for (IAdapter a : al.values())
			if (a instanceof IXRay) {
				xray = (IXRay) a;
				break;
			}
		if (xrp != null && xrp.getDeltaBeta() != 0)
			gesamt = (int) Math.round((xrp.getBeta2() - xrp.getBeta1())
					/ xrp.getDeltaBeta() + 1.0);
		else if (xrp != null)
			gesamt = 1;
		wiederholung = gesamt;
		geschrieben = 0;
	}

	@Override
	public void init() throws EAdapterException {
		daten = new Double[gesamt][grZahl + UNABHAENGIG];
		gelesen = new int[grZahl];
		Arrays.fill(gelesen, 0);
		super.init();
		// t, n,f, beta, lambda, E Spalten füllen
		for (int i = 0; i < gesamt; i++) {
			daten[i][T.getIndex()] = (double) i;
			daten[i][N.getIndex()] = (double) (i + 1);
			daten[i][F.getIndex()] = i == 0 ? 1.0 : 1.0 / daten[i][T.getIndex()];
			daten[i][BETA.getIndex()] = xrp.getBeta1() + i * xrp.getDeltaBeta();
			daten[i][LAMBDA.getIndex()] = mp.getXrg().getD_pm()
					* Math.sin(Math.toRadians(daten[i][BETA.getIndex()]));
			daten[i][ENERGIE.getIndex()] = mp.getXrg().getE_keV()
					* daten[i][LAMBDA.getIndex()];
		}
		for (int i = 0; i < grZahl; i++) {
			if (gl.get(i).getMessreihe() != mrNr)
				gelesen[i] = gesamt;
		}
		initialized = true;
	}

	@Override
	public void initGroesse(int i, Groesse g) {
		super.initGroesse(i, g);
		if (g.getMessreihe() != mrNr)
			return;
		switch (g.getArt()) {
		case Messgroesse:
			if (!g.isAktiv())
				return;
			IAdapter a = al.get(g.getNr());
			if( a != null) {
				eingang[i] = a.getEingang(g);
				messkanal[i] = a.getMesskanal(g);
			}
			if (messkanal[i] != null && messkanal[i] instanceof IWaveform && g.isSynchron())
				startFG = true;
			break;
		default:
			super.initGroesse(i, g);
			break;
		case Parameter:
		case Konstante:
			for (int j = 0; j < gesamt; j++)
				daten[j][i + UNABHAENGIG] = g.getWert();
			gelesen[i] = gesamt;
			break;
		case Integral:
			daten[0][i + UNABHAENGIG] = 0.0;
			gelesen[i] = 1;
			break;
		}
	}

	@Override
	public void restart() {
		initialized = false;
		geschrieben = 0;
		super.restart();
	}

	@Override
	public void start() throws EAdapterException {
		if (!initialized)
			throw new EAdapterException(
					"[XRayTimerTask] nicht initialisiert, soll aber gestartet werden.");
		if (wiederholung == 1)
			return;
		if (startFG)
			cassys.starteOszilloskope();
		if (xray != null) {
			stopScan();
			xray.setMode(new XRayModes(XRayModes.Scan));
			if (xray.getMode().isAndEmpty(XRayModes.Scan))
				throw new EAdapterException(
						"[XRayTimerTask] - start: Scan konnte nicht gestartet werden, Türen korrekt geschlossen?");
		}
		started = true;
	}

	@Override
	public void stop() {
		try {
			if (startFG)
				cassys.stoppeOszilloskope();
		} catch (EAdapterException e) {
			if (DEBUG)
				e.printStackTrace();
		}
		if (xray != null) {
			XRayModes alt = xray.getMode();
			if (!alt.isAndEmpty(XRayModes.Scan))
				xray.setMode(new XRayModes(XRayModes.Scan));
		}
	}

	private int werteEinlesen(int i, Groesse g) {
		int bisher = gelesen[i];
		switch (g.getArt()) {
		case Messgroesse:
			if (g.isAktiv() && messkanal[i] != null) {
				try {
					int n = messkanal[i].fuellePuffer(gelesen[i], gesamt
							- gelesen[i]);
					if (DEBUG)
						System.out.println("[XRayTimerTask] - werteEinlesen("
								+ n + ")");
					for (int n1 = 0; n1 < n; n1++) {
						try {
							daten[gelesen[i]][i + UNABHAENGIG] = messkanal[i]
									.lesePufferWert(n1);
						} catch (InterruptedException e) {
							if (DEBUG)
								e.printStackTrace();
							daten[gelesen[i]][i + UNABHAENGIG] = null;
						} catch (EAdapterException e) {
							if (DEBUG)
								e.printStackTrace();
							daten[gelesen[i]][i + UNABHAENGIG] = null;
						}
						gelesen[i]++;
					}
				} catch (InterruptedException e) {
					if (DEBUG)
						e.printStackTrace();
				} catch (EAdapterException e) {
					if (DEBUG)
						e.printStackTrace();
				}
			} else if (!g.isAktiv()) {
				if (gelesen[i] < gesamt) {
					for (int ii = gelesen[i]; ii < gesamt; ii++)
						daten[ii][i + UNABHAENGIG] = null;
					gelesen[i] = gesamt;
				}
			}
			break;
		case Ableitung:
			int bezug = gl.indexOf(g.getBezug()) + UNABHAENGIG;
			if (bezug >= UNABHAENGIG) {
				while (gelesen[i] < gelesen[bezug])
					daten[++gelesen[i]][i + UNABHAENGIG] = (daten[gelesen[i]][bezug] - daten[gelesen[i] - 1][bezug])
							/ (daten[gelesen[i]][T.getIndex()] - daten[gelesen[i] - 1][T.getIndex()]); // Differenzenquotient
			}
			break;
		case Integral:
			bezug = gl.indexOf(g.getBezug()) + UNABHAENGIG;
			if (bezug >= UNABHAENGIG) {
				while (gelesen[i] < gelesen[bezug])
					daten[++gelesen[i]][i + UNABHAENGIG] = daten[gelesen[i] - 1][i
							+ UNABHAENGIG]
							+ (daten[gelesen[i]][bezug] + daten[gelesen[i] - 1][bezug])
							* (daten[gelesen[i]][T.getIndex()] - daten[gelesen[i] - 1][T.getIndex()])
							/ 2.0; // Trapezformel
			}
			break;
		default:
			break;
		}
		return gelesen[i] - bisher;
	}

	public void werteEinlesen() {
		for (int i = 0; i < grZahl; i++) {
			Groesse g = gl.get(i);
			if (g.getMessreihe() != mrNr)
				continue;
			werteEinlesen(i, g);
		}
	}

	public void werteSchreiben() {
		if (daten == null || gt == null)
			return;
		int minGelesen = minimum(gelesen);
		if (DEBUG)
			System.out.println("[XRayTimerTask] - werteSchreiben(" + minGelesen
					+ ")");
		if (minGelesen <= geschrieben)
			return;
		@SuppressWarnings("rawtypes")
		Vector<Vector> v = new Vector<Vector>(minGelesen
				- geschrieben);
		for (int i = geschrieben; i < minGelesen; i++) {
			Vector<Double> row = new Vector<Double>(grZahl + UNABHAENGIG);
			for (int j = 0; j < grZahl + UNABHAENGIG; j++)
				row.add(daten[i][j]);
			v.add(row);
		}
		gt.replaceRows(gtOffset + ((Double) v.get(0).get(N.getIndex())).intValue() - 1, v,
				false, true);
		geschrieben = minGelesen;
	}

	public int minimum(int[] feld) {
		int result;
		if (feld == null || feld.length < 1)
			return -1;
		result = feld[0];
		for (int i = 1; i < feld.length; i++)
			result = Math.min(result, feld[i]);
		return result;
	}

	public boolean fertig() {
		return geschrieben >= wiederholung || gesamt > 0
				&& geschrieben >= gesamt;
	}

	public boolean hatTrue(boolean[] feld) {
		if (feld == null)
			return false;
		for (int i = 0; i < feld.length; i++)
			if (feld[i])
				return true;
		return false;
	}

	public void setWerteTabelle(WerteTabelle w) {
		this.gt = w;
	}

	public void setGroessenliste(Groessenliste g) {
		this.gl = g;
	}

	@Override
	public int prozentFertig() {
		double prozentWerte = geschrieben * 100.0 / gesamt;
		return (int) (fertig() ? 100 : Math.round(prozentWerte));
	}

	@Override
	public void run() {
		if (gt != null && gl != null && mp != null && al != null) {
			if (!started)
				try {
					start();
				} catch (EAdapterException e) {
					if (DEBUG)
						e.printStackTrace();
				}
			if(cassys.oszilloskopStarted != null) {
				if( cassys.oszilloskopStarted.getCount() > 0)
					return;
				else
					cassys.oszilloskopStarted = null;
			}
			werteEinlesen();
			werteSchreiben();
			if (fertig()) {
				if (mp.isWiederholungEin())
					restart();
				else
					this.cancel();
			}
		}
	}

	@Override
	public boolean messBedingungErfuellt() {
		return true;
	}

	@Override
	public boolean stoppBedingungErfuellt() {
		return false;
	}

	@Override
	public boolean triggerBedingungErfuellt() {
		return true;
	}

	/**
	 * Zusatzmethode stopScan: stoppt einen ev. laufenden Scan
	 */
	private void stopScan() {
		if (xray == null)
			return;
		XRayModes mode = xray.getMode();
		if (!mode.isAndEmpty(XRayModes.Scan)) {
			xray.setMode(new XRayModes(XRayModes.Scan));
			try {
				Thread.sleep(4000);
			} catch (InterruptedException e) {
			}
		}
	}

}
