/* 
 * GHMessen
 * --------
 * 
 * GHMessen is a measuring program to use in connection with a computer 
 * measuring interface. It's designed for school usage.
 * 
 * Copyright (C) 2010-2011  Frank Schütte - Gymnasium Himmelsthür
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms
 * of the GNU General Public License as published by the Free Software Foundation; either 
 * version 3 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with this program; 
 * if not, see <http://www.gnu.org/licenses/>.
 * 
 * Frank Schütte fschuett AT gymnasium-himmelsthuer DOT de
 * 
 */

package ghmessen.messtasks;

import static ghmessen.groesse.wertetabelle.XAchseTyp.BETA;
import static ghmessen.groesse.wertetabelle.XAchseTyp.ENERGIE;
import static ghmessen.groesse.wertetabelle.XAchseTyp.F;
import static ghmessen.groesse.wertetabelle.XAchseTyp.LAMBDA;
import static ghmessen.groesse.wertetabelle.XAchseTyp.N;
import static ghmessen.groesse.wertetabelle.XAchseTyp.T;
import static ghmessen.groesse.wertetabelle.XAchseTyp.UNABHAENGIG;
import ghmessen.Messparameter;
import ghmessen.adapter.AdapterListe;
import ghmessen.adapter.EAdapterException;
import ghmessen.adapter.IAdapter;
import ghmessen.adapter.IMesskanal;
import ghmessen.adapter.MesskanalException;
import ghmessen.groesse.Groesse;
import ghmessen.groesse.Groessenliste;
import ghmessen.groesse.WerteTabelle;

import java.lang.reflect.InvocationTargetException;
import java.util.TimerTask;
import java.util.Vector;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.ScheduledThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

public class ZeitTimerTask extends MessTimerTask {
	private static final boolean DEBUG = false;
	private long debugStartzeit;
	private static final int PUFFERGROESSE = 20000;
	Vector<Vector> puffer;
	Object monitor;
	ScheduledFuture<?> schreibProzess;
	ScheduledThreadPoolExecutor schreibTimer;
	protected boolean startFG;
	private boolean stoppen;
	private boolean schreiben;
	protected double messBeginn;

	@SuppressWarnings("rawtypes")
	public ZeitTimerTask(WerteTabelle gt, Groessenliste gl, Messparameter mp,
			ScheduledThreadPoolExecutor timer, AdapterListe al) {
		super(gt, gl, mp, al);
		debugStartzeit = 0;
		puffer = null;
		monitor = null;
		schreibProzess = null;
		schreibTimer = null;
		startFG = false;
		stoppen = false;
		messBeginn = 0;
		if (mp != null && !mp.isEinzelEin() && mp.getIntervallMS() < 100
				&& !gl.hatArrayZugriff()) {
			puffer = new Vector<Vector>(PUFFERGROESSE);
			monitor = new Object();
			schreibTimer = timer;
		}
		schreiben = true;
		cassys.setOszilloskopeWiederholung((int) wiederholung);
		cassys.setOszilloskopeIntervall(mp.getIntervall());
	}

	/*
	 * CASSY setzt Messmodus auf Fast-/SlowOscilloscope
	 * @see ghmessen.messtasks.MessTimerTask#init()
	 */
	@Override
	public void init() throws EAdapterException {
		if (DEBUG)
			System.out.println("[ZeitTimerTask] - init...");
		cassys.init(mp.getIntervall());
		super.init();
	}

	public void finalize() {
		if (schreibProzess != null) {
			schreibProzess.cancel(true);
			schreibProzess = null;
		}
	}

	protected static Vector<Double> convertToVector(Double[] anArray) {
		if (anArray == null) {
			return null;
		}
		Vector<Double> v = new Vector<Double>(anArray.length);
		for (int i = 0; i < anArray.length; i++) {
			v.addElement(anArray[i]);
		}
		return v;
	}

	public void puffern(Double[] zeile) {
		if (puffer != null) {
			synchronized (monitor) {
				puffer.add(convertToVector(zeile));
			}
		}
	}

	@Override
	protected void initGroesse(int i, Groesse g) {
		switch (g.getArt()) {
		case Messgroesse:
			if (!g.isAktiv())
				return;
			IAdapter a = al.get(g.getNr());
			if(a != null) {
				eingang[i] = a.getEingang(g);
				messkanal[i] = a.getMesskanal(g);
			}
			if (messkanal[i] != null) {
				try {
					messkanal[i].setAbstand(0.5 * mp.getIntervallMS());
				} catch (MesskanalException e) {
					if (DEBUG)
						e.printStackTrace();
				}
				if(!mp.isTriggerEin() && g.isAusloeserTriggert() 
						&& messkanal[i].istTriggerKanal()) {
					cassys.setTriggerQuantity(g.getSymbol());
					cassys.setTrigger(true);
				}
			}
			break;
		default:
			super.initGroesse(i, g);
			break;
		}
	}

	/*
	 * Startet eventuell den Pufferprozess, startet alle Messkanäle
	 * und schließlich die CASSY-Oszilloskope.
	 * Notiert den messBeginn in Millisekunden.
	 * @see ghmessen.messtasks.MessTimerTask#start()
	 */
	@Override
	public void start() throws EAdapterException {
		if (DEBUG)
			System.out.println("[ZeitTimerTask] - start...");
		super.start();
		try {
			debugStartzeit = System.currentTimeMillis();
			if (puffer != null) {
				schreibProzess = schreibTimer.scheduleWithFixedDelay(
						new SchreibProzess(), 500, 500, TimeUnit.MILLISECONDS);
			}
			for (IMesskanal m : messkanal)
				if (m != null)
					m.start();
			cassys.starteOszilloskope();
		} catch (MesskanalException e) {
			if (DEBUG)
				e.printStackTrace();
		}
		messBeginn = System.currentTimeMillis();
	}

	@Override
	public void restart() {
		if (DEBUG)
			System.out.println("[ZeitTimerTask] - restart...");
		try {
			cassys.stoppeOszilloskope();
		} catch (EAdapterException e) {
			if (DEBUG)
				e.printStackTrace();
		}
		if (DEBUG)
			System.out.println("[ZeitTimerTask] - Messzeit("
					+ ((System.currentTimeMillis() - debugStartzeit) / 1000.0)
					+ ")");
		messBeginn = 0;
		super.restart();
	}

	@Override
	public void stop() {
		if (DEBUG)
			System.out.println("[ZeitTimerTask] - stop...");
		for (IMesskanal m : messkanal)
			try {
				if (m != null)
					m.stop();
				cassys.stoppeOszilloskope();
			} catch (MesskanalException e) {
				if (DEBUG)
					e.printStackTrace();
			} catch (EAdapterException e) {
				if (DEBUG)
					e.printStackTrace();
			}
		cancel();
	}

	@Override
	public boolean cancel() {
		if (DEBUG)
			System.out.println("[ZeitTimerTask] - cancel...");
		if (puffer != null && schreibTimer != null && schreibProzess != null) {
			schreibProzess.cancel(false);
			schreibProzess = null;
			if (!puffer.isEmpty())
				gt.replaceRows(gtOffset + ((Double) puffer.get(0).get(N.getIndex())).intValue()
						- 1, puffer, false, true);
		}

		if (DEBUG)
			System.out.println("[ZeitTimerTask] - Messzeit("
					+ ((System.currentTimeMillis() - debugStartzeit) / 1000.0)
					+ ")");
		return super.cancel();
	}

	@Override
	protected double getTime() {
		int anzahl;
		if (cassys.isQuantitiesSelected() && (anzahl = cassys.getTimes()) > 0) {
			return cassys.getTimes(anzahl - 1);
		} 
		else {
			return super.getTime();
		}
	}
	
	@Override
	protected void messeGroesse(int i, Groesse g, Double[] d)
			throws MesskanalException, InterruptedException, EAdapterException {
		switch (g.getArt()) {
		default:
			super.messeGroesse(i, g, d);
			break;
		case Messgroesse:
			if (g.isAktiv() && messkanal[i] != null) {
				d[i + UNABHAENGIG] = messkanal[i].einlesen();
				if (messkanal[i].istUeberschritten())
					throw new EAdapterException("Nr(" + g.getNr()
							+ ") Messbereich überschritten.");
				if (messkanal[i].istLangsamerZeitKanal()) {
					d[T.getIndex()] = messkanal[i].einlesenZeit();
					d[F.getIndex()] = 0.0;
				}
			} else if (!g.isAktiv()) {
				d[i + UNABHAENGIG] = null;
			} else
				throw new MesskanalException(
						"[ZeitTimerTask] Messkanal nicht initialisiert.");
			break;
		}
	}

	@Override
	public void schreiben(Double[] d) {
		schreiben = true;
		if (warteAufSoftTrigger && !triggerBedingungErfuellt())
			return;
		ersterMesswert = false;
		if (mp.isBedingungEin() && !messBedingungErfuellt())
			schreiben = false;
		if (schreiben && d != null) {
			if (puffer != null)
				puffern(d);
			else
				gt.replaceRow(gtOffset + d[N.getIndex()].intValue() - 1, d, false,
						true);
			dAlt = d;
		}
		if (mp.isStoppEin() && stoppBedingungErfuellt())
			stoppen = true;
	}

	protected boolean hatTrue(boolean[] feld) {
		if (feld == null)
			return false;
		for (int i = 0; i < feld.length; i++)
			if (feld[i])
				return true;
		return false;
	}

	public boolean istMesswertVorhanden() {
		return wiederholung == 1 
				|| !cassys.isQuantitiesSelected()
				|| cassys.istMesswertVorhanden();
	}

	@Override
	public int prozentFertig() {
		if (mp.isTimerBox()) {
			return (int) Math.round((System.currentTimeMillis() - messBeginn)
					/ mp.getMesszeit() / 10.0);
		} else
			return (int) Math.round((stoppen ? wiederholung : anzahl) * 100.0
					/ wiederholung);
	}

	@Override
	public void run() {
		super.run();
		try {
			while (!istMesswertVorhanden())
				Thread.sleep(100); // abwarten, bis der erste Messwert vorliegt
			if (gt != null && gl != null && mp != null && al != null) {
				einzelMessung();
				anzahl++;
				if (prozentFertig() >= 100) {
					if (mp.isWiederholungEin())
						restart();
					else
						this.cancel();
				}
			}
		} catch (InterruptedException e) {
			if (DEBUG)
				e.printStackTrace();
		}

	}

	/*
	 * PUFFER - Thread
	 */

	private class SchreibProzess extends TimerTask {
		@SuppressWarnings({ "unchecked", "rawtypes" })
		@Override
		public void run() {
			if (puffer != null && gt != null) {
				if (!puffer.isEmpty()) {
					@SuppressWarnings("rawtypes")
					Vector<Vector> kopie = null;
					synchronized (monitor) {
						kopie = (Vector<Vector>) puffer.clone();
						puffer.clear();
					}
					gt.replaceRows(gtOffset
							+ ((Double) kopie.get(0).get(N.getIndex())).intValue() - 1, kopie,
							false, true);
				}
			}
		}
	}

	@Override
	public boolean messBedingungErfuellt() {
		if (!mp.isBedingungEin() || mp.getIntervallMS() < 10 || d == null)
			return true;
		try {
			Double resultat = ((Double) mp.getBedingungEE().evaluate(
					new Object[] { d[T.getIndex()], d[N.getIndex()], d[F.getIndex()],
							d[BETA.getIndex()], d[LAMBDA.getIndex()], d[ENERGIE.getIndex()],
							Double.valueOf(mp.getIntervall()), datum, uhrzeit, d,
							gt }));
			return resultat != null && resultat != 0;
		} catch (InvocationTargetException e) {
			if (DEBUG)
				e.printStackTrace();
			return true;
		}
	}

	@Override
	public boolean stoppBedingungErfuellt() {
		if (!mp.isStoppEin() || mp.getIntervallMS() < 10 || d == null)
			return false;
		try {
			return ((Double) mp.getStoppEE().evaluate(
					new Object[] { d[T.getIndex()], d[N.getIndex()], d[F.getIndex()],
							d[BETA.getIndex()], d[LAMBDA.getIndex()], d[ENERGIE.getIndex()],
							Double.valueOf(mp.getIntervall()), datum, uhrzeit, d,
							gt })) != 0;
		} catch (InvocationTargetException e) {
			if (DEBUG)
				e.printStackTrace();
			return false;
		}
	}

	@Override
	public boolean triggerBedingungErfuellt() {
		if (!mp.isTriggerEin() || mp.getIntervallMS() < 10 || d == null)
			return true;
		try {
			Double resultat = ((Double) mp.getTriggerEE().evaluate(
					new Object[] { d[T.getIndex()], d[N.getIndex()], d[F.getIndex()],
							d[BETA.getIndex()], d[LAMBDA.getIndex()], d[ENERGIE.getIndex()],
							new Double(mp.getIntervall()), datum, uhrzeit, d,
							gt }));
			return resultat != null && resultat != 0;
		} catch (InvocationTargetException e) {
			if (DEBUG)
				e.printStackTrace();
			return true;
		}
	}
}
