package ghmessen.util;

import java.awt.geom.Path2D;

import myflanagan.CubicInterpolation;

/**
 * This interpolator uses Akima cubic Splines<br>
 * <br>
 * 
 * @author Leandro Jose
 * @author Rodrigo Bruno
 * 
 */
public class AkimaInterpolation {

	private CubicInterpolation ci;
	private int segmente;
	private double[] xval, yval;
	private double[] xs, ys;
	private Path2D pfad;
	
	public AkimaInterpolation(Double[] x, Double[] y) throws Exception {
		pfad = null;
		if(x == null || y == null)
			return;
		int anfang, ende;
		anfang = 0;
		while(anfang < x.length){
			while(anfang < x.length-1 && (x[anfang] == null || y[anfang] == null))
				anfang++;
			ende = anfang+1;
			while(ende < x.length && (x[ende] != null && y[ende] != null))
				ende++;
			if(anfang < x.length && x[anfang] != null && y[anfang] != null){
				double[] xx = new double[ende-anfang];
				double[] yy = new double[ende-anfang];
				for(int i=anfang;i<ende;i++){
					xx[i-anfang]=x[i];
					yy[i-anfang]=y[i];
				}
				doInterpolation(xx,yy);
			}
			anfang = ende;
		}
	}
	
	public AkimaInterpolation(double[] x, double[] y) throws Exception {
		pfad = null;
		doInterpolation(x,y);
	}
	
	public void doInterpolation(double[] x, double[] y) throws Exception {
		if (x.length <= 4 || y.length <= 4) {
			throw new Exception("At least five points must be given");
		}
		// prepare array: ensure i != i+1 => xval[i] != xval[i+1]
		int doppelte = 0;
		for(int i=0;i<x.length-1;i++)
			if(x[i] == x[i+1])
				doppelte++;
		int n = x.length - doppelte;
		
		xval = new double[n];
		yval = new double[n];
		doppelte = 0;
		xval[0] = x[0];
		yval[0] = y[0];
		for(int i=1;i<x.length;i++){
			if(x[i-1] == x[i])
				doppelte++;
			else {
				xval[i-doppelte] = x[i];
				yval[i-doppelte] = y[i];
			}
		}

		double[] d = new double[n];
		// Weights array
		double[] w = new double[n - 1];
		// differences array
		double[] diff = new double[n - 1];

		for (int i = 0; i <= n - 2; i++) {
			diff[i] = (yval[i + 1] - yval[i]) / (xval[i + 1] - xval[i]);
		}
		for (int i = 1; i <= n - 2; i++) {
			w[i] = Math.abs(diff[i] - diff[i - 1]);
		}

		//
		// Prepare Hermite interpolation scheme
		//

		for (int i = 2; i <= n - 3; i++) {
			if (Math.abs(w[i - 1]) + Math.abs(w[i + 1]) != 0) {
				d[i] = (w[i + 1] * diff[i - 1] + w[i - 1] * diff[i])
						/ (w[i + 1] + w[i - 1]);
			} else {
				d[i] = ((xval[i + 1] - xval[i]) * diff[i - 1] + (xval[i] - xval[i - 1])
						* diff[i])
						/ (xval[i + 1] - xval[i - 1]);
			}
		}

		d[0] = diffthreepoint(xval[0], xval[0], yval[0], xval[1], yval[1],
				xval[2], yval[2]);
		d[1] = diffthreepoint(xval[1], xval[0], yval[0], xval[1], yval[1],
				xval[2], yval[2]);
		d[n - 2] = diffthreepoint(xval[n - 2], xval[n - 3], yval[n - 3],
				xval[n - 2], yval[n - 2], xval[n - 1], yval[n - 1]);
		d[n - 1] = diffthreepoint(xval[n - 1], xval[n - 3], yval[n - 3],
				xval[n - 2], yval[n - 2], xval[n - 1], yval[n - 1]);

		//
		// Build Akima spline using Hermite interpolation scheme
		//

		ci = new CubicInterpolation(xval, yval, d);
		segmente = n-1;
		calculatePoints();
		calculatePath();
	}

	public double getX(int index){
		return xs[index];
	}
	
	public double getY(int index){
		return ys[index];
	}
	
	public int getSegmente(){
		return segmente;
	}
	
	public Path2D getPfad(){
		return pfad;
	}
	
	private void calculatePoints(){
		xs = new double[3*segmente+1];
		ys = new double[3*segmente+1];
		for(int i=0;i<segmente;i++)
			calculateSegment(i);
		xs[xs.length-1] = xval[xval.length-1];
		ys[ys.length-1] = yval[yval.length-1];
	}
	
	private void calculateSegment(int i){
		double[] yv;
		ci.interpolate(xval[i]); // y, dydx, dydx2, x
		yv = ci.getInterpolatedValues();
		int index = 3*i;
		xs[index] = xval[i]; ys[index++]=yval[i];
		xs[index] = xval[i] + 0.333*(xval[i+1]-xval[i]); ys[index++]=0.333*yv[1]+yval[i];
		xs[index] = xval[i] + 0.666*(xval[i+1]-xval[i]); ys[index++]=1./6.*yv[2]+0.666*yv[1]+yval[i];
	}
	
	private void calculatePath(){
		if(pfad == null)
			pfad = new Path2D.Double();
		pfad.moveTo(xs[0],ys[0]);
		for (int i = 0; i < 3 * segmente; i += 3){
			pfad.curveTo(xs[i+1],ys[i+1],xs[i+2],ys[i+2],xs[i+3],ys[i+3]);
		}
	}
	
	private static double diffthreepoint(double t, double x0, double f0,
			double x1, double f1, double x2, double f2) {
		double result = 0;
		double a = 0;
		double b = 0;

		t = t - x0;
		x1 = x1 - x0;
		x2 = x2 - x0;
		a = (f2 - f0 - x2 / x1 * (f1 - f0)) / (Math.pow(x2, 2) - x1 * x2);
		b = (f1 - f0 - a * Math.pow(x1, 2)) / x1;
		result = 2 * a * t + b;
		return result;
	}

}