/* 
 * GHMessen
 * --------
 * 
 * GHMessen is a measuring program to use in connection with a computer 
 * measuring interface. It's designed for school usage.
 * 
 * Copyright (C) 2010-2011  Frank Schütte - Gymnasium Himmelsthür
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms
 * of the GNU General Public License as published by the Free Software Foundation; either 
 * version 3 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with this program; 
 * if not, see <http://www.gnu.org/licenses/>.
 * 
 * Frank Schütte fschuett AT gymnasium-himmelsthuer DOT de
 * 
 */

package libGoIO;

import java.io.Serializable;
import java.util.Iterator;

import javax.xml.stream.XMLEventReader;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.events.Attribute;
import javax.xml.stream.events.StartElement;
import javax.xml.stream.events.XMLEvent;

import libGoIO.GoIOLibrary.EEquationType;
import ghmessen.components.Intervall;
import ghmessen.groesse.Einheit;
import ghmessen.groesse.Messbereich;
import static ghmessen.groesse.Einheit.*;

public class GoIOCalibration implements Serializable {
	/**
	 * 
	 */
	private static final long serialVersionUID = -9162085854988914078L;
	private int sensorNumber;
	private int ddsIndex;
	private Messbereich messbereich;
	private EEquationType equation;
	private double k0;
	private double k1;
	private double k2;

	public GoIOCalibration() {
		sensorNumber = 0;
		ddsIndex = 0;
		messbereich = new Messbereich(0);
		equation = EEquationType.kEquationType_None;
		k0 = k1 = k2 = 0.0;
	}

	private static final String CALIBRATION = "Calibration";
	private static final String CALUNITS = "Units";
	private static final String CALPRETTYUNITS = "PrettyUnits";
	private static final String CALDDSINDEX = "DDSIndex";
	private static final String CALSENSORNUMBER = "IDValue";
	private static final String EQUATION = "Equation";
	private static final String DISPLAYRANGE = "DisplayRange";
	private static final String DECIMALPLACES = "DecimalPlaces";

	public static GoIOCalibration readFromXml(XMLEventReader eventReader,
			StartElement startElement) throws XMLStreamException {
		GoIOCalibration gc = new GoIOCalibration();
		gc.parseCalibrationAttribs(startElement);
		gc.parseCalibration(eventReader);
		return gc;
	}

	public double calibrateData(double volts){
		switch(equation){
			default:
			case kEquationType_None:
				return volts;
		/// linear, y = bx + a
			case kEquationType_Linear:
				return k1*volts+k0;
		/// y = cx^2 + bx + a
			case kEquationType_Quadratic:
				return k2*volts*volts+k1*volts+k0;
		/// y = a * x^b
			case kEquationType_Power:
				return k0*Math.pow(volts, k1);
		/// y = a * b^x
			case kEquationType_ModifiedPower:
				return k0*Math.pow(k1, volts);
		/// y = a + b * ln(x)
			case kEquationType_Logarithmic:
				return k0+k1*Math.log(volts);
		/// y = a + b * ln(1/x)
			case kEquationType_ModifiedLogarithmic:
				return k0+k1*Math.log(1.0/volts);
		/// y = a * e^(b * x)
			case kEquationType_Exponential:
				return k0*Math.exp(k1*volts);
		/// y = a * e^(b / x)
			case kEquationType_ModifiedExponential:
				return k0*Math.exp(k1/volts);
		/// y = a * x^(b * x)
			case kEquationType_Geometric:
				return k0*Math.pow(volts, k1*volts);
		/// y = a * x^(b / x)
			case kEquationType_ModifiedGeometric:
				return k0*Math.pow(volts, k1/volts);
		/// y = 1 / (a + b * ln(c * x))
			case kEquationType_ReciprocalLog:
				return 1.0/(k0+k1*Math.log(k2*volts));
		/// y = 1 / (a + b * ln(1000*x) + c * (ln(1000*x)^3) )
			case kEquationType_SteinhartHart:
				double r = 15000.0/(5.0/volts-1.0);
				double t = 1.0/(k0+k1*Math.log(r)+k2*Math.log(r)*Math.log(r)*Math.log(r));
				return konvertiereTemperatur(t, K, messbereich.getEinheit());
		}
	}
	
	private double konvertiereTemperatur(double wert, Einheit from, Einheit to){
		if(from == to)
			return wert;
		if(from == F) {
			if(to == _C)
				return (wert - 32) / 1.8;
			else
				return (wert + 459.67) / 1.8;
		} else if(from == _C) {
			if(to == F)
				return wert * 1.8 + 32;
			else
				return wert + 273.15;
		} else if(from == K) {
			if(to == _C)
				return wert - 273.15;
			else
				return wert * 1.8 - 459.67;
		} else 
			return wert;
	}
	
	private void parseCalibrationAttribs(StartElement startElement)
			throws XMLStreamException {
		@SuppressWarnings("unchecked")
		Iterator<Attribute> attributes = startElement.getAttributes();
		String units = null, prettyUnits = null;
		while (attributes.hasNext()) {
			Attribute attribute = attributes.next();
			String s = attribute.getName().toString();
			if (s.equals(CALUNITS))
				units = attribute.getValue();
			else if(s.equals(CALPRETTYUNITS))
				prettyUnits = attribute.getValue();
			else if (s.equals(CALDDSINDEX))
				setDdsIndex(Integer.parseInt(attribute.getValue()));
			else if (s.equals(CALSENSORNUMBER))
				setSensorNumber(Integer.parseInt(attribute.getValue()));
			else if (s.equals(DECIMALPLACES)){
				try {
					int stellen = Integer.parseInt(attribute.getValue());
					messbereich.setStellen(stellen == -1 ? 3 : stellen);
				} catch(NumberFormatException e){ }
			}
		}
		messbereich.setEinheit(Einheit.toEinheit(prettyUnits));
		if(messbereich.getEinheit() == Einheit.Keine)
			messbereich.setEinheit(Einheit.toEinheit(units));
	}

	private void parseCalibration(XMLEventReader eventReader)
			throws XMLStreamException {
		while (eventReader.hasNext()) {
			XMLEvent event = eventReader.nextEvent();
			if (event.isStartElement()) {
				StartElement startElement = event.asStartElement();
				if (startElement.getName().getLocalPart() == (EQUATION)) {
					@SuppressWarnings("unchecked")
					Iterator<Attribute> attributes = startElement
							.getAttributes();
					while (attributes.hasNext()) {
						Attribute attribute = attributes.next();
						String s = attribute.getName().toString();
						if (s.equals("Type"))
							equation = convertToEEquationType(attribute
									.getValue());
						else if (s.equals("K0"))
							k0 = Double.parseDouble(attribute.getValue());
						else if (s.equals("K1"))
							k1 = Double.parseDouble(attribute.getValue());
						else if (s.equals("K2"))
							k2 = Double.parseDouble(attribute.getValue());
					}
					continue;
				} else if (startElement.getName().getLocalPart() == (DISPLAYRANGE)) {
					@SuppressWarnings("unchecked")
					Iterator<Attribute> attributes = startElement
							.getAttributes();
					Double min = null, max = null;
					while (attributes.hasNext()) {
						Attribute attribute = attributes.next();
						String s = attribute.getName().toString();
						if (s.equals("YMin"))
							min = Double.parseDouble(attribute.getValue());
						else if (s.equals("YMax"))
							max = Double.parseDouble(attribute.getValue());
					}
					if (min != null && max != null){
						messbereich.setIntervall(new Intervall(min, max));
						messbereich.setBereichName("" + min + messbereich.getEinheit().toString()+".."
								+max+ messbereich.getEinheit().toString());
					}
					continue;
				}
			}

			if (event.isEndElement()
					&& event.asEndElement().getName().getLocalPart() == (CALIBRATION))
				break;
		}
	}

	private EEquationType convertToEEquationType(String s){
		int nr = Integer.parseInt(s);
		for(EEquationType ee : EEquationType.values())
			if(ee.value == nr)
				return ee;
		return null;
	}
	
	public int getSensorNumber() {
		return sensorNumber;
	}
	
	public void setSensorNumber(int sensorNumber) {
		this.sensorNumber = sensorNumber;
	}
	
	public int getDdsIndex() {
		return ddsIndex;
	}

	public void setDdsIndex(int ddsIndex) {
		this.ddsIndex = ddsIndex;
	}

	public Messbereich getMessbereich() {
		return messbereich;
	}

	public void setMessbereich(Messbereich messbereich) {
		this.messbereich = messbereich;
	}

	public Einheit getEinheit() {
		return messbereich != null ? messbereich.getEinheit() : null;
	}

	public void setEinheit(Einheit e) {
		if (messbereich == null)
			messbereich = new Messbereich(0);
		messbereich.setEinheit(e);
	}

	public EEquationType getEquation() {
		return equation;
	}

	public void setEquation(EEquationType equation) {
		this.equation = equation;
	}

	public double getK0() {
		return k0;
	}

	public void setK0(double k0) {
		this.k0 = k0;
	}

	public double getK1() {
		return k1;
	}

	public void setK1(double k1) {
		this.k1 = k1;
	}

	public double getK2() {
		return k2;
	}

	public void setK2(double k2) {
		this.k2 = k2;
	}

}
