/* 
 * GHMessen
 * --------
 * 
 * GHMessen is a measuring program to use in connection with a computer 
 * measuring interface. It's designed for school usage.
 * 
 * Copyright (C) 2010-2011  Frank Schütte - Gymnasium Himmelsthür
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms
 * of the GNU General Public License as published by the Free Software Foundation; either 
 * version 3 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with this program; 
 * if not, see <http://www.gnu.org/licenses/>.
 * 
 * Frank Schütte fschuett AT gymnasium-himmelsthuer DOT de
 * 
 */

package libGoIO;

import ghmessen.groesse.Messbereich;

import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.io.Serializable;
import java.nio.charset.Charset;
import java.util.Iterator;
import java.util.Vector;

import javax.xml.stream.XMLEventReader;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.events.Attribute;
import javax.xml.stream.events.StartElement;
import javax.xml.stream.events.XMLEvent;

import libGoIO.GoIOLibrary.EProbeType;

public class GoIOSensor implements Serializable {

	/**
	 * 
	 */
	private static final long serialVersionUID = 1198520986140075803L;

	private static final boolean DEBUG = false;

	private static transient SensorMap strings;

	/* Attribute von Sensor */
	private String sensorId;
	private int sensorNumber;
	private int numAutoIds;
	private EProbeType sensorType;
	private boolean dds;
	private boolean legacy;
	/* Felder von Sensor */
	private String sensorFamily;
	private static final String SENSORFAMILY = "SensorFamily";
	private String localName;
	private static final String LOCALNAME = "LocalName";
	private String lname_base;
	private static final String LNAME_BASE = "lname_base";
	private boolean reverseSupported;
	private static final String REVERSESUPPORTED = "ReverseSupported";
	private boolean zeroSupported;
	private static final String ZEROSUPPORTED = "ZeroSupported";
	private boolean showZeroButton;
	private static final String SHOWZEROBUTTON = "ShowZeroButton";
	private int voltageDecimalPlaces;
	private static final String VOLTAGEDECIMALPLACES = "VoltageDecimalPlaces";
	private boolean singlePointCalibrationSupported;
	private static final String SINGLEPOINTCALIBRATIONSUPPORTED = "SinglePointCalibrationSupported";
	private double singlePointCalibrationDefault;
	private static final String SINGLEPOINTCALIBRATIONDEFAULT = "SinglePointCalibrationDefault";
	/* Hardware */
	private double warmUpTime;
	private static final String WARMUPTIME = "WarmUpTime";
	/* CollectDefaults */
	private double delta;
	private static final String DELTA = "Delta";
	private double maxDelta;
	private static final String MAXDELTA = "MaxDelta";
	private int numSamples;
	private static final String NUMSAMPLES = "NumSamples";
	private boolean averaging;
	private static final String AVERAGING = "Averaging";
	/* CalibrationList */
	private int defaultIndex;
	private GoIOCalibrationVector calibrationList;

	/* DDS Record */
	private GoIOSensorDDSRec sensorDDSRec;

	public GoIOSensor() {
		sensorId = "";
		sensorNumber = 0;
		numAutoIds = 0;
		sensorType = EProbeType.kProbeTypeAnalog5V;
		dds = false;
		legacy = false;
		sensorFamily = "";
		localName = "";
		lname_base = "";
		reverseSupported = false;
		zeroSupported = false;
		showZeroButton = false;
		voltageDecimalPlaces = 2;
		singlePointCalibrationSupported = true;
		singlePointCalibrationDefault = 0.0;
		warmUpTime = 0.0;
		delta = 0.02;
		maxDelta = 0.001;
		numSamples = 100;
		averaging = false;
		defaultIndex = 0;
		calibrationList = new GoIOCalibrationVector();
		sensorDDSRec = new GoIOSensorDDSRec();
	}

	public GoIOSensor(GoIOSensor src) {
		sensorId = src.sensorId;
		sensorNumber = src.sensorNumber;
		numAutoIds = src.numAutoIds;
		sensorType = src.sensorType;
		dds = src.dds;
		legacy = src.legacy;
		sensorFamily = src.sensorFamily;
		localName = src.localName;
		lname_base = src.lname_base;
		reverseSupported = src.reverseSupported;
		zeroSupported = src.zeroSupported;
		showZeroButton = src.showZeroButton;
		voltageDecimalPlaces = src.voltageDecimalPlaces;
		singlePointCalibrationSupported = src.singlePointCalibrationSupported;
		singlePointCalibrationDefault = src.singlePointCalibrationDefault;
		warmUpTime = src.warmUpTime;
		delta = src.delta;
		maxDelta = src.maxDelta;
		numSamples = src.numSamples;
		averaging = src.averaging;
		defaultIndex = src.defaultIndex;
		calibrationList = (GoIOCalibrationVector) src.calibrationList.clone();
		sensorDDSRec = (GoIOSensorDDSRec) src.sensorDDSRec.clone();
	}

	public static GoIOSensor readFromXml(String xml) {
		GoIOSensor ret = new GoIOSensor();
		try {
			InputStream in;
			in = new ByteArrayInputStream(
					xml.getBytes(Charset.defaultCharset()));
			if (strings == null)
				strings = new SensorMap();
			ret.parseXmlStream(in);
		} catch (XMLStreamException e) {
			if (DEBUG)
				e.printStackTrace();
		}
		return ret;
	}

	@Override
	public GoIOSensor clone() {
		GoIOSensor ret = new GoIOSensor(this);
		return ret;
	}

	/* XML Elements */
	/* Sensor */
	private static final String SENSOR = "Sensor";
	/* Sensor-Attribute */
	private static final String SENSORID = "ID_Name";
	private static final String SENSORNUMBER = "BaseID";
	private static final String SENSORTYPE = "Type";
	private static final String SENSORDDS = "DDS";
	private static final String SENSORLEGACY = "Legacy";
	private static final String SENSORNUMAUTOIDS = "NumAutoIDs";

	/* CalibrationList */
	private static final String CALIBRATIONLIST = "CalibrationList";
	private static final String CALIBRATIONLISTDEFAULTINDEX = "DefaultIndex";

	/* SensorBurnRec */
	private static final String SENSORBURNREC = "SensorBurnRec";

	private void parseSensorAttribs(StartElement startElement)
			throws XMLStreamException {
		@SuppressWarnings("unchecked")
		Iterator<Attribute> attributes = startElement.getAttributes();
		while (attributes.hasNext()) {
			Attribute attribute = attributes.next();
			String s = attribute.getName().toString();
			if (s.equals(SENSORID))
				sensorId = attribute.getValue();
			else if (s.equals(SENSORNUMBER))
				sensorNumber = Integer.parseInt(attribute.getValue());
			else if (s.equals(SENSORTYPE))
				sensorType = convertToType(attribute.getValue());
			else if (s.equals(SENSORDDS))
				dds = Integer.parseInt(attribute.getValue()) != 0;
			else if (s.equals(SENSORLEGACY))
				legacy = Integer.parseInt(attribute.getValue()) != 0;
			else if (s.equals(SENSORNUMAUTOIDS))
				numAutoIds = Integer.parseInt(attribute.getValue());
		}

	}

	private void parseCalibrationListAttribs(StartElement startElement)
			throws XMLStreamException {
		@SuppressWarnings("unchecked")
		Iterator<Attribute> attributes = startElement.getAttributes();
		while (attributes.hasNext()) {
			Attribute attribute = attributes.next();
			String s = attribute.getName().toString();
			if (s.equals(CALIBRATIONLISTDEFAULTINDEX))
				defaultIndex = Integer.parseInt(attribute.getValue());
			break;
		}

	}

	private static final String CALIBRATION = "Calibration";

	private void parseCalibrationList(XMLEventReader eventReader)
			throws XMLStreamException {
		GoIOCalibration gc;
		while (eventReader.hasNext()) {
			XMLEvent event = eventReader.nextEvent();
			if (event.isStartElement()) {
				StartElement startElement = event.asStartElement();
				if (startElement.getName().getLocalPart() == (CALIBRATION)) {
					gc = GoIOCalibration.readFromXml(eventReader, startElement);
					calibrationList.add(gc);
					if (!dds && numAutoIds < 2)
						gc.setDdsIndex(calibrationList.indexOf(gc));
					else if (!dds)
						System.err
								.println("[GoIOSensor] - parseCalibrationList: dds("
										+ dds
										+ ") numAutoIds("
										+ numAutoIds
										+ ") darf nicht größer als 1 sein.");
				}
			}
			if (event.isEndElement()
					&& event.asEndElement().getName().getLocalPart() == (CALIBRATIONLIST))
				break;
		}
	}

	private void parseXmlStream(InputStream in) throws XMLStreamException {
		if (in == null)
			throw new XMLStreamException(
					"[GoIOSensor] - parseXmlStream: Der InputStream ist leer.");
		XMLInputFactory inputFactory = XMLInputFactory.newInstance();
		XMLEventReader eventReader = inputFactory.createXMLEventReader(in);
		while (eventReader.hasNext()) {
			XMLEvent event = eventReader.nextEvent();

			if (event.isStartElement()) {
				StartElement startElement = event.asStartElement();
				if (startElement.getName().getLocalPart() == (SENSOR)) {
					parseSensorAttribs(startElement);
					continue;
				} else if (startElement.getName().getLocalPart() == (CALIBRATIONLIST)) {
					parseCalibrationListAttribs(startElement);
					parseCalibrationList(eventReader);
					continue;
				} else if (startElement.getName().getLocalPart() == (SENSORBURNREC)) {
					sensorDDSRec = GoIOSensorDDSRec.readFromXml(eventReader);
					continue;
				} else if (startElement.getName().getLocalPart()
						.equals("Hardware"))
					continue;
				else if (startElement.getName().getLocalPart()
						.equals("CollectDefaults"))
					continue;
				else if (startElement.getName().getLocalPart()
						.equals("BurnParams"))
					continue;

				event = eventReader.nextEvent();
				parseSensorElement(startElement.getName().getLocalPart(), event);
			}
			if (event.isEndElement()
					&& event.asEndElement().getName().getLocalPart() == (SENSOR))
				break;
		}
	}

	private void parseSensorElement(String n, XMLEvent e) {
		try {
			if (n == (SENSORFAMILY)) {
				sensorFamily = e.toString();
				sensorFamily = strings.getString(sensorFamily);
			} else if (n == (LOCALNAME)) {
				localName = e.toString();
				localName = strings.getString(localName);
			} else if (n == (LNAME_BASE)) {
				lname_base = e.toString();
				lname_base = strings.getString(lname_base);
			} else if (n == (REVERSESUPPORTED))
				reverseSupported = convertToBoolean(e.toString());
			else if (n == (ZEROSUPPORTED))
				zeroSupported = convertToBoolean(e.toString());
			else if (n == (SHOWZEROBUTTON))
				showZeroButton = convertToBoolean(e.toString());
			else if (n == (VOLTAGEDECIMALPLACES))
				voltageDecimalPlaces = convertToInt(e.toString());
			else if (n == (SINGLEPOINTCALIBRATIONSUPPORTED))
				singlePointCalibrationSupported = convertToBoolean(e.toString());
			else if (n == (SINGLEPOINTCALIBRATIONDEFAULT))
				singlePointCalibrationDefault = convertToDouble(e.toString());
			else if (n == (WARMUPTIME))
				warmUpTime = convertToDouble(e.toString());
			else if (n == (DELTA))
				delta = convertToDouble(e.toString());
			else if (n == (MAXDELTA))
				maxDelta = convertToDouble(e.toString());
			else if (n == (NUMSAMPLES))
				numSamples = convertToInt(e.toString());
			else if (n == (AVERAGING))
				averaging = convertToBoolean(e.toString());
		} catch (NumberFormatException ex) {
			if (DEBUG)
				System.err
						.println("[GoIOSensor] - parseSensorElement: NumberFormatException(name("
								+ n + ") inhalt(" + e + "))");
		}
	}

	public EProbeType convertToType(String s) {
		try {
			if (s.equals("Photogate"))
				s = "PhotoGate";// FIX for error in sensormap.xml
			return EProbeType.valueOf("kProbeType" + s);
		} catch (Exception e) {
			if (DEBUG)
				e.printStackTrace();
		}
		return sensorType;
	}

	private boolean convertToBoolean(String s) {
		return Integer.parseInt(s) != 0;
	}

	private int convertToInt(String s) {
		return Integer.parseInt(s);
	}

	private double convertToDouble(String s) {
		return Double.parseDouble(s);
	}

	public String getSensorId() {
		return sensorId;
	}

	public void setSensorId(String sensorId) {
		this.sensorId = sensorId;
	}

	public int getSensorNumber() {
		return sensorNumber;
	}

	public void setSensorNumber(int sensorNumber) {
		this.sensorNumber = sensorNumber;
	}

	public int getNumAutoIds() {
		return numAutoIds;
	}

	public void setNumAutoIds(int numAutoIds) {
		this.numAutoIds = numAutoIds;
	}

	public EProbeType getSensorType() {
		return sensorType;
	}

	public void setSensorType(EProbeType sensorType) {
		this.sensorType = sensorType;
	}

	public boolean isDds() {
		return dds;
	}

	public void setDds(boolean dds) {
		this.dds = dds;
	}

	public boolean isLegacy() {
		return legacy;
	}

	public void setLegacy(boolean legacy) {
		this.legacy = legacy;
	}

	public String getLocalName() {
		return localName;
	}

	public void setLocalName(String localName) {
		this.localName = localName;
	}

	public String getRubrik() {
		return (sensorFamily == null || "".equals(sensorFamily)) ? lname_base
				: sensorFamily;
	}

	public String getLname_base() {
		return lname_base;
	}

	public void setLname_base(String lname_base) {
		this.lname_base = lname_base;
	}

	public String getMessgroesse() {
		return lname_base;
	}

	public boolean isReverseSupported() {
		return reverseSupported;
	}

	public void setReverseSupported(boolean reverseSupported) {
		this.reverseSupported = reverseSupported;
	}

	public boolean isZeroSupported() {
		return zeroSupported;
	}

	public void setZeroSupported(boolean zeroSupported) {
		this.zeroSupported = zeroSupported;
	}

	public boolean isShowZeroButton() {
		return showZeroButton;
	}

	public void setShowZeroButton(boolean showZeroButton) {
		this.showZeroButton = showZeroButton;
	}

	public int getVoltageDecimalPlaces() {
		return voltageDecimalPlaces;
	}

	public void setVoltageDecimalPlaces(int voltageDecimalPlaces) {
		this.voltageDecimalPlaces = voltageDecimalPlaces;
	}

	public boolean isSinglePointCalibrationSupported() {
		return singlePointCalibrationSupported;
	}

	public void setSinglePointCalibrationSupported(
			boolean singlePointCalibrationSupported) {
		this.singlePointCalibrationSupported = singlePointCalibrationSupported;
	}

	public double getSinglePointCalibrationDefault() {
		return singlePointCalibrationDefault;
	}

	public void setSinglePointCalibrationDefault(
			double singlePointCalibrationDefault) {
		this.singlePointCalibrationDefault = singlePointCalibrationDefault;
	}

	public double getWarmUpTime() {
		return warmUpTime;
	}

	public void setWarmUpTime(double warmUpTime) {
		this.warmUpTime = warmUpTime;
	}

	public double getDelta() {
		return delta;
	}

	public void setDelta(double delta) {
		this.delta = delta;
	}

	public double getMaxDelta() {
		return maxDelta;
	}

	public void setMaxDelta(double maxDelta) {
		this.maxDelta = maxDelta;
	}

	public int getNumSamples() {
		return numSamples;
	}

	public void setNumSamples(int numSamples) {
		this.numSamples = numSamples;
	}

	public boolean isAveraging() {
		return averaging;
	}

	public void setAveraging(boolean averaging) {
		this.averaging = averaging;
	}

	public int getDefaultIndex() {
		return defaultIndex;
	}

	public void setDefaultIndex(int defaultIndex) {
		this.defaultIndex = defaultIndex;
	}

	public GoIOCalibrationVector getCalibrationList() {
		return calibrationList;
	}

	public void setCalibrationList(GoIOCalibrationVector calibrationList) {
		this.calibrationList = calibrationList;
	}

	public GoIOCalibrationVector getCalibrationListForSensorNumber() {
		return calibrationList.getCalibrationVector(sensorNumber);
	}

	public Vector<Messbereich> getMessbereicheForSensorNumber() {
		return calibrationList.getMessbereiche(sensorNumber);
	}

	public GoIOSensorDDSRec getSensorDDSRec() {
		return sensorDDSRec;
	}

	public void setSensorDDSRec(GoIOSensorDDSRec sensorDDSRec) {
		this.sensorDDSRec = sensorDDSRec;
	}

	public String getSensorFamily() {
		return sensorFamily;
	}

	public void setSensorFamily(String sensorFamily) {
		this.sensorFamily = sensorFamily;
	}

	public String toString() {
		return localName + "(" + sensorId + "|" + sensorNumber + ")";
	}

	public static void main(String[] args) {
		String s = "	<Sensor ID_Name=\"ACC\" BaseID=\"27\" NumAutoIDs=\"1\" Product=\"ACC-BTA\" DDS=\"1\" Obnoxious=\"0\" Type=\"Analog5V\" CHType=\"analog-in\" Legacy=\"1\">"
				+ "		<LocalName>25g Accelerometer</LocalName>"
				+ "		<lname_base>Acceleration</lname_base>"
				+ "		<sname_base>Acc</sname_base>"
				+ "		<URL>http://www.vernier.com/faqs/acc-din.html</URL>"
				+ "		<SensorFamily>Accelerometer</SensorFamily>"
				+ "		<UnitConversionFamily>Accelerometer</UnitConversionFamily>"
				+ "		<UserCalibrationSupported>1</UserCalibrationSupported>"
				+ "		<ReverseSupported>0</ReverseSupported>"
				+ "		<ZeroSupported>1</ZeroSupported>"
				+ "		<ShowZeroButton>1</ShowZeroButton>"
				+ "		<UserRemoteSupported>1</UserRemoteSupported>"
				+ "		<GroupGraphs>0</GroupGraphs>"
				+ "		<Oversample>0</Oversample>"
				+ "		<SampleAtTimeZeroSupported>1</SampleAtTimeZeroSupported>"
				+ "		<ForceTimePlaces>0</ForceTimePlaces>"
				+ "		<VoltageDecimalPlaces>2</VoltageDecimalPlaces>"
				+ "		<LabProSourceSupported>1</LabProSourceSupported>"
				+ "		<SkipSourceSupported>1</SkipSourceSupported>"
				+ "		<USBDirectTempSourceSupported>0</USBDirectTempSourceSupported>"
				+ "		<CyclopsSourceSupported>0</CyclopsSourceSupported>"
				+ "		<OhausScoutProSourceSupported>0</OhausScoutProSourceSupported>"
				+ "		<OceanOpticsSourceSupported>0</OceanOpticsSourceSupported>"
				+ "		<GPSGarminSourceSupported>0</GPSGarminSourceSupported>"
				+ "		<SensorDAQSourceSupported>1</SensorDAQSourceSupported>"
				+ "		<LabQuestSourceSupported>1</LabQuestSourceSupported>"
				+ "		<SinglePointCalibrationSupported>1</SinglePointCalibrationSupported>"
				+ "		<SinglePointCalibrationDefault>0</SinglePointCalibrationDefault>"
				+ "		<AutoAddDigitalMeter>1</AutoAddDigitalMeter>"
				+ "		<AutoAddGauge>0</AutoAddGauge>"
				+ "		<AutoAddThermometer>0</AutoAddThermometer>"
				+ "		<Hardware>"
				+ "			<Uncertainty>0</Uncertainty>"
				+ "			<AvgCurrent>9</AvgCurrent>"
				+ "			<WarmUpTime>0</WarmUpTime>"
				+ "		</Hardware>"
				+ "		<CollectDefaults>"
				+ "			<Delta>0.02</Delta>"
				+ "			<MaxDelta>0.001</MaxDelta>"
				+ "			<NumSamples>100</NumSamples>"
				+ "			<Averaging>1</Averaging>"
				+ "			<ColumnName>Acceleration</ColumnName>"
				+ "			<DefaultColumnShortName>Acc</DefaultColumnShortName>"
				+ "		</CollectDefaults>"
				+ "		<CalibrationList DefaultIndex=\"0\" DefaultCalIDName=\"ACC\">"
				+ "			<Calibration Units=\"m/s^2\" PrettyUnits=\"m/s&#xB2;\" IDName=\"ACC\" UserName=\"25 g Acc m/s&#xB2;\" IDValue=\"27\" DDSIndex=\"0\" RangeDesc=\"\" ColumnName=\"\" ShortColumnName=\"\" ExperimentType=\"1\" DecimalPlaces=\"-1\" Trace=\"1\" Trace_LL=\"1\" LiveReadouts=\"1\">"
				+ "				<Equation Type=\"1\" LabProOp=\"14\" K0=\"-287.8\" K1=\"127.9\"/>"
				+ "				<DisplayRange YMin=\"-250\" YMax=\"250\"/>"
				+ "			</Calibration>"
				+ "			<Calibration Units=\"N/kg\" PrettyUnits=\"N/kg\" IDName=\"ACCNKG\" UserName=\"25 g Acc N/kg\" IDValue=\"27\" DDSIndex=\"0\" RangeDesc=\"\" ColumnName=\"\" ShortColumnName=\"\" ExperimentType=\"1\" DecimalPlaces=\"-1\" Trace=\"1\" Trace_LL=\"1\" LiveReadouts=\"1\">"
				+ "				<Equation Type=\"1\" LabProOp=\"14\" K0=\"-287.8\" K1=\"127.9\"/>"
				+ "				<DisplayRange YMin=\"-250\" YMax=\"250\"/>"
				+ "			</Calibration>"
				+ "			<Calibration Units=\"g\" PrettyUnits=\"g\" IDName=\"ACCG\" UserName=\"25 g Acc g\" IDValue=\"27\" DDSIndex=\"0\" RangeDesc=\"\" ColumnName=\"\" ShortColumnName=\"\" ExperimentType=\"1\" DecimalPlaces=\"-1\" Trace=\"1\" Trace_LL=\"1\" LiveReadouts=\"1\">"
				+ "				<Equation Type=\"1\" LabProOp=\"14\" K0=\"-29.37\" K1=\"13.05\"/>"
				+ "				<DisplayRange YMin=\"-25\" YMax=\"25\"/>"
				+ "			</Calibration>"
				+ "		</CalibrationList>"
				+ "		<BurnParams>"
				+ "			<SensorBurnRec SensorNumber=\"27\">"
				+ "				<MemMapVersion>1</MemMapVersion>"
				+ "				<ManufacturerID>0</ManufacturerID>"
				+ "				<SensorLongName>25G ACCEL</SensorLongName>"
				+ "				<SensorShortName>ACCEL</SensorShortName>"
				+ "				<Uncertainty>0</Uncertainty>"
				+ "				<SignificantFigures>3</SignificantFigures>"
				+ "				<Precision>3</Precision>"
				+ "				<CurrentRequirement>9</CurrentRequirement>"
				+ "				<Averaging>1</Averaging>"
				+ "				<MinSamplePeriod>0.001</MinSamplePeriod>"
				+ "				<TypSamplePeriod>0.05</TypSamplePeriod>"
				+ "				<TypNumberofSamples>100</TypNumberofSamples>"
				+ "				<WarmUpTime>0</WarmUpTime>"
				+ "				<ExperimentType>1</ExperimentType>"
				+ "				<OperationType>14</OperationType>"
				+ "				<CalibrationEquation>1</CalibrationEquation>"
				+ "				<YminValue>-250</YminValue>"
				+ "				<YmaxValue>250</YmaxValue>"
				+ "				<Yscale>10</Yscale>"
				+ "				<ActiveCalPage>0</ActiveCalPage>"
				+ "				<CalibrationList>"
				+ "					<Calibration DDSIndex=\"0\">"
				+ "						<CalibrationCoefficientA>-287.8</CalibrationCoefficientA>"
				+ "						<CalibrationCoefficientB>127.9</CalibrationCoefficientB>"
				+ "						<CalibrationCoefficientC>0</CalibrationCoefficientC>"
				+ "						<Units>m/s^2</Units>"
				+ "					</Calibration>"
				+ "				</CalibrationList>"
				+ "			</SensorBurnRec>"
				+ "		</BurnParams>" + "	</Sensor>";

		GoIOSensor gs = readFromXml(s);

		System.out.println(gs);
		GoIOCalibrationPage gcp;
		if (gs.getSensorDDSRec().getHighestValidCalPageIndex() > -1) {
			gcp = gs.getSensorDDSRec().getCalibrationPage().firstElement();
			System.out.println(gcp);
			System.out.println(gcp.getUnits());
		}
	}
}
