/* 
 * GHMessen
 * --------
 * 
 * GHMessen is a measuring program to use in connection with a computer 
 * measuring interface. It's designed for school usage.
 * 
 * Copyright (C) 2010-2011  Frank Schütte - Gymnasium Himmelsthür
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms
 * of the GNU General Public License as published by the Free Software Foundation; either 
 * version 3 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with this program; 
 * if not, see <http://www.gnu.org/licenses/>.
 * 
 * Frank Schütte fschuett AT gymnasium-himmelsthuer DOT de
 * 
 */

package libGoIO;

import static libGoIO.GoIOLibrary.*;
import ghmessen.adapter.EAdapterException;
import ghmessen.adapter.MesskanalException;
import libGoIO.GoIOLibrary.ESensorIdNumber;

import org.bridj.Pointer;

public class LibGoIO {

	private static boolean DEBUG = false;

	public static int DLLVERSIONMAJOR = 2;
	public static int DLLVERSIONMINOR = 53;

	private static final int MAX_VERSUCHE = 3;

	public static int VERNIER_DEFAULT_VENDOR_ID = 0x08F7;
	public static int LABPRO_DEFAULT_PRODUCT_ID = 0x0001,
			USB_DIRECT_TEMP_DEFAULT_PRODUCT_ID = 0x0002, // aka GoTemp
			SKIP_DEFAULT_PRODUCT_ID = 0x0003, // aka GoLink
			CYCLOPS_DEFAULT_PRODUCT_ID = 0x0004, // aka GoMotion
			NGI_DEFAULT_PRODUCT_ID = 0x0005, // aka LabQuest
			LOWCOST_SPEC_DEFAULT_PRODUCT_ID = 0x0006, // aka CK Spectrometer
			MINI_GC_DEFAULT_PRODUCT_ID = 0x0007, // aka Vernier Mini Gas
													// Chromatograph
			STANDALONE_DAQ_DEFAULT_PRODUCT_ID = 0x0008;
	// Identifikationsnummer zur Unterscheidung von den
	// Cassy-Identifikationsnummern 524...
	public static final int VERNIER_EINGANG = (VERNIER_DEFAULT_VENDOR_ID << 16)
			+ VERNIER_DEFAULT_VENDOR_ID;
	// Konstanten aus verschiedenen Dateien
	private static final byte SKIP_CMD_ID_START_MEASUREMENTS = 0x18;
	private static final byte SKIP_CMD_ID_STOP_MEASUREMENTS = 0x19;
	private static final byte SKIP_CMD_ID_SET_ANALOG_INPUT_CHANNEL = 0x29;
	private static final byte SKIP_ANALOG_INPUT_CHANNEL_VIN = 1;
	private static final byte SKIP_ANALOG_INPUT_CHANNEL_VIN_LOW = 2;

	private boolean initialisiert;

	public LibGoIO() throws EAdapterException {
		super();
		initialisiert = false;
		init();
		if (DEBUG)
			System.out.println("[LibGoIO] - Konstruktor aufgerufen...");
	}

	public void finalize() {
		uninit();
	}

	private void init() throws EAdapterException {
		initialisiert = goIOInit() == 0;
		if (!initialisiert)
			throw new EAdapterException(
					"[LibGoIO] - init: Der GoIO-Adapter konnte nicht initialisiert werden.");
		Pointer<Short> major = Pointer.allocateShort();
		Pointer<Short> minor = Pointer.allocateShort();
		if (goIOGetDLLVersion(major, minor) != 0)
			throw new EAdapterException(
					"[LibGoIO] - init: Die Adapter-DLL-Version("
							+ DLLVERSIONMAJOR + "." + DLLVERSIONMINOR
							+ ") konnte nicht ermittelt werden.");
		if (major.get() < DLLVERSIONMAJOR || minor.get() < DLLVERSIONMINOR)
			throw new EAdapterException(
					"[LibGoIO] - init: Die Version der geladenen DLL(libGoIO.so) ist "
							+ major.get() + "." + minor.get()
							+ ", sollte aber mindestens " + DLLVERSIONMAJOR
							+ "." + DLLVERSIONMINOR + " sein.");
		if ((major.get() != DLLVERSIONMAJOR || minor.get() != DLLVERSIONMINOR))
			System.out
					.println("[LibGoIO] - init: Die Version der geladenen DLL(libGoIO.so) ist "
							+ major.get()
							+ "."
							+ minor.get()
							+ " anstatt "
							+ DLLVERSIONMAJOR + "." + DLLVERSIONMINOR + ".");

	}

	public void uninit() {
		if (initialisiert)
			goIOUninit();
		initialisiert = false;
	}

	public Pointer<?> open(String deviceName)
			throws EAdapterException {
		int n = 0;
		Pointer<?> hDevice = null;
		int vendor = VERNIER_DEFAULT_VENDOR_ID;
		int product = SKIP_DEFAULT_PRODUCT_ID;
		if ((n = goIOUpdateListOfAvailableDevices(vendor, product)) <= 0)
			throw new EAdapterException(
					"[LibGoIO] - open: Es wurde kein Geräte mit Vendor("
							+ vendor + ") Product(" + product + ") gefunden.");
		if (DEBUG)
			System.out.println("[LibGoIO] - open: Es wurden " + n
					+ " passende Geräte gefunden.");
		int PCHARSIZE = 20;
		Pointer<Byte> pchar = Pointer.allocateBytes(PCHARSIZE);
		for (int index = 0; index < n; index++) {
			goIOGetNthAvailableDeviceName(pchar, PCHARSIZE, vendor, product,
					index);
			if (DEBUG)
				System.out.println("[LibGoIO] - open: Das " + (index + 1)
						+ ". Gerät hat den Namen " + pchar.getCString() + ".");
			if (deviceName.equals(pchar.getCString())) {
				int versuch = 0;
				while (versuch < MAX_VERSUCHE && hDevice == null) {
					hDevice = goIOSensorOpen(pchar, vendor, product, 0);
					if (DEBUG && hDevice == null)
						System.err.println("[LibGoIO] - open: " + (versuch + 1)
								+ ". Versuch ist fehlgeschlagen...");
					versuch++;
				}
				if (DEBUG)
					System.out.println("[LibGoIO] - open: Das Gerät "
							+ deviceName
							+ (hDevice != null ? " wurde geöffnet."
									: " konnte nicht geöffnet werden."));
				break;
			}
		}
		if (hDevice != null) {
			if (goIOSensorUnlock(hDevice) != 0 && DEBUG)
				System.err
						.println("[LibGoIO] - open: Unlock nach dem Öffnen fehlgeschlagen.");
		}
		return hDevice;
	}

	public void close(Pointer<?> hDevice) {
		int versuch = 0;
		int ergebnis = -1;
		while (versuch < MAX_VERSUCHE && ergebnis != 0) {
			ergebnis = goIOSensorClose(hDevice);
			if (ergebnis != 0 && DEBUG)
				System.err.println("[LibGoIO] - close: " + (versuch + 1)
						+ ". Versuch ist fehlgeschlagen...");
			try {
				Thread.sleep(500);
			} catch (InterruptedException e) {
				if(DEBUG)
					e.printStackTrace();
			}
			versuch++;
		}
		if (DEBUG)
			System.out.println("[LibGoIO] - close: Das Gerät("
					+ hDevice
					+ ") "
					+ (ergebnis == 0 ? " wurde geschlossen. "
							: "konte nicht geschlossen werden."));
	}

	public GoIOSensorDDSRec getSensorDDSRec(Pointer<?> hDevice)
			throws EAdapterException {
		if (hDevice == null)
			throw new EAdapterException(
					"[LibGoIO] - getSensorDDSRed: Es ist kein Sensor geöffnet.");
		Pointer<Byte> sensorNumber = Pointer.allocateByte();
		if (goIOSensorDDSMemGetSensorNumber(hDevice, sensorNumber, 0, 0) != 0)
			throw new EAdapterException(
					"[LibGoIO] - getSensorDDSRec: Die Sensornummer konnte nicht gelesen werden.");
		Pointer<tagGSensorDDSRec> ddsrec = Pointer
				.allocate(tagGSensorDDSRec.class);
		if (goIOSensorDDSMemGetRecord(hDevice, ddsrec) != 0)
			throw new EAdapterException(
					"[LibGoIO] - getSensorDDSRec: Der Sensor-DDS-Record konnte nicht gelesen werden.");
		return new GoIOSensorDDSRec(ddsrec.get());
	}

	public void setSensorNumber(Pointer<?> hDevice, int productId, int sensorId)
			throws EAdapterException {
		if (hDevice == null)
			throw new EAdapterException(
					"[LibGoIO - setSensorNumber: Es ist kein Sensor geöffnet.");
		Pointer<Byte> old = Pointer.allocateByte();
		if (goIOSensorDDSMemGetSensorNumber(hDevice, old, 0, 0) != 0)
			throw new EAdapterException(
					"[LibGoIO - setSensorNumber: Die alte Sensornummer kann nicht ermittelt werden.");
		if (old.get() != 0 && old.get() != sensorId)
			throw new EAdapterException(
					"[LibGoIO - setSensorNumber: Die alte Sensornummer("
							+ old.get()
							+ ") weist auf einen smarten Sensor hin (neue Sensornummer("
							+ sensorId + "))");
		int oldProbeType = goIOSensorGetProbeType(hDevice);
		if (goIOSensorDDSMemSetSensorNumber(hDevice, (byte) sensorId) != 0)
			throw new EAdapterException(
					"[LibGoIO - setSensorNumber: Kein Go! Link-Device.");
		if (productId == SKIP_DEFAULT_PRODUCT_ID
				|| productId == MINI_GC_DEFAULT_PRODUCT_ID) {
			int newProbeType;
			if ((newProbeType = goIOSensorGetProbeType(hDevice)) != oldProbeType) {
				Pointer<Byte> pParams = Pointer.allocateByte();
				pParams.set(EProbeType.kProbeTypeAnalog10V.value == newProbeType ? SKIP_ANALOG_INPUT_CHANNEL_VIN
						: SKIP_ANALOG_INPUT_CHANNEL_VIN_LOW);
				Pointer<Byte> pResponse = Pointer.allocateBytes(8);
				Pointer<Integer> pNumResponse = Pointer.allocateInt();
				pNumResponse.set((int) pResponse.getValidBytes());
				if (goIOSensorSendCmdAndGetResponse(hDevice,
						SKIP_CMD_ID_SET_ANALOG_INPUT_CHANNEL, pParams, 1,
						pResponse, pNumResponse, SKIP_TIMEOUT_MS_DEFAULT) != 0)
					throw new EAdapterException(
							"[LibGoIO] - setSensorNumber: SKIP_CMD_ID_SET_ANALOG_INPUT_CHANNEL fehlgeschlagen.");
			}
		}
	}

	public void setSensorDDSRec(Pointer<?> hDevice, GoIOSensorDDSRec rec)
			throws EAdapterException {
		if (hDevice == null)
			throw new EAdapterException(
					"[LibGoIO] - setSensorDDSRec: Es ist kein Sensor geöffnet.");
		Pointer<tagGSensorDDSRec> ptRec = Pointer
				.allocate(tagGSensorDDSRec.class);
		ptRec.set(rec.toGSensorDDSRec());
		if (goIOSensorDDSMemSetRecord(hDevice, ptRec) != 0)
			throw new EAdapterException(
					"[LibGoIO] - setSensorDDSRec: Der DDSRec konnte nicht geschrieben werden.");
	}

	public static boolean isSmartSensor(int sensorNumber) {
		return sensorNumber >= ESensorIdNumber.kSensorIdNumber_FirstSmartSensor.value;
	}

	public boolean isSmartSensor(Pointer<?> hDevice) throws EAdapterException {
		if (hDevice == null)
			throw new EAdapterException(
					"[LibGoIO] - setSensorDDSRec: Es ist kein Sensor geöffnet.");
		Pointer<Byte> sensorNumber = Pointer.allocateByte();
		goIOSensorDDSMemGetSensorNumber(hDevice, sensorNumber, 0, 0);
		return isSmartSensor(sensorNumber.get());
	}

	public double getLatestVoltageMeasurement(Pointer<?> hDevice)
			throws MesskanalException {
		int n = goIOSensorGetNumMeasurementsAvailable(hDevice);
		if (n <= 0)
			throw new MesskanalException(
					"[LibGoIO] - getLatestMeasurement: Es sind keine Messwerte verfügbar.");
		int raw = goIOSensorGetLatestRawMeasurement(hDevice);
		if (DEBUG)
			System.out.println("[LibGoIO] - getLatestRawMeasurement: raw("
					+ raw + ")");
		return goIOSensorConvertToVoltage(hDevice, raw);

	}

	public double getLatestMeasurement(Pointer<?> hDevice)
			throws MesskanalException {
		double data = getLatestVoltageMeasurement(hDevice);
		data = goIOSensorCalibrateData(hDevice, data);
		return data;
	}

	public int getNumMeasurementsAvailable(Pointer<?> hDevice) {
		return goIOSensorGetNumMeasurementsAvailable(hDevice);
	}

	private double calibrateVoltage(Pointer<?> hDevice, double volts) {
		return goIOSensorCalibrateData(hDevice, volts);
	}

	public int getVoltageMeasurements(Pointer<?> hDevice, double[] puffer,
			int maxLaenge) throws EAdapterException {
		if (maxLaenge > 0) {
			Pointer<Integer> pInt = Pointer.allocateInts(maxLaenge);
			maxLaenge = goIOSensorReadRawMeasurements(hDevice, pInt, maxLaenge);
			for (int index = 0; index < maxLaenge; index++)
				puffer[index] = goIOSensorConvertToVoltage(hDevice,
						pInt.get(index));
		}
		return maxLaenge;
	}

	public int getCalibratedMeasurements(Pointer<?> hDevice, double[] puffer,
			int maxLaenge) throws EAdapterException {
		maxLaenge = getVoltageMeasurements(hDevice, puffer, maxLaenge);
		if (maxLaenge > 0) {
			for (int index = 0; index < maxLaenge; index++)
				puffer[index] = calibrateVoltage(hDevice, puffer[index]);
		}
		return maxLaenge;
	}

	public void startMeasurements(Pointer<?> hDevice) throws EAdapterException {
		if (goIOSensorClearIO(hDevice) != 0)
			throw new EAdapterException(
					"[LibGoIO] - startMeasurements: Ausnahme bei ClearIO.");
		if (goIOSensorSendCmdAndGetResponse(hDevice,
				SKIP_CMD_ID_START_MEASUREMENTS, null, 0, null, null, 100) != 0)
			throw new EAdapterException(
					"[LibGoIO] - stopMeasurements: Ausnahme beim Starten aufgetreten.");
		if (DEBUG)
			System.out.println("[LibGoIO] - startMeasurements...");
	}

	public void stopMeasurements(Pointer<?> hDevice) throws EAdapterException {
		if (goIOSensorSendCmdAndGetResponse(hDevice,
				SKIP_CMD_ID_STOP_MEASUREMENTS, null, 0, null, null, 100) != 0)
			throw new EAdapterException(
					"[LibGoIO] - stopMeasurements: Ausnahme beim Stoppen aufgetreten.");
		if (DEBUG)
			System.out.println("[LibGoIO] - stopMeasurements...");
	}

	public double getAbstand(Pointer<?> hDevice) throws EAdapterException {
		return goIOSensorGetMeasurementPeriod(hDevice, 100);
	}

	public double setAbstandMS(Pointer<?> hDevice, double abstand)
			throws EAdapterException {
		abstand /= 1000.0;
		if (goIOSensorSetMeasurementPeriod(hDevice, abstand, 100) != 0)
			throw new EAdapterException(
					"[LibGoIO] - setAbstand: Abstand setzen fehlgeschlagen.");
		if (DEBUG)
			System.out.println("[LibGoIO] - setAbstand: abstand(" + abstand
					+ ")");
		return goIOSensorGetMeasurementPeriod(hDevice, 100);
	}

	public int getSensorNumber(Pointer<?> hDevice) throws EAdapterException {
		Pointer<Byte> pSensorNumber = Pointer.allocateByte();
		if (goIOSensorDDSMemGetSensorNumber(hDevice, pSensorNumber, 0, 0) != 0)
			throw new EAdapterException(
					"[LibGoIO] - getSensorNumber: die Sensornummer konnte nicht ermittelt werden.");
		return pSensorNumber.get();
	}

	public int getCurrentSensorNumber(Pointer<?> hDevice)
			throws EAdapterException {
		Pointer<Byte> pSensorNumber = Pointer.allocateByte();
		if (goIOSensorDDSMemGetSensorNumber(hDevice, pSensorNumber, 1,
				SKIP_TIMEOUT_MS_DEFAULT) != 0)
			throw new EAdapterException(
					"[LibGoIO] - getCurrentSensorNumber: die neue Sensornummer konnte nicht ermittelt werden.");
		return pSensorNumber.get();
	}

	public void setActiveCalPage(Pointer<?> hDevice, int page)
			throws EAdapterException {
		if (goIOSensorDDSMemSetActiveCalPage(hDevice, (byte) page) != 0)
			throw new EAdapterException(
					"[LibGoIO] - setActiveCalPage: die neue CalibrationPage("
							+ page + ") konnte nicht gesetzt werden.");
	}
}
