/* 
 * GHMessen
 * --------
 * 
 * GHMessen is a measuring program to use in connection with a computer 
 * measuring interface. It's designed for school usage.
 * 
 * Copyright (C) 2010-2011  Frank Schütte - Gymnasium Himmelsthür
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms
 * of the GNU General Public License as published by the Free Software Foundation; either 
 * version 3 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with this program; 
 * if not, see <http://www.gnu.org/licenses/>.
 * 
 * Frank Schütte fschuett AT gymnasium-himmelsthuer DOT de
 * 
 */

package test;

import static libGoIO.GoIOLibrary.GOIO_MAX_SIZE_DEVICE_NAME;
import static libGoIO.GoIOLibrary.SKIP_TIMEOUT_MS_DEFAULT;
import static libGoIO.GoIOLibrary.goIOGetDLLVersion;
import static libGoIO.GoIOLibrary.goIOGetNthAvailableDeviceName;
import static libGoIO.GoIOLibrary.goIOInit;
import static libGoIO.GoIOLibrary.goIOSensorCalibrateData;
import static libGoIO.GoIOLibrary.goIOSensorClose;
import static libGoIO.GoIOLibrary.goIOSensorConvertToVoltage;
import static libGoIO.GoIOLibrary.goIOSensorDDSMemGetActiveCalPage;
import static libGoIO.GoIOLibrary.goIOSensorDDSMemGetCalPage;
import static libGoIO.GoIOLibrary.goIOSensorDDSMemGetCalibrationEquation;
import static libGoIO.GoIOLibrary.goIOSensorDDSMemGetLongName;
import static libGoIO.GoIOLibrary.goIOSensorDDSMemGetSensorNumber;
import static libGoIO.GoIOLibrary.goIOSensorOpen;
import static libGoIO.GoIOLibrary.goIOSensorReadRawMeasurements;
import static libGoIO.GoIOLibrary.goIOSensorSendCmdAndGetResponse;
import static libGoIO.GoIOLibrary.goIOSensorSetMeasurementPeriod;
import static libGoIO.GoIOLibrary.goIOUninit;
import static libGoIO.GoIOLibrary.goIOUpdateListOfAvailableDevices;
import static libGoIO.GoIOLibrary.EEquationType.kEquationType_Linear;

import java.awt.Dimension;

import javax.swing.JTextArea;

import org.bridj.Pointer;
import org.jdesktop.application.Application;
import org.jdesktop.application.FrameView;

public class TestGoIOView extends FrameView {
	public static int MAX_NUM_MEASUREMENTS = 100;
	public static String[] deviceDesc = new String[] { "?", "?", "Go! Temp",
			"Go! Link", "Go! Motion", "?", "?", "Mini GC" };

	public static int VERNIER_DEFAULT_VENDOR_ID = 0x08F7;
	public static int LABPRO_DEFAULT_PRODUCT_ID = 0x0001,
			USB_DIRECT_TEMP_DEFAULT_PRODUCT_ID = 0x0002, // aka GoTemp
			SKIP_DEFAULT_PRODUCT_ID = 0x0003, // aka GoLink
			CYCLOPS_DEFAULT_PRODUCT_ID = 0x0004, // aka GoMotion
			NGI_DEFAULT_PRODUCT_ID = 0x0005, // aka LabQuest
			LOWCOST_SPEC_DEFAULT_PRODUCT_ID = 0x0006, // aka CK Spectrometer
			MINI_GC_DEFAULT_PRODUCT_ID = 0x0007, // aka Vernier Mini Gas
													// Chromatograph
			STANDALONE_DAQ_DEFAULT_PRODUCT_ID = 0x0008;
	public static byte SKIP_CMD_ID_START_MEASUREMENTS = 0x18;

	/**
	 * @wbp.nonvisual location=198,140
	 */
	private final JTextArea textPane = new JTextArea();

	public TestGoIOView(Application application) {
		super(application);
		getRootPane().setContentPane(textPane);
		getRootPane().getParent().setMinimumSize(new Dimension(800, 600));
		fillTextPane();
	}

	private void fillTextPane() {
		Pointer<Byte> deviceName;
		deviceName = Pointer.allocateBytes(GOIO_MAX_SIZE_DEVICE_NAME);
		Pointer<Integer> vendorId = Pointer.allocateInt(); // USB vendor id
		Pointer<Integer> productId = Pointer.allocateInt(); // USB product id
		Pointer<Byte> tmpstring = Pointer.allocateBytes(100);
		Pointer<Short> MajorVersion = Pointer.allocateShort();
		Pointer<Short> MinorVersion = Pointer.allocateShort();
		Pointer<Byte> units = Pointer.allocateBytes(20);
		Pointer<Byte> equationType = Pointer.allocateByte();
		equationType.set((byte) 0);

		Pointer<Integer> rawMeasurements = Pointer
				.allocateInts(MAX_NUM_MEASUREMENTS);
		double[] volts = new double[MAX_NUM_MEASUREMENTS];
		double[] calbMeasurements = new double[MAX_NUM_MEASUREMENTS];
		int numMeasurements, i;
		double averageCalbMeasurement;

		textPane.append(String.format("goIODeviceCheck version 1.0\n"));

		goIOInit();

		goIOGetDLLVersion(MajorVersion, MinorVersion);
		textPane.append(String.format(
				"This app is linked to GoIO lib version %d.%d .\n",
				MajorVersion.get(), MinorVersion.get()));

		boolean bFoundDevice = GetAvailableDeviceName(deviceName,
				GOIO_MAX_SIZE_DEVICE_NAME, vendorId, productId);
		if (!bFoundDevice)
			textPane.append(String.format("No Go devices found.\n"));
		else {
			Pointer<?> hDevice = goIOSensorOpen(deviceName, vendorId.get(),
					productId.get(), 0);
			if (hDevice != null) {
				textPane.append(String.format(
						"Successfully opened %s device %s .\n",
						deviceDesc[productId.get()], deviceName));

				Pointer<Byte> charId = Pointer.allocateByte();
				goIOSensorDDSMemGetSensorNumber(hDevice, charId, 0, 0);
				textPane.append(String.format("Sensor id = %d", charId.get()));

				goIOSensorDDSMemGetLongName(hDevice, tmpstring,
						(short) tmpstring.getValidBytes());
				if (tmpstring.getCString().length() != 0)
					textPane.append(String.format("(%s)",
							tmpstring.getCString()));
				textPane.append(String.format("\n"));

				goIOSensorSetMeasurementPeriod(hDevice, 0.040,
						SKIP_TIMEOUT_MS_DEFAULT);// 40 milliseconds measurement
													// period.
				goIOSensorSendCmdAndGetResponse(hDevice,
						SKIP_CMD_ID_START_MEASUREMENTS, null, 0, null, null,
						SKIP_TIMEOUT_MS_DEFAULT);
				OSSleep(1000); // Wait 1 second.

				numMeasurements = goIOSensorReadRawMeasurements(hDevice,
						rawMeasurements, MAX_NUM_MEASUREMENTS);
				textPane.append(String.format(
						"%d measurements received after about 1 second.\n",
						numMeasurements));
				averageCalbMeasurement = 0.0;
				for (i = 0; i < numMeasurements; i++) {
					volts[i] = goIOSensorConvertToVoltage(hDevice,
							rawMeasurements.get(i));
					calbMeasurements[i] = goIOSensorCalibrateData(hDevice,
							volts[i]);
					averageCalbMeasurement += calbMeasurements[i];
				}
				if (numMeasurements > 1)
					averageCalbMeasurement = averageCalbMeasurement
							/ numMeasurements;

				goIOSensorDDSMemGetCalibrationEquation(hDevice, equationType);
				if (equationType.get() != kEquationType_Linear.ordinal())
					units.setCString("volts");
				else {
					Pointer<Float> a = Pointer.allocateFloat(), b = Pointer
							.allocateFloat(), c = Pointer.allocateFloat();
					Pointer<Byte> activeCalPage = Pointer.allocateByte();
					activeCalPage.set((byte) 0);
					goIOSensorDDSMemGetActiveCalPage(hDevice, activeCalPage);
					goIOSensorDDSMemGetCalPage(hDevice, activeCalPage.get(), a,
							b, c, units, (short) units.getValidBytes());
				}
				textPane.append(String.format(
						"Average measurement = %8.3f %s .\n",
						averageCalbMeasurement, units.getCString()));

				goIOSensorClose(hDevice);
			}
		}
		goIOUninit();
	}

	public static void OSSleep(int ms) {
		try {
			Thread.sleep(ms);
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
	}

	public static boolean GetAvailableDeviceName(Pointer<Byte> deviceName,
			int nameLength, Pointer<Integer> pVendorId,
			Pointer<Integer> pProductId) {
		boolean bFoundDevice = false;
		deviceName.set((byte) 0);
		int numSkips = goIOUpdateListOfAvailableDevices(
				VERNIER_DEFAULT_VENDOR_ID, SKIP_DEFAULT_PRODUCT_ID);
		int numJonahs = goIOUpdateListOfAvailableDevices(
				VERNIER_DEFAULT_VENDOR_ID, USB_DIRECT_TEMP_DEFAULT_PRODUCT_ID);
		int numCyclopses = goIOUpdateListOfAvailableDevices(
				VERNIER_DEFAULT_VENDOR_ID, CYCLOPS_DEFAULT_PRODUCT_ID);
		int numMiniGCs = goIOUpdateListOfAvailableDevices(
				VERNIER_DEFAULT_VENDOR_ID, MINI_GC_DEFAULT_PRODUCT_ID);

		if (numSkips > 0) {
			goIOGetNthAvailableDeviceName(deviceName, nameLength,
					VERNIER_DEFAULT_VENDOR_ID, SKIP_DEFAULT_PRODUCT_ID, 0);
			pVendorId.set(VERNIER_DEFAULT_VENDOR_ID);
			pProductId.set(SKIP_DEFAULT_PRODUCT_ID);
			bFoundDevice = true;
		} else if (numJonahs > 0) {
			goIOGetNthAvailableDeviceName(deviceName, nameLength,
					VERNIER_DEFAULT_VENDOR_ID,
					USB_DIRECT_TEMP_DEFAULT_PRODUCT_ID, 0);
			pVendorId.set(VERNIER_DEFAULT_VENDOR_ID);
			pProductId.set(USB_DIRECT_TEMP_DEFAULT_PRODUCT_ID);
			bFoundDevice = true;
		} else if (numCyclopses > 0) {
			goIOGetNthAvailableDeviceName(deviceName, nameLength,
					VERNIER_DEFAULT_VENDOR_ID, CYCLOPS_DEFAULT_PRODUCT_ID, 0);
			pVendorId.set(VERNIER_DEFAULT_VENDOR_ID);
			pProductId.set(CYCLOPS_DEFAULT_PRODUCT_ID);
			bFoundDevice = true;
		} else if (numMiniGCs > 0) {
			goIOGetNthAvailableDeviceName(deviceName, nameLength,
					VERNIER_DEFAULT_VENDOR_ID, MINI_GC_DEFAULT_PRODUCT_ID, 0);
			pVendorId.set(VERNIER_DEFAULT_VENDOR_ID);
			pProductId.set(MINI_GC_DEFAULT_PRODUCT_ID);
			bFoundDevice = true;
		}

		return bFoundDevice;
	}

}
