/* 
 * GHMessen
 * --------
 * 
 * GHMessen is a measuring program to use in connection with a computer 
 * measuring interface. It's designed for school usage.
 * 
 * Copyright (C) 2010-2011  Frank Schütte - Gymnasium Himmelsthür
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms
 * of the GNU General Public License as published by the Free Software Foundation; either 
 * version 3 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with this program; 
 * if not, see <http://www.gnu.org/licenses/>.
 * 
 * Frank Schütte fschuett AT gymnasium-himmelsthuer DOT de
 * 
 */

package udev;

import java.util.HashSet;
import java.util.Set;
import java.util.Vector;

import javax.swing.event.EventListenerList;

public class Udev implements Runnable {

	private static final boolean DEBUG = false;

	private UdevLibrary.udev _udev;
	private UdevLibrary.udev_monitor _monitor;
	private String[] subsystem;
	private String[] devName;
	protected transient EventListenerList listener = new EventListenerList();

	public Udev(String[] subsystem, String[] devName) {
		this.subsystem = subsystem;
		this.devName = devName;
		_udev = UdevLibrary.INSTANCE.udev_new();
		if (_udev == null)
			throw new NullPointerException(
					"UdevLibrary: Kontext konnte nicht erzeugt werden.");
		_monitor = UdevLibrary.INSTANCE.udev_monitor_new_from_netlink(_udev,
				"udev");
		/* subsystem needs to be unique for monitoring */
		for (String s : unique(subsystem))
			UdevLibrary.INSTANCE
					.udev_monitor_filter_add_match_subsystem_devtype(_monitor,
							s, null);
		UdevLibrary.INSTANCE.udev_monitor_enable_receiving(_monitor);
	}

	public void finalize() {
		if (_monitor != null)
			_monitor = null;
		if (_udev != null)
			UdevLibrary.INSTANCE.udev_unref(_udev);
	}

	public void addDeviceEventListener(DeviceEventListener listener) {
		this.listener.add(DeviceEventListener.class, listener);
	}

	public void removeDeviceEventListener(DeviceEventListener listener) {
		this.listener.remove(DeviceEventListener.class, listener);
	}

	void fireDeviceAddedEvent(DeviceAddedEvent evt) {
		Object[] listeners = listener.getListenerList();
		for (int i = 0; i < listeners.length; i += 2) {
			if (listeners[i] == DeviceEventListener.class) {
				((DeviceEventListener) listeners[i + 1]).deviceAdded(evt);
			}
		}
	}

	void fireDeviceRemovedEvent(DeviceRemovedEvent evt) {
		Object[] listeners = listener.getListenerList();
		for (int i = 0; i < listeners.length; i += 2) {
			if (listeners[i] == DeviceEventListener.class) {
				((DeviceEventListener) listeners[i + 1]).deviceRemoved(evt);
			}
		}
	}

	public Vector<UdevDevice> getDevices() {
		UdevLibrary.udev_enumerate enumerate = null;
		Vector<UdevDevice> result = null;
		try {
			if (_udev == null)
				return result;
			result = new Vector<UdevDevice>();
			for (int i = 0; i < subsystem.length; i++) {
				enumerate = UdevLibrary.INSTANCE.udev_enumerate_new(_udev);
				UdevLibrary.INSTANCE.udev_enumerate_add_match_sysname(
						enumerate, devName[i]);
				UdevLibrary.INSTANCE.udev_enumerate_scan_devices(enumerate);
				UdevLibrary.udev_list_entry listEntry = UdevLibrary.INSTANCE
						.udev_enumerate_get_list_entry(enumerate);
				if(listEntry == null)
					continue;
				while (listEntry != null) {
					String path = UdevLibrary.INSTANCE
							.udev_list_entry_get_name(listEntry);
					UdevLibrary.udev_device dev = UdevLibrary.INSTANCE
							.udev_device_new_from_syspath(_udev, path);
					UdevDevice device = new UdevDevice(dev, subsystem[i]);
					result.add(device);
					// UdevLibrary.INSTANCE.udev_device_unref(dev);
					listEntry = UdevLibrary.INSTANCE
							.udev_list_entry_get_next(listEntry);

				}
				if (enumerate != null){
					UdevLibrary.INSTANCE.udev_enumerate_unref(enumerate);
					enumerate = null;
				}
			}
		} finally {
			if (enumerate != null){
				UdevLibrary.INSTANCE.udev_enumerate_unref(enumerate);
				enumerate = null;
			}
		}
		return result;
	}

	public static void main(String[] args) {
		System.out.println("Zugriff auf Structure mittels JNA");
		System.out.println("============================");
		System.out.println("Finde GoIO-Gerät");
		System.out.println("------------------------");
		Udev udev = new Udev(new String[]{"usb"}, new String[]{"*-*"});
		Vector<UdevDevice> result = udev.getDevices();
		System.out.println(result);
		Thread daemon = new Thread(udev);
		daemon.start();
		try {
			while (true) {
				System.out.print(".");
				Thread.sleep(5000);
			}
		} catch (InterruptedException e) {
			e.printStackTrace();
		}

	}

	private int name_match(String name){
		for(int i=0;i<devName.length;i++)
			if(name.matches(devName[i].replace("*","\\d")))
				return i;
		return -1;
	}
	
	/**
	 * create array without duplicates
	 * @param in array with duplicates
	 * @return array without duplicates
	 */
	private String[] unique(String[] in){
		Set<String> sSet = new HashSet<String>();
		for(String s : in)
			sSet.add(s);
		return sSet.toArray(new String[sSet.size()]);
	}
	
	@Override
	public void run() {
		while (true) {
			UdevLibrary.udev_device device = UdevLibrary.INSTANCE
					.udev_monitor_receive_device(_monitor);
			if (device == null){
				try {
					Thread.sleep(1000);
				} catch(InterruptedException e){ }
				continue;
			}
			try {
				String name = UdevLibrary.INSTANCE
						.udev_device_get_sysname(device);
				int i = name_match(name);
				if(i < 0 || i >= devName.length)
					continue;
				String action = UdevLibrary.INSTANCE
						.udev_device_get_action(device);
				String node = UdevLibrary.INSTANCE
						.udev_device_get_devnode(device);
				if (action == null)
					continue;
				if (action.equals("add")) {
					if (DEBUG)
						System.out.println("[Udev] device added(name(" + name
								+ "),node(" + node + "))");
					fireDeviceAddedEvent(new DeviceAddedEvent(this,
							new UdevDevice(device, subsystem[i])));
				} else if (action.equals("remove")) {
					if (DEBUG)
						System.out.println("[Udev] device removed(name(" + name
								+ "),node(" + node + "))");
					fireDeviceRemovedEvent(new DeviceRemovedEvent(this, name));
				}
			} finally {
				if (device != null)
					UdevLibrary.INSTANCE.udev_device_unref(device);
			}
		}
	}
}
