/* 
 * GHMessen
 * --------
 * 
 * GHMessen is a measuring program to use in connection with a computer 
 * measuring interface. It's designed for school usage.
 * 
 * Copyright (C) 2010-2011  Frank Schütte - Gymnasium Himmelsthür
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms
 * of the GNU General Public License as published by the Free Software Foundation; either 
 * version 3 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with this program; 
 * if not, see <http://www.gnu.org/licenses/>.
 * 
 * Frank Schütte fschuett AT gymnasium-himmelsthuer DOT de
 * 
 */

package udev;

import java.util.Vector;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class UdevDevice {
	private String name;
	private String vendorID;
	private String manufacturer;
	private String productID;
	private String productName;
	private String node;
	private String devtype;
	private String usbname;
	private String bcdrevision;
	private String port;
	private int ibcdrevision;

	private static Pattern pUsb = Pattern.compile("usb");

	public UdevDevice(String name, String vendor, String manufacturer,
			String product, String productName, String node, String devtype, String revision, String usbname) {
		this.name = name;
		this.vendorID = vendor;
		this.manufacturer = manufacturer;
		this.productID = product;
		this.productName = name;
		this.node = node;
		this.devtype = devtype;
		this.usbname = usbname;
		this.bcdrevision = revision;
		this.ibcdrevision = convertToInt(revision);
		Matcher mUsb = pUsb.matcher(name);
		this.port = mUsb.find() ? "USB" : "COM";
	}

	public UdevDevice(UdevLibrary.udev_device device, String subsystem) {
		this();
		if (device == null)
			return;
		node = UdevLibrary.INSTANCE.udev_device_get_devnode(device);
		name = UdevLibrary.INSTANCE.udev_device_get_sysname(device);
		devtype = UdevLibrary.INSTANCE.udev_device_get_devtype(device);
		UdevLibrary.udev_device dev = null;
		if ("usb_device".equals(devtype)) {
			dev = device;
		} 
		else if (subsystem.equals("usb") || subsystem.equals("usbmisc")) {
			dev = UdevLibrary.INSTANCE
					.udev_device_get_parent_with_subsystem_devtype(device,
							subsystem, "usb_device");
		}
		else if (name.startsWith("ttyACM")) {
			dev = UdevLibrary.INSTANCE
					.udev_device_get_parent(UdevLibrary.INSTANCE
							.udev_device_get_parent(device));
		}
		else
			dev = null;
		if (dev == null)
			return;
		vendorID = UdevLibrary.INSTANCE.udev_device_get_sysattr_value(dev,
				"idVendor");
		productID = UdevLibrary.INSTANCE.udev_device_get_sysattr_value(dev,
				"idProduct");
		manufacturer = UdevLibrary.INSTANCE.udev_device_get_sysattr_value(dev,
				"manufacturer");
		productName = UdevLibrary.INSTANCE.udev_device_get_sysattr_value(dev,
				"productID");
		if(productName == null)
			productName = UdevLibrary.INSTANCE.udev_device_get_sysattr_value(dev,"product");
		bcdrevision = UdevLibrary.INSTANCE.udev_device_get_sysattr_value(dev,
				"bcdDevice");
		usbname = UdevLibrary.INSTANCE.udev_device_get_sysattr_value(dev, "busnum");
		usbname += ":" + UdevLibrary.INSTANCE.udev_device_get_sysattr_value(dev, "devnum");
		this.ibcdrevision = convertToInt(bcdrevision);
		// UdevLibrary.INSTANCE.udev_device_unref(dev);
		Matcher mUsb = pUsb.matcher(name);
		this.port = mUsb.find() ? "USB" : "COM";
	}

	@Override
	public String toString() {
		return super.toString() + "(Name(" + name + ") Vendor(" + vendorID
				+ ") Manufacturer(" + manufacturer + ") Product(" + productID
				+ ") Name(" + productName + ") Node(" + node + ") Usbname("
				+ usbname + ") Revision("	+ bcdrevision + ")";
	}

	public UdevDevice() {
		this("", "", "", "", "", "", "", "", "");
	}

	public static UdevDevice getDeviceByNode(Vector<UdevDevice> list,
			String node) {
		if (list == null || list.size() == 0)
			return null;
		for (UdevDevice dev : list.toArray(new UdevDevice[1]))
			if (dev.node.equals(node))
				return dev;
		return null;
	}

	public void setName(String name) {
		this.name = name;
	}

	public String getName() {
		return name;
	}

	public void setVendorID(String vendor) {
		this.vendorID = vendor;
	}

	public String getVendorIDString() {
		return vendorID;
	}

	public int getVendorID() {
		return convertToInt(vendorID, 16);
	}

	public void setManufacturer(String manufacturer) {
		this.manufacturer = manufacturer;
	}

	public String getManufacturer() {
		return manufacturer;
	}

	public void setProductID(String product) {
		this.productID = product;
	}

	public String getProductIDString() {
		return productID;
	}

	public int getProductID() {
		try {
			return Integer.parseInt(productID, 16);
		} catch (NumberFormatException e) {
			return -1;
		}
	}

	public void setProductName(String productName) {
		this.productName = productName;
	}

	public String getProductName() {
		return productName;
	}

	public void setNode(String node) {
		this.node = node;
	}

	public String getNode() {
		return node;
	}

	public void setDevType(String devtype) {
		this.devtype = devtype;
	}
	
	public String getDevType() {
		return devtype;
	}
	
	public void setUSBName(String usbname) {
		this.usbname = usbname;
	}
	
	public String getUSBName() {
		return usbname;
	}
	
	public void setbcdRevision(String revision) {
		this.bcdrevision = revision;
	}

	public String getbcdRevisionString() {
		return bcdrevision;
	}

	public void setPort(String port) {
		this.port = port;
	}

	public String getPort() {
		return port;
	}

	public int getVersion() {
		int version = 1000 * (ibcdrevision / 4096 % 16) + 100
				* (ibcdrevision / 256 % 16) + 10 * (ibcdrevision / 16 % 16)
				+ (ibcdrevision % 16);
		return version;
	}

	private int convertToInt(String s) {
		return convertToInt(s, 10);
	}

	private int convertToInt(String s, int basis) {
		int result;
		try {
			result = Integer.parseInt(s, basis);
		} catch (NumberFormatException e) {
			result = 0;
		}
		return result;
	}

}
